/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "acqdp.utility.opt_einsum_paths",
        "sources": [
            "acqdp/utility/opt_einsum_paths.pyx"
        ]
    },
    "module_name": "acqdp.utility.opt_einsum_paths"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__acqdp__utility__opt_einsum_paths
#define __PYX_HAVE_API__acqdp__utility__opt_einsum_paths
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "acqdp/utility/opt_einsum_paths.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr;
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr;

/* "acqdp/utility/opt_einsum_paths.pyx":52
 * 
 * 
 * def ssa_to_linear(ssa_path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with static single assignment ids to a path with recycled
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear {
  PyObject_HEAD
  PyObject *__pyx_v_ids;
  PyObject *__pyx_v_ssa_ids;
};


/* "acqdp/utility/opt_einsum_paths.pyx":63
 *     path = []
 *     for ssa_ids in ssa_path:
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))             # <<<<<<<<<<<<<<
 *         for ssa_id in ssa_ids:
 *             ids[ssa_id:] -= 1
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *__pyx_outer_scope;
  PyObject *__pyx_v_ssa_id;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "acqdp/utility/opt_einsum_paths.pyx":69
 * 
 * 
 * def linear_to_ssa(path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with recycled linear ids to a path with static single
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa {
  PyObject_HEAD
  PyObject *__pyx_v_ids;
  PyObject *__pyx_v_linear_to_ssa;
};


/* "acqdp/utility/opt_einsum_paths.pyx":82
 *     ssa_path = []
 *     for ids in path:
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))             # <<<<<<<<<<<<<<
 *         for id_ in sorted(ids, reverse=True):
 *             del linear_to_ssa[id_]
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *__pyx_outer_scope;
  PyObject *__pyx_v_id_;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "acqdp/utility/opt_einsum_paths.pyx":141
 * 
 * 
 * def optimal(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """
 *     Computes all possible pair contractions in a depth-first recursive manner,
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal {
  PyObject_HEAD
  PyObject *__pyx_v__optimal_iterate;
  PyObject *__pyx_v_best;
  PyObject *__pyx_v_memory_limit;
  PyObject *__pyx_v_output;
  PyObject *__pyx_v_result_cache;
  PyObject *__pyx_v_size_cache;
  PyObject *__pyx_v_size_dict;
};


/* "acqdp/utility/opt_einsum_paths.pyx":316
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 * 
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ {
  PyObject_HEAD
  PyObject *__pyx_v__branch_iterate;
  PyObject *__pyx_v_memory_limit;
  PyObject *__pyx_v_output;
  PyObject *__pyx_v_result_cache;
  PyObject *__pyx_v_self;
  PyObject *__pyx_v_size_cache;
  PyObject *__pyx_v_size_dict;
};


/* "acqdp/utility/opt_einsum_paths.pyx":350
 *         result_cache = {}
 * 
 *         def _branch_iterate(path, inputs, remaining, flops, size):             # <<<<<<<<<<<<<<
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *__pyx_outer_scope;
  PyObject *__pyx_v_flops;
  PyObject *__pyx_v_inputs;
  PyObject *__pyx_v_path;
  PyObject *__pyx_v_remaining;
  PyObject *__pyx_v_size;
};


/* "acqdp/utility/opt_einsum_paths.pyx":469
 * 
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):             # <<<<<<<<<<<<<<
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate {
  PyObject_HEAD
  PyObject *__pyx_v_cost_fn;
  PyObject *__pyx_v_dim_ref_counts;
  PyObject *__pyx_v_footprints;
  PyObject *__pyx_v_k1;
  PyObject *__pyx_v_k2s;
  PyObject *__pyx_v_output;
  PyObject *__pyx_v_remaining;
  PyObject *__pyx_v_sizes;
};


/* "acqdp/utility/opt_einsum_paths.pyx":470
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)             # <<<<<<<<<<<<<<
 *     if push_all:
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *__pyx_outer_scope;
  PyObject *__pyx_v_k2;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "acqdp/utility/opt_einsum_paths.pyx":502
 * 
 * 
 * def ssa_greedy_optimize(inputs, output, sizes, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     This is the core function for :func:`greedy` but produces a path with
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize {
  PyObject_HEAD
  PyObject *__pyx_8genexpr4__pyx_v_count;
  PyObject *__pyx_v_dim_to_keys;
  PyObject *__pyx_v_k1;
};


/* "acqdp/utility/opt_einsum_paths.pyx":551
 *     # ref counts of >=2 or >=3.
 *     dim_ref_counts = {
 *         count: set(dim for dim, keys in dim_to_keys.items() if len(keys) >= count) - output             # <<<<<<<<<<<<<<
 *         for count in [2, 3]}
 * 
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *__pyx_outer_scope;
  PyObject *__pyx_v_dim;
  PyObject *__pyx_v_keys;
};


/* "acqdp/utility/opt_einsum_paths.pyx":591
 *         # Find new candidate contractions.
 *         k1 = k12
 *         k2s = set(k2 for dim in k1 for k2 in dim_to_keys[dim])             # <<<<<<<<<<<<<<
 *         k2s.discard(k1)
 *         if k2s:
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *__pyx_outer_scope;
  PyObject *__pyx_v_dim;
  PyObject *__pyx_v_k2;
};


/* "acqdp/utility/opt_einsum_paths.pyx":658
 * 
 * 
 * def _tree_to_sequence(c):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a contraction tree to a contraction path as it has to be
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence {
  PyObject_HEAD
  PyObject *__pyx_v_i;
  PyObject *__pyx_v_t;
};


/* "acqdp/utility/opt_einsum_paths.pyx":706
 * 
 *         for i in sorted([i for i in j if type(i) == int]):
 *             s[0] += (sum(1 for q in t if q < i),)             # <<<<<<<<<<<<<<
 *             t.insert(s[0][-1], i)
 * 
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *__pyx_outer_scope;
  PyObject *__pyx_v_q;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "acqdp/utility/opt_einsum_paths.pyx":764
 * 
 * 
 * def _bitmapset_indices(s):             # <<<<<<<<<<<<<<
 *     """
 *     Returns a generator object allowing to iterate over the elements contained
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices {
  PyObject_HEAD
  PyObject *__pyx_v_j;
  PyObject *__pyx_v_s;
};


/* "acqdp/utility/opt_einsum_paths.pyx":811
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ {
  PyObject_HEAD
  PyObject *__pyx_v_g;
  PyObject *__pyx_v_i;
  PyObject *__pyx_v_inputs;
  PyObject *__pyx_v_output;
  PyObject *__pyx_v_r;
  PyObject *__pyx_v_symbol2int;
};


/* "acqdp/utility/opt_einsum_paths.pyx":851
 *         # convert all indices to integers (makes set operations ~10 % faster)
 *         symbol2int = {c: j for j, c in enumerate(set.union(*inputs) | output)}
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]             # <<<<<<<<<<<<<<
 *         output = set(symbol2int[c] for c in output)
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_outer_scope;
  PyObject *__pyx_v_c;
};


/* "acqdp/utility/opt_einsum_paths.pyx":852
 *         symbol2int = {c: j for j, c in enumerate(set.union(*inputs) | output)}
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]
 *         output = set(symbol2int[c] for c in output)             # <<<<<<<<<<<<<<
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}
 *         size_dict = [size_dict[j] for j in range(len(size_dict))]
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_outer_scope;
  PyObject *__pyx_v_c;
};


/* "acqdp/utility/opt_einsum_paths.pyx":858
 *         # all summation indices occurring exactly in one input:
 *         i_single = set(
 *             c for c in set.union(*inputs) - output             # <<<<<<<<<<<<<<
 *             if sum(1 for i in inputs if c in i) == 1
 *         )
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_outer_scope;
  PyObject *__pyx_v_c;
  PyObject *__pyx_v_genexpr;
};


/* "acqdp/utility/opt_einsum_paths.pyx":859
 *         i_single = set(
 *             c for c in set.union(*inputs) - output
 *             if sum(1 for i in inputs if c in i) == 1             # <<<<<<<<<<<<<<
 *         )
 * 
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *__pyx_outer_scope;
  PyObject *__pyx_v_i;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "acqdp/utility/opt_einsum_paths.pyx":894
 * 
 *             # convert set of tensors g to a bitmap set:
 *             g = functools.reduce(lambda x, y: x | y, (1 << j for j in g))             # <<<<<<<<<<<<<<
 * 
 *             # the bitmap set of all tensors is computed as it is needed to
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_outer_scope;
  PyObject *__pyx_v_j;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "acqdp/utility/opt_einsum_paths.pyx":906
 *             # output index dimensions as initial cost_cap
 *             cost_cap = helpers.compute_size_by_dict(
 *                 set.union(*(inputs[j] for j in _bitmapset_indices(g))) & output,             # <<<<<<<<<<<<<<
 *                 size_dict
 *             )
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_outer_scope;
  PyObject *__pyx_v_j;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "acqdp/utility/opt_einsum_paths.pyx":939
 * 
 *                                                     # indices of remaining indices:
 *                                                     i_r = (set.union(*(inputs[j] for j in _bitmapset_indices(r)))             # <<<<<<<<<<<<<<
 *                                                            if r != 0 else set())
 * 
 */
struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr {
  PyObject_HEAD
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_outer_scope;
  PyObject *__pyx_v_j;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* pop_index.proto */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix);
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix);
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (likely(PyList_CheckExact(L) && __Pyx_fits_Py_ssize_t(ix, type, is_signed))) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    __Pyx_fits_Py_ssize_t(ix, type, is_signed) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#else
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)\
    __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
        __Pyx__PyObject_PopIndex(L, py_ix))
#endif

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AndObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAnd(op1, op2) : PyNumber_And(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RshiftObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_RshiftObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceRshift(op1, op2) : PyNumber_Rshift(op1, op2))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_FloorDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceFloorDivide(op1, op2) : PyNumber_FloorDivide(op1, op2))
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'acqdp.utility.opt_einsum_paths' */
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr = 0;
static PyTypeObject *__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr = 0;
#define __Pyx_MODULE_NAME "acqdp.utility.opt_einsum_paths"
extern int __pyx_module_is_main_acqdp__utility__opt_einsum_paths;
int __pyx_module_is_main_acqdp__utility__opt_einsum_paths = 0;

/* Implementation of 'acqdp.utility.opt_einsum_paths' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_property;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_zip;
static const char __pyx_k_N[] = "N";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_q[] = "q";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k__6[] = "_";
static const char __pyx_k_bi[] = "bi";
static const char __pyx_k_dp[] = "dp";
static const char __pyx_k_fn[] = "fn";
static const char __pyx_k_i1[] = "i1";
static const char __pyx_k_i2[] = "i2";
static const char __pyx_k_id[] = "id_";
static const char __pyx_k_k1[] = "k1";
static const char __pyx_k_k2[] = "k2";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_s1[] = "s1";
static const char __pyx_k_s2[] = "s2";
static const char __pyx_k_xn[] = "xn";
static const char __pyx_k__13[] = "";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_con[] = "con";
static const char __pyx_k_dim[] = "dim";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_i_r[] = "i_r";
static const char __pyx_k_id1[] = "id1";
static const char __pyx_k_id2[] = "id2";
static const char __pyx_k_ids[] = "ids";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_k12[] = "k12";
static const char __pyx_k_k2s[] = "k2s";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_len[] = "len";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_mem[] = "mem";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_one[] = "one";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_two[] = "two";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_auto[] = "auto";
static const char __pyx_k_best[] = "best";
static const char __pyx_k_call[] = "__call__";
static const char __pyx_k_cost[] = "cost";
static const char __pyx_k_dims[] = "dims";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_keep[] = "keep";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tree[] = "tree";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_cost1[] = "cost1";
static const char __pyx_k_cost2[] = "cost2";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_eager[] = "eager";
static const char __pyx_k_flops[] = "flops";
static const char __pyx_k_gauss[] = "gauss";
static const char __pyx_k_heapq[] = "heapq";
static const char __pyx_k_i_sum[] = "i_sum";
static const char __pyx_k_i_tmp[] = "i_tmp";
static const char __pyx_k_inner[] = "inner";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_lower[] = "lower";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_queue[] = "queue";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_size1[] = "size1";
static const char __pyx_k_size2[] = "size2";
static const char __pyx_k_sizes[] = "sizes";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_union[] = "union";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_arange[] = "arange";
static const char __pyx_k_better[] = "better";
static const char __pyx_k_branch[] = "branch";
static const char __pyx_k_either[] = "either";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_greedy[] = "greedy";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_inputs[] = "inputs";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_output[] = "output";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "reduce";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_shared[] = "shared";
static const char __pyx_k_size12[] = "size12";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_ssa_id[] = "ssa_id";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_argsort[] = "argsort";
static const char __pyx_k_cntrct1[] = "cntrct1";
static const char __pyx_k_cntrct2[] = "cntrct2";
static const char __pyx_k_cost_fn[] = "cost_fn";
static const char __pyx_k_discard[] = "discard";
static const char __pyx_k_flops12[] = "flops12";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_getitem[] = "__getitem__";
static const char __pyx_k_heapify[] = "heapify";
static const char __pyx_k_heappop[] = "heappop";
static const char __pyx_k_helpers[] = "helpers";
static const char __pyx_k_nbranch[] = "nbranch";
static const char __pyx_k_new_ids[] = "new_ids";
static const char __pyx_k_optimal[] = "optimal";
static const char __pyx_k_partial[] = "partial";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_ssa_id1[] = "ssa_id1";
static const char __pyx_k_ssa_id2[] = "ssa_id2";
static const char __pyx_k_ssa_ids[] = "ssa_ids";
static const char __pyx_k_COST_FNS[] = "_COST_FNS";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_branch_1[] = "branch_1";
static const char __pyx_k_branch_2[] = "branch_2";
static const char __pyx_k_cost_cap[] = "cost_cap";
static const char __pyx_k_heappush[] = "heappush";
static const char __pyx_k_i_cntrct[] = "i_cntrct";
static const char __pyx_k_i_single[] = "i_single";
static const char __pyx_k_minimize[] = "minimize";
static const char __pyx_k_new_size[] = "new_size";
static const char __pyx_k_property[] = "property";
static const char __pyx_k_push_all[] = "push_all";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_ssa_id12[] = "ssa_id12";
static const char __pyx_k_ssa_path[] = "ssa_path";
static const char __pyx_k_best_size[] = "best_size";
static const char __pyx_k_candidate[] = "candidate";
static const char __pyx_k_choose_fn[] = "choose_fn";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_functools[] = "functools";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_new_flops[] = "new_flops";
static const char __pyx_k_num_terms[] = "num_terms";
static const char __pyx_k_optimizer[] = "optimizer";
static const char __pyx_k_path_type[] = "path_type";
static const char __pyx_k_remaining[] = "remaining";
static const char __pyx_k_size_dict[] = "size_dict";
static const char __pyx_k_subgraphs[] = "subgraphs";
static const char __pyx_k_BETTER_FNS[] = "_BETTER_FNS";
static const char __pyx_k_best_flops[] = "best_flops";
static const char __pyx_k_branch_1_2[] = "branch-1";
static const char __pyx_k_branch_2_2[] = "branch-2";
static const char __pyx_k_branch_all[] = "branch_all";
static const char __pyx_k_candidates[] = "candidates";
static const char __pyx_k_flop_count[] = "flop_count";
static const char __pyx_k_footprints[] = "footprints";
static const char __pyx_k_isdisjoint[] = "isdisjoint";
static const char __pyx_k_num_inputs[] = "num_inputs";
static const char __pyx_k_size_cache[] = "size_cache";
static const char __pyx_k_symbol2int[] = "symbol2int";
static const char __pyx_k_BranchBound[] = "BranchBound";
static const char __pyx_k_all_tensors[] = "all_tensors";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_contraction[] = "contraction";
static const char __pyx_k_defaultdict[] = "defaultdict";
static const char __pyx_k_dim_to_keys[] = "dim_to_keys";
static const char __pyx_k_get_path_fn[] = "get_path_fn";
static const char __pyx_k_heappushpop[] = "heappushpop";
static const char __pyx_k_i1_union_i2[] = "i1_union_i2";
static const char __pyx_k_inputs_done[] = "inputs_done";
static const char __pyx_k_AUTO_CHOICES[] = "_AUTO_CHOICES";
static const char __pyx_k_PATH_OPTIONS[] = "_PATH_OPTIONS";
static const char __pyx_k_branch_all_2[] = "branch-all";
static const char __pyx_k_combinations[] = "combinations";
static const char __pyx_k_intersection[] = "intersection";
static const char __pyx_k_memory_limit[] = "memory_limit";
static const char __pyx_k_result_cache[] = "result_cache";
static const char __pyx_k_PathOptimizer[] = "PathOptimizer";
static const char __pyx_k_UNLIMITED_MEM[] = "_UNLIMITED_MEM";
static const char __pyx_k_best_progress[] = "best_progress";
static const char __pyx_k_get_better_fn[] = "get_better_fn";
static const char __pyx_k_get_candidate[] = "_get_candidate";
static const char __pyx_k_linear_to_ssa[] = "linear_to_ssa";
static const char __pyx_k_opportunistic[] = "opportunistic";
static const char __pyx_k_ssa_to_linear[] = "ssa_to_linear";
static const char __pyx_k_unused_inputs[] = "unused_inputs";
static const char __pyx_k_branch_iterate[] = "_branch_iterate";
static const char __pyx_k_calc_k12_flops[] = "calc_k12_flops";
static const char __pyx_k_dim_ref_counts[] = "dim_ref_counts";
static const char __pyx_k_memory_removed[] = "memory-removed";
static const char __pyx_k_push_candidate[] = "_push_candidate";
static const char __pyx_k_simple_chooser[] = "_simple_chooser";
static const char __pyx_k_greedy_line_612[] = "greedy (line 612)";
static const char __pyx_k_idx_contraction[] = "idx_contraction";
static const char __pyx_k_optimal_iterate[] = "_optimal_iterate";
static const char __pyx_k_BranchBound_path[] = "BranchBound.path";
static const char __pyx_k_assess_candidate[] = "_assess_candidate";
static const char __pyx_k_optimal_line_141[] = "optimal (line 141)";
static const char __pyx_k_optimizer_kwargs[] = "optimizer_kwargs";
static const char __pyx_k_register_path_fn[] = "register_path_fn";
static const char __pyx_k_tree_to_sequence[] = "_tree_to_sequence";
static const char __pyx_k_better_size_first[] = "better_size_first";
static const char __pyx_k_bitmapset_indices[] = "_bitmapset_indices";
static const char __pyx_k_difference_update[] = "difference_update";
static const char __pyx_k_update_ref_counts[] = "_update_ref_counts";
static const char __pyx_k_BranchBound___call[] = "BranchBound.__call__";
static const char __pyx_k_BranchBound___init[] = "BranchBound.__init__";
static const char __pyx_k_better_flops_first[] = "better_flops_first";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_opt_einsum_helpers[] = "opt_einsum_helpers";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_cost_memory_removed[] = "cost_memory_removed";
static const char __pyx_k_cutoff_flops_factor[] = "cutoff_flops_factor";
static const char __pyx_k_dynamic_programming[] = "dynamic_programming";
static const char __pyx_k_i1_cut_i2_wo_output[] = "i1_cut_i2_wo_output";
static const char __pyx_k_inputs_contractions[] = "inputs_contractions";
static const char __pyx_k_ssa_greedy_optimize[] = "ssa_greedy_optimize";
static const char __pyx_k_PathOptimizer___call[] = "PathOptimizer.__call__";
static const char __pyx_k_compute_size_by_dict[] = "compute_size_by_dict";
static const char __pyx_k_dynamic_programming_2[] = "dynamic-programming";
static const char __pyx_k_linear_to_ssa_line_69[] = "linear_to_ssa (line 69)";
static const char __pyx_k_memory_removed_jitter[] = "memory-removed-jitter";
static const char __pyx_k_ssa_to_linear_line_52[] = "ssa_to_linear (line 52)";
static const char __pyx_k_subgraph_contractions[] = "subgraph_contractions";
static const char __pyx_k_compute_oversize_flops[] = "_compute_oversize_flops";
static const char __pyx_k_tree_to_sequence_line_658[] = "_tree_to_sequence (line 658)";
static const char __pyx_k_bitmapset_indices_line_764[] = "_bitmapset_indices (line 764)";
static const char __pyx_k_cost_memory_removed_jitter[] = "cost_memory_removed_jitter";
static const char __pyx_k_subgraph_contractions_size[] = "subgraph_contractions_size";
static const char __pyx_k_DynamicProgrammingOptimizer[] = "DynamicProgrammingOptimizer";
static const char __pyx_k_find_disconnected_subgraphs[] = "_find_disconnected_subgraphs";
static const char __pyx_k_linear_to_ssa_locals_genexpr[] = "linear_to_ssa.<locals>.genexpr";
static const char __pyx_k_ssa_to_linear_locals_genexpr[] = "ssa_to_linear.<locals>.genexpr";
static const char __pyx_k_BranchBound___call___line_316[] = "BranchBound.__call__ (line 316)";
static const char __pyx_k_Path_optimizer_already_exists[] = "Path optimizer '{}' already exists.";
static const char __pyx_k_push_candidate_locals_genexpr[] = "_push_candidate.<locals>.genexpr";
static const char __pyx_k_Converts_a_contraction_tree_to[] = "\n    Converts a contraction tree to a contraction path as it has to be\n    returned by path optimizers. A contraction tree can either be an int\n    (=no contraction) or a tuple containing the terms to be contracted. An\n    arbitrary number (>= 1) of terms can be contracted at once. Note that\n    contractions are commutative, e.g. (j, k, l) = (k, l, j). Note that in\n    general, solutions are not unique.\n\n    Parameters\n    ----------\n    c : tuple or int\n        Contraction tree\n\n    Returns\n    -------\n    path : list[set[int]]\n        Contraction path\n\n    Examples\n    --------\n    >>> _tree_to_sequence(((1,2),(0,(4,5,3))))\n    [(1, 2), (1, 2, 3), (0, 2), (0, 1)]\n    ";
static const char __pyx_k_Parameters_inputs_list_List_of[] = "\n        Parameters\n        ----------\n        inputs : list\n            List of sets that represent the lhs side of the einsum subscript\n        output : set\n            Set that represents the rhs side of the overall einsum subscript\n        size_dict : dictionary\n            Dictionary of index sizes\n        memory_limit : int\n            The maximum number of elements in a temporary array\n\n        Returns\n        -------\n        path : list\n            The contraction order (a list of tuples of ints).\n\n        Examples\n        --------\n        >>> n_in = 3  # exponential scaling\n        >>> n_out = 2 # linear scaling\n        >>> s = dict()\n        >>> i_all = []\n        >>> for _ in range(n_out):\n        >>>     i = [set() for _ in range(n_in)]\n        >>>     for j in range(n_in):\n        >>>         for k in range(j+1, n_in):\n        >>>             c = oe.get_symbol(len(s))\n        >>>             i[j].add(c)\n        >>>             i[k].add(c)\n        >>>             s[c] = 2\n        >>>     i_all.extend(i)\n        >>> o = DynamicProgrammingOptimizer()\n        >>> o(i_all, set(), s)\n        [(1, 2), (0, 4), (1, 2), (0, 2), (0, 1)]\n        ";
static const char __pyx_k_acqdp_utility_opt_einsum_paths[] = "acqdp.utility.opt_einsum_paths";
static const char __pyx_k_Computes_all_possible_pair_cont[] = "\n    Computes all possible pair contractions in a depth-first recursive manner,\n    sieving results based on ``memory_limit`` and the best path found so far.\n    Returns the lowest cost path. This algorithm scales factoriallly with\n    respect to the elements in the list ``input_sets``.\n\n    Parameters\n    ----------\n    inputs : list\n        List of sets that represent the lhs side of the einsum subscript.\n    output : set\n        Set that represents the rhs side of the overall einsum subscript.\n    size_dict : dictionary\n        Dictionary of index sizes.\n    memory_limit : int\n        The maximum number of elements in a temporary array.\n\n    Returns\n    -------\n    path : list\n        The optimal contraction order within the memory limit constraint.\n\n    Examples\n    --------\n    >>> isets = [set('abd'), set('ac'), set('bdc')]\n    >>> oset = set('')\n    >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}\n    >>> optimal(isets, oset, idx_sizes, 5000)\n    [(0, 2), (0, 1)]\n    ";
static const char __pyx_k_Contains_the_path_technology_be[] = "\nContains the path technology behind opt_einsum in addition to several path helpers\n";
static const char __pyx_k_Convert_a_path_with_recycled_li[] = "\n    Convert a path with recycled linear ids to a path with static single\n    assignment ids. For example::\n\n        >>> linear_to_ssa([(0, 3), (1, 2), (0, 1)])\n        [(0, 3), (2, 4), (1, 5)]\n    ";
static const char __pyx_k_Convert_a_path_with_static_sing[] = "\n    Convert a path with static single assignment ids to a path with recycled\n    linear ids. For example::\n\n        >>> ssa_to_linear([(0, 3), (2, 4), (1, 5)])\n        [(0, 3), (1, 2), (0, 1)]\n    ";
static const char __pyx_k_Explores_possible_pair_contract[] = "\n    Explores possible pair contractions in a depth-first recursive manner like\n    the ``optimal`` approach, but with extra heuristic early pruning of branches\n    as well sieving by ``memory_limit`` and the best path found so far. Returns\n    the lowest cost path. This algorithm still scales factorially with respect\n    to the elements in the list ``input_sets`` if ``nbranch`` is not set, but it\n    scales exponentially like ``nbranch**len(input_sets)`` otherwise.\n\n    Parameters\n    ----------\n    nbranch : None or int, optional\n        How many branches to explore at each contraction step. If None, explore\n        all possible branches. If an integer, branch into this many paths at\n        each step. Defaults to None.\n    cutoff_flops_factor : float, optional\n        If at any point, a path is doing this much worse than the best path\n        found so far was, terminate it. The larger this is made, the more paths\n        will be fully explored and the slower the algorithm. Defaults to 4.\n    minimize : {'flops', 'size'}, optional\n        Whether to optimize the path with regard primarily to the total\n        estimated flop-count, or the size of the largest intermediate. The\n        option not chosen will still be used as a secondary criterion.\n    cost_fn : callable, optional\n        A function that returns a heuristic 'cost' of a potential contraction\n        with which to sort candidates. Should have signature\n        ``cost_fn(size12, size1, size2, k12, k1, k2)``.\n    ";
static const char __pyx_k_Finds_disconnected_subgraphs_in[] = "\n    Finds disconnected subgraphs in the given list of inputs. Inputs are\n    connected if they share summation indices. Note: Disconnected subgraphs\n    can be contracted independently before forming outer products.\n\n    Parameters\n    ----------\n    inputs : list[set]\n        List of sets that represent the lhs side of the einsum subscript\n    output : set\n        Set that represents the rhs side of the overall einsum subscript\n\n    Returns\n    -------\n    subgraphs : list[set[int]]\n        List containing sets of indices for each subgraph\n\n    Examples\n    --------\n    >>> _find_disconnected_subgraphs([set(\"ab\"), set(\"c\"), set(\"ad\")], set(\"bd\"))\n    [{0, 2}, {1}]\n\n    >>> _find_disconnected_subgraphs([set(\"ab\"), set(\"c\"), set(\"ad\")], set(\"abd\"))\n    [{0}, {1}, {2}]\n    ";
static const char __pyx_k_Finds_the_optimal_path_of_pairw[] = "\n    Finds the optimal path of pairwise contractions without intermediate outer\n    products based a dynamic programming approach presented in\n    Phys. Rev. E 90, 033315 (2014) (the corresponding preprint is publically\n    available at https://arxiv.org/abs/1304.6112). This method is especially\n    well-suited in the area of tensor network states, where it usually\n    outperforms all the other optimization strategies.\n\n    This algorithm shows exponential scaling with the number of inputs\n    in the worst case scenario (see example below). If the graph to be\n    contracted consists of disconnected subgraphs, the algorithm scales\n    linearly in the number of disconnected subgraphs and only exponentially\n    with the number of inputs per subgraph.\n    ";
static const char __pyx_k_Finds_the_path_by_a_three_stage[] = "\n    Finds the path by a three stage algorithm:\n\n    1. Eagerly compute Hadamard products.\n    2. Greedily compute contractions to maximize ``removed_size``\n    3. Greedily compute outer products.\n\n    This algorithm scales quadratically with respect to the\n    maximum number of elements sharing a common dim.\n\n    Parameters\n    ----------\n    inputs : list\n        List of sets that represent the lhs side of the einsum subscript\n    output : set\n        Set that represents the rhs side of the overall einsum subscript\n    size_dict : dictionary\n        Dictionary of index sizes\n    memory_limit : int\n        The maximum number of elements in a temporary array\n    choose_fn : callable, optional\n        A function that chooses which contraction to perform from the queu\n    cost_fn : callable, optional\n        A function that assigns a potential contraction a cost.\n\n    Returns\n    -------\n    path : list\n        The contraction order (a list of tuples of ints).\n\n    Examples\n    --------\n    >>> isets = [set('abd'), set('ac'), set('bdc')]\n    >>> oset = set('')\n    >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}\n    >>> greedy(isets, oset, idx_sizes)\n    [(0, 2), (0, 1)]\n    ";
static const char __pyx_k_Parameters_input_sets_list_List[] = "\n\n        Parameters\n        ----------\n        input_sets : list\n            List of sets that represent the lhs side of the einsum subscript\n        output_set : set\n            Set that represents the rhs side of the overall einsum subscript\n        idx_dict : dictionary\n            Dictionary of index sizes\n        memory_limit : int\n            The maximum number of elements in a temporary array\n\n        Returns\n        -------\n        path : list\n            The contraction order within the memory limit constraint.\n\n        Examples\n        --------\n        >>> isets = [set('abd'), set('ac'), set('bdc')]\n        >>> oset = set('')\n        >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}\n        >>> optimal(isets, oset, idx_sizes, 5000)\n        [(0, 2), (0, 1)]\n        ";
static const char __pyx_k_Returns_a_generator_object_allo[] = "\n    Returns a generator object allowing to iterate over the elements contained\n    in a bitmap set.\n\n    Parameters\n    ----------\n    s : int\n        The bitmap set to iterate over\n\n    Returns\n    -------\n    path : generator\n        Generator object to iterate over the elements in s\n\n    Examples\n    --------\n    >>> type(_bitmapset_indices(0b1001011))\n    generator\n\n    >>> list(_bitmapset_indices(0b1001011))\n    [0, 1, 3, 6]\n    ";
static const char __pyx_k_find_disconnected_subgraphs_lin[] = "_find_disconnected_subgraphs (line 716)";
static const char __pyx_k_optimal_locals__optimal_iterate[] = "optimal.<locals>._optimal_iterate";
static const char __pyx_k_tree_to_sequence_locals_genexpr[] = "_tree_to_sequence.<locals>.genexpr";
static const char __pyx_k_Base_class_for_different_path_op[] = "Base class for different path optimizers to inherit from.\n\n    Subclassed optimizers should define a call method with signature::\n\n        def __call__(self, inputs, output, size_dict, memory_limit=None):\n            \"\"\"\n            Parameters\n            ----------\n            inputs : list[set[str]]\n                The indices of each input array.\n            outputs : set[str]\n                The output indices\n            size_dict : dict[str, int]\n                The size of each index\n            memory_limit : int, optional\n                If given, the maximum allowed memory.\n            \"\"\"\n            # ... compute path here ...\n            return path\n\n    where ``path`` is a list of int-tuples specifiying a contraction order.\n    ";
static const char __pyx_k_BranchBound___call___locals__bra[] = "BranchBound.__call__.<locals>._branch_iterate.<locals>._assess_candidate";
static const char __pyx_k_BranchBound___init___locals_lamb[] = "BranchBound.__init__.<locals>.<lambda>";
static const char __pyx_k_DynamicProgrammingOptimizer___ca[] = "DynamicProgrammingOptimizer.__call__.<locals>.genexpr";
static const char __pyx_k_Path_optimizer_not_found_valid_o[] = "Path optimizer '{}' not found, valid options are {}.";
static const char __pyx_k_acqdp_utility_opt_einsum_paths_p[] = "acqdp/utility/opt_einsum_paths.pyx";
static const char __pyx_k_ssa_greedy_optimize_locals_genex[] = "ssa_greedy_optimize.<locals>.genexpr";
static const char __pyx_k_BranchBound___call___locals__bra_2[] = "BranchBound.__call__.<locals>._branch_iterate";
static const char __pyx_k_DynamicProgrammingOptimizer___ca_2[] = "DynamicProgrammingOptimizer.__call__.<locals>.genexpr.<locals>.genexpr";
static const char __pyx_k_DynamicProgrammingOptimizer___ca_3[] = "DynamicProgrammingOptimizer.__call__.<locals>.<lambda>";
static const char __pyx_k_DynamicProgrammingOptimizer___ca_4[] = "DynamicProgrammingOptimizer.__call__";
static const char __pyx_k_DynamicProgrammingOptimizer___ca_5[] = "DynamicProgrammingOptimizer.__call__ (line 811)";
static PyObject *__pyx_n_s_AUTO_CHOICES;
static PyObject *__pyx_n_s_BETTER_FNS;
static PyObject *__pyx_kp_s_Base_class_for_different_path_op;
static PyObject *__pyx_n_s_BranchBound;
static PyObject *__pyx_n_s_BranchBound___call;
static PyObject *__pyx_kp_u_BranchBound___call___line_316;
static PyObject *__pyx_n_s_BranchBound___call___locals__bra;
static PyObject *__pyx_n_s_BranchBound___call___locals__bra_2;
static PyObject *__pyx_n_s_BranchBound___init;
static PyObject *__pyx_n_s_BranchBound___init___locals_lamb;
static PyObject *__pyx_n_s_BranchBound_path;
static PyObject *__pyx_n_s_COST_FNS;
static PyObject *__pyx_kp_u_Computes_all_possible_pair_cont;
static PyObject *__pyx_kp_u_Convert_a_path_with_recycled_li;
static PyObject *__pyx_kp_u_Convert_a_path_with_static_sing;
static PyObject *__pyx_kp_u_Converts_a_contraction_tree_to;
static PyObject *__pyx_n_s_DynamicProgrammingOptimizer;
static PyObject *__pyx_n_s_DynamicProgrammingOptimizer___ca;
static PyObject *__pyx_n_s_DynamicProgrammingOptimizer___ca_2;
static PyObject *__pyx_n_s_DynamicProgrammingOptimizer___ca_3;
static PyObject *__pyx_n_s_DynamicProgrammingOptimizer___ca_4;
static PyObject *__pyx_kp_u_DynamicProgrammingOptimizer___ca_5;
static PyObject *__pyx_kp_s_Explores_possible_pair_contract;
static PyObject *__pyx_kp_u_Finds_disconnected_subgraphs_in;
static PyObject *__pyx_kp_s_Finds_the_optimal_path_of_pairw;
static PyObject *__pyx_kp_u_Finds_the_path_by_a_three_stage;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_n_s_N;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_PATH_OPTIONS;
static PyObject *__pyx_kp_u_Parameters_input_sets_list_List;
static PyObject *__pyx_kp_u_Parameters_inputs_list_List_of;
static PyObject *__pyx_n_s_PathOptimizer;
static PyObject *__pyx_n_s_PathOptimizer___call;
static PyObject *__pyx_kp_s_Path_optimizer_already_exists;
static PyObject *__pyx_kp_s_Path_optimizer_not_found_valid_o;
static PyObject *__pyx_kp_u_Returns_a_generator_object_allo;
static PyObject *__pyx_n_s_UNLIMITED_MEM;
static PyObject *__pyx_n_s__13;
static PyObject *__pyx_n_s__6;
static PyObject *__pyx_n_s_acqdp_utility_opt_einsum_paths;
static PyObject *__pyx_kp_s_acqdp_utility_opt_einsum_paths_p;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_all_tensors;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_arange;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_argsort;
static PyObject *__pyx_n_s_assess_candidate;
static PyObject *__pyx_n_s_auto;
static PyObject *__pyx_n_s_best;
static PyObject *__pyx_n_s_best_flops;
static PyObject *__pyx_n_s_best_progress;
static PyObject *__pyx_n_s_best_size;
static PyObject *__pyx_n_s_better;
static PyObject *__pyx_n_s_better_flops_first;
static PyObject *__pyx_n_s_better_size_first;
static PyObject *__pyx_n_s_bi;
static PyObject *__pyx_n_s_bitmapset_indices;
static PyObject *__pyx_kp_u_bitmapset_indices_line_764;
static PyObject *__pyx_n_s_branch;
static PyObject *__pyx_n_s_branch_1;
static PyObject *__pyx_kp_s_branch_1_2;
static PyObject *__pyx_n_s_branch_2;
static PyObject *__pyx_kp_s_branch_2_2;
static PyObject *__pyx_n_s_branch_all;
static PyObject *__pyx_kp_s_branch_all_2;
static PyObject *__pyx_n_s_branch_iterate;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_calc_k12_flops;
static PyObject *__pyx_n_s_call;
static PyObject *__pyx_n_s_candidate;
static PyObject *__pyx_n_s_candidates;
static PyObject *__pyx_n_s_choose_fn;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cntrct1;
static PyObject *__pyx_n_s_cntrct2;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_combinations;
static PyObject *__pyx_n_s_compute_oversize_flops;
static PyObject *__pyx_n_s_compute_size_by_dict;
static PyObject *__pyx_n_s_con;
static PyObject *__pyx_n_s_contraction;
static PyObject *__pyx_n_s_cost;
static PyObject *__pyx_n_s_cost1;
static PyObject *__pyx_n_s_cost2;
static PyObject *__pyx_n_s_cost_cap;
static PyObject *__pyx_n_s_cost_fn;
static PyObject *__pyx_n_s_cost_memory_removed;
static PyObject *__pyx_n_s_cost_memory_removed_jitter;
static PyObject *__pyx_n_s_count;
static PyObject *__pyx_n_s_cutoff_flops_factor;
static PyObject *__pyx_n_s_defaultdict;
static PyObject *__pyx_n_s_difference_update;
static PyObject *__pyx_n_s_dim;
static PyObject *__pyx_n_s_dim_ref_counts;
static PyObject *__pyx_n_s_dim_to_keys;
static PyObject *__pyx_n_s_dims;
static PyObject *__pyx_n_s_discard;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dp;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dynamic_programming;
static PyObject *__pyx_kp_s_dynamic_programming_2;
static PyObject *__pyx_n_s_eager;
static PyObject *__pyx_n_s_either;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_find_disconnected_subgraphs;
static PyObject *__pyx_kp_u_find_disconnected_subgraphs_lin;
static PyObject *__pyx_n_s_flop_count;
static PyObject *__pyx_n_s_flops;
static PyObject *__pyx_n_s_flops12;
static PyObject *__pyx_n_s_fn;
static PyObject *__pyx_n_s_footprints;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_functools;
static PyObject *__pyx_n_s_g;
static PyObject *__pyx_n_s_gauss;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_better_fn;
static PyObject *__pyx_n_s_get_candidate;
static PyObject *__pyx_n_s_get_path_fn;
static PyObject *__pyx_n_s_getitem;
static PyObject *__pyx_n_s_greedy;
static PyObject *__pyx_kp_u_greedy_line_612;
static PyObject *__pyx_n_s_heapify;
static PyObject *__pyx_n_s_heappop;
static PyObject *__pyx_n_s_heappush;
static PyObject *__pyx_n_s_heappushpop;
static PyObject *__pyx_n_s_heapq;
static PyObject *__pyx_n_s_helpers;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_i1;
static PyObject *__pyx_n_s_i1_cut_i2_wo_output;
static PyObject *__pyx_n_s_i1_union_i2;
static PyObject *__pyx_n_s_i2;
static PyObject *__pyx_n_s_i_cntrct;
static PyObject *__pyx_n_s_i_r;
static PyObject *__pyx_n_s_i_single;
static PyObject *__pyx_n_s_i_sum;
static PyObject *__pyx_n_s_i_tmp;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_id1;
static PyObject *__pyx_n_s_id2;
static PyObject *__pyx_n_s_ids;
static PyObject *__pyx_n_s_idx_contraction;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_inner;
static PyObject *__pyx_n_s_inputs;
static PyObject *__pyx_n_s_inputs_contractions;
static PyObject *__pyx_n_s_inputs_done;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_intersection;
static PyObject *__pyx_n_s_isdisjoint;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_k1;
static PyObject *__pyx_n_s_k12;
static PyObject *__pyx_n_s_k2;
static PyObject *__pyx_n_s_k2s;
static PyObject *__pyx_n_s_keep;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_len;
static PyObject *__pyx_n_s_linear_to_ssa;
static PyObject *__pyx_kp_u_linear_to_ssa_line_69;
static PyObject *__pyx_n_s_linear_to_ssa_locals_genexpr;
static PyObject *__pyx_n_s_lower;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_mem;
static PyObject *__pyx_n_s_memory_limit;
static PyObject *__pyx_kp_s_memory_removed;
static PyObject *__pyx_kp_s_memory_removed_jitter;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_minimize;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_nbranch;
static PyObject *__pyx_n_s_new_flops;
static PyObject *__pyx_n_s_new_ids;
static PyObject *__pyx_n_s_new_size;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_num_inputs;
static PyObject *__pyx_n_s_num_terms;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_one;
static PyObject *__pyx_n_s_opportunistic;
static PyObject *__pyx_n_s_opt_einsum_helpers;
static PyObject *__pyx_n_s_optimal;
static PyObject *__pyx_n_s_optimal_iterate;
static PyObject *__pyx_kp_u_optimal_line_141;
static PyObject *__pyx_n_s_optimal_locals__optimal_iterate;
static PyObject *__pyx_n_s_optimizer;
static PyObject *__pyx_n_s_optimizer_kwargs;
static PyObject *__pyx_n_s_output;
static PyObject *__pyx_n_s_partial;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_path_type;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_property;
static PyObject *__pyx_n_s_push_all;
static PyObject *__pyx_n_s_push_candidate;
static PyObject *__pyx_n_s_push_candidate_locals_genexpr;
static PyObject *__pyx_n_s_q;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_queue;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_register_path_fn;
static PyObject *__pyx_n_s_remaining;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_result_cache;
static PyObject *__pyx_n_s_reverse;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_s1;
static PyObject *__pyx_n_s_s2;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_shared;
static PyObject *__pyx_n_s_simple_chooser;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_size1;
static PyObject *__pyx_n_s_size12;
static PyObject *__pyx_n_s_size2;
static PyObject *__pyx_n_s_size_cache;
static PyObject *__pyx_n_s_size_dict;
static PyObject *__pyx_n_s_sizes;
static PyObject *__pyx_n_s_sorted;
static PyObject *__pyx_n_s_ssa_greedy_optimize;
static PyObject *__pyx_n_s_ssa_greedy_optimize_locals_genex;
static PyObject *__pyx_n_s_ssa_id;
static PyObject *__pyx_n_s_ssa_id1;
static PyObject *__pyx_n_s_ssa_id12;
static PyObject *__pyx_n_s_ssa_id2;
static PyObject *__pyx_n_s_ssa_ids;
static PyObject *__pyx_n_s_ssa_path;
static PyObject *__pyx_n_s_ssa_to_linear;
static PyObject *__pyx_kp_u_ssa_to_linear_line_52;
static PyObject *__pyx_n_s_ssa_to_linear_locals_genexpr;
static PyObject *__pyx_n_s_subgraph_contractions;
static PyObject *__pyx_n_s_subgraph_contractions_size;
static PyObject *__pyx_n_s_subgraphs;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_symbol2int;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tree;
static PyObject *__pyx_n_s_tree_to_sequence;
static PyObject *__pyx_kp_u_tree_to_sequence_line_658;
static PyObject *__pyx_n_s_tree_to_sequence_locals_genexpr;
static PyObject *__pyx_n_s_two;
static PyObject *__pyx_n_s_union;
static PyObject *__pyx_n_s_unused_inputs;
static PyObject *__pyx_n_s_update_ref_counts;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_xn;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer___call__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_inputs, CYTHON_UNUSED PyObject *__pyx_v_output, CYTHON_UNUSED PyObject *__pyx_v_size_dict, CYTHON_UNUSED PyObject *__pyx_v_memory_limit); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_13ssa_to_linear_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_ssa_to_linear(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ssa_path); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_13linear_to_ssa_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_2linear_to_ssa(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_4calc_k12_flops(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_remaining, int __pyx_v_i, int __pyx_v_j, PyObject *__pyx_v_size_dict); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_6_compute_oversize_flops(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_remaining, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_7optimal__optimal_iterate(PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_remaining, PyObject *__pyx_v_inputs, PyObject *__pyx_v_flops); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_8optimal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_10better_flops_first(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flops, PyObject *__pyx_v_size, PyObject *__pyx_v_best_flops, PyObject *__pyx_v_best_size); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_12better_size_first(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flops, PyObject *__pyx_v_size, PyObject *__pyx_v_best_flops, PyObject *__pyx_v_best_size); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_14get_better_fn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_16cost_memory_removed(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_size12, PyObject *__pyx_v_size1, PyObject *__pyx_v_size2, CYTHON_UNUSED PyObject *__pyx_v_k12, CYTHON_UNUSED PyObject *__pyx_v_k1, CYTHON_UNUSED PyObject *__pyx_v_k2); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_18cost_memory_removed_jitter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_size12, PyObject *__pyx_v_size1, PyObject *__pyx_v_size2, CYTHON_UNUSED PyObject *__pyx_v_k12, CYTHON_UNUSED PyObject *__pyx_v_k1, CYTHON_UNUSED PyObject *__pyx_v_k2); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda2(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nbranch, PyObject *__pyx_v_cutoff_flops_factor, PyObject *__pyx_v_minimize, PyObject *__pyx_v_cost_fn); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_2path(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate__assess_candidate(PyObject *__pyx_self, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2, PyObject *__pyx_v_i, PyObject *__pyx_v_j); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call____branch_iterate(PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_inputs, PyObject *__pyx_v_remaining, PyObject *__pyx_v_flops, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_4__call__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_20branch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit, PyObject *__pyx_v_optimizer_kwargs); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_22_get_candidate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_output, PyObject *__pyx_v_sizes, PyObject *__pyx_v_remaining, PyObject *__pyx_v_footprints, PyObject *__pyx_v_dim_ref_counts, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2, PyObject *__pyx_v_cost_fn); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_15_push_candidate_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_24_push_candidate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_output, PyObject *__pyx_v_sizes, PyObject *__pyx_v_remaining, PyObject *__pyx_v_footprints, PyObject *__pyx_v_dim_ref_counts, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2s, PyObject *__pyx_v_queue, PyObject *__pyx_v_push_all, PyObject *__pyx_v_cost_fn); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_26_update_ref_counts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dim_to_keys, PyObject *__pyx_v_dim_ref_counts, PyObject *__pyx_v_dims); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_28_simple_chooser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_queue, PyObject *__pyx_v_remaining); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_8genexpr4_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_30ssa_greedy_optimize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_sizes, PyObject *__pyx_v_choose_fn, PyObject *__pyx_v_cost_fn); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_32greedy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit, PyObject *__pyx_v_choose_fn, PyObject *__pyx_v_cost_fn); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_17_tree_to_sequence_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_34_tree_to_sequence(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_36_find_disconnected_subgraphs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_38_bitmapset_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___7genexpr_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___6genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda11(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___10genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___13genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___16genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda15(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer___call__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_41dynamic_programming(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_43auto(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_45register_path_fn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_fn); /* proto */
static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_47get_path_fn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path_type); /* proto */
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, &__pyx_n_s_items, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PySet_Type_difference_update = {0, &__pyx_n_s_difference_update, 0, 0, 0};
static PyObject *__pyx_float_1_0;
static PyObject *__pyx_float_0_01;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_codeobj__2;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__8;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__69;
static PyObject *__pyx_codeobj__71;
static PyObject *__pyx_codeobj__73;
static PyObject *__pyx_codeobj__75;
/* Late includes */

/* "acqdp/utility/opt_einsum_paths.pyx":48
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer_1__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer_1__call__ = {"__call__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer_1__call__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer_1__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_inputs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_output = 0;
  CYTHON_UNUSED PyObject *__pyx_v_size_dict = 0;
  CYTHON_UNUSED PyObject *__pyx_v_memory_limit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[4] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 1); __PYX_ERR(0, 48, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 2); __PYX_ERR(0, 48, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 3); __PYX_ERR(0, 48, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(0, 48, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_inputs = values[1];
    __pyx_v_output = values[2];
    __pyx_v_size_dict = values[3];
    __pyx_v_memory_limit = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 48, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.PathOptimizer.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer___call__(__pyx_self, __pyx_v_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer___call__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_inputs, CYTHON_UNUSED PyObject *__pyx_v_output, CYTHON_UNUSED PyObject *__pyx_v_size_dict, CYTHON_UNUSED PyObject *__pyx_v_memory_limit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":49
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 49, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":48
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.PathOptimizer.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":52
 * 
 * 
 * def ssa_to_linear(ssa_path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with static single assignment ids to a path with recycled
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_1ssa_to_linear(PyObject *__pyx_self, PyObject *__pyx_v_ssa_path); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_ssa_to_linear[] = "\n    Convert a path with static single assignment ids to a path with recycled\n    linear ids. For example::\n\n        >>> ssa_to_linear([(0, 3), (2, 4), (1, 5)])\n        [(0, 3), (1, 2), (0, 1)]\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_1ssa_to_linear = {"ssa_to_linear", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_1ssa_to_linear, METH_O, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_ssa_to_linear};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_1ssa_to_linear(PyObject *__pyx_self, PyObject *__pyx_v_ssa_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ssa_to_linear (wrapper)", 0);
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_ssa_to_linear(__pyx_self, ((PyObject *)__pyx_v_ssa_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_13ssa_to_linear_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":63
 *     path = []
 *     for ssa_ids in ssa_path:
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))             # <<<<<<<<<<<<<<
 *         for ssa_id in ssa_ids:
 *             ids[ssa_id:] -= 1
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_13ssa_to_linear_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 63, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_13ssa_to_linear_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ssa_to_linear_locals_genexpr, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.ssa_to_linear.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_13ssa_to_linear_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 63, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ssa_ids)) { __Pyx_RaiseClosureNameError("ssa_ids"); __PYX_ERR(0, 63, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ssa_ids)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ssa_ids)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_ssa_ids; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ssa_ids); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 63, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 63, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 63, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ssa_id);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ssa_id, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ids)) { __Pyx_RaiseClosureNameError("ids"); __PYX_ERR(0, 63, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ids, __pyx_cur_scope->__pyx_v_ssa_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 63, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":52
 * 
 * 
 * def ssa_to_linear(ssa_path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with static single assignment ids to a path with recycled
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_ssa_to_linear(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ssa_path) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *__pyx_cur_scope;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_ssa_id = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ssa_to_linear", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 52, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":60
 *         [(0, 3), (1, 2), (0, 1)]
 *     """
 *     ids = np.arange(1 + max(map(max, ssa_path)), dtype=np.int32)             # <<<<<<<<<<<<<<
 *     path = []
 *     for ssa_ids in ssa_path:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_builtin_max);
  __Pyx_GIVEREF(__pyx_builtin_max);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_builtin_max);
  __Pyx_INCREF(__pyx_v_ssa_path);
  __Pyx_GIVEREF(__pyx_v_ssa_path);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_ssa_path);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_AddCObj(__pyx_int_1, __pyx_t_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int32); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_cur_scope->__pyx_v_ids = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":61
 *     """
 *     ids = np.arange(1 + max(map(max, ssa_path)), dtype=np.int32)
 *     path = []             # <<<<<<<<<<<<<<
 *     for ssa_ids in ssa_path:
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_path = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":62
 *     ids = np.arange(1 + max(map(max, ssa_path)), dtype=np.int32)
 *     path = []
 *     for ssa_ids in ssa_path:             # <<<<<<<<<<<<<<
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))
 *         for ssa_id in ssa_ids:
 */
  if (likely(PyList_CheckExact(__pyx_v_ssa_path)) || PyTuple_CheckExact(__pyx_v_ssa_path)) {
    __pyx_t_5 = __pyx_v_ssa_path; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_ssa_path); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 62, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 62, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 62, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_5);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 62, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ssa_ids);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ssa_ids, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":63
 *     path = []
 *     for ssa_ids in ssa_path:
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))             # <<<<<<<<<<<<<<
 *         for ssa_id in ssa_ids:
 *             ids[ssa_id:] -= 1
 */
    __pyx_t_3 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_13ssa_to_linear_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_path, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":64
 *     for ssa_ids in ssa_path:
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))
 *         for ssa_id in ssa_ids:             # <<<<<<<<<<<<<<
 *             ids[ssa_id:] -= 1
 *     return path
 */
    if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_ssa_ids)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_ssa_ids)) {
      __pyx_t_1 = __pyx_cur_scope->__pyx_v_ssa_ids; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_ssa_ids); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 64, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 64, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 64, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_10(__pyx_t_1);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 64, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_ssa_id, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":65
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))
 *         for ssa_id in ssa_ids:
 *             ids[ssa_id:] -= 1             # <<<<<<<<<<<<<<
 *     return path
 * 
 */
      __pyx_t_3 = PySlice_New(__pyx_v_ssa_id, Py_None, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 65, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_ids, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(PyObject_SetItem(__pyx_cur_scope->__pyx_v_ids, __pyx_t_3, __pyx_t_4) < 0)) __PYX_ERR(0, 65, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":64
 *     for ssa_ids in ssa_path:
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))
 *         for ssa_id in ssa_ids:             # <<<<<<<<<<<<<<
 *             ids[ssa_id:] -= 1
 *     return path
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":62
 *     ids = np.arange(1 + max(map(max, ssa_path)), dtype=np.int32)
 *     path = []
 *     for ssa_ids in ssa_path:             # <<<<<<<<<<<<<<
 *         path.append(tuple(int(ids[ssa_id]) for ssa_id in ssa_ids))
 *         for ssa_id in ssa_ids:
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":66
 *         for ssa_id in ssa_ids:
 *             ids[ssa_id:] -= 1
 *     return path             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_path);
  __pyx_r = __pyx_v_path;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":52
 * 
 * 
 * def ssa_to_linear(ssa_path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with static single assignment ids to a path with recycled
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.ssa_to_linear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_ssa_id);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":69
 * 
 * 
 * def linear_to_ssa(path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with recycled linear ids to a path with static single
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_3linear_to_ssa(PyObject *__pyx_self, PyObject *__pyx_v_path); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_2linear_to_ssa[] = "\n    Convert a path with recycled linear ids to a path with static single\n    assignment ids. For example::\n\n        >>> linear_to_ssa([(0, 3), (1, 2), (0, 1)])\n        [(0, 3), (2, 4), (1, 5)]\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_3linear_to_ssa = {"linear_to_ssa", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_3linear_to_ssa, METH_O, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_2linear_to_ssa};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_3linear_to_ssa(PyObject *__pyx_self, PyObject *__pyx_v_path) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("linear_to_ssa (wrapper)", 0);
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_2linear_to_ssa(__pyx_self, ((PyObject *)__pyx_v_path));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_13linear_to_ssa_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":82
 *     ssa_path = []
 *     for ids in path:
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))             # <<<<<<<<<<<<<<
 *         for id_ in sorted(ids, reverse=True):
 *             del linear_to_ssa[id_]
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_13linear_to_ssa_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 82, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_13linear_to_ssa_2generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_linear_to_ssa_locals_genexpr, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.linear_to_ssa.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_13linear_to_ssa_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 82, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ids)) { __Pyx_RaiseClosureNameError("ids"); __PYX_ERR(0, 82, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ids)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ids)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_ids; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ids); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 82, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 82, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 82, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_id_);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_id_, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_linear_to_ssa)) { __Pyx_RaiseClosureNameError("linear_to_ssa"); __PYX_ERR(0, 82, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_linear_to_ssa == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 82, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_linear_to_ssa, __pyx_cur_scope->__pyx_v_id_); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 82, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":69
 * 
 * 
 * def linear_to_ssa(path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with recycled linear ids to a path with static single
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_2linear_to_ssa(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *__pyx_cur_scope;
  PyObject *__pyx_v_num_inputs = NULL;
  PyObject *__pyx_v_new_ids = NULL;
  PyObject *__pyx_v_ssa_path = NULL;
  PyObject *__pyx_v_id_ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("linear_to_ssa", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 69, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":77
 *         [(0, 3), (2, 4), (1, 5)]
 *     """
 *     num_inputs = sum(map(len, path)) - len(path) + 1             # <<<<<<<<<<<<<<
 *     linear_to_ssa = list(range(num_inputs))
 *     new_ids = itertools.count(num_inputs)
 */
  __pyx_t_1 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_path);
  __Pyx_GIVEREF(__pyx_v_path);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_path);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = PyObject_Length(__pyx_v_path); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Subtract(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_num_inputs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":78
 *     """
 *     num_inputs = sum(map(len, path)) - len(path) + 1
 *     linear_to_ssa = list(range(num_inputs))             # <<<<<<<<<<<<<<
 *     new_ids = itertools.count(num_inputs)
 *     ssa_path = []
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_num_inputs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_linear_to_ssa = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":79
 *     num_inputs = sum(map(len, path)) - len(path) + 1
 *     linear_to_ssa = list(range(num_inputs))
 *     new_ids = itertools.count(num_inputs)             # <<<<<<<<<<<<<<
 *     ssa_path = []
 *     for ids in path:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_itertools); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_num_inputs) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_num_inputs);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_new_ids = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":80
 *     linear_to_ssa = list(range(num_inputs))
 *     new_ids = itertools.count(num_inputs)
 *     ssa_path = []             # <<<<<<<<<<<<<<
 *     for ids in path:
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_ssa_path = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":81
 *     new_ids = itertools.count(num_inputs)
 *     ssa_path = []
 *     for ids in path:             # <<<<<<<<<<<<<<
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))
 *         for id_ in sorted(ids, reverse=True):
 */
  if (likely(PyList_CheckExact(__pyx_v_path)) || PyTuple_CheckExact(__pyx_v_path)) {
    __pyx_t_4 = __pyx_v_path; __Pyx_INCREF(__pyx_t_4); __pyx_t_3 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 81, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 81, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_4);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 81, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ids);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ids, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":82
 *     ssa_path = []
 *     for ids in path:
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))             # <<<<<<<<<<<<<<
 *         for id_ in sorted(ids, reverse=True):
 *             del linear_to_ssa[id_]
 */
    __pyx_t_2 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_13linear_to_ssa_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_ssa_path, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":83
 *     for ids in path:
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))
 *         for id_ in sorted(ids, reverse=True):             # <<<<<<<<<<<<<<
 *             del linear_to_ssa[id_]
 *         linear_to_ssa.append(next(new_ids))
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ids);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ids);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_v_ids);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_reverse, Py_True) < 0) __PYX_ERR(0, 83, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_2 = __pyx_t_7; __Pyx_INCREF(__pyx_t_2); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 83, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 83, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 83, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_7); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 83, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 83, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_9(__pyx_t_2);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 83, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_id_, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":84
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))
 *         for id_ in sorted(ids, reverse=True):
 *             del linear_to_ssa[id_]             # <<<<<<<<<<<<<<
 *         linear_to_ssa.append(next(new_ids))
 *     return ssa_path
 */
      if (unlikely(PyObject_DelItem(__pyx_cur_scope->__pyx_v_linear_to_ssa, __pyx_v_id_) < 0)) __PYX_ERR(0, 84, __pyx_L1_error)

      /* "acqdp/utility/opt_einsum_paths.pyx":83
 *     for ids in path:
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))
 *         for id_ in sorted(ids, reverse=True):             # <<<<<<<<<<<<<<
 *             del linear_to_ssa[id_]
 *         linear_to_ssa.append(next(new_ids))
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":85
 *         for id_ in sorted(ids, reverse=True):
 *             del linear_to_ssa[id_]
 *         linear_to_ssa.append(next(new_ids))             # <<<<<<<<<<<<<<
 *     return ssa_path
 * 
 */
    __pyx_t_2 = __Pyx_PyIter_Next(__pyx_v_new_ids); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_linear_to_ssa, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":81
 *     new_ids = itertools.count(num_inputs)
 *     ssa_path = []
 *     for ids in path:             # <<<<<<<<<<<<<<
 *         ssa_path.append(tuple(linear_to_ssa[id_] for id_ in ids))
 *         for id_ in sorted(ids, reverse=True):
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":86
 *             del linear_to_ssa[id_]
 *         linear_to_ssa.append(next(new_ids))
 *     return ssa_path             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ssa_path);
  __pyx_r = __pyx_v_ssa_path;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":69
 * 
 * 
 * def linear_to_ssa(path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with recycled linear ids to a path with static single
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.linear_to_ssa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_num_inputs);
  __Pyx_XDECREF(__pyx_v_new_ids);
  __Pyx_XDECREF(__pyx_v_ssa_path);
  __Pyx_XDECREF(__pyx_v_id_);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":89
 * 
 * 
 * def calc_k12_flops(inputs, output, remaining, int i, int j, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Calculate the resulting indices and flops for a potential pairwise
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_5calc_k12_flops(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_4calc_k12_flops[] = "\n    Calculate the resulting indices and flops for a potential pairwise\n    contraction - used in the recursive (optimal/branch) algorithms.\n\n    Parameters\n    ----------\n    inputs : tuple[frozenset[str]]\n        The indices of each tensor in this contraction, note this includes\n        tensors unavaiable to contract as static single assignment is used ->\n        contracted tensors are not removed from the list.\n    output : frozenset[str]\n        The set of output indices for the whole contraction.\n    remaining : frozenset[int]\n        The set of indices (corresponding to ``inputs``) of tensors still\n        available to contract.\n    i : int\n        Index of potential tensor to contract.\n    j : int\n        Index of potential tensor to contract.\n    size_dict dict[str, int]\n        Size mapping of all the indices.\n\n    Returns\n    -------\n    k12 : frozenset\n        The resulting indices of the potential tensor.\n    cost : int\n        Estimated flop count of operation.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_5calc_k12_flops = {"calc_k12_flops", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_5calc_k12_flops, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_4calc_k12_flops};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_5calc_k12_flops(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_remaining = 0;
  int __pyx_v_i;
  int __pyx_v_j;
  PyObject *__pyx_v_size_dict = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calc_k12_flops (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_remaining,&__pyx_n_s_i,&__pyx_n_s_j,&__pyx_n_s_size_dict,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_k12_flops", 1, 6, 6, 1); __PYX_ERR(0, 89, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_remaining)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_k12_flops", 1, 6, 6, 2); __PYX_ERR(0, 89, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_k12_flops", 1, 6, 6, 3); __PYX_ERR(0, 89, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_j)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_k12_flops", 1, 6, 6, 4); __PYX_ERR(0, 89, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("calc_k12_flops", 1, 6, 6, 5); __PYX_ERR(0, 89, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "calc_k12_flops") < 0)) __PYX_ERR(0, 89, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
    __pyx_v_remaining = values[2];
    __pyx_v_i = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L3_error)
    __pyx_v_j = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_j == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L3_error)
    __pyx_v_size_dict = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("calc_k12_flops", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 89, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.calc_k12_flops", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_4calc_k12_flops(__pyx_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_remaining, __pyx_v_i, __pyx_v_j, __pyx_v_size_dict);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_4calc_k12_flops(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_remaining, int __pyx_v_i, int __pyx_v_j, PyObject *__pyx_v_size_dict) {
  PyObject *__pyx_v_k1 = NULL;
  PyObject *__pyx_v_k2 = NULL;
  PyObject *__pyx_v_either = NULL;
  PyObject *__pyx_v_shared = NULL;
  PyObject *__pyx_v_keep = NULL;
  PyObject *__pyx_v_k12 = NULL;
  PyObject *__pyx_v_cost = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calc_k12_flops", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":119
 *         Estimated flop count of operation.
 *     """
 *     k1, k2 = inputs[i], inputs[j]             # <<<<<<<<<<<<<<
 *     either = k1 | k2
 *     shared = k1 & k2
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_inputs, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_inputs, __pyx_v_j, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_k1 = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_k2 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":120
 *     """
 *     k1, k2 = inputs[i], inputs[j]
 *     either = k1 | k2             # <<<<<<<<<<<<<<
 *     shared = k1 & k2
 *     keep = frozenset.union(output, *map(inputs.__getitem__, remaining - {i, j}))
 */
  __pyx_t_2 = PyNumber_Or(__pyx_v_k1, __pyx_v_k2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_either = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":121
 *     k1, k2 = inputs[i], inputs[j]
 *     either = k1 | k2
 *     shared = k1 & k2             # <<<<<<<<<<<<<<
 *     keep = frozenset.union(output, *map(inputs.__getitem__, remaining - {i, j}))
 * 
 */
  __pyx_t_2 = PyNumber_And(__pyx_v_k1, __pyx_v_k2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_shared = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":122
 *     either = k1 | k2
 *     shared = k1 & k2
 *     keep = frozenset.union(output, *map(inputs.__getitem__, remaining - {i, j}))             # <<<<<<<<<<<<<<
 * 
 *     k12 = either & keep
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyFrozenSet_Type)), __pyx_n_s_union); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_output);
  __Pyx_GIVEREF(__pyx_v_output);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_output);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_inputs, __pyx_n_s_getitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PySet_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PySet_Add(__pyx_t_6, __pyx_t_4) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PySet_Add(__pyx_t_6, __pyx_t_5) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyNumber_Subtract(__pyx_v_remaining, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyNumber_Add(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_keep = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":124
 *     keep = frozenset.union(output, *map(inputs.__getitem__, remaining - {i, j}))
 * 
 *     k12 = either & keep             # <<<<<<<<<<<<<<
 *     cost = helpers.flop_count(either, shared - keep, 2, size_dict)
 * 
 */
  __pyx_t_6 = PyNumber_And(__pyx_v_either, __pyx_v_keep); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_k12 = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":125
 * 
 *     k12 = either & keep
 *     cost = helpers.flop_count(either, shared - keep, 2, size_dict)             # <<<<<<<<<<<<<<
 * 
 *     return k12, cost
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_helpers); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_flop_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyNumber_Subtract(__pyx_v_shared, __pyx_v_keep); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_either, __pyx_t_5, __pyx_int_2, __pyx_v_size_dict};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_1, __pyx_v_either, __pyx_t_5, __pyx_int_2, __pyx_v_size_dict};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 4+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(4+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_either);
    __Pyx_GIVEREF(__pyx_v_either);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_v_either);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_t_5);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_7, __pyx_int_2);
    __Pyx_INCREF(__pyx_v_size_dict);
    __Pyx_GIVEREF(__pyx_v_size_dict);
    PyTuple_SET_ITEM(__pyx_t_3, 3+__pyx_t_7, __pyx_v_size_dict);
    __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cost = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":127
 *     cost = helpers.flop_count(either, shared - keep, 2, size_dict)
 * 
 *     return k12, cost             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_k12);
  __Pyx_GIVEREF(__pyx_v_k12);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_k12);
  __Pyx_INCREF(__pyx_v_cost);
  __Pyx_GIVEREF(__pyx_v_cost);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_cost);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":89
 * 
 * 
 * def calc_k12_flops(inputs, output, remaining, int i, int j, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Calculate the resulting indices and flops for a potential pairwise
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.calc_k12_flops", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k1);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XDECREF(__pyx_v_either);
  __Pyx_XDECREF(__pyx_v_shared);
  __Pyx_XDECREF(__pyx_v_keep);
  __Pyx_XDECREF(__pyx_v_k12);
  __Pyx_XDECREF(__pyx_v_cost);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":130
 * 
 * 
 * def _compute_oversize_flops(inputs, remaining, output, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Compute the flop count for a contraction of all remaining arguments. This
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_7_compute_oversize_flops(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_6_compute_oversize_flops[] = "\n    Compute the flop count for a contraction of all remaining arguments. This\n    is used when a memory limit means that no pairwise contractions can be made.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_7_compute_oversize_flops = {"_compute_oversize_flops", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_7_compute_oversize_flops, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_6_compute_oversize_flops};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_7_compute_oversize_flops(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_remaining = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_compute_oversize_flops (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_remaining,&__pyx_n_s_output,&__pyx_n_s_size_dict,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_remaining)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_compute_oversize_flops", 1, 4, 4, 1); __PYX_ERR(0, 130, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_compute_oversize_flops", 1, 4, 4, 2); __PYX_ERR(0, 130, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_compute_oversize_flops", 1, 4, 4, 3); __PYX_ERR(0, 130, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_compute_oversize_flops") < 0)) __PYX_ERR(0, 130, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_inputs = values[0];
    __pyx_v_remaining = values[1];
    __pyx_v_output = values[2];
    __pyx_v_size_dict = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_compute_oversize_flops", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 130, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._compute_oversize_flops", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_6_compute_oversize_flops(__pyx_self, __pyx_v_inputs, __pyx_v_remaining, __pyx_v_output, __pyx_v_size_dict);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_6_compute_oversize_flops(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_remaining, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict) {
  PyObject *__pyx_v_idx_contraction = NULL;
  PyObject *__pyx_v_inner = NULL;
  Py_ssize_t __pyx_v_num_terms;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compute_oversize_flops", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":135
 *     is used when a memory limit means that no pairwise contractions can be made.
 *     """
 *     idx_contraction = frozenset.union(*map(inputs.__getitem__, remaining))             # <<<<<<<<<<<<<<
 *     inner = idx_contraction - output
 *     num_terms = len(remaining)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyFrozenSet_Type)), __pyx_n_s_union); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_inputs, __pyx_n_s_getitem); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_remaining);
  __Pyx_GIVEREF(__pyx_v_remaining);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_remaining);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_idx_contraction = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":136
 *     """
 *     idx_contraction = frozenset.union(*map(inputs.__getitem__, remaining))
 *     inner = idx_contraction - output             # <<<<<<<<<<<<<<
 *     num_terms = len(remaining)
 *     return helpers.flop_count(idx_contraction, inner, num_terms, size_dict)
 */
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_idx_contraction, __pyx_v_output); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_inner = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":137
 *     idx_contraction = frozenset.union(*map(inputs.__getitem__, remaining))
 *     inner = idx_contraction - output
 *     num_terms = len(remaining)             # <<<<<<<<<<<<<<
 *     return helpers.flop_count(idx_contraction, inner, num_terms, size_dict)
 * 
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_remaining); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 137, __pyx_L1_error)
  __pyx_v_num_terms = __pyx_t_4;

  /* "acqdp/utility/opt_einsum_paths.pyx":138
 *     inner = idx_contraction - output
 *     num_terms = len(remaining)
 *     return helpers.flop_count(idx_contraction, inner, num_terms, size_dict)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_helpers); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_flop_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_num_terms); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_idx_contraction, __pyx_v_inner, __pyx_t_3, __pyx_v_size_dict};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_idx_contraction, __pyx_v_inner, __pyx_t_3, __pyx_v_size_dict};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_idx_contraction);
    __Pyx_GIVEREF(__pyx_v_idx_contraction);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_idx_contraction);
    __Pyx_INCREF(__pyx_v_inner);
    __Pyx_GIVEREF(__pyx_v_inner);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_inner);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_size_dict);
    __Pyx_GIVEREF(__pyx_v_size_dict);
    PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_size_dict);
    __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":130
 * 
 * 
 * def _compute_oversize_flops(inputs, remaining, output, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Compute the flop count for a contraction of all remaining arguments. This
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._compute_oversize_flops", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx_contraction);
  __Pyx_XDECREF(__pyx_v_inner);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":141
 * 
 * 
 * def optimal(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """
 *     Computes all possible pair contractions in a depth-first recursive manner,
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_9optimal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_8optimal[] = "\n    Computes all possible pair contractions in a depth-first recursive manner,\n    sieving results based on ``memory_limit`` and the best path found so far.\n    Returns the lowest cost path. This algorithm scales factoriallly with\n    respect to the elements in the list ``input_sets``.\n\n    Parameters\n    ----------\n    inputs : list\n        List of sets that represent the lhs side of the einsum subscript.\n    output : set\n        Set that represents the rhs side of the overall einsum subscript.\n    size_dict : dictionary\n        Dictionary of index sizes.\n    memory_limit : int\n        The maximum number of elements in a temporary array.\n\n    Returns\n    -------\n    path : list\n        The optimal contraction order within the memory limit constraint.\n\n    Examples\n    --------\n    >>> isets = [set('abd'), set('ac'), set('bdc')]\n    >>> oset = set('')\n    >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}\n    >>> optimal(isets, oset, idx_sizes, 5000)\n    [(0, 2), (0, 1)]\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_9optimal = {"optimal", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_9optimal, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_8optimal};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_9optimal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  PyObject *__pyx_v_memory_limit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("optimal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("optimal", 0, 3, 4, 1); __PYX_ERR(0, 141, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("optimal", 0, 3, 4, 2); __PYX_ERR(0, 141, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "optimal") < 0)) __PYX_ERR(0, 141, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
    __pyx_v_size_dict = values[2];
    __pyx_v_memory_limit = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("optimal", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 141, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.optimal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_8optimal(__pyx_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":179
 *     result_cache = {}
 * 
 *     def _optimal_iterate(path, remaining, inputs, flops):             # <<<<<<<<<<<<<<
 * 
 *         # reached end of path (only ever get here if flops is best found so far)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_7optimal_1_optimal_iterate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_7optimal_1_optimal_iterate = {"_optimal_iterate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_7optimal_1_optimal_iterate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_7optimal_1_optimal_iterate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_remaining = 0;
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_flops = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_optimal_iterate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_remaining,&__pyx_n_s_inputs,&__pyx_n_s_flops,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_remaining)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_optimal_iterate", 1, 4, 4, 1); __PYX_ERR(0, 179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_optimal_iterate", 1, 4, 4, 2); __PYX_ERR(0, 179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flops)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_optimal_iterate", 1, 4, 4, 3); __PYX_ERR(0, 179, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_optimal_iterate") < 0)) __PYX_ERR(0, 179, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_path = values[0];
    __pyx_v_remaining = values[1];
    __pyx_v_inputs = values[2];
    __pyx_v_flops = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_optimal_iterate", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 179, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.optimal._optimal_iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_7optimal__optimal_iterate(__pyx_self, __pyx_v_path, __pyx_v_remaining, __pyx_v_inputs, __pyx_v_flops);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_7optimal__optimal_iterate(PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_remaining, PyObject *__pyx_v_inputs, PyObject *__pyx_v_flops) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *__pyx_cur_scope;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *__pyx_outer_scope;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_k12 = NULL;
  PyObject *__pyx_v_flops12 = NULL;
  PyObject *__pyx_v_new_flops = NULL;
  PyObject *__pyx_v_size12 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_optimal_iterate", 0);
  __pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "acqdp/utility/opt_einsum_paths.pyx":182
 * 
 *         # reached end of path (only ever get here if flops is best found so far)
 *         if len(remaining) == 1:             # <<<<<<<<<<<<<<
 *             best['flops'] = flops
 *             best['ssa_path'] = path
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_remaining); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 182, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 == 1) != 0);
  if (__pyx_t_2) {

    /* "acqdp/utility/opt_einsum_paths.pyx":183
 *         # reached end of path (only ever get here if flops is best found so far)
 *         if len(remaining) == 1:
 *             best['flops'] = flops             # <<<<<<<<<<<<<<
 *             best['ssa_path'] = path
 *             return
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_best)) { __Pyx_RaiseClosureNameError("best"); __PYX_ERR(0, 183, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_best == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 183, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_best, __pyx_n_s_flops, __pyx_v_flops) < 0)) __PYX_ERR(0, 183, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":184
 *         if len(remaining) == 1:
 *             best['flops'] = flops
 *             best['ssa_path'] = path             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_best)) { __Pyx_RaiseClosureNameError("best"); __PYX_ERR(0, 184, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_best == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 184, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_best, __pyx_n_s_ssa_path, __pyx_v_path) < 0)) __PYX_ERR(0, 184, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":185
 *             best['flops'] = flops
 *             best['ssa_path'] = path
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # check all possible remaining paths
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":182
 * 
 *         # reached end of path (only ever get here if flops is best found so far)
 *         if len(remaining) == 1:             # <<<<<<<<<<<<<<
 *             best['flops'] = flops
 *             best['ssa_path'] = path
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":188
 * 
 *         # check all possible remaining paths
 *         for i, j in itertools.combinations(remaining, 2):             # <<<<<<<<<<<<<<
 *             if i > j:
 *                 i, j = j, i
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_itertools); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_combinations); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_remaining, __pyx_int_2};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_remaining, __pyx_int_2};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_remaining);
    __Pyx_GIVEREF(__pyx_v_remaining);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_remaining);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_2);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5); __pyx_t_1 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 188, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 188, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 188, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_5);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 188, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 188, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 188, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 188, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":189
 *         # check all possible remaining paths
 *         for i, j in itertools.combinations(remaining, 2):
 *             if i > j:             # <<<<<<<<<<<<<<
 *                 i, j = j, i
 *             key = (inputs[i], inputs[j])
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_i, __pyx_v_j, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 189, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "acqdp/utility/opt_einsum_paths.pyx":190
 *         for i, j in itertools.combinations(remaining, 2):
 *             if i > j:
 *                 i, j = j, i             # <<<<<<<<<<<<<<
 *             key = (inputs[i], inputs[j])
 *             try:
 */
      __pyx_t_11 = __pyx_v_j;
      __pyx_t_12 = __pyx_v_i;
      __pyx_v_i = __pyx_t_11;
      __pyx_t_11 = 0;
      __pyx_v_j = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":189
 *         # check all possible remaining paths
 *         for i, j in itertools.combinations(remaining, 2):
 *             if i > j:             # <<<<<<<<<<<<<<
 *                 i, j = j, i
 *             key = (inputs[i], inputs[j])
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":191
 *             if i > j:
 *                 i, j = j, i
 *             key = (inputs[i], inputs[j])             # <<<<<<<<<<<<<<
 *             try:
 *                 k12, flops12 = result_cache[key]
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_inputs, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_inputs, __pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key, ((PyObject*)__pyx_t_7));
    __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":192
 *                 i, j = j, i
 *             key = (inputs[i], inputs[j])
 *             try:             # <<<<<<<<<<<<<<
 *                 k12, flops12 = result_cache[key]
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_11, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_13);
      /*try:*/ {

        /* "acqdp/utility/opt_einsum_paths.pyx":193
 *             key = (inputs[i], inputs[j])
 *             try:
 *                 k12, flops12 = result_cache[key]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 k12, flops12 = result_cache[key] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_result_cache)) { __Pyx_RaiseClosureNameError("result_cache"); __PYX_ERR(0, 193, __pyx_L9_error) }
        if (unlikely(__pyx_cur_scope->__pyx_v_result_cache == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 193, __pyx_L9_error)
        }
        __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_result_cache, __pyx_v_key); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 193, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_7);
        if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
          PyObject* sequence = __pyx_t_7;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 193, __pyx_L9_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          #else
          __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 193, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_9 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 193, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
          index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          index = 1; __pyx_t_3 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_3)) goto __pyx_L17_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_3);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 193, __pyx_L9_error)
          __pyx_t_10 = NULL;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L18_unpacking_done;
          __pyx_L17_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_10 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 193, __pyx_L9_error)
          __pyx_L18_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_k12, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_flops12, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":192
 *                 i, j = j, i
 *             key = (inputs[i], inputs[j])
 *             try:             # <<<<<<<<<<<<<<
 *                 k12, flops12 = result_cache[key]
 *             except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L16_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":194
 *             try:
 *                 k12, flops12 = result_cache[key]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 k12, flops12 = result_cache[key] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)
 * 
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.optimal._optimal_iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 194, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);

        /* "acqdp/utility/opt_einsum_paths.pyx":195
 *                 k12, flops12 = result_cache[key]
 *             except KeyError:
 *                 k12, flops12 = result_cache[key] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)             # <<<<<<<<<<<<<<
 * 
 *             # sieve based on current best flops
 */
        __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_calc_k12_flops); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 195, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        if (unlikely(!__pyx_cur_scope->__pyx_v_output)) { __Pyx_RaiseClosureNameError("output"); __PYX_ERR(0, 195, __pyx_L11_except_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_size_dict)) { __Pyx_RaiseClosureNameError("size_dict"); __PYX_ERR(0, 195, __pyx_L11_except_error) }
        __pyx_t_15 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_14, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_14)) {
          PyObject *__pyx_temp[7] = {__pyx_t_15, __pyx_v_inputs, __pyx_cur_scope->__pyx_v_output, __pyx_v_remaining, __pyx_v_i, __pyx_v_j, __pyx_cur_scope->__pyx_v_size_dict};
          __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_6, 6+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
          PyObject *__pyx_temp[7] = {__pyx_t_15, __pyx_v_inputs, __pyx_cur_scope->__pyx_v_output, __pyx_v_remaining, __pyx_v_i, __pyx_v_j, __pyx_cur_scope->__pyx_v_size_dict};
          __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_6, 6+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(6+__pyx_t_6); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_15) {
            __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_15); __pyx_t_15 = NULL;
          }
          __Pyx_INCREF(__pyx_v_inputs);
          __Pyx_GIVEREF(__pyx_v_inputs);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_6, __pyx_v_inputs);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_output);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_output);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_6, __pyx_cur_scope->__pyx_v_output);
          __Pyx_INCREF(__pyx_v_remaining);
          __Pyx_GIVEREF(__pyx_v_remaining);
          PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_6, __pyx_v_remaining);
          __Pyx_INCREF(__pyx_v_i);
          __Pyx_GIVEREF(__pyx_v_i);
          PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_6, __pyx_v_i);
          __Pyx_INCREF(__pyx_v_j);
          __Pyx_GIVEREF(__pyx_v_j);
          PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_6, __pyx_v_j);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size_dict);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size_dict);
          PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_6, __pyx_cur_scope->__pyx_v_size_dict);
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_16, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
          PyObject* sequence = __pyx_t_9;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 195, __pyx_L11_except_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_14 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_14 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_16);
          #else
          __pyx_t_14 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          #endif
        } else {
          Py_ssize_t index = -1;
          __pyx_t_15 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_10 = Py_TYPE(__pyx_t_15)->tp_iternext;
          index = 0; __pyx_t_14 = __pyx_t_10(__pyx_t_15); if (unlikely(!__pyx_t_14)) goto __pyx_L21_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_14);
          index = 1; __pyx_t_16 = __pyx_t_10(__pyx_t_15); if (unlikely(!__pyx_t_16)) goto __pyx_L21_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_16);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_15), 2) < 0) __PYX_ERR(0, 195, __pyx_L11_except_error)
          __pyx_t_10 = NULL;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          goto __pyx_L22_unpacking_done;
          __pyx_L21_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_10 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 195, __pyx_L11_except_error)
          __pyx_L22_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_k12, __pyx_t_14);
        __pyx_t_14 = 0;
        __Pyx_XDECREF_SET(__pyx_v_flops12, __pyx_t_16);
        __pyx_t_16 = 0;
        if (unlikely(!__pyx_cur_scope->__pyx_v_result_cache)) { __Pyx_RaiseClosureNameError("result_cache"); __PYX_ERR(0, 195, __pyx_L11_except_error) }
        if (unlikely(__pyx_cur_scope->__pyx_v_result_cache == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 195, __pyx_L11_except_error)
        }
        if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_result_cache, __pyx_v_key, __pyx_t_9) < 0)) __PYX_ERR(0, 195, __pyx_L11_except_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L10_exception_handled;
      }
      goto __pyx_L11_except_error;
      __pyx_L11_except_error:;

      /* "acqdp/utility/opt_einsum_paths.pyx":192
 *                 i, j = j, i
 *             key = (inputs[i], inputs[j])
 *             try:             # <<<<<<<<<<<<<<
 *                 k12, flops12 = result_cache[key]
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_13);
      goto __pyx_L1_error;
      __pyx_L10_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_11, __pyx_t_13);
      __pyx_L16_try_end:;
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":198
 * 
 *             # sieve based on current best flops
 *             new_flops = flops + flops12             # <<<<<<<<<<<<<<
 *             if new_flops >= best['flops']:
 *                 continue
 */
    __pyx_t_4 = PyNumber_Add(__pyx_v_flops, __pyx_v_flops12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_new_flops, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":199
 *             # sieve based on current best flops
 *             new_flops = flops + flops12
 *             if new_flops >= best['flops']:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_best)) { __Pyx_RaiseClosureNameError("best"); __PYX_ERR(0, 199, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_v_best == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 199, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_best, __pyx_n_s_flops); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_new_flops, __pyx_t_4, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "acqdp/utility/opt_einsum_paths.pyx":200
 *             new_flops = flops + flops12
 *             if new_flops >= best['flops']:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             # sieve based on memory limit
 */
      goto __pyx_L4_continue;

      /* "acqdp/utility/opt_einsum_paths.pyx":199
 *             # sieve based on current best flops
 *             new_flops = flops + flops12
 *             if new_flops >= best['flops']:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":203
 * 
 *             # sieve based on memory limit
 *             if memory_limit not in _UNLIMITED_MEM:             # <<<<<<<<<<<<<<
 *                 try:
 *                     size12 = size_cache[k12]
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_memory_limit)) { __Pyx_RaiseClosureNameError("memory_limit"); __PYX_ERR(0, 203, __pyx_L1_error) }
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_UNLIMITED_MEM); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_v_memory_limit, __pyx_t_3, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_17 = (__pyx_t_2 != 0);
    if (__pyx_t_17) {

      /* "acqdp/utility/opt_einsum_paths.pyx":204
 *             # sieve based on memory limit
 *             if memory_limit not in _UNLIMITED_MEM:
 *                 try:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {

          /* "acqdp/utility/opt_einsum_paths.pyx":205
 *             if memory_limit not in _UNLIMITED_MEM:
 *                 try:
 *                     size12 = size_cache[k12]             # <<<<<<<<<<<<<<
 *                 except KeyError:
 *                     size12 = size_cache[k12] = helpers.compute_size_by_dict(k12, size_dict)
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_size_cache)) { __Pyx_RaiseClosureNameError("size_cache"); __PYX_ERR(0, 205, __pyx_L25_error) }
          if (unlikely(__pyx_cur_scope->__pyx_v_size_cache == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 205, __pyx_L25_error)
          }
          __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_size_cache, __pyx_v_k12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L25_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_XDECREF_SET(__pyx_v_size12, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "acqdp/utility/opt_einsum_paths.pyx":204
 *             # sieve based on memory limit
 *             if memory_limit not in _UNLIMITED_MEM:
 *                 try:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 */
        }
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L32_try_end;
        __pyx_L25_error:;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":206
 *                 try:
 *                     size12 = size_cache[k12]
 *                 except KeyError:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12] = helpers.compute_size_by_dict(k12, size_dict)
 * 
 */
        __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
        if (__pyx_t_6) {
          __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.optimal._optimal_iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 206, __pyx_L27_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_7);

          /* "acqdp/utility/opt_einsum_paths.pyx":207
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 *                     size12 = size_cache[k12] = helpers.compute_size_by_dict(k12, size_dict)             # <<<<<<<<<<<<<<
 * 
 *                 # possibly terminate this path with an all-terms einsum
 */
          __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_helpers); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 207, __pyx_L27_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 207, __pyx_L27_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (unlikely(!__pyx_cur_scope->__pyx_v_size_dict)) { __Pyx_RaiseClosureNameError("size_dict"); __PYX_ERR(0, 207, __pyx_L27_except_error) }
          __pyx_t_16 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
              __pyx_t_6 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_14)) {
            PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_v_k12, __pyx_cur_scope->__pyx_v_size_dict};
            __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 207, __pyx_L27_except_error)
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_GOTREF(__pyx_t_9);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
            PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_v_k12, __pyx_cur_scope->__pyx_v_size_dict};
            __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 207, __pyx_L27_except_error)
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_GOTREF(__pyx_t_9);
          } else
          #endif
          {
            __pyx_t_15 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 207, __pyx_L27_except_error)
            __Pyx_GOTREF(__pyx_t_15);
            if (__pyx_t_16) {
              __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_16); __pyx_t_16 = NULL;
            }
            __Pyx_INCREF(__pyx_v_k12);
            __Pyx_GIVEREF(__pyx_v_k12);
            PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_6, __pyx_v_k12);
            __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size_dict);
            __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size_dict);
            PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_6, __pyx_cur_scope->__pyx_v_size_dict);
            __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_15, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 207, __pyx_L27_except_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_XDECREF_SET(__pyx_v_size12, __pyx_t_9);
          if (unlikely(!__pyx_cur_scope->__pyx_v_size_cache)) { __Pyx_RaiseClosureNameError("size_cache"); __PYX_ERR(0, 207, __pyx_L27_except_error) }
          if (unlikely(__pyx_cur_scope->__pyx_v_size_cache == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 207, __pyx_L27_except_error)
          }
          if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_size_cache, __pyx_v_k12, __pyx_t_9) < 0)) __PYX_ERR(0, 207, __pyx_L27_except_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L26_exception_handled;
        }
        goto __pyx_L27_except_error;
        __pyx_L27_except_error:;

        /* "acqdp/utility/opt_einsum_paths.pyx":204
 *             # sieve based on memory limit
 *             if memory_limit not in _UNLIMITED_MEM:
 *                 try:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 */
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_11, __pyx_t_12);
        goto __pyx_L1_error;
        __pyx_L26_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_11, __pyx_t_12);
        __pyx_L32_try_end:;
      }

      /* "acqdp/utility/opt_einsum_paths.pyx":210
 * 
 *                 # possibly terminate this path with an all-terms einsum
 *                 if size12 > memory_limit:             # <<<<<<<<<<<<<<
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < best['flops']:
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_memory_limit)) { __Pyx_RaiseClosureNameError("memory_limit"); __PYX_ERR(0, 210, __pyx_L1_error) }
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_size12, __pyx_cur_scope->__pyx_v_memory_limit, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 210, __pyx_L1_error)
      __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 210, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_17) {

        /* "acqdp/utility/opt_einsum_paths.pyx":211
 *                 # possibly terminate this path with an all-terms einsum
 *                 if size12 > memory_limit:
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)             # <<<<<<<<<<<<<<
 *                     if new_flops < best['flops']:
 *                         best['flops'] = new_flops
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_compute_oversize_flops); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(!__pyx_cur_scope->__pyx_v_output)) { __Pyx_RaiseClosureNameError("output"); __PYX_ERR(0, 211, __pyx_L1_error) }
        if (unlikely(!__pyx_cur_scope->__pyx_v_size_dict)) { __Pyx_RaiseClosureNameError("size_dict"); __PYX_ERR(0, 211, __pyx_L1_error) }
        __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_inputs, __pyx_v_remaining, __pyx_cur_scope->__pyx_v_output, __pyx_cur_scope->__pyx_v_size_dict};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 211, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_inputs, __pyx_v_remaining, __pyx_cur_scope->__pyx_v_output, __pyx_cur_scope->__pyx_v_size_dict};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 211, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        {
          __pyx_t_9 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 211, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(__pyx_v_inputs);
          __Pyx_GIVEREF(__pyx_v_inputs);
          PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_6, __pyx_v_inputs);
          __Pyx_INCREF(__pyx_v_remaining);
          __Pyx_GIVEREF(__pyx_v_remaining);
          PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_6, __pyx_v_remaining);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_output);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_output);
          PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_6, __pyx_cur_scope->__pyx_v_output);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size_dict);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size_dict);
          PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_6, __pyx_cur_scope->__pyx_v_size_dict);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 211, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Add(__pyx_v_flops, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v_new_flops, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":212
 *                 if size12 > memory_limit:
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < best['flops']:             # <<<<<<<<<<<<<<
 *                         best['flops'] = new_flops
 *                         best['ssa_path'] = path + (tuple(remaining),)
 */
        if (unlikely(!__pyx_cur_scope->__pyx_v_best)) { __Pyx_RaiseClosureNameError("best"); __PYX_ERR(0, 212, __pyx_L1_error) }
        if (unlikely(__pyx_cur_scope->__pyx_v_best == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 212, __pyx_L1_error)
        }
        __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_best, __pyx_n_s_flops); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 212, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = PyObject_RichCompare(__pyx_v_new_flops, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 212, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_17 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_17 < 0)) __PYX_ERR(0, 212, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_17) {

          /* "acqdp/utility/opt_einsum_paths.pyx":213
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < best['flops']:
 *                         best['flops'] = new_flops             # <<<<<<<<<<<<<<
 *                         best['ssa_path'] = path + (tuple(remaining),)
 *                     continue
 */
          if (unlikely(!__pyx_cur_scope->__pyx_v_best)) { __Pyx_RaiseClosureNameError("best"); __PYX_ERR(0, 213, __pyx_L1_error) }
          if (unlikely(__pyx_cur_scope->__pyx_v_best == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 213, __pyx_L1_error)
          }
          if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_best, __pyx_n_s_flops, __pyx_v_new_flops) < 0)) __PYX_ERR(0, 213, __pyx_L1_error)

          /* "acqdp/utility/opt_einsum_paths.pyx":214
 *                     if new_flops < best['flops']:
 *                         best['flops'] = new_flops
 *                         best['ssa_path'] = path + (tuple(remaining),)             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
          __pyx_t_7 = __Pyx_PySequence_Tuple(__pyx_v_remaining); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 214, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 214, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_7);
          __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_Add(__pyx_v_path, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 214, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_cur_scope->__pyx_v_best)) { __Pyx_RaiseClosureNameError("best"); __PYX_ERR(0, 214, __pyx_L1_error) }
          if (unlikely(__pyx_cur_scope->__pyx_v_best == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
            __PYX_ERR(0, 214, __pyx_L1_error)
          }
          if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_v_best, __pyx_n_s_ssa_path, __pyx_t_7) < 0)) __PYX_ERR(0, 214, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "acqdp/utility/opt_einsum_paths.pyx":212
 *                 if size12 > memory_limit:
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < best['flops']:             # <<<<<<<<<<<<<<
 *                         best['flops'] = new_flops
 *                         best['ssa_path'] = path + (tuple(remaining),)
 */
        }

        /* "acqdp/utility/opt_einsum_paths.pyx":215
 *                         best['flops'] = new_flops
 *                         best['ssa_path'] = path + (tuple(remaining),)
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *             # add contraction and recurse into all remaining
 */
        goto __pyx_L4_continue;

        /* "acqdp/utility/opt_einsum_paths.pyx":210
 * 
 *                 # possibly terminate this path with an all-terms einsum
 *                 if size12 > memory_limit:             # <<<<<<<<<<<<<<
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < best['flops']:
 */
      }

      /* "acqdp/utility/opt_einsum_paths.pyx":203
 * 
 *             # sieve based on memory limit
 *             if memory_limit not in _UNLIMITED_MEM:             # <<<<<<<<<<<<<<
 *                 try:
 *                     size12 = size_cache[k12]
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":218
 * 
 *             # add contraction and recurse into all remaining
 *             _optimal_iterate(path=path + ((i, j),),             # <<<<<<<<<<<<<<
 *                              inputs=inputs + (k12,),
 *                              remaining=remaining - {i, j} | {len(inputs)},
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v__optimal_iterate)) { __Pyx_RaiseClosureNameError("_optimal_iterate"); __PYX_ERR(0, 218, __pyx_L1_error) }
    __pyx_t_7 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_j);
    __Pyx_GIVEREF(__pyx_v_j);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_j);
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_v_path, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_path, __pyx_t_4) < 0) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":219
 *             # add contraction and recurse into all remaining
 *             _optimal_iterate(path=path + ((i, j),),
 *                              inputs=inputs + (k12,),             # <<<<<<<<<<<<<<
 *                              remaining=remaining - {i, j} | {len(inputs)},
 *                              flops=new_flops)
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_k12);
    __Pyx_GIVEREF(__pyx_v_k12);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_k12);
    __pyx_t_9 = PyNumber_Add(__pyx_v_inputs, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_inputs, __pyx_t_9) < 0) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":220
 *             _optimal_iterate(path=path + ((i, j),),
 *                              inputs=inputs + (k12,),
 *                              remaining=remaining - {i, j} | {len(inputs)},             # <<<<<<<<<<<<<<
 *                              flops=new_flops)
 * 
 */
    __pyx_t_9 = PySet_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PySet_Add(__pyx_t_9, __pyx_v_i) < 0) __PYX_ERR(0, 220, __pyx_L1_error)
    if (PySet_Add(__pyx_t_9, __pyx_v_j) < 0) __PYX_ERR(0, 220, __pyx_L1_error)
    __pyx_t_4 = PyNumber_Subtract(__pyx_v_remaining, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_18 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_18 == ((Py_ssize_t)-1))) __PYX_ERR(0, 220, __pyx_L1_error)
    __pyx_t_9 = PyInt_FromSsize_t(__pyx_t_18); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PySet_Add(__pyx_t_3, __pyx_t_9) < 0) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Or(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_remaining, __pyx_t_9) < 0) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":221
 *                              inputs=inputs + (k12,),
 *                              remaining=remaining - {i, j} | {len(inputs)},
 *                              flops=new_flops)             # <<<<<<<<<<<<<<
 * 
 *     _optimal_iterate(path=(),
 */
    if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_flops, __pyx_v_new_flops) < 0) __PYX_ERR(0, 218, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":218
 * 
 *             # add contraction and recurse into all remaining
 *             _optimal_iterate(path=path + ((i, j),),             # <<<<<<<<<<<<<<
 *                              inputs=inputs + (k12,),
 *                              remaining=remaining - {i, j} | {len(inputs)},
 */
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_v__optimal_iterate, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":188
 * 
 *         # check all possible remaining paths
 *         for i, j in itertools.combinations(remaining, 2):             # <<<<<<<<<<<<<<
 *             if i > j:
 *                 i, j = j, i
 */
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":179
 *     result_cache = {}
 * 
 *     def _optimal_iterate(path, remaining, inputs, flops):             # <<<<<<<<<<<<<<
 * 
 *         # reached end of path (only ever get here if flops is best found so far)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.optimal._optimal_iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_k12);
  __Pyx_XDECREF(__pyx_v_flops12);
  __Pyx_XDECREF(__pyx_v_new_flops);
  __Pyx_XDECREF(__pyx_v_size12);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":141
 * 
 * 
 * def optimal(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """
 *     Computes all possible pair contractions in a depth-first recursive manner,
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_8optimal(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("optimal", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 141, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_output = __pyx_v_output;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_output);
  __pyx_cur_scope->__pyx_v_size_dict = __pyx_v_size_dict;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size_dict);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size_dict);
  __pyx_cur_scope->__pyx_v_memory_limit = __pyx_v_memory_limit;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_memory_limit);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_memory_limit);
  __Pyx_INCREF(__pyx_v_inputs);

  /* "acqdp/utility/opt_einsum_paths.pyx":172
 *     [(0, 2), (0, 1)]
 *     """
 *     inputs = tuple(map(frozenset, inputs))             # <<<<<<<<<<<<<<
 *     output = frozenset(output)
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&PyFrozenSet_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyFrozenSet_Type)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyFrozenSet_Type)));
  __Pyx_INCREF(__pyx_v_inputs);
  __Pyx_GIVEREF(__pyx_v_inputs);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_inputs);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_inputs, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":173
 *     """
 *     inputs = tuple(map(frozenset, inputs))
 *     output = frozenset(output)             # <<<<<<<<<<<<<<
 * 
 *     best = {'flops': float('inf'), 'ssa_path': (tuple(range(len(inputs))),)}
 */
  __pyx_t_1 = __Pyx_PyFrozenSet_New(__pyx_cur_scope->__pyx_v_output); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_output, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":175
 *     output = frozenset(output)
 * 
 *     best = {'flops': float('inf'), 'ssa_path': (tuple(range(len(inputs))),)}             # <<<<<<<<<<<<<<
 *     size_cache = {}
 *     result_cache = {}
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_n_s_inf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_flops, __pyx_t_2) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_ssa_path, __pyx_t_4) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_best = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":176
 * 
 *     best = {'flops': float('inf'), 'ssa_path': (tuple(range(len(inputs))),)}
 *     size_cache = {}             # <<<<<<<<<<<<<<
 *     result_cache = {}
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_size_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":177
 *     best = {'flops': float('inf'), 'ssa_path': (tuple(range(len(inputs))),)}
 *     size_cache = {}
 *     result_cache = {}             # <<<<<<<<<<<<<<
 * 
 *     def _optimal_iterate(path, remaining, inputs, flops):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_result_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":179
 *     result_cache = {}
 * 
 *     def _optimal_iterate(path, remaining, inputs, flops):             # <<<<<<<<<<<<<<
 * 
 *         # reached end of path (only ever get here if flops is best found so far)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_7optimal_1_optimal_iterate, 0, __pyx_n_s_optimal_locals__optimal_iterate, ((PyObject*)__pyx_cur_scope), __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v__optimal_iterate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":223
 *                              flops=new_flops)
 * 
 *     _optimal_iterate(path=(),             # <<<<<<<<<<<<<<
 *                      inputs=inputs,
 *                      remaining=set(range(len(inputs))),
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_path, __pyx_empty_tuple) < 0) __PYX_ERR(0, 223, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":224
 * 
 *     _optimal_iterate(path=(),
 *                      inputs=inputs,             # <<<<<<<<<<<<<<
 *                      remaining=set(range(len(inputs))),
 *                      flops=0)
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_inputs, __pyx_v_inputs) < 0) __PYX_ERR(0, 223, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":225
 *     _optimal_iterate(path=(),
 *                      inputs=inputs,
 *                      remaining=set(range(len(inputs))),             # <<<<<<<<<<<<<<
 *                      flops=0)
 * 
 */
  __pyx_t_3 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 225, __pyx_L1_error)
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PySet_New(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_remaining, __pyx_t_4) < 0) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_flops, __pyx_int_0) < 0) __PYX_ERR(0, 223, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":223
 *                              flops=new_flops)
 * 
 *     _optimal_iterate(path=(),             # <<<<<<<<<<<<<<
 *                      inputs=inputs,
 *                      remaining=set(range(len(inputs))),
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_v__optimal_iterate, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":228
 *                      flops=0)
 * 
 *     return ssa_to_linear(best['ssa_path'])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ssa_to_linear); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_best, __pyx_n_s_ssa_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":141
 * 
 * 
 * def optimal(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """
 *     Computes all possible pair contractions in a depth-first recursive manner,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.optimal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_inputs);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":233
 * # functions for comparing which of two paths is 'better'
 * 
 * def better_flops_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (flops, size) < (best_flops, best_size)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11better_flops_first(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11better_flops_first = {"better_flops_first", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11better_flops_first, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11better_flops_first(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_flops = 0;
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_v_best_flops = 0;
  PyObject *__pyx_v_best_size = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("better_flops_first (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_flops,&__pyx_n_s_size,&__pyx_n_s_best_flops,&__pyx_n_s_best_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flops)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("better_flops_first", 1, 4, 4, 1); __PYX_ERR(0, 233, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_best_flops)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("better_flops_first", 1, 4, 4, 2); __PYX_ERR(0, 233, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_best_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("better_flops_first", 1, 4, 4, 3); __PYX_ERR(0, 233, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "better_flops_first") < 0)) __PYX_ERR(0, 233, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_flops = values[0];
    __pyx_v_size = values[1];
    __pyx_v_best_flops = values[2];
    __pyx_v_best_size = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("better_flops_first", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 233, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.better_flops_first", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_10better_flops_first(__pyx_self, __pyx_v_flops, __pyx_v_size, __pyx_v_best_flops, __pyx_v_best_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_10better_flops_first(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flops, PyObject *__pyx_v_size, PyObject *__pyx_v_best_flops, PyObject *__pyx_v_best_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("better_flops_first", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":234
 * 
 * def better_flops_first(flops, size, best_flops, best_size):
 *     return (flops, size) < (best_flops, best_size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_flops);
  __Pyx_GIVEREF(__pyx_v_flops);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_flops);
  __Pyx_INCREF(__pyx_v_size);
  __Pyx_GIVEREF(__pyx_v_size);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_size);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_best_flops);
  __Pyx_GIVEREF(__pyx_v_best_flops);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_best_flops);
  __Pyx_INCREF(__pyx_v_best_size);
  __Pyx_GIVEREF(__pyx_v_best_size);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_best_size);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":233
 * # functions for comparing which of two paths is 'better'
 * 
 * def better_flops_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (flops, size) < (best_flops, best_size)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.better_flops_first", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":237
 * 
 * 
 * def better_size_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (size, flops) < (best_size, best_flops)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_13better_size_first(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_13better_size_first = {"better_size_first", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_13better_size_first, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_13better_size_first(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_flops = 0;
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_v_best_flops = 0;
  PyObject *__pyx_v_best_size = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("better_size_first (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_flops,&__pyx_n_s_size,&__pyx_n_s_best_flops,&__pyx_n_s_best_size,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flops)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("better_size_first", 1, 4, 4, 1); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_best_flops)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("better_size_first", 1, 4, 4, 2); __PYX_ERR(0, 237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_best_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("better_size_first", 1, 4, 4, 3); __PYX_ERR(0, 237, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "better_size_first") < 0)) __PYX_ERR(0, 237, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_flops = values[0];
    __pyx_v_size = values[1];
    __pyx_v_best_flops = values[2];
    __pyx_v_best_size = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("better_size_first", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.better_size_first", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_12better_size_first(__pyx_self, __pyx_v_flops, __pyx_v_size, __pyx_v_best_flops, __pyx_v_best_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_12better_size_first(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flops, PyObject *__pyx_v_size, PyObject *__pyx_v_best_flops, PyObject *__pyx_v_best_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("better_size_first", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":238
 * 
 * def better_size_first(flops, size, best_flops, best_size):
 *     return (size, flops) < (best_size, best_flops)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_size);
  __Pyx_GIVEREF(__pyx_v_size);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_size);
  __Pyx_INCREF(__pyx_v_flops);
  __Pyx_GIVEREF(__pyx_v_flops);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_flops);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_best_size);
  __Pyx_GIVEREF(__pyx_v_best_size);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_best_size);
  __Pyx_INCREF(__pyx_v_best_flops);
  __Pyx_GIVEREF(__pyx_v_best_flops);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_best_flops);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":237
 * 
 * 
 * def better_size_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (size, flops) < (best_size, best_flops)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.better_size_first", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":247
 * 
 * 
 * def get_better_fn(key):             # <<<<<<<<<<<<<<
 *     return _BETTER_FNS[key]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_15get_better_fn(PyObject *__pyx_self, PyObject *__pyx_v_key); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_15get_better_fn = {"get_better_fn", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_15get_better_fn, METH_O, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_15get_better_fn(PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_better_fn (wrapper)", 0);
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_14get_better_fn(__pyx_self, ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_14get_better_fn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_better_fn", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":248
 * 
 * def get_better_fn(key):
 *     return _BETTER_FNS[key]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_BETTER_FNS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":247
 * 
 * 
 * def get_better_fn(key):             # <<<<<<<<<<<<<<
 *     return _BETTER_FNS[key]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.get_better_fn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":253
 * # functions for assigning a heuristic 'cost' to a potential contraction
 * 
 * def cost_memory_removed(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """The default heuristic cost, corresponding to the total reduction in
 *     memory of performing a contraction.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_17cost_memory_removed(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_16cost_memory_removed[] = "The default heuristic cost, corresponding to the total reduction in\n    memory of performing a contraction.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_17cost_memory_removed = {"cost_memory_removed", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_17cost_memory_removed, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_16cost_memory_removed};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_17cost_memory_removed(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_size12 = 0;
  PyObject *__pyx_v_size1 = 0;
  PyObject *__pyx_v_size2 = 0;
  CYTHON_UNUSED PyObject *__pyx_v_k12 = 0;
  CYTHON_UNUSED PyObject *__pyx_v_k1 = 0;
  CYTHON_UNUSED PyObject *__pyx_v_k2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cost_memory_removed (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size12,&__pyx_n_s_size1,&__pyx_n_s_size2,&__pyx_n_s_k12,&__pyx_n_s_k1,&__pyx_n_s_k2,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size12)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed", 1, 6, 6, 1); __PYX_ERR(0, 253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed", 1, 6, 6, 2); __PYX_ERR(0, 253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k12)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed", 1, 6, 6, 3); __PYX_ERR(0, 253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed", 1, 6, 6, 4); __PYX_ERR(0, 253, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed", 1, 6, 6, 5); __PYX_ERR(0, 253, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cost_memory_removed") < 0)) __PYX_ERR(0, 253, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_size12 = values[0];
    __pyx_v_size1 = values[1];
    __pyx_v_size2 = values[2];
    __pyx_v_k12 = values[3];
    __pyx_v_k1 = values[4];
    __pyx_v_k2 = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cost_memory_removed", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 253, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.cost_memory_removed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_16cost_memory_removed(__pyx_self, __pyx_v_size12, __pyx_v_size1, __pyx_v_size2, __pyx_v_k12, __pyx_v_k1, __pyx_v_k2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_16cost_memory_removed(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_size12, PyObject *__pyx_v_size1, PyObject *__pyx_v_size2, CYTHON_UNUSED PyObject *__pyx_v_k12, CYTHON_UNUSED PyObject *__pyx_v_k1, CYTHON_UNUSED PyObject *__pyx_v_k2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cost_memory_removed", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":257
 *     memory of performing a contraction.
 *     """
 *     return size12 - size1 - size2             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_size12, __pyx_v_size1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_size2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":253
 * # functions for assigning a heuristic 'cost' to a potential contraction
 * 
 * def cost_memory_removed(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """The default heuristic cost, corresponding to the total reduction in
 *     memory of performing a contraction.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.cost_memory_removed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":260
 * 
 * 
 * def cost_memory_removed_jitter(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """Like memory-removed, but with a slight amount of noise that breaks ties
 *     and thus jumbles the contractions a bit.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_19cost_memory_removed_jitter(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_18cost_memory_removed_jitter[] = "Like memory-removed, but with a slight amount of noise that breaks ties\n    and thus jumbles the contractions a bit.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_19cost_memory_removed_jitter = {"cost_memory_removed_jitter", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_19cost_memory_removed_jitter, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_18cost_memory_removed_jitter};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_19cost_memory_removed_jitter(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_size12 = 0;
  PyObject *__pyx_v_size1 = 0;
  PyObject *__pyx_v_size2 = 0;
  CYTHON_UNUSED PyObject *__pyx_v_k12 = 0;
  CYTHON_UNUSED PyObject *__pyx_v_k1 = 0;
  CYTHON_UNUSED PyObject *__pyx_v_k2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cost_memory_removed_jitter (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size12,&__pyx_n_s_size1,&__pyx_n_s_size2,&__pyx_n_s_k12,&__pyx_n_s_k1,&__pyx_n_s_k2,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size12)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed_jitter", 1, 6, 6, 1); __PYX_ERR(0, 260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed_jitter", 1, 6, 6, 2); __PYX_ERR(0, 260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k12)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed_jitter", 1, 6, 6, 3); __PYX_ERR(0, 260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed_jitter", 1, 6, 6, 4); __PYX_ERR(0, 260, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cost_memory_removed_jitter", 1, 6, 6, 5); __PYX_ERR(0, 260, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cost_memory_removed_jitter") < 0)) __PYX_ERR(0, 260, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_size12 = values[0];
    __pyx_v_size1 = values[1];
    __pyx_v_size2 = values[2];
    __pyx_v_k12 = values[3];
    __pyx_v_k1 = values[4];
    __pyx_v_k2 = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cost_memory_removed_jitter", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 260, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.cost_memory_removed_jitter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_18cost_memory_removed_jitter(__pyx_self, __pyx_v_size12, __pyx_v_size1, __pyx_v_size2, __pyx_v_k12, __pyx_v_k1, __pyx_v_k2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_18cost_memory_removed_jitter(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_size12, PyObject *__pyx_v_size1, PyObject *__pyx_v_size2, CYTHON_UNUSED PyObject *__pyx_v_k12, CYTHON_UNUSED PyObject *__pyx_v_k1, CYTHON_UNUSED PyObject *__pyx_v_k2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cost_memory_removed_jitter", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":264
 *     and thus jumbles the contractions a bit.
 *     """
 *     return random.gauss(1.0, 0.01) * (size12 - size1 - size2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_random); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_gauss); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_size12, __pyx_v_size1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_2, __pyx_v_size2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":260
 * 
 * 
 * def cost_memory_removed_jitter(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """Like memory-removed, but with a slight amount of noise that breaks ties
 *     and thus jumbles the contractions a bit.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.cost_memory_removed_jitter", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":302
 *     """
 * 
 *     def __init__(self, nbranch=None, cutoff_flops_factor=4, minimize='flops', cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *         self.nbranch = nbranch
 *         self.cutoff_flops_factor = cutoff_flops_factor
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_1__init__ = {"__init__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_nbranch = 0;
  PyObject *__pyx_v_cutoff_flops_factor = 0;
  PyObject *__pyx_v_minimize = 0;
  PyObject *__pyx_v_cost_fn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_nbranch,&__pyx_n_s_cutoff_flops_factor,&__pyx_n_s_minimize,&__pyx_n_s_cost_fn,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject *)((PyObject *)Py_None));
    values[2] = ((PyObject *)((PyObject *)__pyx_int_4));
    values[3] = ((PyObject *)((PyObject*)__pyx_n_s_flops));
    values[4] = ((PyObject *)((PyObject*)__pyx_kp_s_memory_removed));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nbranch);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cutoff_flops_factor);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minimize);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cost_fn);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 302, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_nbranch = values[1];
    __pyx_v_cutoff_flops_factor = values[2];
    __pyx_v_minimize = values[3];
    __pyx_v_cost_fn = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 302, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound___init__(__pyx_self, __pyx_v_self, __pyx_v_nbranch, __pyx_v_cutoff_flops_factor, __pyx_v_minimize, __pyx_v_cost_fn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":310
 *         self.better = get_better_fn(minimize)
 *         self.best = {'flops': float('inf'), 'size': float('inf')}
 *         self.best_progress = defaultdict(lambda: float('inf'))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__init___lambda2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__init___lambda2 = {"lambda2", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__init___lambda2, METH_NOARGS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__init___lambda2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda2 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda2(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda2(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda2", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_n_s_inf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__init__.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":302
 *     """
 * 
 *     def __init__(self, nbranch=None, cutoff_flops_factor=4, minimize='flops', cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *         self.nbranch = nbranch
 *         self.cutoff_flops_factor = cutoff_flops_factor
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_nbranch, PyObject *__pyx_v_cutoff_flops_factor, PyObject *__pyx_v_minimize, PyObject *__pyx_v_cost_fn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":303
 * 
 *     def __init__(self, nbranch=None, cutoff_flops_factor=4, minimize='flops', cost_fn='memory-removed'):
 *         self.nbranch = nbranch             # <<<<<<<<<<<<<<
 *         self.cutoff_flops_factor = cutoff_flops_factor
 *         self.minimize = minimize
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_nbranch, __pyx_v_nbranch) < 0) __PYX_ERR(0, 303, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":304
 *     def __init__(self, nbranch=None, cutoff_flops_factor=4, minimize='flops', cost_fn='memory-removed'):
 *         self.nbranch = nbranch
 *         self.cutoff_flops_factor = cutoff_flops_factor             # <<<<<<<<<<<<<<
 *         self.minimize = minimize
 *         self.cost_fn = _COST_FNS.get(cost_fn, cost_fn)
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_cutoff_flops_factor, __pyx_v_cutoff_flops_factor) < 0) __PYX_ERR(0, 304, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":305
 *         self.nbranch = nbranch
 *         self.cutoff_flops_factor = cutoff_flops_factor
 *         self.minimize = minimize             # <<<<<<<<<<<<<<
 *         self.cost_fn = _COST_FNS.get(cost_fn, cost_fn)
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_minimize, __pyx_v_minimize) < 0) __PYX_ERR(0, 305, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":306
 *         self.cutoff_flops_factor = cutoff_flops_factor
 *         self.minimize = minimize
 *         self.cost_fn = _COST_FNS.get(cost_fn, cost_fn)             # <<<<<<<<<<<<<<
 * 
 *         self.better = get_better_fn(minimize)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_COST_FNS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cost_fn, __pyx_v_cost_fn};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_cost_fn, __pyx_v_cost_fn};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cost_fn);
    __Pyx_GIVEREF(__pyx_v_cost_fn);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_cost_fn);
    __Pyx_INCREF(__pyx_v_cost_fn);
    __Pyx_GIVEREF(__pyx_v_cost_fn);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_cost_fn);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_cost_fn, __pyx_t_1) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":308
 *         self.cost_fn = _COST_FNS.get(cost_fn, cost_fn)
 * 
 *         self.better = get_better_fn(minimize)             # <<<<<<<<<<<<<<
 *         self.best = {'flops': float('inf'), 'size': float('inf')}
 *         self.best_progress = defaultdict(lambda: float('inf'))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_get_better_fn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_minimize) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_minimize);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_better, __pyx_t_1) < 0) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":309
 * 
 *         self.better = get_better_fn(minimize)
 *         self.best = {'flops': float('inf'), 'size': float('inf')}             # <<<<<<<<<<<<<<
 *         self.best_progress = defaultdict(lambda: float('inf'))
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_n_s_inf); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_flops, __pyx_t_3) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_n_s_inf); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_size, __pyx_t_3) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_best, __pyx_t_1) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":310
 *         self.better = get_better_fn(minimize)
 *         self.best = {'flops': float('inf'), 'size': float('inf')}
 *         self.best_progress = defaultdict(lambda: float('inf'))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__init___lambda2, 0, __pyx_n_s_BranchBound___init___locals_lamb, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_best_progress, __pyx_t_1) < 0) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":302
 *     """
 * 
 *     def __init__(self, nbranch=None, cutoff_flops_factor=4, minimize='flops', cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *         self.nbranch = nbranch
 *         self.cutoff_flops_factor = cutoff_flops_factor
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":313
 * 
 *     @property
 *     def path(self):             # <<<<<<<<<<<<<<
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_3path(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_3path = {"path", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_3path, METH_O, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_3path(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("path (wrapper)", 0);
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_2path(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_2path(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("path", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":314
 *     @property
 *     def path(self):
 *         return ssa_to_linear(self.best['ssa_path'])             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ssa_to_linear); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_s_ssa_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":313
 * 
 *     @property
 *     def path(self):             # <<<<<<<<<<<<<<
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":316
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_5__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_11BranchBound_4__call__[] = "\n\n        Parameters\n        ----------\n        input_sets : list\n            List of sets that represent the lhs side of the einsum subscript\n        output_set : set\n            Set that represents the rhs side of the overall einsum subscript\n        idx_dict : dictionary\n            Dictionary of index sizes\n        memory_limit : int\n            The maximum number of elements in a temporary array\n\n        Returns\n        -------\n        path : list\n            The contraction order within the memory limit constraint.\n\n        Examples\n        --------\n        >>> isets = [set('abd'), set('ac'), set('bdc')]\n        >>> oset = set('')\n        >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}\n        >>> optimal(isets, oset, idx_sizes, 5000)\n        [(0, 2), (0, 1)]\n        ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_5__call__ = {"__call__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_5__call__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_11BranchBound_4__call__};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_5__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  PyObject *__pyx_v_memory_limit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[4] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 1); __PYX_ERR(0, 316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 2); __PYX_ERR(0, 316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 3); __PYX_ERR(0, 316, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(0, 316, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_inputs = values[1];
    __pyx_v_output = values[2];
    __pyx_v_size_dict = values[3];
    __pyx_v_memory_limit = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 316, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_4__call__(__pyx_self, __pyx_v_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":350
 *         result_cache = {}
 * 
 *         def _branch_iterate(path, inputs, remaining, flops, size):             # <<<<<<<<<<<<<<
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___1_branch_iterate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___1_branch_iterate = {"_branch_iterate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___1_branch_iterate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___1_branch_iterate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_remaining = 0;
  PyObject *__pyx_v_flops = 0;
  PyObject *__pyx_v_size = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_branch_iterate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_inputs,&__pyx_n_s_remaining,&__pyx_n_s_flops,&__pyx_n_s_size,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_branch_iterate", 1, 5, 5, 1); __PYX_ERR(0, 350, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_remaining)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_branch_iterate", 1, 5, 5, 2); __PYX_ERR(0, 350, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flops)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_branch_iterate", 1, 5, 5, 3); __PYX_ERR(0, 350, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_branch_iterate", 1, 5, 5, 4); __PYX_ERR(0, 350, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_branch_iterate") < 0)) __PYX_ERR(0, 350, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_path = values[0];
    __pyx_v_inputs = values[1];
    __pyx_v_remaining = values[2];
    __pyx_v_flops = values[3];
    __pyx_v_size = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_branch_iterate", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 350, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__._branch_iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call____branch_iterate(__pyx_self, __pyx_v_path, __pyx_v_inputs, __pyx_v_remaining, __pyx_v_flops, __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":359
 *                 return
 * 
 *             def _assess_candidate(k1, k2, i, j):             # <<<<<<<<<<<<<<
 *                 # find resulting indices and flops
 *                 try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate_1_assess_candidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate_1_assess_candidate = {"_assess_candidate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate_1_assess_candidate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate_1_assess_candidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_k1 = 0;
  PyObject *__pyx_v_k2 = 0;
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_v_j = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_assess_candidate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_k1,&__pyx_n_s_k2,&__pyx_n_s_i,&__pyx_n_s_j,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_assess_candidate", 1, 4, 4, 1); __PYX_ERR(0, 359, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_assess_candidate", 1, 4, 4, 2); __PYX_ERR(0, 359, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_j)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_assess_candidate", 1, 4, 4, 3); __PYX_ERR(0, 359, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_assess_candidate") < 0)) __PYX_ERR(0, 359, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_k1 = values[0];
    __pyx_v_k2 = values[1];
    __pyx_v_i = values[2];
    __pyx_v_j = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_assess_candidate", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 359, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__._branch_iterate._assess_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate__assess_candidate(__pyx_self, __pyx_v_k1, __pyx_v_k2, __pyx_v_i, __pyx_v_j);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate__assess_candidate(PyObject *__pyx_self, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2, PyObject *__pyx_v_i, PyObject *__pyx_v_j) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *__pyx_cur_scope;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *__pyx_outer_scope;
  PyObject *__pyx_v_k12 = NULL;
  PyObject *__pyx_v_flops12 = NULL;
  PyObject *__pyx_v_size12 = NULL;
  PyObject *__pyx_v_new_flops = NULL;
  PyObject *__pyx_v_new_size = NULL;
  PyObject *__pyx_v_size1 = NULL;
  PyObject *__pyx_v_size2 = NULL;
  PyObject *__pyx_v_cost = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_assess_candidate", 0);
  __pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "acqdp/utility/opt_einsum_paths.pyx":361
 *             def _assess_candidate(k1, k2, i, j):
 *                 # find resulting indices and flops
 *                 try:             # <<<<<<<<<<<<<<
 *                     k12, flops12 = result_cache[k1, k2]
 *                 except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "acqdp/utility/opt_einsum_paths.pyx":362
 *                 # find resulting indices and flops
 *                 try:
 *                     k12, flops12 = result_cache[k1, k2]             # <<<<<<<<<<<<<<
 *                 except KeyError:
 *                     k12, flops12 = result_cache[k1, k2] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)
 */
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_result_cache)) { __Pyx_RaiseClosureNameError("result_cache"); __PYX_ERR(0, 362, __pyx_L3_error) }
      if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_result_cache == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 362, __pyx_L3_error)
      }
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_k1);
      __Pyx_GIVEREF(__pyx_v_k1);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_k1);
      __Pyx_INCREF(__pyx_v_k2);
      __Pyx_GIVEREF(__pyx_v_k2);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_k2);
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_result_cache, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 362, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 362, __pyx_L3_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 362, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 362, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
        index = 0; __pyx_t_4 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_4)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 362, __pyx_L3_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 362, __pyx_L3_error)
        __pyx_L10_unpacking_done:;
      }
      __pyx_v_k12 = __pyx_t_4;
      __pyx_t_4 = 0;
      __pyx_v_flops12 = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":361
 *             def _assess_candidate(k1, k2, i, j):
 *                 # find resulting indices and flops
 *                 try:             # <<<<<<<<<<<<<<
 *                     k12, flops12 = result_cache[k1, k2]
 *                 except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":363
 *                 try:
 *                     k12, flops12 = result_cache[k1, k2]
 *                 except KeyError:             # <<<<<<<<<<<<<<
 *                     k12, flops12 = result_cache[k1, k2] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__._branch_iterate._assess_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_4) < 0) __PYX_ERR(0, 363, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_4);

      /* "acqdp/utility/opt_einsum_paths.pyx":364
 *                     k12, flops12 = result_cache[k1, k2]
 *                 except KeyError:
 *                     k12, flops12 = result_cache[k1, k2] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)             # <<<<<<<<<<<<<<
 * 
 *                 try:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_calc_k12_flops); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 364, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (unlikely(!__pyx_cur_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 364, __pyx_L5_except_error) }
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output)) { __Pyx_RaiseClosureNameError("output"); __PYX_ERR(0, 364, __pyx_L5_except_error) }
      if (unlikely(!__pyx_cur_scope->__pyx_v_remaining)) { __Pyx_RaiseClosureNameError("remaining"); __PYX_ERR(0, 364, __pyx_L5_except_error) }
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict)) { __Pyx_RaiseClosureNameError("size_dict"); __PYX_ERR(0, 364, __pyx_L5_except_error) }
      __pyx_t_11 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[7] = {__pyx_t_11, __pyx_cur_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output, __pyx_cur_scope->__pyx_v_remaining, __pyx_v_i, __pyx_v_j, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict};
        __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 6+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[7] = {__pyx_t_11, __pyx_cur_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output, __pyx_cur_scope->__pyx_v_remaining, __pyx_v_i, __pyx_v_j, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict};
        __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 6+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(6+__pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_11) {
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11); __pyx_t_11 = NULL;
        }
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_inputs);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_inputs);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_9, __pyx_cur_scope->__pyx_v_inputs);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_9, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_remaining);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_remaining);
        PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_9, __pyx_cur_scope->__pyx_v_remaining);
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_9, __pyx_v_i);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_12, 4+__pyx_t_9, __pyx_v_j);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
        PyTuple_SET_ITEM(__pyx_t_12, 5+__pyx_t_9, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 364, __pyx_L5_except_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_12 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_12 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_12);
        #else
        __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_12 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_11 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_8 = Py_TYPE(__pyx_t_11)->tp_iternext;
        index = 0; __pyx_t_10 = __pyx_t_8(__pyx_t_11); if (unlikely(!__pyx_t_10)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_10);
        index = 1; __pyx_t_12 = __pyx_t_8(__pyx_t_11); if (unlikely(!__pyx_t_12)) goto __pyx_L13_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_12);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_11), 2) < 0) __PYX_ERR(0, 364, __pyx_L5_except_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L14_unpacking_done;
        __pyx_L13_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 364, __pyx_L5_except_error)
        __pyx_L14_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_k12, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_flops12, __pyx_t_12);
      __pyx_t_12 = 0;
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_result_cache)) { __Pyx_RaiseClosureNameError("result_cache"); __PYX_ERR(0, 364, __pyx_L5_except_error) }
      if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_result_cache == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 364, __pyx_L5_except_error)
      }
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 364, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(__pyx_v_k1);
      __Pyx_GIVEREF(__pyx_v_k1);
      PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_k1);
      __Pyx_INCREF(__pyx_v_k2);
      __Pyx_GIVEREF(__pyx_v_k2);
      PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_k2);
      if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_result_cache, __pyx_t_12, __pyx_t_7) < 0)) __PYX_ERR(0, 364, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "acqdp/utility/opt_einsum_paths.pyx":361
 *             def _assess_candidate(k1, k2, i, j):
 *                 # find resulting indices and flops
 *                 try:             # <<<<<<<<<<<<<<
 *                     k12, flops12 = result_cache[k1, k2]
 *                 except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":366
 *                     k12, flops12 = result_cache[k1, k2] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_2, &__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_1);
    /*try:*/ {

      /* "acqdp/utility/opt_einsum_paths.pyx":367
 * 
 *                 try:
 *                     size12 = size_cache[k12]             # <<<<<<<<<<<<<<
 *                 except KeyError:
 *                     size12 = size_cache[k12] = helpers.compute_size_by_dict(k12, size_dict)
 */
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache)) { __Pyx_RaiseClosureNameError("size_cache"); __PYX_ERR(0, 367, __pyx_L15_error) }
      if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 367, __pyx_L15_error)
      }
      __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache, __pyx_v_k12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 367, __pyx_L15_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_size12 = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":366
 *                     k12, flops12 = result_cache[k1, k2] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L20_try_end;
    __pyx_L15_error:;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":368
 *                 try:
 *                     size12 = size_cache[k12]
 *                 except KeyError:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12] = helpers.compute_size_by_dict(k12, size_dict)
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__._branch_iterate._assess_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 368, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_5);

      /* "acqdp/utility/opt_einsum_paths.pyx":369
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 *                     size12 = size_cache[k12] = helpers.compute_size_by_dict(k12, size_dict)             # <<<<<<<<<<<<<<
 * 
 *                 new_flops = flops + flops12
 */
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_helpers); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 369, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 369, __pyx_L17_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict)) { __Pyx_RaiseClosureNameError("size_dict"); __PYX_ERR(0, 369, __pyx_L17_except_error) }
      __pyx_t_12 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_k12, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict};
        __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 369, __pyx_L17_except_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_k12, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict};
        __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 369, __pyx_L17_except_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_7);
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 369, __pyx_L17_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_12) {
          __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_12); __pyx_t_12 = NULL;
        }
        __Pyx_INCREF(__pyx_v_k12);
        __Pyx_GIVEREF(__pyx_v_k12);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_9, __pyx_v_k12);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_9, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_11, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 369, __pyx_L17_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_size12, __pyx_t_7);
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache)) { __Pyx_RaiseClosureNameError("size_cache"); __PYX_ERR(0, 369, __pyx_L17_except_error) }
      if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 369, __pyx_L17_except_error)
      }
      if (unlikely(PyDict_SetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache, __pyx_v_k12, __pyx_t_7) < 0)) __PYX_ERR(0, 369, __pyx_L17_except_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L16_exception_handled;
    }
    goto __pyx_L17_except_error;
    __pyx_L17_except_error:;

    /* "acqdp/utility/opt_einsum_paths.pyx":366
 *                     k12, flops12 = result_cache[k1, k2] = calc_k12_flops(inputs, output, remaining, i, j, size_dict)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     size12 = size_cache[k12]
 *                 except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    goto __pyx_L1_error;
    __pyx_L16_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_2, __pyx_t_1);
    __pyx_L20_try_end:;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":371
 *                     size12 = size_cache[k12] = helpers.compute_size_by_dict(k12, size_dict)
 * 
 *                 new_flops = flops + flops12             # <<<<<<<<<<<<<<
 *                 new_size = max(size, size12)
 * 
 */
  if (unlikely(!__pyx_cur_scope->__pyx_v_flops)) { __Pyx_RaiseClosureNameError("flops"); __PYX_ERR(0, 371, __pyx_L1_error) }
  __pyx_t_5 = PyNumber_Add(__pyx_cur_scope->__pyx_v_flops, __pyx_v_flops12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_new_flops = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":372
 * 
 *                 new_flops = flops + flops12
 *                 new_size = max(size, size12)             # <<<<<<<<<<<<<<
 * 
 *                 # sieve based on current best i.e. check flops and size still better
 */
  __Pyx_INCREF(__pyx_v_size12);
  __pyx_t_5 = __pyx_v_size12;
  if (unlikely(!__pyx_cur_scope->__pyx_v_size)) { __Pyx_RaiseClosureNameError("size"); __PYX_ERR(0, 372, __pyx_L1_error) }
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size);
  __pyx_t_6 = __pyx_cur_scope->__pyx_v_size;
  __pyx_t_7 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 372, __pyx_L1_error)
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_13) {
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_4 = __pyx_t_5;
  } else {
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_4 = __pyx_t_6;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_new_size = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":375
 * 
 *                 # sieve based on current best i.e. check flops and size still better
 *                 if not self.better(new_flops, new_size, self.best['flops'], self.best['size']):             # <<<<<<<<<<<<<<
 *                     return None
 * 
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 375, __pyx_L1_error) }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_better); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 375, __pyx_L1_error) }
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_s_flops); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 375, __pyx_L1_error) }
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_10 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_s_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_new_flops, __pyx_v_new_size, __pyx_t_7, __pyx_t_10};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[5] = {__pyx_t_6, __pyx_v_new_flops, __pyx_v_new_size, __pyx_t_7, __pyx_t_10};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_v_new_flops);
    __Pyx_GIVEREF(__pyx_v_new_flops);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_9, __pyx_v_new_flops);
    __Pyx_INCREF(__pyx_v_new_size);
    __Pyx_GIVEREF(__pyx_v_new_size);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_9, __pyx_v_new_size);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_9, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_9, __pyx_t_10);
    __pyx_t_7 = 0;
    __pyx_t_10 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_11, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_14 = ((!__pyx_t_13) != 0);
  if (__pyx_t_14) {

    /* "acqdp/utility/opt_einsum_paths.pyx":376
 *                 # sieve based on current best i.e. check flops and size still better
 *                 if not self.better(new_flops, new_size, self.best['flops'], self.best['size']):
 *                     return None             # <<<<<<<<<<<<<<
 * 
 *                 # compare to how the best method was doing as this point
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":375
 * 
 *                 # sieve based on current best i.e. check flops and size still better
 *                 if not self.better(new_flops, new_size, self.best['flops'], self.best['size']):             # <<<<<<<<<<<<<<
 *                     return None
 * 
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":379
 * 
 *                 # compare to how the best method was doing as this point
 *                 if new_flops < self.best_progress[len(inputs)]:             # <<<<<<<<<<<<<<
 *                     self.best_progress[len(inputs)] = new_flops
 *                 # sieve based on current progress relative to best
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 379, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best_progress); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(!__pyx_cur_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 379, __pyx_L1_error) }
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_inputs;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_15 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_5, __pyx_t_15, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_RichCompare(__pyx_v_new_flops, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_14) {

    /* "acqdp/utility/opt_einsum_paths.pyx":380
 *                 # compare to how the best method was doing as this point
 *                 if new_flops < self.best_progress[len(inputs)]:
 *                     self.best_progress[len(inputs)] = new_flops             # <<<<<<<<<<<<<<
 *                 # sieve based on current progress relative to best
 *                 elif new_flops > self.cutoff_flops_factor * self.best_progress[len(inputs)]:
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 380, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best_progress); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(!__pyx_cur_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 380, __pyx_L1_error) }
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_inputs;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_15 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(__Pyx_SetItemInt(__pyx_t_5, __pyx_t_15, __pyx_v_new_flops, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1) < 0)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":379
 * 
 *                 # compare to how the best method was doing as this point
 *                 if new_flops < self.best_progress[len(inputs)]:             # <<<<<<<<<<<<<<
 *                     self.best_progress[len(inputs)] = new_flops
 *                 # sieve based on current progress relative to best
 */
    goto __pyx_L24;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":382
 *                     self.best_progress[len(inputs)] = new_flops
 *                 # sieve based on current progress relative to best
 *                 elif new_flops > self.cutoff_flops_factor * self.best_progress[len(inputs)]:             # <<<<<<<<<<<<<<
 *                     return None
 * 
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 382, __pyx_L1_error) }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_cutoff_flops_factor); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 382, __pyx_L1_error) }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best_progress); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_cur_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 382, __pyx_L1_error) }
  __pyx_t_11 = __pyx_cur_scope->__pyx_v_inputs;
  __Pyx_INCREF(__pyx_t_11);
  __pyx_t_15 = PyObject_Length(__pyx_t_11); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_GetItemInt(__pyx_t_4, __pyx_t_15, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_5, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = PyObject_RichCompare(__pyx_v_new_flops, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (__pyx_t_14) {

    /* "acqdp/utility/opt_einsum_paths.pyx":383
 *                 # sieve based on current progress relative to best
 *                 elif new_flops > self.cutoff_flops_factor * self.best_progress[len(inputs)]:
 *                     return None             # <<<<<<<<<<<<<<
 * 
 *                 # sieve based on memory limit
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":382
 *                     self.best_progress[len(inputs)] = new_flops
 *                 # sieve based on current progress relative to best
 *                 elif new_flops > self.cutoff_flops_factor * self.best_progress[len(inputs)]:             # <<<<<<<<<<<<<<
 *                     return None
 * 
 */
  }
  __pyx_L24:;

  /* "acqdp/utility/opt_einsum_paths.pyx":386
 * 
 *                 # sieve based on memory limit
 *                 if (memory_limit not in _UNLIMITED_MEM) and (size12 > memory_limit):             # <<<<<<<<<<<<<<
 *                     # terminate path here, but check all-terms contract first
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_memory_limit)) { __Pyx_RaiseClosureNameError("memory_limit"); __PYX_ERR(0, 386, __pyx_L1_error) }
  __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_UNLIMITED_MEM); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_13 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_memory_limit, __pyx_t_11, Py_NE)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_16 = (__pyx_t_13 != 0);
  if (__pyx_t_16) {
  } else {
    __pyx_t_14 = __pyx_t_16;
    goto __pyx_L26_bool_binop_done;
  }
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_memory_limit)) { __Pyx_RaiseClosureNameError("memory_limit"); __PYX_ERR(0, 386, __pyx_L1_error) }
  __pyx_t_11 = PyObject_RichCompare(__pyx_v_size12, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_memory_limit, Py_GT); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 386, __pyx_L1_error)
  __pyx_t_16 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_16 < 0)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_14 = __pyx_t_16;
  __pyx_L26_bool_binop_done:;
  if (__pyx_t_14) {

    /* "acqdp/utility/opt_einsum_paths.pyx":388
 *                 if (memory_limit not in _UNLIMITED_MEM) and (size12 > memory_limit):
 *                     # terminate path here, but check all-terms contract first
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)             # <<<<<<<<<<<<<<
 *                     if new_flops < self.best['flops']:
 *                         self.best['flops'] = new_flops
 */
    if (unlikely(!__pyx_cur_scope->__pyx_v_flops)) { __Pyx_RaiseClosureNameError("flops"); __PYX_ERR(0, 388, __pyx_L1_error) }
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_compute_oversize_flops); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 388, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_v_remaining)) { __Pyx_RaiseClosureNameError("remaining"); __PYX_ERR(0, 388, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output)) { __Pyx_RaiseClosureNameError("output"); __PYX_ERR(0, 388, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict)) { __Pyx_RaiseClosureNameError("size_dict"); __PYX_ERR(0, 388, __pyx_L1_error) }
    __pyx_t_5 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_v_remaining, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict};
      __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_11);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_v_remaining, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict};
      __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_9, 4+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_11);
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(4+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_inputs);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_inputs);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_cur_scope->__pyx_v_inputs);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_remaining);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_remaining);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_cur_scope->__pyx_v_remaining);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
      PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
      PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_dict);
      __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_cur_scope->__pyx_v_flops, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF_SET(__pyx_v_new_flops, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":389
 *                     # terminate path here, but check all-terms contract first
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < self.best['flops']:             # <<<<<<<<<<<<<<
 *                         self.best['flops'] = new_flops
 *                         self.best['ssa_path'] = path + (tuple(remaining),)
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 389, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_11 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_n_s_flops); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_new_flops, __pyx_t_11, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_14) {

      /* "acqdp/utility/opt_einsum_paths.pyx":390
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < self.best['flops']:
 *                         self.best['flops'] = new_flops             # <<<<<<<<<<<<<<
 *                         self.best['ssa_path'] = path + (tuple(remaining),)
 *                     return None
 */
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 390, __pyx_L1_error) }
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_n_s_flops, __pyx_v_new_flops) < 0)) __PYX_ERR(0, 390, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":391
 *                     if new_flops < self.best['flops']:
 *                         self.best['flops'] = new_flops
 *                         self.best['ssa_path'] = path + (tuple(remaining),)             # <<<<<<<<<<<<<<
 *                     return None
 * 
 */
      if (unlikely(!__pyx_cur_scope->__pyx_v_path)) { __Pyx_RaiseClosureNameError("path"); __PYX_ERR(0, 391, __pyx_L1_error) }
      if (unlikely(!__pyx_cur_scope->__pyx_v_remaining)) { __Pyx_RaiseClosureNameError("remaining"); __PYX_ERR(0, 391, __pyx_L1_error) }
      __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_cur_scope->__pyx_v_remaining); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Add(__pyx_cur_scope->__pyx_v_path, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 391, __pyx_L1_error) }
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (unlikely(PyObject_SetItem(__pyx_t_11, __pyx_n_s_ssa_path, __pyx_t_4) < 0)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":389
 *                     # terminate path here, but check all-terms contract first
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 *                     if new_flops < self.best['flops']:             # <<<<<<<<<<<<<<
 *                         self.best['flops'] = new_flops
 *                         self.best['ssa_path'] = path + (tuple(remaining),)
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":392
 *                         self.best['flops'] = new_flops
 *                         self.best['ssa_path'] = path + (tuple(remaining),)
 *                     return None             # <<<<<<<<<<<<<<
 * 
 *                 # set cost heuristic in order to locally sort possible contractions
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":386
 * 
 *                 # sieve based on memory limit
 *                 if (memory_limit not in _UNLIMITED_MEM) and (size12 > memory_limit):             # <<<<<<<<<<<<<<
 *                     # terminate path here, but check all-terms contract first
 *                     new_flops = flops + _compute_oversize_flops(inputs, remaining, output, size_dict)
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":395
 * 
 *                 # set cost heuristic in order to locally sort possible contractions
 *                 size1, size2 = size_cache[inputs[i]], size_cache[inputs[j]]             # <<<<<<<<<<<<<<
 *                 cost = self.cost_fn(size12, size1, size2, k12, k1, k2)
 * 
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache)) { __Pyx_RaiseClosureNameError("size_cache"); __PYX_ERR(0, 395, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 395, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 395, __pyx_L1_error) }
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_inputs, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache)) { __Pyx_RaiseClosureNameError("size_cache"); __PYX_ERR(0, 395, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 395, __pyx_L1_error)
  }
  if (unlikely(!__pyx_cur_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 395, __pyx_L1_error) }
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_inputs, __pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_size_cache, __pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_size1 = __pyx_t_11;
  __pyx_t_11 = 0;
  __pyx_v_size2 = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":396
 *                 # set cost heuristic in order to locally sort possible contractions
 *                 size1, size2 = size_cache[inputs[i]], size_cache[inputs[j]]
 *                 cost = self.cost_fn(size12, size1, size2, k12, k1, k2)             # <<<<<<<<<<<<<<
 * 
 *                 return cost, flops12, new_flops, new_size, (i, j), k12
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 396, __pyx_L1_error) }
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_cost_fn); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_11);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_11, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_11)) {
    PyObject *__pyx_temp[7] = {__pyx_t_4, __pyx_v_size12, __pyx_v_size1, __pyx_v_size2, __pyx_v_k12, __pyx_v_k1, __pyx_v_k2};
    __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_9, 6+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
    PyObject *__pyx_temp[7] = {__pyx_t_4, __pyx_v_size12, __pyx_v_size1, __pyx_v_size2, __pyx_v_k12, __pyx_v_k1, __pyx_v_k2};
    __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_9, 6+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_10);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(6+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_size12);
    __Pyx_GIVEREF(__pyx_v_size12);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_9, __pyx_v_size12);
    __Pyx_INCREF(__pyx_v_size1);
    __Pyx_GIVEREF(__pyx_v_size1);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_9, __pyx_v_size1);
    __Pyx_INCREF(__pyx_v_size2);
    __Pyx_GIVEREF(__pyx_v_size2);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_9, __pyx_v_size2);
    __Pyx_INCREF(__pyx_v_k12);
    __Pyx_GIVEREF(__pyx_v_k12);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_9, __pyx_v_k12);
    __Pyx_INCREF(__pyx_v_k1);
    __Pyx_GIVEREF(__pyx_v_k1);
    PyTuple_SET_ITEM(__pyx_t_5, 4+__pyx_t_9, __pyx_v_k1);
    __Pyx_INCREF(__pyx_v_k2);
    __Pyx_GIVEREF(__pyx_v_k2);
    PyTuple_SET_ITEM(__pyx_t_5, 5+__pyx_t_9, __pyx_v_k2);
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_5, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_v_cost = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":398
 *                 cost = self.cost_fn(size12, size1, size2, k12, k1, k2)
 * 
 *                 return cost, flops12, new_flops, new_size, (i, j), k12             # <<<<<<<<<<<<<<
 * 
 *             # check all possible remaining paths
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_v_i);
  __Pyx_GIVEREF(__pyx_v_i);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_i);
  __Pyx_INCREF(__pyx_v_j);
  __Pyx_GIVEREF(__pyx_v_j);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_v_j);
  __pyx_t_11 = PyTuple_New(6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_v_cost);
  __Pyx_GIVEREF(__pyx_v_cost);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_cost);
  __Pyx_INCREF(__pyx_v_flops12);
  __Pyx_GIVEREF(__pyx_v_flops12);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_flops12);
  __Pyx_INCREF(__pyx_v_new_flops);
  __Pyx_GIVEREF(__pyx_v_new_flops);
  PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_new_flops);
  __Pyx_INCREF(__pyx_v_new_size);
  __Pyx_GIVEREF(__pyx_v_new_size);
  PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_v_new_size);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_11, 4, __pyx_t_10);
  __Pyx_INCREF(__pyx_v_k12);
  __Pyx_GIVEREF(__pyx_v_k12);
  PyTuple_SET_ITEM(__pyx_t_11, 5, __pyx_v_k12);
  __pyx_t_10 = 0;
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":359
 *                 return
 * 
 *             def _assess_candidate(k1, k2, i, j):             # <<<<<<<<<<<<<<
 *                 # find resulting indices and flops
 *                 try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__._branch_iterate._assess_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k12);
  __Pyx_XDECREF(__pyx_v_flops12);
  __Pyx_XDECREF(__pyx_v_size12);
  __Pyx_XDECREF(__pyx_v_new_flops);
  __Pyx_XDECREF(__pyx_v_new_size);
  __Pyx_XDECREF(__pyx_v_size1);
  __Pyx_XDECREF(__pyx_v_size2);
  __Pyx_XDECREF(__pyx_v_cost);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":350
 *         result_cache = {}
 * 
 *         def _branch_iterate(path, inputs, remaining, flops, size):             # <<<<<<<<<<<<<<
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call____branch_iterate(PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_inputs, PyObject *__pyx_v_remaining, PyObject *__pyx_v_flops, PyObject *__pyx_v_size) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *__pyx_cur_scope;
  PyObject *__pyx_v__assess_candidate = 0;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_k1 = NULL;
  PyObject *__pyx_v_k2 = NULL;
  PyObject *__pyx_v_candidate = NULL;
  PyObject *__pyx_v_bi = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_new_flops = NULL;
  PyObject *__pyx_v_new_size = NULL;
  PyObject *__pyx_v_k12 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_branch_iterate", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 350, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_v_path = __pyx_v_path;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_path);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_path);
  __pyx_cur_scope->__pyx_v_inputs = __pyx_v_inputs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_inputs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_inputs);
  __pyx_cur_scope->__pyx_v_remaining = __pyx_v_remaining;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_remaining);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_remaining);
  __pyx_cur_scope->__pyx_v_flops = __pyx_v_flops;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_flops);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_flops);
  __pyx_cur_scope->__pyx_v_size = __pyx_v_size;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size);

  /* "acqdp/utility/opt_einsum_paths.pyx":353
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 *             if len(remaining) == 1:             # <<<<<<<<<<<<<<
 *                 self.best['size'] = size
 *                 self.best['flops'] = flops
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_remaining;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((__pyx_t_2 == 1) != 0);
  if (__pyx_t_3) {

    /* "acqdp/utility/opt_einsum_paths.pyx":354
 *             # reached end of path (only ever get here if flops is best found so far)
 *             if len(remaining) == 1:
 *                 self.best['size'] = size             # <<<<<<<<<<<<<<
 *                 self.best['flops'] = flops
 *                 self.best['ssa_path'] = path
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 354, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_n_s_size, __pyx_cur_scope->__pyx_v_size) < 0)) __PYX_ERR(0, 354, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":355
 *             if len(remaining) == 1:
 *                 self.best['size'] = size
 *                 self.best['flops'] = flops             # <<<<<<<<<<<<<<
 *                 self.best['ssa_path'] = path
 *                 return
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 355, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_n_s_flops, __pyx_cur_scope->__pyx_v_flops) < 0)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":356
 *                 self.best['size'] = size
 *                 self.best['flops'] = flops
 *                 self.best['ssa_path'] = path             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 356, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_best); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_n_s_ssa_path, __pyx_cur_scope->__pyx_v_path) < 0)) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":357
 *                 self.best['flops'] = flops
 *                 self.best['ssa_path'] = path
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             def _assess_candidate(k1, k2, i, j):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":353
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 *             if len(remaining) == 1:             # <<<<<<<<<<<<<<
 *                 self.best['size'] = size
 *                 self.best['flops'] = flops
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":359
 *                 return
 * 
 *             def _assess_candidate(k1, k2, i, j):             # <<<<<<<<<<<<<<
 *                 # find resulting indices and flops
 *                 try:
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate_1_assess_candidate, 0, __pyx_n_s_BranchBound___call___locals__bra, ((PyObject*)__pyx_cur_scope), __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__assess_candidate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":401
 * 
 *             # check all possible remaining paths
 *             candidates = []             # <<<<<<<<<<<<<<
 *             for i, j in itertools.combinations(remaining, 2):
 *                 if i > j:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_candidates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":402
 *             # check all possible remaining paths
 *             candidates = []
 *             for i, j in itertools.combinations(remaining, 2):             # <<<<<<<<<<<<<<
 *                 if i > j:
 *                     i, j = j, i
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_itertools); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_combinations); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_remaining, __pyx_int_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_remaining, __pyx_int_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_remaining);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_remaining);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_cur_scope->__pyx_v_remaining);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_int_2);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_5 = __pyx_t_1; __Pyx_INCREF(__pyx_t_5); __pyx_t_2 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 402, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 402, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_8(__pyx_t_5);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 402, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 402, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_7)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 402, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 402, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":403
 *             candidates = []
 *             for i, j in itertools.combinations(remaining, 2):
 *                 if i > j:             # <<<<<<<<<<<<<<
 *                     i, j = j, i
 *                 k1, k2 = inputs[i], inputs[j]
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_i, __pyx_v_j, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "acqdp/utility/opt_einsum_paths.pyx":404
 *             for i, j in itertools.combinations(remaining, 2):
 *                 if i > j:
 *                     i, j = j, i             # <<<<<<<<<<<<<<
 *                 k1, k2 = inputs[i], inputs[j]
 * 
 */
      __pyx_t_11 = __pyx_v_j;
      __pyx_t_12 = __pyx_v_i;
      __pyx_v_i = __pyx_t_11;
      __pyx_t_11 = 0;
      __pyx_v_j = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":403
 *             candidates = []
 *             for i, j in itertools.combinations(remaining, 2):
 *                 if i > j:             # <<<<<<<<<<<<<<
 *                     i, j = j, i
 *                 k1, k2 = inputs[i], inputs[j]
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":405
 *                 if i > j:
 *                     i, j = j, i
 *                 k1, k2 = inputs[i], inputs[j]             # <<<<<<<<<<<<<<
 * 
 *                 # initially ignore outer products
 */
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_inputs, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_inputs, __pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_k1, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k2, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":408
 * 
 *                 # initially ignore outer products
 *                 if k1.isdisjoint(k2):             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_k1, __pyx_n_s_isdisjoint); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_v_k2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_k2);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "acqdp/utility/opt_einsum_paths.pyx":409
 *                 # initially ignore outer products
 *                 if k1.isdisjoint(k2):
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 candidate = _assess_candidate(k1, k2, i, j)
 */
      goto __pyx_L4_continue;

      /* "acqdp/utility/opt_einsum_paths.pyx":408
 * 
 *                 # initially ignore outer products
 *                 if k1.isdisjoint(k2):             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":411
 *                     continue
 * 
 *                 candidate = _assess_candidate(k1, k2, i, j)             # <<<<<<<<<<<<<<
 *                 if candidate:
 *                     heapq.heappush(candidates, candidate)
 */
    __pyx_t_4 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate__assess_candidate(__pyx_v__assess_candidate, __pyx_v_k1, __pyx_v_k2, __pyx_v_i, __pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_candidate, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":412
 * 
 *                 candidate = _assess_candidate(k1, k2, i, j)
 *                 if candidate:             # <<<<<<<<<<<<<<
 *                     heapq.heappush(candidates, candidate)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_candidate); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 412, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "acqdp/utility/opt_einsum_paths.pyx":413
 *                 candidate = _assess_candidate(k1, k2, i, j)
 *                 if candidate:
 *                     heapq.heappush(candidates, candidate)             # <<<<<<<<<<<<<<
 * 
 *             # assess outer products if nothing left
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_heapq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_heappush); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_candidates, __pyx_v_candidate};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_candidates, __pyx_v_candidate};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_INCREF(__pyx_v_candidates);
        __Pyx_GIVEREF(__pyx_v_candidates);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_6, __pyx_v_candidates);
        __Pyx_INCREF(__pyx_v_candidate);
        __Pyx_GIVEREF(__pyx_v_candidate);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_6, __pyx_v_candidate);
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":412
 * 
 *                 candidate = _assess_candidate(k1, k2, i, j)
 *                 if candidate:             # <<<<<<<<<<<<<<
 *                     heapq.heappush(candidates, candidate)
 * 
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":402
 *             # check all possible remaining paths
 *             candidates = []
 *             for i, j in itertools.combinations(remaining, 2):             # <<<<<<<<<<<<<<
 *                 if i > j:
 *                     i, j = j, i
 */
    __pyx_L4_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":416
 * 
 *             # assess outer products if nothing left
 *             if not candidates:             # <<<<<<<<<<<<<<
 *                 for i, j in itertools.combinations(remaining, 2):
 *                     if i > j:
 */
  __pyx_t_3 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_13 = ((!__pyx_t_3) != 0);
  if (__pyx_t_13) {

    /* "acqdp/utility/opt_einsum_paths.pyx":417
 *             # assess outer products if nothing left
 *             if not candidates:
 *                 for i, j in itertools.combinations(remaining, 2):             # <<<<<<<<<<<<<<
 *                     if i > j:
 *                         i, j = j, i
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_itertools); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_combinations); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_remaining, __pyx_int_2};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 417, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_cur_scope->__pyx_v_remaining, __pyx_int_2};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 417, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_remaining);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_remaining);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_6, __pyx_cur_scope->__pyx_v_remaining);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_6, __pyx_int_2);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_7 = __pyx_t_5; __Pyx_INCREF(__pyx_t_7); __pyx_t_2 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_2 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 417, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 417, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 417, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 417, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 417, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_8(__pyx_t_7);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 417, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 417, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_4);
        #else
        __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_10 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_9 = __pyx_t_10(__pyx_t_1); if (unlikely(!__pyx_t_9)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_1); if (unlikely(!__pyx_t_4)) goto __pyx_L14_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_1), 2) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L15_unpacking_done;
        __pyx_L14_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 417, __pyx_L1_error)
        __pyx_L15_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":418
 *             if not candidates:
 *                 for i, j in itertools.combinations(remaining, 2):
 *                     if i > j:             # <<<<<<<<<<<<<<
 *                         i, j = j, i
 *                     k1, k2 = inputs[i], inputs[j]
 */
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_i, __pyx_v_j, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 418, __pyx_L1_error)
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 418, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_13) {

        /* "acqdp/utility/opt_einsum_paths.pyx":419
 *                 for i, j in itertools.combinations(remaining, 2):
 *                     if i > j:
 *                         i, j = j, i             # <<<<<<<<<<<<<<
 *                     k1, k2 = inputs[i], inputs[j]
 *                     candidate = _assess_candidate(k1, k2, i, j)
 */
        __pyx_t_12 = __pyx_v_j;
        __pyx_t_11 = __pyx_v_i;
        __pyx_v_i = __pyx_t_12;
        __pyx_t_12 = 0;
        __pyx_v_j = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":418
 *             if not candidates:
 *                 for i, j in itertools.combinations(remaining, 2):
 *                     if i > j:             # <<<<<<<<<<<<<<
 *                         i, j = j, i
 *                     k1, k2 = inputs[i], inputs[j]
 */
      }

      /* "acqdp/utility/opt_einsum_paths.pyx":420
 *                     if i > j:
 *                         i, j = j, i
 *                     k1, k2 = inputs[i], inputs[j]             # <<<<<<<<<<<<<<
 *                     candidate = _assess_candidate(k1, k2, i, j)
 *                     if candidate:
 */
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_inputs, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_inputs, __pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_k1, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_k2, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":421
 *                         i, j = j, i
 *                     k1, k2 = inputs[i], inputs[j]
 *                     candidate = _assess_candidate(k1, k2, i, j)             # <<<<<<<<<<<<<<
 *                     if candidate:
 *                         heapq.heappush(candidates, candidate)
 */
      __pyx_t_4 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___15_branch_iterate__assess_candidate(__pyx_v__assess_candidate, __pyx_v_k1, __pyx_v_k2, __pyx_v_i, __pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_candidate, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":422
 *                     k1, k2 = inputs[i], inputs[j]
 *                     candidate = _assess_candidate(k1, k2, i, j)
 *                     if candidate:             # <<<<<<<<<<<<<<
 *                         heapq.heappush(candidates, candidate)
 * 
 */
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_v_candidate); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 422, __pyx_L1_error)
      if (__pyx_t_13) {

        /* "acqdp/utility/opt_einsum_paths.pyx":423
 *                     candidate = _assess_candidate(k1, k2, i, j)
 *                     if candidate:
 *                         heapq.heappush(candidates, candidate)             # <<<<<<<<<<<<<<
 * 
 *             # recurse into all or some of the best candidate contractions
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_heapq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_heappush); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 423, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_6 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_candidates, __pyx_v_candidate};
          __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_candidates, __pyx_v_candidate};
          __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        {
          __pyx_t_1 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_5) {
            __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
          }
          __Pyx_INCREF(__pyx_v_candidates);
          __Pyx_GIVEREF(__pyx_v_candidates);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_6, __pyx_v_candidates);
          __Pyx_INCREF(__pyx_v_candidate);
          __Pyx_GIVEREF(__pyx_v_candidate);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_6, __pyx_v_candidate);
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":422
 *                     k1, k2 = inputs[i], inputs[j]
 *                     candidate = _assess_candidate(k1, k2, i, j)
 *                     if candidate:             # <<<<<<<<<<<<<<
 *                         heapq.heappush(candidates, candidate)
 * 
 */
      }

      /* "acqdp/utility/opt_einsum_paths.pyx":417
 *             # assess outer products if nothing left
 *             if not candidates:
 *                 for i, j in itertools.combinations(remaining, 2):             # <<<<<<<<<<<<<<
 *                     if i > j:
 *                         i, j = j, i
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":416
 * 
 *             # assess outer products if nothing left
 *             if not candidates:             # <<<<<<<<<<<<<<
 *                 for i, j in itertools.combinations(remaining, 2):
 *                     if i > j:
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":426
 * 
 *             # recurse into all or some of the best candidate contractions
 *             bi = 0             # <<<<<<<<<<<<<<
 *             while (self.nbranch is None or bi < self.nbranch) and candidates:
 *                 _, _, new_flops, new_size, (i, j), k12 = heapq.heappop(candidates)
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_bi = __pyx_int_0;

  /* "acqdp/utility/opt_einsum_paths.pyx":427
 *             # recurse into all or some of the best candidate contractions
 *             bi = 0
 *             while (self.nbranch is None or bi < self.nbranch) and candidates:             # <<<<<<<<<<<<<<
 *                 _, _, new_flops, new_size, (i, j), k12 = heapq.heappop(candidates)
 *                 _branch_iterate(path=path + ((i, j),),
 */
  while (1) {
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 427, __pyx_L1_error) }
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_nbranch); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = (__pyx_t_7 == Py_None);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_14 = (__pyx_t_3 != 0);
    if (!__pyx_t_14) {
    } else {
      goto __pyx_L21_next_and;
    }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 427, __pyx_L1_error) }
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self, __pyx_n_s_nbranch); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_bi, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_14) {
    } else {
      __pyx_t_13 = __pyx_t_14;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_L21_next_and:;
    __pyx_t_14 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
    __pyx_t_13 = __pyx_t_14;
    __pyx_L20_bool_binop_done:;
    if (!__pyx_t_13) break;

    /* "acqdp/utility/opt_einsum_paths.pyx":428
 *             bi = 0
 *             while (self.nbranch is None or bi < self.nbranch) and candidates:
 *                 _, _, new_flops, new_size, (i, j), k12 = heapq.heappop(candidates)             # <<<<<<<<<<<<<<
 *                 _branch_iterate(path=path + ((i, j),),
 *                                 inputs=inputs + (k12,),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_heapq); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_heappop); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_7, __pyx_v_candidates) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_candidates);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 6)) {
        if (size > 6) __Pyx_RaiseTooManyValuesError(6);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 428, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
        __pyx_t_15 = PyTuple_GET_ITEM(sequence, 4); 
        __pyx_t_16 = PyTuple_GET_ITEM(sequence, 5); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
        __pyx_t_15 = PyList_GET_ITEM(sequence, 4); 
        __pyx_t_16 = PyList_GET_ITEM(sequence, 5); 
      }
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_16);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[6] = {&__pyx_t_9,&__pyx_t_7,&__pyx_t_1,&__pyx_t_5,&__pyx_t_15,&__pyx_t_16};
        for (i=0; i < 6; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 428, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[6] = {&__pyx_t_9,&__pyx_t_7,&__pyx_t_1,&__pyx_t_5,&__pyx_t_15,&__pyx_t_16};
      __pyx_t_17 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_17)->tp_iternext;
      for (index=0; index < 6; index++) {
        PyObject* item = __pyx_t_10(__pyx_t_17); if (unlikely(!item)) goto __pyx_L23_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_17), 6) < 0) __PYX_ERR(0, 428, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      goto __pyx_L24_unpacking_done;
      __pyx_L23_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 428, __pyx_L1_error)
      __pyx_L24_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_9);
    __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_flops, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_size, __pyx_t_5);
    __pyx_t_5 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_15))) || (PyList_CheckExact(__pyx_t_15))) {
      PyObject* sequence = __pyx_t_15;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 428, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_18 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_17 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_18 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_18);
      #else
      __pyx_t_17 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_18 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      #endif
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_19 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_19)->tp_iternext;
      index = 0; __pyx_t_17 = __pyx_t_10(__pyx_t_19); if (unlikely(!__pyx_t_17)) goto __pyx_L25_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_17);
      index = 1; __pyx_t_18 = __pyx_t_10(__pyx_t_19); if (unlikely(!__pyx_t_18)) goto __pyx_L25_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_18);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_19), 2) < 0) __PYX_ERR(0, 428, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      goto __pyx_L26_unpacking_done;
      __pyx_L25_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 428, __pyx_L1_error)
      __pyx_L26_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_17);
    __pyx_t_17 = 0;
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_18);
    __pyx_t_18 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k12, __pyx_t_16);
    __pyx_t_16 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":429
 *             while (self.nbranch is None or bi < self.nbranch) and candidates:
 *                 _, _, new_flops, new_size, (i, j), k12 = heapq.heappop(candidates)
 *                 _branch_iterate(path=path + ((i, j),),             # <<<<<<<<<<<<<<
 *                                 inputs=inputs + (k12,),
 *                                 remaining=(remaining - {i, j}) | {len(inputs)},
 */
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v__branch_iterate)) { __Pyx_RaiseClosureNameError("_branch_iterate"); __PYX_ERR(0, 429, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_j);
    __Pyx_GIVEREF(__pyx_v_j);
    PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_v_j);
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_16);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_16);
    __pyx_t_16 = 0;
    __pyx_t_16 = PyNumber_Add(__pyx_cur_scope->__pyx_v_path, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_path, __pyx_t_16) < 0) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":430
 *                 _, _, new_flops, new_size, (i, j), k12 = heapq.heappop(candidates)
 *                 _branch_iterate(path=path + ((i, j),),
 *                                 inputs=inputs + (k12,),             # <<<<<<<<<<<<<<
 *                                 remaining=(remaining - {i, j}) | {len(inputs)},
 *                                 flops=new_flops,
 */
    __pyx_t_16 = PyTuple_New(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_v_k12);
    __Pyx_GIVEREF(__pyx_v_k12);
    PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_k12);
    __pyx_t_15 = PyNumber_Add(__pyx_cur_scope->__pyx_v_inputs, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_inputs, __pyx_t_15) < 0) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":431
 *                 _branch_iterate(path=path + ((i, j),),
 *                                 inputs=inputs + (k12,),
 *                                 remaining=(remaining - {i, j}) | {len(inputs)},             # <<<<<<<<<<<<<<
 *                                 flops=new_flops,
 *                                 size=new_size)
 */
    __pyx_t_15 = PySet_New(0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (PySet_Add(__pyx_t_15, __pyx_v_i) < 0) __PYX_ERR(0, 431, __pyx_L1_error)
    if (PySet_Add(__pyx_t_15, __pyx_v_j) < 0) __PYX_ERR(0, 431, __pyx_L1_error)
    __pyx_t_16 = PyNumber_Subtract(__pyx_cur_scope->__pyx_v_remaining, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __pyx_cur_scope->__pyx_v_inputs;
    __Pyx_INCREF(__pyx_t_15);
    __pyx_t_2 = PyObject_Length(__pyx_t_15); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_5 = PySet_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PySet_Add(__pyx_t_5, __pyx_t_15) < 0) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = PyNumber_Or(__pyx_t_16, __pyx_t_5); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_remaining, __pyx_t_15) < 0) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":432
 *                                 inputs=inputs + (k12,),
 *                                 remaining=(remaining - {i, j}) | {len(inputs)},
 *                                 flops=new_flops,             # <<<<<<<<<<<<<<
 *                                 size=new_size)
 *                 bi += 1
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_flops, __pyx_v_new_flops) < 0) __PYX_ERR(0, 429, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":433
 *                                 remaining=(remaining - {i, j}) | {len(inputs)},
 *                                 flops=new_flops,
 *                                 size=new_size)             # <<<<<<<<<<<<<<
 *                 bi += 1
 * 
 */
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_size, __pyx_v_new_size) < 0) __PYX_ERR(0, 429, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":429
 *             while (self.nbranch is None or bi < self.nbranch) and candidates:
 *                 _, _, new_flops, new_size, (i, j), k12 = heapq.heappop(candidates)
 *                 _branch_iterate(path=path + ((i, j),),             # <<<<<<<<<<<<<<
 *                                 inputs=inputs + (k12,),
 *                                 remaining=(remaining - {i, j}) | {len(inputs)},
 */
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_outer_scope->__pyx_v__branch_iterate, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":434
 *                                 flops=new_flops,
 *                                 size=new_size)
 *                 bi += 1             # <<<<<<<<<<<<<<
 * 
 *         _branch_iterate(path=(),
 */
    __pyx_t_15 = __Pyx_PyInt_AddObjC(__pyx_v_bi, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF_SET(__pyx_v_bi, __pyx_t_15);
    __pyx_t_15 = 0;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":350
 *         result_cache = {}
 * 
 *         def _branch_iterate(path, inputs, remaining, flops, size):             # <<<<<<<<<<<<<<
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__._branch_iterate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__assess_candidate);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_k1);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XDECREF(__pyx_v_candidate);
  __Pyx_XDECREF(__pyx_v_bi);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_new_flops);
  __Pyx_XDECREF(__pyx_v_new_size);
  __Pyx_XDECREF(__pyx_v_k12);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":316
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 * 
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_11BranchBound_4__call__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *__pyx_cur_scope;
  PyObject *__pyx_8genexpr2__pyx_v_k = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 316, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_output = __pyx_v_output;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_output);
  __pyx_cur_scope->__pyx_v_size_dict = __pyx_v_size_dict;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size_dict);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size_dict);
  __pyx_cur_scope->__pyx_v_memory_limit = __pyx_v_memory_limit;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_memory_limit);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_memory_limit);
  __Pyx_INCREF(__pyx_v_inputs);

  /* "acqdp/utility/opt_einsum_paths.pyx":344
 *         """
 * 
 *         inputs = tuple(map(frozenset, inputs))             # <<<<<<<<<<<<<<
 *         output = frozenset(output)
 * 
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&PyFrozenSet_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyFrozenSet_Type)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyFrozenSet_Type)));
  __Pyx_INCREF(__pyx_v_inputs);
  __Pyx_GIVEREF(__pyx_v_inputs);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_inputs);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PySequence_Tuple(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_inputs, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":345
 * 
 *         inputs = tuple(map(frozenset, inputs))
 *         output = frozenset(output)             # <<<<<<<<<<<<<<
 * 
 *         size_cache = {k: helpers.compute_size_by_dict(k, size_dict) for k in inputs}
 */
  __pyx_t_1 = __Pyx_PyFrozenSet_New(__pyx_cur_scope->__pyx_v_output); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_output, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":347
 *         output = frozenset(output)
 * 
 *         size_cache = {k: helpers.compute_size_by_dict(k, size_dict) for k in inputs}             # <<<<<<<<<<<<<<
 *         result_cache = {}
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_inputs)) || PyTuple_CheckExact(__pyx_v_inputs)) {
      __pyx_t_2 = __pyx_v_inputs; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_inputs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 347, __pyx_L5_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 347, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 347, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 347, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_k, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_helpers); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 347, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 347, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_8genexpr2__pyx_v_k, __pyx_cur_scope->__pyx_v_size_dict};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L5_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_8genexpr2__pyx_v_k, __pyx_cur_scope->__pyx_v_size_dict};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L5_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 347, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_8genexpr2__pyx_v_k);
        __Pyx_GIVEREF(__pyx_8genexpr2__pyx_v_k);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_8genexpr2__pyx_v_k);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_size_dict);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_size_dict);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_cur_scope->__pyx_v_size_dict);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_k, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 347, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k); __pyx_8genexpr2__pyx_v_k = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k); __pyx_8genexpr2__pyx_v_k = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_size_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":348
 * 
 *         size_cache = {k: helpers.compute_size_by_dict(k, size_dict) for k in inputs}
 *         result_cache = {}             # <<<<<<<<<<<<<<
 * 
 *         def _branch_iterate(path, inputs, remaining, flops, size):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_result_cache = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":350
 *         result_cache = {}
 * 
 *         def _branch_iterate(path, inputs, remaining, flops, size):             # <<<<<<<<<<<<<<
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_8__call___1_branch_iterate, 0, __pyx_n_s_BranchBound___call___locals__bra_2, ((PyObject*)__pyx_cur_scope), __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v__branch_iterate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":436
 *                 bi += 1
 * 
 *         _branch_iterate(path=(),             # <<<<<<<<<<<<<<
 *                         inputs=inputs,
 *                         remaining=set(range(len(inputs))),
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_path, __pyx_empty_tuple) < 0) __PYX_ERR(0, 436, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":437
 * 
 *         _branch_iterate(path=(),
 *                         inputs=inputs,             # <<<<<<<<<<<<<<
 *                         remaining=set(range(len(inputs))),
 *                         flops=0,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_inputs, __pyx_v_inputs) < 0) __PYX_ERR(0, 436, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":438
 *         _branch_iterate(path=(),
 *                         inputs=inputs,
 *                         remaining=set(range(len(inputs))),             # <<<<<<<<<<<<<<
 *                         flops=0,
 *                         size=0)
 */
  __pyx_t_3 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 438, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySet_New(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_remaining, __pyx_t_2) < 0) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_flops, __pyx_int_0) < 0) __PYX_ERR(0, 436, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_size, __pyx_int_0) < 0) __PYX_ERR(0, 436, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":436
 *                 bi += 1
 * 
 *         _branch_iterate(path=(),             # <<<<<<<<<<<<<<
 *                         inputs=inputs,
 *                         remaining=set(range(len(inputs))),
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_cur_scope->__pyx_v__branch_iterate, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":442
 *                         size=0)
 * 
 *         return self.path             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_self, __pyx_n_s_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":316
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.BranchBound.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_inputs);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":445
 * 
 * 
 * def branch(inputs, output, size_dict, memory_limit=None, **optimizer_kwargs):             # <<<<<<<<<<<<<<
 *     optimizer = BranchBound(**optimizer_kwargs)
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_21branch(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_21branch = {"branch", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_21branch, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_21branch(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  PyObject *__pyx_v_memory_limit = 0;
  PyObject *__pyx_v_optimizer_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("branch (wrapper)", 0);
  __pyx_v_optimizer_kwargs = PyDict_New(); if (unlikely(!__pyx_v_optimizer_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_optimizer_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("branch", 0, 3, 4, 1); __PYX_ERR(0, 445, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("branch", 0, 3, 4, 2); __PYX_ERR(0, 445, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_optimizer_kwargs, values, pos_args, "branch") < 0)) __PYX_ERR(0, 445, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
    __pyx_v_size_dict = values[2];
    __pyx_v_memory_limit = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("branch", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 445, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_optimizer_kwargs); __pyx_v_optimizer_kwargs = 0;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.branch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_20branch(__pyx_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit, __pyx_v_optimizer_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_optimizer_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_20branch(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit, PyObject *__pyx_v_optimizer_kwargs) {
  PyObject *__pyx_v_optimizer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("branch", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":446
 * 
 * def branch(inputs, output, size_dict, memory_limit=None, **optimizer_kwargs):
 *     optimizer = BranchBound(**optimizer_kwargs)             # <<<<<<<<<<<<<<
 *     return optimizer(inputs, output, size_dict, memory_limit)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_BranchBound); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_v_optimizer_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_optimizer = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":447
 * def branch(inputs, output, size_dict, memory_limit=None, **optimizer_kwargs):
 *     optimizer = BranchBound(**optimizer_kwargs)
 *     return optimizer(inputs, output, size_dict, memory_limit)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_optimizer);
  __pyx_t_1 = __pyx_v_optimizer; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_inputs);
    __Pyx_GIVEREF(__pyx_v_inputs);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_inputs);
    __Pyx_INCREF(__pyx_v_output);
    __Pyx_GIVEREF(__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_output);
    __Pyx_INCREF(__pyx_v_size_dict);
    __Pyx_GIVEREF(__pyx_v_size_dict);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_size_dict);
    __Pyx_INCREF(__pyx_v_memory_limit);
    __Pyx_GIVEREF(__pyx_v_memory_limit);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_memory_limit);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":445
 * 
 * 
 * def branch(inputs, output, size_dict, memory_limit=None, **optimizer_kwargs):             # <<<<<<<<<<<<<<
 *     optimizer = BranchBound(**optimizer_kwargs)
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.branch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_optimizer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":455
 * 
 * 
 * def _get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn):             # <<<<<<<<<<<<<<
 *     cdef int id1, id2
 *     either = k1 | k2
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_23_get_candidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_23_get_candidate = {"_get_candidate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_23_get_candidate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_23_get_candidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_sizes = 0;
  PyObject *__pyx_v_remaining = 0;
  PyObject *__pyx_v_footprints = 0;
  PyObject *__pyx_v_dim_ref_counts = 0;
  PyObject *__pyx_v_k1 = 0;
  PyObject *__pyx_v_k2 = 0;
  PyObject *__pyx_v_cost_fn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_candidate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_output,&__pyx_n_s_sizes,&__pyx_n_s_remaining,&__pyx_n_s_footprints,&__pyx_n_s_dim_ref_counts,&__pyx_n_s_k1,&__pyx_n_s_k2,&__pyx_n_s_cost_fn,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sizes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, 1); __PYX_ERR(0, 455, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_remaining)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, 2); __PYX_ERR(0, 455, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_footprints)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, 3); __PYX_ERR(0, 455, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dim_ref_counts)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, 4); __PYX_ERR(0, 455, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, 5); __PYX_ERR(0, 455, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, 6); __PYX_ERR(0, 455, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cost_fn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, 7); __PYX_ERR(0, 455, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_get_candidate") < 0)) __PYX_ERR(0, 455, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_output = values[0];
    __pyx_v_sizes = values[1];
    __pyx_v_remaining = values[2];
    __pyx_v_footprints = values[3];
    __pyx_v_dim_ref_counts = values[4];
    __pyx_v_k1 = values[5];
    __pyx_v_k2 = values[6];
    __pyx_v_cost_fn = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_candidate", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 455, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._get_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_22_get_candidate(__pyx_self, __pyx_v_output, __pyx_v_sizes, __pyx_v_remaining, __pyx_v_footprints, __pyx_v_dim_ref_counts, __pyx_v_k1, __pyx_v_k2, __pyx_v_cost_fn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_22_get_candidate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_output, PyObject *__pyx_v_sizes, PyObject *__pyx_v_remaining, PyObject *__pyx_v_footprints, PyObject *__pyx_v_dim_ref_counts, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2, PyObject *__pyx_v_cost_fn) {
  int __pyx_v_id1;
  int __pyx_v_id2;
  PyObject *__pyx_v_either = NULL;
  PyObject *__pyx_v_two = NULL;
  PyObject *__pyx_v_one = NULL;
  PyObject *__pyx_v_k12 = NULL;
  PyObject *__pyx_v_cost = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_candidate", 0);
  __Pyx_INCREF(__pyx_v_k1);
  __Pyx_INCREF(__pyx_v_k2);

  /* "acqdp/utility/opt_einsum_paths.pyx":457
 * def _get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn):
 *     cdef int id1, id2
 *     either = k1 | k2             # <<<<<<<<<<<<<<
 *     two = k1 & k2
 *     one = either - two
 */
  __pyx_t_1 = PyNumber_Or(__pyx_v_k1, __pyx_v_k2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_either = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":458
 *     cdef int id1, id2
 *     either = k1 | k2
 *     two = k1 & k2             # <<<<<<<<<<<<<<
 *     one = either - two
 *     k12 = (either & output) | (two & dim_ref_counts[3]) | (one & dim_ref_counts[2])
 */
  __pyx_t_1 = PyNumber_And(__pyx_v_k1, __pyx_v_k2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_two = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":459
 *     either = k1 | k2
 *     two = k1 & k2
 *     one = either - two             # <<<<<<<<<<<<<<
 *     k12 = (either & output) | (two & dim_ref_counts[3]) | (one & dim_ref_counts[2])
 *     cost = cost_fn(helpers.compute_size_by_dict(k12, sizes), footprints[k1], footprints[k2], k12, k1, k2)
 */
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_either, __pyx_v_two); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_one = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":460
 *     two = k1 & k2
 *     one = either - two
 *     k12 = (either & output) | (two & dim_ref_counts[3]) | (one & dim_ref_counts[2])             # <<<<<<<<<<<<<<
 *     cost = cost_fn(helpers.compute_size_by_dict(k12, sizes), footprints[k1], footprints[k2], k12, k1, k2)
 *     id1 = remaining[k1]
 */
  __pyx_t_1 = PyNumber_And(__pyx_v_either, __pyx_v_output); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_And(__pyx_v_two, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Or(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_And(__pyx_v_one, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Or(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_k12 = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":461
 *     one = either - two
 *     k12 = (either & output) | (two & dim_ref_counts[3]) | (one & dim_ref_counts[2])
 *     cost = cost_fn(helpers.compute_size_by_dict(k12, sizes), footprints[k1], footprints[k2], k12, k1, k2)             # <<<<<<<<<<<<<<
 *     id1 = remaining[k1]
 *     id2 = remaining[k2]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_helpers); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_k12, __pyx_v_sizes};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_v_k12, __pyx_v_sizes};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_v_k12);
    __Pyx_GIVEREF(__pyx_v_k12);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_k12);
    __Pyx_INCREF(__pyx_v_sizes);
    __Pyx_GIVEREF(__pyx_v_sizes);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_sizes);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_footprints, __pyx_v_k1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_footprints, __pyx_v_k2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_cost_fn);
  __pyx_t_2 = __pyx_v_cost_fn; __pyx_t_7 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[7] = {__pyx_t_7, __pyx_t_1, __pyx_t_4, __pyx_t_6, __pyx_v_k12, __pyx_v_k1, __pyx_v_k2};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 6+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[7] = {__pyx_t_7, __pyx_t_1, __pyx_t_4, __pyx_t_6, __pyx_v_k12, __pyx_v_k1, __pyx_v_k2};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 6+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(6+__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_5, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_5, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_5, __pyx_t_6);
    __Pyx_INCREF(__pyx_v_k12);
    __Pyx_GIVEREF(__pyx_v_k12);
    PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_5, __pyx_v_k12);
    __Pyx_INCREF(__pyx_v_k1);
    __Pyx_GIVEREF(__pyx_v_k1);
    PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_5, __pyx_v_k1);
    __Pyx_INCREF(__pyx_v_k2);
    __Pyx_GIVEREF(__pyx_v_k2);
    PyTuple_SET_ITEM(__pyx_t_8, 5+__pyx_t_5, __pyx_v_k2);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cost = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":462
 *     k12 = (either & output) | (two & dim_ref_counts[3]) | (one & dim_ref_counts[2])
 *     cost = cost_fn(helpers.compute_size_by_dict(k12, sizes), footprints[k1], footprints[k2], k12, k1, k2)
 *     id1 = remaining[k1]             # <<<<<<<<<<<<<<
 *     id2 = remaining[k2]
 *     if id1 > id2:
 */
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_remaining, __pyx_v_k1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_id1 = __pyx_t_5;

  /* "acqdp/utility/opt_einsum_paths.pyx":463
 *     cost = cost_fn(helpers.compute_size_by_dict(k12, sizes), footprints[k1], footprints[k2], k12, k1, k2)
 *     id1 = remaining[k1]
 *     id2 = remaining[k2]             # <<<<<<<<<<<<<<
 *     if id1 > id2:
 *         k1, id1, k2, id2 = k2, id2, k1, id1
 */
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_remaining, __pyx_v_k2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_id2 = __pyx_t_5;

  /* "acqdp/utility/opt_einsum_paths.pyx":464
 *     id1 = remaining[k1]
 *     id2 = remaining[k2]
 *     if id1 > id2:             # <<<<<<<<<<<<<<
 *         k1, id1, k2, id2 = k2, id2, k1, id1
 *     return (cost, id2, id1), k1, k2, k12  # break ties to ensure determinism
 */
  __pyx_t_9 = ((__pyx_v_id1 > __pyx_v_id2) != 0);
  if (__pyx_t_9) {

    /* "acqdp/utility/opt_einsum_paths.pyx":465
 *     id2 = remaining[k2]
 *     if id1 > id2:
 *         k1, id1, k2, id2 = k2, id2, k1, id1             # <<<<<<<<<<<<<<
 *     return (cost, id2, id1), k1, k2, k12  # break ties to ensure determinism
 * 
 */
    __pyx_t_3 = __pyx_v_k2;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = __pyx_v_id2;
    __pyx_t_2 = __pyx_v_k1;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_10 = __pyx_v_id1;
    __Pyx_DECREF_SET(__pyx_v_k1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_v_id1 = __pyx_t_5;
    __Pyx_DECREF_SET(__pyx_v_k2, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_v_id2 = __pyx_t_10;

    /* "acqdp/utility/opt_einsum_paths.pyx":464
 *     id1 = remaining[k1]
 *     id2 = remaining[k2]
 *     if id1 > id2:             # <<<<<<<<<<<<<<
 *         k1, id1, k2, id2 = k2, id2, k1, id1
 *     return (cost, id2, id1), k1, k2, k12  # break ties to ensure determinism
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":466
 *     if id1 > id2:
 *         k1, id1, k2, id2 = k2, id2, k1, id1
 *     return (cost, id2, id1), k1, k2, k12  # break ties to ensure determinism             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_id2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_id1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_cost);
  __Pyx_GIVEREF(__pyx_v_cost);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_cost);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8);
  __Pyx_INCREF(__pyx_v_k1);
  __Pyx_GIVEREF(__pyx_v_k1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_k1);
  __Pyx_INCREF(__pyx_v_k2);
  __Pyx_GIVEREF(__pyx_v_k2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_k2);
  __Pyx_INCREF(__pyx_v_k12);
  __Pyx_GIVEREF(__pyx_v_k12);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_k12);
  __pyx_t_8 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":455
 * 
 * 
 * def _get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn):             # <<<<<<<<<<<<<<
 *     cdef int id1, id2
 *     either = k1 | k2
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._get_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_either);
  __Pyx_XDECREF(__pyx_v_two);
  __Pyx_XDECREF(__pyx_v_one);
  __Pyx_XDECREF(__pyx_v_k12);
  __Pyx_XDECREF(__pyx_v_cost);
  __Pyx_XDECREF(__pyx_v_k1);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":469
 * 
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):             # <<<<<<<<<<<<<<
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_25_push_candidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_25_push_candidate = {"_push_candidate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_25_push_candidate, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_25_push_candidate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_sizes = 0;
  PyObject *__pyx_v_remaining = 0;
  PyObject *__pyx_v_footprints = 0;
  PyObject *__pyx_v_dim_ref_counts = 0;
  PyObject *__pyx_v_k1 = 0;
  PyObject *__pyx_v_k2s = 0;
  PyObject *__pyx_v_queue = 0;
  PyObject *__pyx_v_push_all = 0;
  PyObject *__pyx_v_cost_fn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_push_candidate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_output,&__pyx_n_s_sizes,&__pyx_n_s_remaining,&__pyx_n_s_footprints,&__pyx_n_s_dim_ref_counts,&__pyx_n_s_k1,&__pyx_n_s_k2s,&__pyx_n_s_queue,&__pyx_n_s_push_all,&__pyx_n_s_cost_fn,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sizes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 1); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_remaining)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 2); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_footprints)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 3); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dim_ref_counts)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 4); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 5); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k2s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 6); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_queue)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 7); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_push_all)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 8); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cost_fn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, 9); __PYX_ERR(0, 469, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_push_candidate") < 0)) __PYX_ERR(0, 469, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_output = values[0];
    __pyx_v_sizes = values[1];
    __pyx_v_remaining = values[2];
    __pyx_v_footprints = values[3];
    __pyx_v_dim_ref_counts = values[4];
    __pyx_v_k1 = values[5];
    __pyx_v_k2s = values[6];
    __pyx_v_queue = values[7];
    __pyx_v_push_all = values[8];
    __pyx_v_cost_fn = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_push_candidate", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 469, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._push_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_24_push_candidate(__pyx_self, __pyx_v_output, __pyx_v_sizes, __pyx_v_remaining, __pyx_v_footprints, __pyx_v_dim_ref_counts, __pyx_v_k1, __pyx_v_k2s, __pyx_v_queue, __pyx_v_push_all, __pyx_v_cost_fn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_15_push_candidate_2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":470
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)             # <<<<<<<<<<<<<<
 *     if push_all:
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_15_push_candidate_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 470, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_15_push_candidate_2generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_push_candidate_locals_genexpr, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._push_candidate.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_15_push_candidate_2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 470, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k2s)) { __Pyx_RaiseClosureNameError("k2s"); __PYX_ERR(0, 470, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k2s)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k2s)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_k2s; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k2s); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 470, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 470, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 470, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 470, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 470, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k2);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k2, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_get_candidate); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output)) { __Pyx_RaiseClosureNameError("output"); __PYX_ERR(0, 470, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_sizes)) { __Pyx_RaiseClosureNameError("sizes"); __PYX_ERR(0, 470, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_remaining)) { __Pyx_RaiseClosureNameError("remaining"); __PYX_ERR(0, 470, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_footprints)) { __Pyx_RaiseClosureNameError("footprints"); __PYX_ERR(0, 470, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_ref_counts)) { __Pyx_RaiseClosureNameError("dim_ref_counts"); __PYX_ERR(0, 470, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1)) { __Pyx_RaiseClosureNameError("k1"); __PYX_ERR(0, 470, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_cost_fn)) { __Pyx_RaiseClosureNameError("cost_fn"); __PYX_ERR(0, 470, __pyx_L1_error) }
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[9] = {__pyx_t_6, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_sizes, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_remaining, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_footprints, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_ref_counts, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1, __pyx_cur_scope->__pyx_v_k2, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_cost_fn};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 8+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 470, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[9] = {__pyx_t_6, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_sizes, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_remaining, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_footprints, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_ref_counts, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1, __pyx_cur_scope->__pyx_v_k2, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_cost_fn};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 8+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 470, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(8+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 470, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_sizes);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_sizes);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_sizes);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_remaining);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_remaining);
      PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_remaining);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_footprints);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_footprints);
      PyTuple_SET_ITEM(__pyx_t_8, 3+__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_footprints);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_ref_counts);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_ref_counts);
      PyTuple_SET_ITEM(__pyx_t_8, 4+__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_ref_counts);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1);
      PyTuple_SET_ITEM(__pyx_t_8, 5+__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k2);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k2);
      PyTuple_SET_ITEM(__pyx_t_8, 6+__pyx_t_7, __pyx_cur_scope->__pyx_v_k2);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_cost_fn);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_cost_fn);
      PyTuple_SET_ITEM(__pyx_t_8, 7+__pyx_t_7, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_cost_fn);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 470, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 470, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":469
 * 
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):             # <<<<<<<<<<<<<<
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_24_push_candidate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_output, PyObject *__pyx_v_sizes, PyObject *__pyx_v_remaining, PyObject *__pyx_v_footprints, PyObject *__pyx_v_dim_ref_counts, PyObject *__pyx_v_k1, PyObject *__pyx_v_k2s, PyObject *__pyx_v_queue, PyObject *__pyx_v_push_all, PyObject *__pyx_v_cost_fn) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *__pyx_cur_scope;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_candidate = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_push_candidate", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 469, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_output = __pyx_v_output;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_output);
  __pyx_cur_scope->__pyx_v_sizes = __pyx_v_sizes;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_sizes);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_sizes);
  __pyx_cur_scope->__pyx_v_remaining = __pyx_v_remaining;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_remaining);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_remaining);
  __pyx_cur_scope->__pyx_v_footprints = __pyx_v_footprints;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_footprints);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_footprints);
  __pyx_cur_scope->__pyx_v_dim_ref_counts = __pyx_v_dim_ref_counts;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_dim_ref_counts);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_dim_ref_counts);
  __pyx_cur_scope->__pyx_v_k1 = __pyx_v_k1;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k1);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k1);
  __pyx_cur_scope->__pyx_v_k2s = __pyx_v_k2s;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k2s);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k2s);
  __pyx_cur_scope->__pyx_v_cost_fn = __pyx_v_cost_fn;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_cost_fn);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_cost_fn);

  /* "acqdp/utility/opt_einsum_paths.pyx":470
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)             # <<<<<<<<<<<<<<
 *     if push_all:
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 */
  __pyx_t_1 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_15_push_candidate_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_candidates = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":471
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:             # <<<<<<<<<<<<<<
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 *         for candidate in candidates:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_push_all); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 471, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "acqdp/utility/opt_einsum_paths.pyx":473
 *     if push_all:
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 *         for candidate in candidates:             # <<<<<<<<<<<<<<
 *             heapq.heappush(queue, candidate)
 *     else:
 */
    if (likely(PyList_CheckExact(__pyx_v_candidates)) || PyTuple_CheckExact(__pyx_v_candidates)) {
      __pyx_t_1 = __pyx_v_candidates; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_candidates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 473, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_4(__pyx_t_1);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 473, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_candidate, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":474
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 *         for candidate in candidates:
 *             heapq.heappush(queue, candidate)             # <<<<<<<<<<<<<<
 *     else:
 *         heapq.heappush(queue, min(candidates))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_heapq); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_heappush); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_queue, __pyx_v_candidate};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_queue, __pyx_v_candidate};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_INCREF(__pyx_v_queue);
        __Pyx_GIVEREF(__pyx_v_queue);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_queue);
        __Pyx_INCREF(__pyx_v_candidate);
        __Pyx_GIVEREF(__pyx_v_candidate);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_candidate);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":473
 *     if push_all:
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 *         for candidate in candidates:             # <<<<<<<<<<<<<<
 *             heapq.heappush(queue, candidate)
 *     else:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":471
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:             # <<<<<<<<<<<<<<
 *         # want to do this if we e.g. are using a custom 'choose_fn'
 *         for candidate in candidates:
 */
    goto __pyx_L3;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":476
 *             heapq.heappush(queue, candidate)
 *     else:
 *         heapq.heappush(queue, min(candidates))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_heapq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_heappush); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_v_candidates); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_queue, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_queue, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(__pyx_v_queue);
      __Pyx_GIVEREF(__pyx_v_queue);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_8, __pyx_v_queue);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_8, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "acqdp/utility/opt_einsum_paths.pyx":469
 * 
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):             # <<<<<<<<<<<<<<
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._push_candidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_candidate);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":479
 * 
 * 
 * def _update_ref_counts(dim_to_keys, dim_ref_counts, dims):             # <<<<<<<<<<<<<<
 *     for dim in dims:
 *         count = len(dim_to_keys[dim])
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27_update_ref_counts(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27_update_ref_counts = {"_update_ref_counts", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27_update_ref_counts, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27_update_ref_counts(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dim_to_keys = 0;
  PyObject *__pyx_v_dim_ref_counts = 0;
  PyObject *__pyx_v_dims = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update_ref_counts (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dim_to_keys,&__pyx_n_s_dim_ref_counts,&__pyx_n_s_dims,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dim_to_keys)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dim_ref_counts)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_ref_counts", 1, 3, 3, 1); __PYX_ERR(0, 479, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dims)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_update_ref_counts", 1, 3, 3, 2); __PYX_ERR(0, 479, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_update_ref_counts") < 0)) __PYX_ERR(0, 479, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_dim_to_keys = values[0];
    __pyx_v_dim_ref_counts = values[1];
    __pyx_v_dims = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_update_ref_counts", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 479, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._update_ref_counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_26_update_ref_counts(__pyx_self, __pyx_v_dim_to_keys, __pyx_v_dim_ref_counts, __pyx_v_dims);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_26_update_ref_counts(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dim_to_keys, PyObject *__pyx_v_dim_ref_counts, PyObject *__pyx_v_dims) {
  PyObject *__pyx_v_dim = NULL;
  Py_ssize_t __pyx_v_count;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update_ref_counts", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":480
 * 
 * def _update_ref_counts(dim_to_keys, dim_ref_counts, dims):
 *     for dim in dims:             # <<<<<<<<<<<<<<
 *         count = len(dim_to_keys[dim])
 *         if count <= 1:
 */
  if (likely(PyList_CheckExact(__pyx_v_dims)) || PyTuple_CheckExact(__pyx_v_dims)) {
    __pyx_t_1 = __pyx_v_dims; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_dims); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 480, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 480, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 480, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 480, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_dim, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":481
 * def _update_ref_counts(dim_to_keys, dim_ref_counts, dims):
 *     for dim in dims:
 *         count = len(dim_to_keys[dim])             # <<<<<<<<<<<<<<
 *         if count <= 1:
 *             dim_ref_counts[2].discard(dim)
 */
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_dim_to_keys, __pyx_v_dim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_count = __pyx_t_5;

    /* "acqdp/utility/opt_einsum_paths.pyx":482
 *     for dim in dims:
 *         count = len(dim_to_keys[dim])
 *         if count <= 1:             # <<<<<<<<<<<<<<
 *             dim_ref_counts[2].discard(dim)
 *             dim_ref_counts[3].discard(dim)
 */
    __pyx_t_6 = ((__pyx_v_count <= 1) != 0);
    if (__pyx_t_6) {

      /* "acqdp/utility/opt_einsum_paths.pyx":483
 *         count = len(dim_to_keys[dim])
 *         if count <= 1:
 *             dim_ref_counts[2].discard(dim)             # <<<<<<<<<<<<<<
 *             dim_ref_counts[3].discard(dim)
 *         elif count == 2:
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_discard); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_dim) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_dim);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":484
 *         if count <= 1:
 *             dim_ref_counts[2].discard(dim)
 *             dim_ref_counts[3].discard(dim)             # <<<<<<<<<<<<<<
 *         elif count == 2:
 *             dim_ref_counts[2].add(dim)
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_discard); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_dim) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_dim);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":482
 *     for dim in dims:
 *         count = len(dim_to_keys[dim])
 *         if count <= 1:             # <<<<<<<<<<<<<<
 *             dim_ref_counts[2].discard(dim)
 *             dim_ref_counts[3].discard(dim)
 */
      goto __pyx_L5;
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":485
 *             dim_ref_counts[2].discard(dim)
 *             dim_ref_counts[3].discard(dim)
 *         elif count == 2:             # <<<<<<<<<<<<<<
 *             dim_ref_counts[2].add(dim)
 *             dim_ref_counts[3].discard(dim)
 */
    __pyx_t_6 = ((__pyx_v_count == 2) != 0);
    if (__pyx_t_6) {

      /* "acqdp/utility/opt_einsum_paths.pyx":486
 *             dim_ref_counts[3].discard(dim)
 *         elif count == 2:
 *             dim_ref_counts[2].add(dim)             # <<<<<<<<<<<<<<
 *             dim_ref_counts[3].discard(dim)
 *         else:
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_add); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_dim) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_dim);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":487
 *         elif count == 2:
 *             dim_ref_counts[2].add(dim)
 *             dim_ref_counts[3].discard(dim)             # <<<<<<<<<<<<<<
 *         else:
 *             dim_ref_counts[2].add(dim)
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_discard); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_dim) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_dim);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 487, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":485
 *             dim_ref_counts[2].discard(dim)
 *             dim_ref_counts[3].discard(dim)
 *         elif count == 2:             # <<<<<<<<<<<<<<
 *             dim_ref_counts[2].add(dim)
 *             dim_ref_counts[3].discard(dim)
 */
      goto __pyx_L5;
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":489
 *             dim_ref_counts[3].discard(dim)
 *         else:
 *             dim_ref_counts[2].add(dim)             # <<<<<<<<<<<<<<
 *             dim_ref_counts[3].add(dim)
 * 
 */
    /*else*/ {
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_add); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_v_dim) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_dim);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":490
 *         else:
 *             dim_ref_counts[2].add(dim)
 *             dim_ref_counts[3].add(dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_dim_ref_counts, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_add); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_4 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_dim) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_dim);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L5:;

    /* "acqdp/utility/opt_einsum_paths.pyx":480
 * 
 * def _update_ref_counts(dim_to_keys, dim_ref_counts, dims):
 *     for dim in dims:             # <<<<<<<<<<<<<<
 *         count = len(dim_to_keys[dim])
 *         if count <= 1:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":479
 * 
 * 
 * def _update_ref_counts(dim_to_keys, dim_ref_counts, dims):             # <<<<<<<<<<<<<<
 *     for dim in dims:
 *         count = len(dim_to_keys[dim])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._update_ref_counts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dim);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":493
 * 
 * 
 * def _simple_chooser(queue, remaining):             # <<<<<<<<<<<<<<
 *     """Default contraction chooser that simply takes the minimum cost option.
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_29_simple_chooser(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_28_simple_chooser[] = "Default contraction chooser that simply takes the minimum cost option.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_29_simple_chooser = {"_simple_chooser", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_29_simple_chooser, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_28_simple_chooser};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_29_simple_chooser(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_queue = 0;
  PyObject *__pyx_v_remaining = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_simple_chooser (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_queue,&__pyx_n_s_remaining,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_queue)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_remaining)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_simple_chooser", 1, 2, 2, 1); __PYX_ERR(0, 493, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_simple_chooser") < 0)) __PYX_ERR(0, 493, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_queue = values[0];
    __pyx_v_remaining = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_simple_chooser", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 493, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._simple_chooser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_28_simple_chooser(__pyx_self, __pyx_v_queue, __pyx_v_remaining);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_28_simple_chooser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_queue, PyObject *__pyx_v_remaining) {
  PyObject *__pyx_v_cost = NULL;
  PyObject *__pyx_v_k1 = NULL;
  PyObject *__pyx_v_k2 = NULL;
  PyObject *__pyx_v_k12 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_simple_chooser", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":496
 *     """Default contraction chooser that simply takes the minimum cost option.
 *     """
 *     cost, k1, k2, k12 = heapq.heappop(queue)             # <<<<<<<<<<<<<<
 *     if k1 not in remaining or k2 not in remaining:
 *         return None  # candidate is obsolete
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_heapq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_heappop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_queue) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_queue);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 496, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 496, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_3,&__pyx_t_2,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 4) < 0) __PYX_ERR(0, 496, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 496, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_cost = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_k1 = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_k2 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_k12 = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":497
 *     """
 *     cost, k1, k2, k12 = heapq.heappop(queue)
 *     if k1 not in remaining or k2 not in remaining:             # <<<<<<<<<<<<<<
 *         return None  # candidate is obsolete
 *     return cost, k1, k2, k12
 */
  __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_v_k1, __pyx_v_remaining, Py_NE)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_t_10 = (__pyx_t_9 != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_8 = __pyx_t_10;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_10 = (__Pyx_PySequence_ContainsTF(__pyx_v_k2, __pyx_v_remaining, Py_NE)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_10 != 0);
  __pyx_t_8 = __pyx_t_9;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_8) {

    /* "acqdp/utility/opt_einsum_paths.pyx":498
 *     cost, k1, k2, k12 = heapq.heappop(queue)
 *     if k1 not in remaining or k2 not in remaining:
 *         return None  # candidate is obsolete             # <<<<<<<<<<<<<<
 *     return cost, k1, k2, k12
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":497
 *     """
 *     cost, k1, k2, k12 = heapq.heappop(queue)
 *     if k1 not in remaining or k2 not in remaining:             # <<<<<<<<<<<<<<
 *         return None  # candidate is obsolete
 *     return cost, k1, k2, k12
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":499
 *     if k1 not in remaining or k2 not in remaining:
 *         return None  # candidate is obsolete
 *     return cost, k1, k2, k12             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_cost);
  __Pyx_GIVEREF(__pyx_v_cost);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_cost);
  __Pyx_INCREF(__pyx_v_k1);
  __Pyx_GIVEREF(__pyx_v_k1);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_k1);
  __Pyx_INCREF(__pyx_v_k2);
  __Pyx_GIVEREF(__pyx_v_k2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_k2);
  __Pyx_INCREF(__pyx_v_k12);
  __Pyx_GIVEREF(__pyx_v_k12);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_k12);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":493
 * 
 * 
 * def _simple_chooser(queue, remaining):             # <<<<<<<<<<<<<<
 *     """Default contraction chooser that simply takes the minimum cost option.
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._simple_chooser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cost);
  __Pyx_XDECREF(__pyx_v_k1);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XDECREF(__pyx_v_k12);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":502
 * 
 * 
 * def ssa_greedy_optimize(inputs, output, sizes, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     This is the core function for :func:`greedy` but produces a path with
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_31ssa_greedy_optimize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_30ssa_greedy_optimize[] = "\n    This is the core function for :func:`greedy` but produces a path with\n    static single assignment ids rather than recycled linear ids.\n    SSA ids are cheaper to work with and easier to reason about.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_31ssa_greedy_optimize = {"ssa_greedy_optimize", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_31ssa_greedy_optimize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_30ssa_greedy_optimize};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_31ssa_greedy_optimize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_sizes = 0;
  PyObject *__pyx_v_choose_fn = 0;
  PyObject *__pyx_v_cost_fn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ssa_greedy_optimize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_sizes,&__pyx_n_s_choose_fn,&__pyx_n_s_cost_fn,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)__pyx_kp_s_memory_removed);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssa_greedy_optimize", 0, 3, 5, 1); __PYX_ERR(0, 502, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sizes)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssa_greedy_optimize", 0, 3, 5, 2); __PYX_ERR(0, 502, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_choose_fn);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cost_fn);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ssa_greedy_optimize") < 0)) __PYX_ERR(0, 502, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
    __pyx_v_sizes = values[2];
    __pyx_v_choose_fn = values[3];
    __pyx_v_cost_fn = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ssa_greedy_optimize", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 502, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.ssa_greedy_optimize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_30ssa_greedy_optimize(__pyx_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_sizes, __pyx_v_choose_fn, __pyx_v_cost_fn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_8genexpr4_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":551
 *     # ref counts of >=2 or >=3.
 *     dim_ref_counts = {
 *         count: set(dim for dim, keys in dim_to_keys.items() if len(keys) >= count) - output             # <<<<<<<<<<<<<<
 *         for count in [2, 3]}
 * 
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_8genexpr4_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 551, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_8genexpr4_2generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ssa_greedy_optimize_locals_genex, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.ssa_greedy_optimize.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_8genexpr4_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 551, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_to_keys)) { __Pyx_RaiseClosureNameError("dim_to_keys"); __PYX_ERR(0, 551, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_to_keys, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 551, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 551, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 551, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 551, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 551, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 551, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_dim);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_dim, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_keys);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_keys, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_9 = PyObject_Length(__pyx_cur_scope->__pyx_v_keys); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 551, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_8genexpr4__pyx_v_count)) { __Pyx_RaiseClosureNameError("count"); __PYX_ERR(0, 551, __pyx_L1_error) }
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_cur_scope->__pyx_outer_scope->__pyx_8genexpr4__pyx_v_count, Py_GE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_10) {
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_dim))) __PYX_ERR(0, 551, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":591
 *         # Find new candidate contractions.
 *         k1 = k12
 *         k2s = set(k2 for dim in k1 for k2 in dim_to_keys[dim])             # <<<<<<<<<<<<<<
 *         k2s.discard(k1)
 *         if k2s:
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 591, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_2generator5, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_ssa_greedy_optimize_locals_genex, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.ssa_greedy_optimize.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 591, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1)) { __Pyx_RaiseClosureNameError("k1"); __PYX_ERR(0, 591, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_k1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 591, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 591, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 591, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_dim);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_dim, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_to_keys)) { __Pyx_RaiseClosureNameError("dim_to_keys"); __PYX_ERR(0, 591, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_dim_to_keys, __pyx_cur_scope->__pyx_v_dim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 591, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 591, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 591, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 591, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k2);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k2, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_k2))) __PYX_ERR(0, 591, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":502
 * 
 * 
 * def ssa_greedy_optimize(inputs, output, sizes, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     This is the core function for :func:`greedy` but produces a path with
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_30ssa_greedy_optimize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_sizes, PyObject *__pyx_v_choose_fn, PyObject *__pyx_v_cost_fn) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *__pyx_cur_scope;
  int __pyx_v_push_all;
  PyObject *__pyx_v_remaining = NULL;
  PyObject *__pyx_v_ssa_ids = NULL;
  PyObject *__pyx_v_ssa_path = NULL;
  PyObject *__pyx_v_ssa_id = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_dim = NULL;
  PyObject *__pyx_v_dim_ref_counts = NULL;
  PyObject *__pyx_v_footprints = NULL;
  PyObject *__pyx_v_queue = NULL;
  PyObject *__pyx_v_keys = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_k2s = NULL;
  PyObject *__pyx_v_con = NULL;
  PyObject *__pyx_v_cost = NULL;
  PyObject *__pyx_v_k2 = NULL;
  PyObject *__pyx_v_k12 = NULL;
  PyObject *__pyx_v_ssa_id1 = NULL;
  PyObject *__pyx_v_ssa_id2 = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_ssa_id12 = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_key = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *(*__pyx_t_17)(PyObject *);
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ssa_greedy_optimize", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 502, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_INCREF(__pyx_v_inputs);
  __Pyx_INCREF(__pyx_v_output);
  __Pyx_INCREF(__pyx_v_choose_fn);
  __Pyx_INCREF(__pyx_v_cost_fn);

  /* "acqdp/utility/opt_einsum_paths.pyx":508
 *     SSA ids are cheaper to work with and easier to reason about.
 *     """
 *     if len(inputs) == 1:             # <<<<<<<<<<<<<<
 *         # Perform a single contraction to match output shape.
 *         return [(0,)]
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 508, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 == 1) != 0);
  if (__pyx_t_2) {

    /* "acqdp/utility/opt_einsum_paths.pyx":510
 *     if len(inputs) == 1:
 *         # Perform a single contraction to match output shape.
 *         return [(0,)]             # <<<<<<<<<<<<<<
 * 
 *     # set the function that assigns a heuristic cost to a possible contraction
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_tuple__9);
    __Pyx_GIVEREF(__pyx_tuple__9);
    PyList_SET_ITEM(__pyx_t_3, 0, __pyx_tuple__9);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":508
 *     SSA ids are cheaper to work with and easier to reason about.
 *     """
 *     if len(inputs) == 1:             # <<<<<<<<<<<<<<
 *         # Perform a single contraction to match output shape.
 *         return [(0,)]
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":513
 * 
 *     # set the function that assigns a heuristic cost to a possible contraction
 *     cost_fn = _COST_FNS.get(cost_fn, cost_fn)             # <<<<<<<<<<<<<<
 * 
 *     # set the function that chooses which contraction to take
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_COST_FNS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_cost_fn, __pyx_v_cost_fn};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_cost_fn, __pyx_v_cost_fn};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cost_fn);
    __Pyx_GIVEREF(__pyx_v_cost_fn);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_cost_fn);
    __Pyx_INCREF(__pyx_v_cost_fn);
    __Pyx_GIVEREF(__pyx_v_cost_fn);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_cost_fn);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_cost_fn, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":516
 * 
 *     # set the function that chooses which contraction to take
 *     if choose_fn is None:             # <<<<<<<<<<<<<<
 *         choose_fn = _simple_chooser
 *         push_all = False
 */
  __pyx_t_2 = (__pyx_v_choose_fn == Py_None);
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "acqdp/utility/opt_einsum_paths.pyx":517
 *     # set the function that chooses which contraction to take
 *     if choose_fn is None:
 *         choose_fn = _simple_chooser             # <<<<<<<<<<<<<<
 *         push_all = False
 *     else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_simple_chooser); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_choose_fn, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":518
 *     if choose_fn is None:
 *         choose_fn = _simple_chooser
 *         push_all = False             # <<<<<<<<<<<<<<
 *     else:
 *         # assume chooser wants access to all possible contractions
 */
    __pyx_v_push_all = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":516
 * 
 *     # set the function that chooses which contraction to take
 *     if choose_fn is None:             # <<<<<<<<<<<<<<
 *         choose_fn = _simple_chooser
 *         push_all = False
 */
    goto __pyx_L4;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":521
 *     else:
 *         # assume chooser wants access to all possible contractions
 *         push_all = True             # <<<<<<<<<<<<<<
 * 
 *     # A dim that is common to all tensors might as well be an output dim, since it
 */
  /*else*/ {
    __pyx_v_push_all = 1;
  }
  __pyx_L4:;

  /* "acqdp/utility/opt_einsum_paths.pyx":527
 *     # comparison to search for possible contractions at each step, leading to speedup
 *     # in many practical problems where all tensors share a common batch dimension.
 *     inputs = list(map(frozenset, inputs))             # <<<<<<<<<<<<<<
 *     output = frozenset(output) | frozenset.intersection(*inputs)
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)(&PyFrozenSet_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyFrozenSet_Type)));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyFrozenSet_Type)));
  __Pyx_INCREF(__pyx_v_inputs);
  __Pyx_GIVEREF(__pyx_v_inputs);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_inputs);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PySequence_List(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_inputs, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":528
 *     # in many practical problems where all tensors share a common batch dimension.
 *     inputs = list(map(frozenset, inputs))
 *     output = frozenset(output) | frozenset.intersection(*inputs)             # <<<<<<<<<<<<<<
 * 
 *     # Deduplicate shapes by eagerly computing Hadamard products.
 */
  __pyx_t_3 = __Pyx_PyFrozenSet_New(__pyx_v_output); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyFrozenSet_Type)), __pyx_n_s_intersection); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PySequence_Tuple(__pyx_v_inputs); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyNumber_Or(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_output, __pyx_t_7);
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":531
 * 
 *     # Deduplicate shapes by eagerly computing Hadamard products.
 *     remaining = {}  # key -> ssa_id             # <<<<<<<<<<<<<<
 *     ssa_ids = itertools.count(len(inputs))
 *     ssa_path = []
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_remaining = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":532
 *     # Deduplicate shapes by eagerly computing Hadamard products.
 *     remaining = {}  # key -> ssa_id
 *     ssa_ids = itertools.count(len(inputs))             # <<<<<<<<<<<<<<
 *     ssa_path = []
 *     for ssa_id, key in enumerate(inputs):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_itertools); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 532, __pyx_L1_error)
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ssa_ids = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":533
 *     remaining = {}  # key -> ssa_id
 *     ssa_ids = itertools.count(len(inputs))
 *     ssa_path = []             # <<<<<<<<<<<<<<
 *     for ssa_id, key in enumerate(inputs):
 *         if key in remaining:
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_ssa_path = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":534
 *     ssa_ids = itertools.count(len(inputs))
 *     ssa_path = []
 *     for ssa_id, key in enumerate(inputs):             # <<<<<<<<<<<<<<
 *         if key in remaining:
 *             ssa_path.append((remaining[key], ssa_id))
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_7 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_inputs)) || PyTuple_CheckExact(__pyx_v_inputs)) {
    __pyx_t_3 = __pyx_v_inputs; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_inputs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 534, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_4); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 534, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 534, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_4); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 534, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 534, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_9(__pyx_t_3);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 534, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_ssa_id, __pyx_t_7);
    __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_7, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7);
    __pyx_t_7 = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":535
 *     ssa_path = []
 *     for ssa_id, key in enumerate(inputs):
 *         if key in remaining:             # <<<<<<<<<<<<<<
 *             ssa_path.append((remaining[key], ssa_id))
 *             remaining[key] = next(ssa_ids)
 */
    __pyx_t_8 = (__Pyx_PyDict_ContainsTF(__pyx_v_key, __pyx_v_remaining, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 535, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_8 != 0);
    if (__pyx_t_2) {

      /* "acqdp/utility/opt_einsum_paths.pyx":536
 *     for ssa_id, key in enumerate(inputs):
 *         if key in remaining:
 *             ssa_path.append((remaining[key], ssa_id))             # <<<<<<<<<<<<<<
 *             remaining[key] = next(ssa_ids)
 *         else:
 */
      __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_remaining, __pyx_v_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
      __Pyx_INCREF(__pyx_v_ssa_id);
      __Pyx_GIVEREF(__pyx_v_ssa_id);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_ssa_id);
      __pyx_t_4 = 0;
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ssa_path, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 536, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":537
 *         if key in remaining:
 *             ssa_path.append((remaining[key], ssa_id))
 *             remaining[key] = next(ssa_ids)             # <<<<<<<<<<<<<<
 *         else:
 *             remaining[key] = ssa_id
 */
      __pyx_t_5 = __Pyx_PyIter_Next(__pyx_v_ssa_ids); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(PyDict_SetItem(__pyx_v_remaining, __pyx_v_key, __pyx_t_5) < 0)) __PYX_ERR(0, 537, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":535
 *     ssa_path = []
 *     for ssa_id, key in enumerate(inputs):
 *         if key in remaining:             # <<<<<<<<<<<<<<
 *             ssa_path.append((remaining[key], ssa_id))
 *             remaining[key] = next(ssa_ids)
 */
      goto __pyx_L7;
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":539
 *             remaining[key] = next(ssa_ids)
 *         else:
 *             remaining[key] = ssa_id             # <<<<<<<<<<<<<<
 * 
 *     # Keep track of possible contraction dims.
 */
    /*else*/ {
      if (unlikely(PyDict_SetItem(__pyx_v_remaining, __pyx_v_key, __pyx_v_ssa_id) < 0)) __PYX_ERR(0, 539, __pyx_L1_error)
    }
    __pyx_L7:;

    /* "acqdp/utility/opt_einsum_paths.pyx":534
 *     ssa_ids = itertools.count(len(inputs))
 *     ssa_path = []
 *     for ssa_id, key in enumerate(inputs):             # <<<<<<<<<<<<<<
 *         if key in remaining:
 *             ssa_path.append((remaining[key], ssa_id))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":542
 * 
 *     # Keep track of possible contraction dims.
 *     dim_to_keys = defaultdict(set)             # <<<<<<<<<<<<<<
 *     for key in remaining:
 *         for dim in key - output:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, ((PyObject *)(&PySet_Type))) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)(&PySet_Type)));
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_cur_scope->__pyx_v_dim_to_keys = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":543
 *     # Keep track of possible contraction dims.
 *     dim_to_keys = defaultdict(set)
 *     for key in remaining:             # <<<<<<<<<<<<<<
 *         for dim in key - output:
 *             dim_to_keys[dim].add(key)
 */
  __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_remaining, 1, ((PyObject *)NULL), (&__pyx_t_11), (&__pyx_t_6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __pyx_t_7 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_7, __pyx_t_11, &__pyx_t_1, &__pyx_t_3, NULL, NULL, __pyx_t_6);
    if (unlikely(__pyx_t_12 == 0)) break;
    if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":544
 *     dim_to_keys = defaultdict(set)
 *     for key in remaining:
 *         for dim in key - output:             # <<<<<<<<<<<<<<
 *             dim_to_keys[dim].add(key)
 * 
 */
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_key, __pyx_v_output); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5); __pyx_t_13 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_13 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 544, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 544, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_13); __Pyx_INCREF(__pyx_t_3); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 544, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_9(__pyx_t_5);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 544, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_dim, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":545
 *     for key in remaining:
 *         for dim in key - output:
 *             dim_to_keys[dim].add(key)             # <<<<<<<<<<<<<<
 * 
 *     # Keep track of the number of tensors using each dim; when the dim is no longer
 */
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_dim_to_keys, __pyx_v_dim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_add); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
        }
      }
      __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_4, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_v_key);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 545, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":544
 *     dim_to_keys = defaultdict(set)
 *     for key in remaining:
 *         for dim in key - output:             # <<<<<<<<<<<<<<
 *             dim_to_keys[dim].add(key)
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":550
 *     # used it can be contracted. Since we specialize to binary ops, we only care about
 *     # ref counts of >=2 or >=3.
 *     dim_ref_counts = {             # <<<<<<<<<<<<<<
 *         count: set(dim for dim, keys in dim_to_keys.items() if len(keys) >= count) - output
 *         for count in [2, 3]}
 */
  { /* enter inner scope */
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "acqdp/utility/opt_einsum_paths.pyx":552
 *     dim_ref_counts = {
 *         count: set(dim for dim, keys in dim_to_keys.items() if len(keys) >= count) - output
 *         for count in [2, 3]}             # <<<<<<<<<<<<<<
 * 
 *     # Compute separable part of the objective function for contractions.
 */
    __pyx_t_5 = __pyx_tuple__10; __Pyx_INCREF(__pyx_t_5); __pyx_t_11 = 0;
    for (;;) {
      if (__pyx_t_11 >= 2) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_11); __Pyx_INCREF(__pyx_t_3); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 552, __pyx_L1_error)
      #else
      __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_8genexpr4__pyx_v_count);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_8genexpr4__pyx_v_count, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":551
 *     # ref counts of >=2 or >=3.
 *     dim_ref_counts = {
 *         count: set(dim for dim, keys in dim_to_keys.items() if len(keys) >= count) - output             # <<<<<<<<<<<<<<
 *         for count in [2, 3]}
 * 
 */
      __pyx_t_3 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_8genexpr4_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = __Pyx_Generator_Next(__pyx_t_3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_Subtract(__pyx_t_14, __pyx_v_output); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_7, (PyObject*)__pyx_cur_scope->__pyx_8genexpr4__pyx_v_count, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":552
 *     dim_ref_counts = {
 *         count: set(dim for dim, keys in dim_to_keys.items() if len(keys) >= count) - output
 *         for count in [2, 3]}             # <<<<<<<<<<<<<<
 * 
 *     # Compute separable part of the objective function for contractions.
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } /* exit inner scope */
  __pyx_v_dim_ref_counts = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":555
 * 
 *     # Compute separable part of the objective function for contractions.
 *     footprints = {key: helpers.compute_size_by_dict(key, sizes) for key in remaining}             # <<<<<<<<<<<<<<
 * 
 *     # Find initial candidate contractions.
 */
  { /* enter inner scope */
    __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 555, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = 0;
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_remaining, 1, ((PyObject *)NULL), (&__pyx_t_1), (&__pyx_t_6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_5, __pyx_t_1, &__pyx_t_11, &__pyx_t_3, NULL, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_12 == 0)) break;
      if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 555, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_helpers); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 555, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 555, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_12 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_8genexpr6__pyx_v_key, __pyx_v_sizes};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L16_error)
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_8genexpr6__pyx_v_key, __pyx_v_sizes};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L16_error)
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_15 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 555, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (__pyx_t_14) {
          __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14); __pyx_t_14 = NULL;
        }
        __Pyx_INCREF(__pyx_8genexpr6__pyx_v_key);
        __Pyx_GIVEREF(__pyx_8genexpr6__pyx_v_key);
        PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_12, __pyx_8genexpr6__pyx_v_key);
        __Pyx_INCREF(__pyx_v_sizes);
        __Pyx_GIVEREF(__pyx_v_sizes);
        PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_12, __pyx_v_sizes);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_15, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 555, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_7, (PyObject*)__pyx_8genexpr6__pyx_v_key, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 555, __pyx_L16_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_key); __pyx_8genexpr6__pyx_v_key = 0;
    goto __pyx_L19_exit_scope;
    __pyx_L16_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_key); __pyx_8genexpr6__pyx_v_key = 0;
    goto __pyx_L1_error;
    __pyx_L19_exit_scope:;
  } /* exit inner scope */
  __pyx_v_footprints = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":558
 * 
 *     # Find initial candidate contractions.
 *     queue = []             # <<<<<<<<<<<<<<
 *     for dim, keys in dim_to_keys.items():
 *         keys = sorted(keys, key=remaining.__getitem__)
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_queue = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":559
 *     # Find initial candidate contractions.
 *     queue = []
 *     for dim, keys in dim_to_keys.items():             # <<<<<<<<<<<<<<
 *         keys = sorted(keys, key=remaining.__getitem__)
 *         for i, k1 in enumerate(keys[:-1]):
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_dim_to_keys, __pyx_n_s_items); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
    __pyx_t_5 = __pyx_t_7; __Pyx_INCREF(__pyx_t_5); __pyx_t_1 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 559, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_7); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 559, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_7); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 559, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_9(__pyx_t_5);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 559, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 559, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_15 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_16 = Py_TYPE(__pyx_t_15)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_3)) goto __pyx_L22_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_16(__pyx_t_15); if (unlikely(!__pyx_t_4)) goto __pyx_L22_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_15), 2) < 0) __PYX_ERR(0, 559, __pyx_L1_error)
      __pyx_t_16 = NULL;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      goto __pyx_L23_unpacking_done;
      __pyx_L22_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_16 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 559, __pyx_L1_error)
      __pyx_L23_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_dim, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_keys, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":560
 *     queue = []
 *     for dim, keys in dim_to_keys.items():
 *         keys = sorted(keys, key=remaining.__getitem__)             # <<<<<<<<<<<<<<
 *         for i, k1 in enumerate(keys[:-1]):
 *             k2s = keys[1 + i:]
 */
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_keys);
    __Pyx_GIVEREF(__pyx_v_keys);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_keys);
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_remaining, __pyx_n_s_getitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_keys, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":561
 *     for dim, keys in dim_to_keys.items():
 *         keys = sorted(keys, key=remaining.__getitem__)
 *         for i, k1 in enumerate(keys[:-1]):             # <<<<<<<<<<<<<<
 *             k2s = keys[1 + i:]
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_3 = __pyx_int_0;
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_keys, 0, -1L, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_7 = __pyx_t_4; __Pyx_INCREF(__pyx_t_7); __pyx_t_11 = 0;
      __pyx_t_17 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_17 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 561, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_17)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_11); __Pyx_INCREF(__pyx_t_4); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 561, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_11); __Pyx_INCREF(__pyx_t_4); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 561, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_7, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_17(__pyx_t_7);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 561, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k1);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k1, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3);
      __pyx_t_3 = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":562
 *         keys = sorted(keys, key=remaining.__getitem__)
 *         for i, k1 in enumerate(keys[:-1]):
 *             k2s = keys[1 + i:]             # <<<<<<<<<<<<<<
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)
 * 
 */
      __pyx_t_4 = __Pyx_PyInt_AddCObj(__pyx_int_1, __pyx_v_i, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_15 = __Pyx_PyObject_GetSlice(__pyx_v_keys, 0, 0, &__pyx_t_4, NULL, NULL, 0, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_k2s, __pyx_t_15);
      __pyx_t_15 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":563
 *         for i, k1 in enumerate(keys[:-1]):
 *             k2s = keys[1 + i:]
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)             # <<<<<<<<<<<<<<
 * 
 *     # Greedily contract pairs of tensors.
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_push_candidate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = __Pyx_PyBool_FromLong(__pyx_v_push_all); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_18 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_18)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[11] = {__pyx_t_18, __pyx_v_output, __pyx_v_sizes, __pyx_v_remaining, __pyx_v_footprints, __pyx_v_dim_ref_counts, __pyx_cur_scope->__pyx_v_k1, __pyx_v_k2s, __pyx_v_queue, __pyx_t_14, __pyx_v_cost_fn};
        __pyx_t_15 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 10+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 563, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[11] = {__pyx_t_18, __pyx_v_output, __pyx_v_sizes, __pyx_v_remaining, __pyx_v_footprints, __pyx_v_dim_ref_counts, __pyx_cur_scope->__pyx_v_k1, __pyx_v_k2s, __pyx_v_queue, __pyx_t_14, __pyx_v_cost_fn};
        __pyx_t_15 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 10+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 563, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      {
        __pyx_t_19 = PyTuple_New(10+__pyx_t_6); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 563, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_19);
        if (__pyx_t_18) {
          __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_18); __pyx_t_18 = NULL;
        }
        __Pyx_INCREF(__pyx_v_output);
        __Pyx_GIVEREF(__pyx_v_output);
        PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_6, __pyx_v_output);
        __Pyx_INCREF(__pyx_v_sizes);
        __Pyx_GIVEREF(__pyx_v_sizes);
        PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_6, __pyx_v_sizes);
        __Pyx_INCREF(__pyx_v_remaining);
        __Pyx_GIVEREF(__pyx_v_remaining);
        PyTuple_SET_ITEM(__pyx_t_19, 2+__pyx_t_6, __pyx_v_remaining);
        __Pyx_INCREF(__pyx_v_footprints);
        __Pyx_GIVEREF(__pyx_v_footprints);
        PyTuple_SET_ITEM(__pyx_t_19, 3+__pyx_t_6, __pyx_v_footprints);
        __Pyx_INCREF(__pyx_v_dim_ref_counts);
        __Pyx_GIVEREF(__pyx_v_dim_ref_counts);
        PyTuple_SET_ITEM(__pyx_t_19, 4+__pyx_t_6, __pyx_v_dim_ref_counts);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k1);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k1);
        PyTuple_SET_ITEM(__pyx_t_19, 5+__pyx_t_6, __pyx_cur_scope->__pyx_v_k1);
        __Pyx_INCREF(__pyx_v_k2s);
        __Pyx_GIVEREF(__pyx_v_k2s);
        PyTuple_SET_ITEM(__pyx_t_19, 6+__pyx_t_6, __pyx_v_k2s);
        __Pyx_INCREF(__pyx_v_queue);
        __Pyx_GIVEREF(__pyx_v_queue);
        PyTuple_SET_ITEM(__pyx_t_19, 7+__pyx_t_6, __pyx_v_queue);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_19, 8+__pyx_t_6, __pyx_t_14);
        __Pyx_INCREF(__pyx_v_cost_fn);
        __Pyx_GIVEREF(__pyx_v_cost_fn);
        PyTuple_SET_ITEM(__pyx_t_19, 9+__pyx_t_6, __pyx_v_cost_fn);
        __pyx_t_14 = 0;
        __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_19, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 563, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":561
 *     for dim, keys in dim_to_keys.items():
 *         keys = sorted(keys, key=remaining.__getitem__)
 *         for i, k1 in enumerate(keys[:-1]):             # <<<<<<<<<<<<<<
 *             k2s = keys[1 + i:]
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":559
 *     # Find initial candidate contractions.
 *     queue = []
 *     for dim, keys in dim_to_keys.items():             # <<<<<<<<<<<<<<
 *         keys = sorted(keys, key=remaining.__getitem__)
 *         for i, k1 in enumerate(keys[:-1]):
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":566
 * 
 *     # Greedily contract pairs of tensors.
 *     while queue:             # <<<<<<<<<<<<<<
 * 
 *         con = choose_fn(queue, remaining)
 */
  while (1) {
    __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_queue) != 0);
    if (!__pyx_t_2) break;

    /* "acqdp/utility/opt_einsum_paths.pyx":568
 *     while queue:
 * 
 *         con = choose_fn(queue, remaining)             # <<<<<<<<<<<<<<
 *         if con is None:
 *             continue  # allow choose_fn to flag all candidates obsolete
 */
    __Pyx_INCREF(__pyx_v_choose_fn);
    __pyx_t_3 = __pyx_v_choose_fn; __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_queue, __pyx_v_remaining};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_queue, __pyx_v_remaining};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_15 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_v_queue);
      __Pyx_GIVEREF(__pyx_v_queue);
      PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_6, __pyx_v_queue);
      __Pyx_INCREF(__pyx_v_remaining);
      __Pyx_GIVEREF(__pyx_v_remaining);
      PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_6, __pyx_v_remaining);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_15, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_con, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":569
 * 
 *         con = choose_fn(queue, remaining)
 *         if con is None:             # <<<<<<<<<<<<<<
 *             continue  # allow choose_fn to flag all candidates obsolete
 *         cost, k1, k2, k12 = con
 */
    __pyx_t_2 = (__pyx_v_con == Py_None);
    __pyx_t_8 = (__pyx_t_2 != 0);
    if (__pyx_t_8) {

      /* "acqdp/utility/opt_einsum_paths.pyx":570
 *         con = choose_fn(queue, remaining)
 *         if con is None:
 *             continue  # allow choose_fn to flag all candidates obsolete             # <<<<<<<<<<<<<<
 *         cost, k1, k2, k12 = con
 * 
 */
      goto __pyx_L26_continue;

      /* "acqdp/utility/opt_einsum_paths.pyx":569
 * 
 *         con = choose_fn(queue, remaining)
 *         if con is None:             # <<<<<<<<<<<<<<
 *             continue  # allow choose_fn to flag all candidates obsolete
 *         cost, k1, k2, k12 = con
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":571
 *         if con is None:
 *             continue  # allow choose_fn to flag all candidates obsolete
 *         cost, k1, k2, k12 = con             # <<<<<<<<<<<<<<
 * 
 *         ssa_id1 = remaining.pop(k1)
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_con))) || (PyList_CheckExact(__pyx_v_con))) {
      PyObject* sequence = __pyx_v_con;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 4)) {
        if (size > 4) __Pyx_RaiseTooManyValuesError(4);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 571, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_15 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 3); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_15 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 3); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_7);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[4] = {&__pyx_t_5,&__pyx_t_3,&__pyx_t_15,&__pyx_t_7};
        for (i=0; i < 4; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 571, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[4] = {&__pyx_t_5,&__pyx_t_3,&__pyx_t_15,&__pyx_t_7};
      __pyx_t_4 = PyObject_GetIter(__pyx_v_con); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_16 = Py_TYPE(__pyx_t_4)->tp_iternext;
      for (index=0; index < 4; index++) {
        PyObject* item = __pyx_t_16(__pyx_t_4); if (unlikely(!item)) goto __pyx_L29_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_4), 4) < 0) __PYX_ERR(0, 571, __pyx_L1_error)
      __pyx_t_16 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L30_unpacking_done;
      __pyx_L29_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_16 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 571, __pyx_L1_error)
      __pyx_L30_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_cost, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k1);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k2, __pyx_t_15);
    __pyx_t_15 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k12, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":573
 *         cost, k1, k2, k12 = con
 * 
 *         ssa_id1 = remaining.pop(k1)             # <<<<<<<<<<<<<<
 *         ssa_id2 = remaining.pop(k2)
 *         for dim in k1 - output:
 */
    __pyx_t_7 = __Pyx_PyDict_Pop(__pyx_v_remaining, __pyx_cur_scope->__pyx_v_k1, ((PyObject *)NULL)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_ssa_id1, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":574
 * 
 *         ssa_id1 = remaining.pop(k1)
 *         ssa_id2 = remaining.pop(k2)             # <<<<<<<<<<<<<<
 *         for dim in k1 - output:
 *             dim_to_keys[dim].remove(k1)
 */
    __pyx_t_7 = __Pyx_PyDict_Pop(__pyx_v_remaining, __pyx_v_k2, ((PyObject *)NULL)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_ssa_id2, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":575
 *         ssa_id1 = remaining.pop(k1)
 *         ssa_id2 = remaining.pop(k2)
 *         for dim in k1 - output:             # <<<<<<<<<<<<<<
 *             dim_to_keys[dim].remove(k1)
 *         for dim in k2 - output:
 */
    __pyx_t_7 = PyNumber_Subtract(__pyx_cur_scope->__pyx_v_k1, __pyx_v_output); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_15 = __pyx_t_7; __Pyx_INCREF(__pyx_t_15); __pyx_t_1 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_1 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_9 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 575, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_15))) {
          if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_15)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_1); __Pyx_INCREF(__pyx_t_7); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 575, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_15, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 575, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_1); __Pyx_INCREF(__pyx_t_7); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 575, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_15, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 575, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_9(__pyx_t_15);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 575, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      __Pyx_XDECREF_SET(__pyx_v_dim, __pyx_t_7);
      __pyx_t_7 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":576
 *         ssa_id2 = remaining.pop(k2)
 *         for dim in k1 - output:
 *             dim_to_keys[dim].remove(k1)             # <<<<<<<<<<<<<<
 *         for dim in k2 - output:
 *             dim_to_keys[dim].remove(k2)
 */
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_dim_to_keys, __pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_remove); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_cur_scope->__pyx_v_k1) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_cur_scope->__pyx_v_k1);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":575
 *         ssa_id1 = remaining.pop(k1)
 *         ssa_id2 = remaining.pop(k2)
 *         for dim in k1 - output:             # <<<<<<<<<<<<<<
 *             dim_to_keys[dim].remove(k1)
 *         for dim in k2 - output:
 */
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":577
 *         for dim in k1 - output:
 *             dim_to_keys[dim].remove(k1)
 *         for dim in k2 - output:             # <<<<<<<<<<<<<<
 *             dim_to_keys[dim].remove(k2)
 *         ssa_path.append((ssa_id1, ssa_id2))
 */
    __pyx_t_15 = PyNumber_Subtract(__pyx_v_k2, __pyx_v_output); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
      __pyx_t_7 = __pyx_t_15; __Pyx_INCREF(__pyx_t_7); __pyx_t_1 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_1 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 577, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_15 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_1); __Pyx_INCREF(__pyx_t_15); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 577, __pyx_L1_error)
          #else
          __pyx_t_15 = PySequence_ITEM(__pyx_t_7, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 577, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          #endif
        } else {
          if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_1); __Pyx_INCREF(__pyx_t_15); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 577, __pyx_L1_error)
          #else
          __pyx_t_15 = PySequence_ITEM(__pyx_t_7, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 577, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          #endif
        }
      } else {
        __pyx_t_15 = __pyx_t_9(__pyx_t_7);
        if (unlikely(!__pyx_t_15)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 577, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_15);
      }
      __Pyx_XDECREF_SET(__pyx_v_dim, __pyx_t_15);
      __pyx_t_15 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":578
 *             dim_to_keys[dim].remove(k1)
 *         for dim in k2 - output:
 *             dim_to_keys[dim].remove(k2)             # <<<<<<<<<<<<<<
 *         ssa_path.append((ssa_id1, ssa_id2))
 *         if k12 in remaining:
 */
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_dim_to_keys, __pyx_v_dim); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_remove); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_15 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_k2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_k2);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 578, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":577
 *         for dim in k1 - output:
 *             dim_to_keys[dim].remove(k1)
 *         for dim in k2 - output:             # <<<<<<<<<<<<<<
 *             dim_to_keys[dim].remove(k2)
 *         ssa_path.append((ssa_id1, ssa_id2))
 */
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":579
 *         for dim in k2 - output:
 *             dim_to_keys[dim].remove(k2)
 *         ssa_path.append((ssa_id1, ssa_id2))             # <<<<<<<<<<<<<<
 *         if k12 in remaining:
 *             ssa_path.append((remaining[k12], next(ssa_ids)))
 */
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_ssa_id1);
    __Pyx_GIVEREF(__pyx_v_ssa_id1);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_ssa_id1);
    __Pyx_INCREF(__pyx_v_ssa_id2);
    __Pyx_GIVEREF(__pyx_v_ssa_id2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_ssa_id2);
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ssa_path, __pyx_t_7); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":580
 *             dim_to_keys[dim].remove(k2)
 *         ssa_path.append((ssa_id1, ssa_id2))
 *         if k12 in remaining:             # <<<<<<<<<<<<<<
 *             ssa_path.append((remaining[k12], next(ssa_ids)))
 *         else:
 */
    __pyx_t_8 = (__Pyx_PyDict_ContainsTF(__pyx_v_k12, __pyx_v_remaining, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 580, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_8 != 0);
    if (__pyx_t_2) {

      /* "acqdp/utility/opt_einsum_paths.pyx":581
 *         ssa_path.append((ssa_id1, ssa_id2))
 *         if k12 in remaining:
 *             ssa_path.append((remaining[k12], next(ssa_ids)))             # <<<<<<<<<<<<<<
 *         else:
 *             for dim in k12 - output:
 */
      __pyx_t_7 = __Pyx_PyDict_GetItem(__pyx_v_remaining, __pyx_v_k12); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_15 = __Pyx_PyIter_Next(__pyx_v_ssa_ids); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_15);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_15);
      __pyx_t_7 = 0;
      __pyx_t_15 = 0;
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ssa_path, __pyx_t_3); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":580
 *             dim_to_keys[dim].remove(k2)
 *         ssa_path.append((ssa_id1, ssa_id2))
 *         if k12 in remaining:             # <<<<<<<<<<<<<<
 *             ssa_path.append((remaining[k12], next(ssa_ids)))
 *         else:
 */
      goto __pyx_L35;
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":583
 *             ssa_path.append((remaining[k12], next(ssa_ids)))
 *         else:
 *             for dim in k12 - output:             # <<<<<<<<<<<<<<
 *                 dim_to_keys[dim].add(k12)
 *         remaining[k12] = next(ssa_ids)
 */
    /*else*/ {
      __pyx_t_3 = PyNumber_Subtract(__pyx_v_k12, __pyx_v_output); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
        __pyx_t_15 = __pyx_t_3; __Pyx_INCREF(__pyx_t_15); __pyx_t_1 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_1 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 583, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_9 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 583, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_15))) {
            if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_15)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 583, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_15, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 583, __pyx_L1_error)
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_15, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 583, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_9(__pyx_t_15);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 583, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v_dim, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":584
 *         else:
 *             for dim in k12 - output:
 *                 dim_to_keys[dim].add(k12)             # <<<<<<<<<<<<<<
 *         remaining[k12] = next(ssa_ids)
 *         _update_ref_counts(dim_to_keys, dim_ref_counts, k1 | k2 - output)
 */
        __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_dim_to_keys, __pyx_v_dim); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_add); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_v_k12) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_k12);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":583
 *             ssa_path.append((remaining[k12], next(ssa_ids)))
 *         else:
 *             for dim in k12 - output:             # <<<<<<<<<<<<<<
 *                 dim_to_keys[dim].add(k12)
 *         remaining[k12] = next(ssa_ids)
 */
      }
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    }
    __pyx_L35:;

    /* "acqdp/utility/opt_einsum_paths.pyx":585
 *             for dim in k12 - output:
 *                 dim_to_keys[dim].add(k12)
 *         remaining[k12] = next(ssa_ids)             # <<<<<<<<<<<<<<
 *         _update_ref_counts(dim_to_keys, dim_ref_counts, k1 | k2 - output)
 *         footprints[k12] = helpers.compute_size_by_dict(k12, sizes)
 */
    __pyx_t_15 = __Pyx_PyIter_Next(__pyx_v_ssa_ids); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (unlikely(PyDict_SetItem(__pyx_v_remaining, __pyx_v_k12, __pyx_t_15) < 0)) __PYX_ERR(0, 585, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":586
 *                 dim_to_keys[dim].add(k12)
 *         remaining[k12] = next(ssa_ids)
 *         _update_ref_counts(dim_to_keys, dim_ref_counts, k1 | k2 - output)             # <<<<<<<<<<<<<<
 *         footprints[k12] = helpers.compute_size_by_dict(k12, sizes)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_update_ref_counts); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_k2, __pyx_v_output); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyNumber_Or(__pyx_cur_scope->__pyx_v_k1, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_dim_to_keys, __pyx_v_dim_ref_counts, __pyx_t_7};
      __pyx_t_15 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_dim_to_keys, __pyx_v_dim_ref_counts, __pyx_t_7};
      __pyx_t_15 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_dim_to_keys);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_dim_to_keys);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_6, __pyx_cur_scope->__pyx_v_dim_to_keys);
      __Pyx_INCREF(__pyx_v_dim_ref_counts);
      __Pyx_GIVEREF(__pyx_v_dim_ref_counts);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_6, __pyx_v_dim_ref_counts);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_4, 2+__pyx_t_6, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":587
 *         remaining[k12] = next(ssa_ids)
 *         _update_ref_counts(dim_to_keys, dim_ref_counts, k1 | k2 - output)
 *         footprints[k12] = helpers.compute_size_by_dict(k12, sizes)             # <<<<<<<<<<<<<<
 * 
 *         # Find new candidate contractions.
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_helpers); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_k12, __pyx_v_sizes};
      __pyx_t_15 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_15);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_k12, __pyx_v_sizes};
      __pyx_t_15 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_15);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_v_k12);
      __Pyx_GIVEREF(__pyx_v_k12);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_k12);
      __Pyx_INCREF(__pyx_v_sizes);
      __Pyx_GIVEREF(__pyx_v_sizes);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_sizes);
      __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(PyDict_SetItem(__pyx_v_footprints, __pyx_v_k12, __pyx_t_15) < 0)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":590
 * 
 *         # Find new candidate contractions.
 *         k1 = k12             # <<<<<<<<<<<<<<
 *         k2s = set(k2 for dim in k1 for k2 in dim_to_keys[dim])
 *         k2s.discard(k1)
 */
    __Pyx_INCREF(__pyx_v_k12);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_k1);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_k1, __pyx_v_k12);
    __Pyx_GIVEREF(__pyx_v_k12);

    /* "acqdp/utility/opt_einsum_paths.pyx":591
 *         # Find new candidate contractions.
 *         k1 = k12
 *         k2s = set(k2 for dim in k1 for k2 in dim_to_keys[dim])             # <<<<<<<<<<<<<<
 *         k2s.discard(k1)
 *         if k2s:
 */
    __pyx_t_15 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_19ssa_greedy_optimize_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k2s, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":592
 *         k1 = k12
 *         k2s = set(k2 for dim in k1 for k2 in dim_to_keys[dim])
 *         k2s.discard(k1)             # <<<<<<<<<<<<<<
 *         if k2s:
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)
 */
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_k2s, __pyx_n_s_discard); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_15))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_15);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_15, function);
      }
    }
    __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_7, __pyx_cur_scope->__pyx_v_k1) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_cur_scope->__pyx_v_k1);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":593
 *         k2s = set(k2 for dim in k1 for k2 in dim_to_keys[dim])
 *         k2s.discard(k1)
 *         if k2s:             # <<<<<<<<<<<<<<
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_k2s); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 593, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "acqdp/utility/opt_einsum_paths.pyx":594
 *         k2s.discard(k1)
 *         if k2s:
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)             # <<<<<<<<<<<<<<
 * 
 *     # Greedily compute pairwise outer products.
 */
      __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_push_candidate); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 594, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_push_all); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 594, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_15);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_15, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_15)) {
        PyObject *__pyx_temp[11] = {__pyx_t_3, __pyx_v_output, __pyx_v_sizes, __pyx_v_remaining, __pyx_v_footprints, __pyx_v_dim_ref_counts, __pyx_cur_scope->__pyx_v_k1, __pyx_v_k2s, __pyx_v_queue, __pyx_t_7, __pyx_v_cost_fn};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_6, 10+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 594, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_15)) {
        PyObject *__pyx_temp[11] = {__pyx_t_3, __pyx_v_output, __pyx_v_sizes, __pyx_v_remaining, __pyx_v_footprints, __pyx_v_dim_ref_counts, __pyx_cur_scope->__pyx_v_k1, __pyx_v_k2s, __pyx_v_queue, __pyx_t_7, __pyx_v_cost_fn};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_6, 10+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 594, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(10+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 594, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_output);
        __Pyx_GIVEREF(__pyx_v_output);
        PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_output);
        __Pyx_INCREF(__pyx_v_sizes);
        __Pyx_GIVEREF(__pyx_v_sizes);
        PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_sizes);
        __Pyx_INCREF(__pyx_v_remaining);
        __Pyx_GIVEREF(__pyx_v_remaining);
        PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_v_remaining);
        __Pyx_INCREF(__pyx_v_footprints);
        __Pyx_GIVEREF(__pyx_v_footprints);
        PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_6, __pyx_v_footprints);
        __Pyx_INCREF(__pyx_v_dim_ref_counts);
        __Pyx_GIVEREF(__pyx_v_dim_ref_counts);
        PyTuple_SET_ITEM(__pyx_t_5, 4+__pyx_t_6, __pyx_v_dim_ref_counts);
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k1);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_k1);
        PyTuple_SET_ITEM(__pyx_t_5, 5+__pyx_t_6, __pyx_cur_scope->__pyx_v_k1);
        __Pyx_INCREF(__pyx_v_k2s);
        __Pyx_GIVEREF(__pyx_v_k2s);
        PyTuple_SET_ITEM(__pyx_t_5, 6+__pyx_t_6, __pyx_v_k2s);
        __Pyx_INCREF(__pyx_v_queue);
        __Pyx_GIVEREF(__pyx_v_queue);
        PyTuple_SET_ITEM(__pyx_t_5, 7+__pyx_t_6, __pyx_v_queue);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_5, 8+__pyx_t_6, __pyx_t_7);
        __Pyx_INCREF(__pyx_v_cost_fn);
        __Pyx_GIVEREF(__pyx_v_cost_fn);
        PyTuple_SET_ITEM(__pyx_t_5, 9+__pyx_t_6, __pyx_v_cost_fn);
        __pyx_t_7 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 594, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":593
 *         k2s = set(k2 for dim in k1 for k2 in dim_to_keys[dim])
 *         k2s.discard(k1)
 *         if k2s:             # <<<<<<<<<<<<<<
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)
 * 
 */
    }
    __pyx_L26_continue:;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":597
 * 
 *     # Greedily compute pairwise outer products.
 *     queue = [(helpers.compute_size_by_dict(key & output, sizes), ssa_id, key)             # <<<<<<<<<<<<<<
 *              for key, ssa_id in remaining.items()]
 *     heapq.heapify(queue)
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "acqdp/utility/opt_einsum_paths.pyx":598
 *     # Greedily compute pairwise outer products.
 *     queue = [(helpers.compute_size_by_dict(key & output, sizes), ssa_id, key)
 *              for key, ssa_id in remaining.items()]             # <<<<<<<<<<<<<<
 *     heapq.heapify(queue)
 *     _, ssa_id1, k1 = heapq.heappop(queue)
 */
  __pyx_t_15 = __Pyx_PyDict_Items(__pyx_v_remaining); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
    __pyx_t_5 = __pyx_t_15; __Pyx_INCREF(__pyx_t_5); __pyx_t_1 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 598, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_15 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_15); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 598, __pyx_L1_error)
        #else
        __pyx_t_15 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 598, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_1); __Pyx_INCREF(__pyx_t_15); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 598, __pyx_L1_error)
        #else
        __pyx_t_15 = PySequence_ITEM(__pyx_t_5, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 598, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        #endif
      }
    } else {
      __pyx_t_15 = __pyx_t_9(__pyx_t_5);
      if (unlikely(!__pyx_t_15)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 598, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_15);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_15))) || (PyList_CheckExact(__pyx_t_15))) {
      PyObject* sequence = __pyx_t_15;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 598, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_19 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_16 = Py_TYPE(__pyx_t_19)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_16(__pyx_t_19); if (unlikely(!__pyx_t_7)) goto __pyx_L41_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_3 = __pyx_t_16(__pyx_t_19); if (unlikely(!__pyx_t_3)) goto __pyx_L41_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_19), 2) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
      __pyx_t_16 = NULL;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      goto __pyx_L42_unpacking_done;
      __pyx_L41_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      __pyx_t_16 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 598, __pyx_L1_error)
      __pyx_L42_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_ssa_id, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":597
 * 
 *     # Greedily compute pairwise outer products.
 *     queue = [(helpers.compute_size_by_dict(key & output, sizes), ssa_id, key)             # <<<<<<<<<<<<<<
 *              for key, ssa_id in remaining.items()]
 *     heapq.heapify(queue)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_helpers); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_And(__pyx_v_key, __pyx_v_output); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_19 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_19)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_19);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_19, __pyx_t_3, __pyx_v_sizes};
      __pyx_t_15 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_19, __pyx_t_3, __pyx_v_sizes};
      __pyx_t_15 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_14 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__pyx_t_19) {
        __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_19); __pyx_t_19 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_6, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_sizes);
      __Pyx_GIVEREF(__pyx_v_sizes);
      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_6, __pyx_v_sizes);
      __pyx_t_3 = 0;
      __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_14, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 597, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_v_ssa_id);
    __Pyx_GIVEREF(__pyx_v_ssa_id);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_ssa_id);
    __Pyx_INCREF(__pyx_v_key);
    __Pyx_GIVEREF(__pyx_v_key);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_key);
    __pyx_t_15 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":598
 *     # Greedily compute pairwise outer products.
 *     queue = [(helpers.compute_size_by_dict(key & output, sizes), ssa_id, key)
 *              for key, ssa_id in remaining.items()]             # <<<<<<<<<<<<<<
 *     heapq.heapify(queue)
 *     _, ssa_id1, k1 = heapq.heappop(queue)
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_queue, ((PyObject*)__pyx_t_4));
  __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":599
 *     queue = [(helpers.compute_size_by_dict(key & output, sizes), ssa_id, key)
 *              for key, ssa_id in remaining.items()]
 *     heapq.heapify(queue)             # <<<<<<<<<<<<<<
 *     _, ssa_id1, k1 = heapq.heappop(queue)
 *     while queue:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_heapq); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_heapify); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_5, __pyx_v_queue) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_queue);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":600
 *              for key, ssa_id in remaining.items()]
 *     heapq.heapify(queue)
 *     _, ssa_id1, k1 = heapq.heappop(queue)             # <<<<<<<<<<<<<<
 *     while queue:
 *         _, ssa_id2, k2 = heapq.heappop(queue)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_heapq); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_heappop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_v_queue) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_queue);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
    PyObject* sequence = __pyx_t_4;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 600, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_15 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_15 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_15);
    #else
    __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_15 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    #endif
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_14 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_16 = Py_TYPE(__pyx_t_14)->tp_iternext;
    index = 0; __pyx_t_5 = __pyx_t_16(__pyx_t_14); if (unlikely(!__pyx_t_5)) goto __pyx_L43_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_5);
    index = 1; __pyx_t_7 = __pyx_t_16(__pyx_t_14); if (unlikely(!__pyx_t_7)) goto __pyx_L43_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_7);
    index = 2; __pyx_t_15 = __pyx_t_16(__pyx_t_14); if (unlikely(!__pyx_t_15)) goto __pyx_L43_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_14), 3) < 0) __PYX_ERR(0, 600, __pyx_L1_error)
    __pyx_t_16 = NULL;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    goto __pyx_L44_unpacking_done;
    __pyx_L43_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_16 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 600, __pyx_L1_error)
    __pyx_L44_unpacking_done:;
  }
  __pyx_v__ = __pyx_t_5;
  __pyx_t_5 = 0;
  __Pyx_XDECREF_SET(__pyx_v_ssa_id1, __pyx_t_7);
  __pyx_t_7 = 0;
  __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k1);
  __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k1, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":601
 *     heapq.heapify(queue)
 *     _, ssa_id1, k1 = heapq.heappop(queue)
 *     while queue:             # <<<<<<<<<<<<<<
 *         _, ssa_id2, k2 = heapq.heappop(queue)
 *         ssa_path.append((min(ssa_id1, ssa_id2), max(ssa_id1, ssa_id2)))
 */
  while (1) {
    __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_queue) != 0);
    if (!__pyx_t_2) break;

    /* "acqdp/utility/opt_einsum_paths.pyx":602
 *     _, ssa_id1, k1 = heapq.heappop(queue)
 *     while queue:
 *         _, ssa_id2, k2 = heapq.heappop(queue)             # <<<<<<<<<<<<<<
 *         ssa_path.append((min(ssa_id1, ssa_id2), max(ssa_id1, ssa_id2)))
 *         k12 = (k1 | k2) & output
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_heapq); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_heappop); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_4 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_15, __pyx_v_queue) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_queue);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 602, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_15 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_15 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_15 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_5 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_14 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 602, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_16 = Py_TYPE(__pyx_t_14)->tp_iternext;
      index = 0; __pyx_t_7 = __pyx_t_16(__pyx_t_14); if (unlikely(!__pyx_t_7)) goto __pyx_L47_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      index = 1; __pyx_t_15 = __pyx_t_16(__pyx_t_14); if (unlikely(!__pyx_t_15)) goto __pyx_L47_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_15);
      index = 2; __pyx_t_5 = __pyx_t_16(__pyx_t_14); if (unlikely(!__pyx_t_5)) goto __pyx_L47_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_14), 3) < 0) __PYX_ERR(0, 602, __pyx_L1_error)
      __pyx_t_16 = NULL;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L48_unpacking_done;
      __pyx_L47_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_16 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 602, __pyx_L1_error)
      __pyx_L48_unpacking_done:;
    }
    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_ssa_id2, __pyx_t_15);
    __pyx_t_15 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k2, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":603
 *     while queue:
 *         _, ssa_id2, k2 = heapq.heappop(queue)
 *         ssa_path.append((min(ssa_id1, ssa_id2), max(ssa_id1, ssa_id2)))             # <<<<<<<<<<<<<<
 *         k12 = (k1 | k2) & output
 *         cost = helpers.compute_size_by_dict(k12, sizes)
 */
    __Pyx_INCREF(__pyx_v_ssa_id2);
    __pyx_t_4 = __pyx_v_ssa_id2;
    __Pyx_INCREF(__pyx_v_ssa_id1);
    __pyx_t_5 = __pyx_v_ssa_id1;
    __pyx_t_7 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 603, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_2) {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_15 = __pyx_t_4;
    } else {
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_15 = __pyx_t_5;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_v_ssa_id2);
    __pyx_t_4 = __pyx_v_ssa_id2;
    __Pyx_INCREF(__pyx_v_ssa_id1);
    __pyx_t_5 = __pyx_v_ssa_id1;
    __pyx_t_14 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_GT); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 603, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (__pyx_t_2) {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_7 = __pyx_t_4;
    } else {
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_7 = __pyx_t_5;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_7);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ssa_path, __pyx_t_4); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 603, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":604
 *         _, ssa_id2, k2 = heapq.heappop(queue)
 *         ssa_path.append((min(ssa_id1, ssa_id2), max(ssa_id1, ssa_id2)))
 *         k12 = (k1 | k2) & output             # <<<<<<<<<<<<<<
 *         cost = helpers.compute_size_by_dict(k12, sizes)
 *         ssa_id12 = next(ssa_ids)
 */
    __pyx_t_4 = PyNumber_Or(__pyx_cur_scope->__pyx_v_k1, __pyx_v_k2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyNumber_And(__pyx_t_4, __pyx_v_output); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k12, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":605
 *         ssa_path.append((min(ssa_id1, ssa_id2), max(ssa_id1, ssa_id2)))
 *         k12 = (k1 | k2) & output
 *         cost = helpers.compute_size_by_dict(k12, sizes)             # <<<<<<<<<<<<<<
 *         ssa_id12 = next(ssa_ids)
 *         _, ssa_id1, k1 = heapq.heappushpop(queue, (cost, ssa_id12, k12))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_helpers); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_15);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_15, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_15)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_k12, __pyx_v_sizes};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_15)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_k12, __pyx_v_sizes};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_k12);
      __Pyx_GIVEREF(__pyx_v_k12);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, __pyx_v_k12);
      __Pyx_INCREF(__pyx_v_sizes);
      __Pyx_GIVEREF(__pyx_v_sizes);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_sizes);
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_5, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_XDECREF_SET(__pyx_v_cost, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":606
 *         k12 = (k1 | k2) & output
 *         cost = helpers.compute_size_by_dict(k12, sizes)
 *         ssa_id12 = next(ssa_ids)             # <<<<<<<<<<<<<<
 *         _, ssa_id1, k1 = heapq.heappushpop(queue, (cost, ssa_id12, k12))
 * 
 */
    __pyx_t_7 = __Pyx_PyIter_Next(__pyx_v_ssa_ids); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 606, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF_SET(__pyx_v_ssa_id12, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":607
 *         cost = helpers.compute_size_by_dict(k12, sizes)
 *         ssa_id12 = next(ssa_ids)
 *         _, ssa_id1, k1 = heapq.heappushpop(queue, (cost, ssa_id12, k12))             # <<<<<<<<<<<<<<
 * 
 *     return ssa_path
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_heapq); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_heappushpop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_v_cost);
    __Pyx_GIVEREF(__pyx_v_cost);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v_cost);
    __Pyx_INCREF(__pyx_v_ssa_id12);
    __Pyx_GIVEREF(__pyx_v_ssa_id12);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_v_ssa_id12);
    __Pyx_INCREF(__pyx_v_k12);
    __Pyx_GIVEREF(__pyx_v_k12);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_k12);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_queue, __pyx_t_15};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_queue, __pyx_t_15};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else
    #endif
    {
      __pyx_t_14 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_queue);
      __Pyx_GIVEREF(__pyx_v_queue);
      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_6, __pyx_v_queue);
      __Pyx_GIVEREF(__pyx_t_15);
      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_6, __pyx_t_15);
      __pyx_t_15 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_14, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 607, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_14 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_15 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_14 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_15 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_15);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_14 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_16 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_16(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L49_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_14 = __pyx_t_16(__pyx_t_4); if (unlikely(!__pyx_t_14)) goto __pyx_L49_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_14);
      index = 2; __pyx_t_15 = __pyx_t_16(__pyx_t_4); if (unlikely(!__pyx_t_15)) goto __pyx_L49_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_15);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_16(__pyx_t_4), 3) < 0) __PYX_ERR(0, 607, __pyx_L1_error)
      __pyx_t_16 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L50_unpacking_done;
      __pyx_L49_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_16 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 607, __pyx_L1_error)
      __pyx_L50_unpacking_done:;
    }
    __Pyx_DECREF_SET(__pyx_v__, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_ssa_id1, __pyx_t_14);
    __pyx_t_14 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_k1);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_k1, __pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    __pyx_t_15 = 0;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":609
 *         _, ssa_id1, k1 = heapq.heappushpop(queue, (cost, ssa_id12, k12))
 * 
 *     return ssa_path             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ssa_path);
  __pyx_r = __pyx_v_ssa_path;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":502
 * 
 * 
 * def ssa_greedy_optimize(inputs, output, sizes, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     This is the core function for :func:`greedy` but produces a path with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.ssa_greedy_optimize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_remaining);
  __Pyx_XDECREF(__pyx_v_ssa_ids);
  __Pyx_XDECREF(__pyx_v_ssa_path);
  __Pyx_XDECREF(__pyx_v_ssa_id);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_dim);
  __Pyx_XDECREF(__pyx_v_dim_ref_counts);
  __Pyx_XDECREF(__pyx_v_footprints);
  __Pyx_XDECREF(__pyx_v_queue);
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_k2s);
  __Pyx_XDECREF(__pyx_v_con);
  __Pyx_XDECREF(__pyx_v_cost);
  __Pyx_XDECREF(__pyx_v_k2);
  __Pyx_XDECREF(__pyx_v_k12);
  __Pyx_XDECREF(__pyx_v_ssa_id1);
  __Pyx_XDECREF(__pyx_v_ssa_id2);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_ssa_id12);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_inputs);
  __Pyx_XDECREF(__pyx_v_output);
  __Pyx_XDECREF(__pyx_v_choose_fn);
  __Pyx_XDECREF(__pyx_v_cost_fn);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":612
 * 
 * 
 * def greedy(inputs, output, size_dict, memory_limit=None, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     Finds the path by a three stage algorithm:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_33greedy(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_32greedy[] = "\n    Finds the path by a three stage algorithm:\n\n    1. Eagerly compute Hadamard products.\n    2. Greedily compute contractions to maximize ``removed_size``\n    3. Greedily compute outer products.\n\n    This algorithm scales quadratically with respect to the\n    maximum number of elements sharing a common dim.\n\n    Parameters\n    ----------\n    inputs : list\n        List of sets that represent the lhs side of the einsum subscript\n    output : set\n        Set that represents the rhs side of the overall einsum subscript\n    size_dict : dictionary\n        Dictionary of index sizes\n    memory_limit : int\n        The maximum number of elements in a temporary array\n    choose_fn : callable, optional\n        A function that chooses which contraction to perform from the queu\n    cost_fn : callable, optional\n        A function that assigns a potential contraction a cost.\n\n    Returns\n    -------\n    path : list\n        The contraction order (a list of tuples of ints).\n\n    Examples\n    --------\n    >>> isets = [set('abd'), set('ac'), set('bdc')]\n    >>> oset = set('')\n    >>> idx_sizes = {'a': 1, 'b':2, 'c':3, 'd':4}\n    >>> greedy(isets, oset, idx_sizes)\n    [(0, 2), (0, 1)]\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_33greedy = {"greedy", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_33greedy, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_32greedy};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_33greedy(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  PyObject *__pyx_v_memory_limit = 0;
  PyObject *__pyx_v_choose_fn = 0;
  PyObject *__pyx_v_cost_fn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("greedy (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,&__pyx_n_s_choose_fn,&__pyx_n_s_cost_fn,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)__pyx_kp_s_memory_removed);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("greedy", 0, 3, 6, 1); __PYX_ERR(0, 612, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("greedy", 0, 3, 6, 2); __PYX_ERR(0, 612, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_choose_fn);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cost_fn);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "greedy") < 0)) __PYX_ERR(0, 612, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
    __pyx_v_size_dict = values[2];
    __pyx_v_memory_limit = values[3];
    __pyx_v_choose_fn = values[4];
    __pyx_v_cost_fn = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("greedy", 0, 3, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 612, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.greedy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_32greedy(__pyx_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit, __pyx_v_choose_fn, __pyx_v_cost_fn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_32greedy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit, PyObject *__pyx_v_choose_fn, PyObject *__pyx_v_cost_fn) {
  PyObject *__pyx_v_ssa_path = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("greedy", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":651
 *     [(0, 2), (0, 1)]
 *     """
 *     if memory_limit not in _UNLIMITED_MEM:             # <<<<<<<<<<<<<<
 *         return branch(inputs, output, size_dict, memory_limit, nbranch=1, cost_fn=cost_fn)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_UNLIMITED_MEM); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_memory_limit, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "acqdp/utility/opt_einsum_paths.pyx":652
 *     """
 *     if memory_limit not in _UNLIMITED_MEM:
 *         return branch(inputs, output, size_dict, memory_limit, nbranch=1, cost_fn=cost_fn)             # <<<<<<<<<<<<<<
 * 
 *     ssa_path = ssa_greedy_optimize(inputs, output, size_dict, cost_fn=cost_fn, choose_fn=choose_fn)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_branch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_inputs);
    __Pyx_GIVEREF(__pyx_v_inputs);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_inputs);
    __Pyx_INCREF(__pyx_v_output);
    __Pyx_GIVEREF(__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_output);
    __Pyx_INCREF(__pyx_v_size_dict);
    __Pyx_GIVEREF(__pyx_v_size_dict);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_size_dict);
    __Pyx_INCREF(__pyx_v_memory_limit);
    __Pyx_GIVEREF(__pyx_v_memory_limit);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_memory_limit);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_nbranch, __pyx_int_1) < 0) __PYX_ERR(0, 652, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_cost_fn, __pyx_v_cost_fn) < 0) __PYX_ERR(0, 652, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":651
 *     [(0, 2), (0, 1)]
 *     """
 *     if memory_limit not in _UNLIMITED_MEM:             # <<<<<<<<<<<<<<
 *         return branch(inputs, output, size_dict, memory_limit, nbranch=1, cost_fn=cost_fn)
 * 
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":654
 *         return branch(inputs, output, size_dict, memory_limit, nbranch=1, cost_fn=cost_fn)
 * 
 *     ssa_path = ssa_greedy_optimize(inputs, output, size_dict, cost_fn=cost_fn, choose_fn=choose_fn)             # <<<<<<<<<<<<<<
 *     return ssa_to_linear(ssa_path)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_ssa_greedy_optimize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_inputs);
  __Pyx_GIVEREF(__pyx_v_inputs);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_inputs);
  __Pyx_INCREF(__pyx_v_output);
  __Pyx_GIVEREF(__pyx_v_output);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_output);
  __Pyx_INCREF(__pyx_v_size_dict);
  __Pyx_GIVEREF(__pyx_v_size_dict);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_size_dict);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_cost_fn, __pyx_v_cost_fn) < 0) __PYX_ERR(0, 654, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_choose_fn, __pyx_v_choose_fn) < 0) __PYX_ERR(0, 654, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_ssa_path = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":655
 * 
 *     ssa_path = ssa_greedy_optimize(inputs, output, size_dict, cost_fn=cost_fn, choose_fn=choose_fn)
 *     return ssa_to_linear(ssa_path)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ssa_to_linear); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_ssa_path) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_ssa_path);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":612
 * 
 * 
 * def greedy(inputs, output, size_dict, memory_limit=None, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     Finds the path by a three stage algorithm:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.greedy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ssa_path);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":658
 * 
 * 
 * def _tree_to_sequence(c):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a contraction tree to a contraction path as it has to be
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_35_tree_to_sequence(PyObject *__pyx_self, PyObject *__pyx_v_c); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_34_tree_to_sequence[] = "\n    Converts a contraction tree to a contraction path as it has to be\n    returned by path optimizers. A contraction tree can either be an int\n    (=no contraction) or a tuple containing the terms to be contracted. An\n    arbitrary number (>= 1) of terms can be contracted at once. Note that\n    contractions are commutative, e.g. (j, k, l) = (k, l, j). Note that in\n    general, solutions are not unique.\n\n    Parameters\n    ----------\n    c : tuple or int\n        Contraction tree\n\n    Returns\n    -------\n    path : list[set[int]]\n        Contraction path\n\n    Examples\n    --------\n    >>> _tree_to_sequence(((1,2),(0,(4,5,3))))\n    [(1, 2), (1, 2, 3), (0, 2), (0, 1)]\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_35_tree_to_sequence = {"_tree_to_sequence", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_35_tree_to_sequence, METH_O, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_34_tree_to_sequence};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_35_tree_to_sequence(PyObject *__pyx_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tree_to_sequence (wrapper)", 0);
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_34_tree_to_sequence(__pyx_self, ((PyObject *)__pyx_v_c));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_17_tree_to_sequence_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":706
 * 
 *         for i in sorted([i for i in j if type(i) == int]):
 *             s[0] += (sum(1 for q in t if q < i),)             # <<<<<<<<<<<<<<
 *             t.insert(s[0][-1], i)
 * 
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_17_tree_to_sequence_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 706, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_17_tree_to_sequence_2generator6, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_tree_to_sequence_locals_genexpr, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 706, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._tree_to_sequence.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_17_tree_to_sequence_2generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 706, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_t)) { __Pyx_RaiseClosureNameError("t"); __PYX_ERR(0, 706, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_t == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 706, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_t; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 706, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 706, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_q);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_q, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_i)) { __Pyx_RaiseClosureNameError("i"); __PYX_ERR(0, 706, __pyx_L1_error) }
    __pyx_t_3 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_q, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_i, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 706, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 706, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {
      __Pyx_INCREF(__pyx_int_1);
      __pyx_r = __pyx_int_1;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 706, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":658
 * 
 * 
 * def _tree_to_sequence(c):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a contraction tree to a contraction path as it has to be
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_34_tree_to_sequence(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_c) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *__pyx_cur_scope;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_tree_to_sequence", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 658, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __Pyx_INCREF(__pyx_v_c);

  /* "acqdp/utility/opt_einsum_paths.pyx":694
 *     # this function iterates through the table shown above from right to left;
 * 
 *     if type(c) == int:             # <<<<<<<<<<<<<<
 *         return []
 * 
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_c)), ((PyObject *)(&PyInt_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "acqdp/utility/opt_einsum_paths.pyx":695
 * 
 *     if type(c) == int:
 *         return []             # <<<<<<<<<<<<<<
 * 
 *     c = [c]  # list of remaining contractions (lower part of columns shown above)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "acqdp/utility/opt_einsum_paths.pyx":694
 *     # this function iterates through the table shown above from right to left;
 * 
 *     if type(c) == int:             # <<<<<<<<<<<<<<
 *         return []
 * 
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":697
 *         return []
 * 
 *     c = [c]  # list of remaining contractions (lower part of columns shown above)             # <<<<<<<<<<<<<<
 *     t = []   # list of elementary tensors (upper part of colums)
 *     s = []   # resulting contraction sequence
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_c);
  __Pyx_GIVEREF(__pyx_v_c);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_c);
  __Pyx_DECREF_SET(__pyx_v_c, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":698
 * 
 *     c = [c]  # list of remaining contractions (lower part of columns shown above)
 *     t = []   # list of elementary tensors (upper part of colums)             # <<<<<<<<<<<<<<
 *     s = []   # resulting contraction sequence
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_t = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":699
 *     c = [c]  # list of remaining contractions (lower part of columns shown above)
 *     t = []   # list of elementary tensors (upper part of colums)
 *     s = []   # resulting contraction sequence             # <<<<<<<<<<<<<<
 * 
 *     while len(c) > 0:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":701
 *     s = []   # resulting contraction sequence
 * 
 *     while len(c) > 0:             # <<<<<<<<<<<<<<
 *         j = c.pop(-1)
 *         s.insert(0, tuple())
 */
  while (1) {
    __pyx_t_3 = PyObject_Length(__pyx_v_c); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 701, __pyx_L1_error)
    __pyx_t_2 = ((__pyx_t_3 > 0) != 0);
    if (!__pyx_t_2) break;

    /* "acqdp/utility/opt_einsum_paths.pyx":702
 * 
 *     while len(c) > 0:
 *         j = c.pop(-1)             # <<<<<<<<<<<<<<
 *         s.insert(0, tuple())
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_PopIndex(__pyx_v_c, __pyx_int_neg_1, -1L, 1, Py_ssize_t, PyInt_FromSsize_t); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":703
 *     while len(c) > 0:
 *         j = c.pop(-1)
 *         s.insert(0, tuple())             # <<<<<<<<<<<<<<
 * 
 *         for i in sorted([i for i in j if type(i) == int]):
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyTuple_Type))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyList_Insert(__pyx_v_s, 0, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":705
 *         s.insert(0, tuple())
 * 
 *         for i in sorted([i for i in j if type(i) == int]):             # <<<<<<<<<<<<<<
 *             s[0] += (sum(1 for q in t if q < i),)
 *             t.insert(s[0][-1], i)
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_v_j)) || PyTuple_CheckExact(__pyx_v_j)) {
      __pyx_t_6 = __pyx_v_j; __Pyx_INCREF(__pyx_t_6); __pyx_t_3 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_j); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 705, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 705, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_6);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 705, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_i)), ((PyObject *)(&PyInt_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 705, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_2) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_cur_scope->__pyx_v_i))) __PYX_ERR(0, 705, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_4 = PyList_Sort(__pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 705, __pyx_L1_error)
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 705, __pyx_L1_error)
    }
    __pyx_t_5 = __pyx_t_1; __Pyx_INCREF(__pyx_t_5); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_5)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 705, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":706
 * 
 *         for i in sorted([i for i in j if type(i) == int]):
 *             s[0] += (sum(1 for q in t if q < i),)             # <<<<<<<<<<<<<<
 *             t.insert(s[0][-1], i)
 * 
 */
      __pyx_t_9 = 0;
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_s, __pyx_t_9, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_17_tree_to_sequence_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(__Pyx_SetItemInt(__pyx_v_s, __pyx_t_9, __pyx_t_8, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0)) __PYX_ERR(0, 706, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":707
 *         for i in sorted([i for i in j if type(i) == int]):
 *             s[0] += (sum(1 for q in t if q < i),)
 *             t.insert(s[0][-1], i)             # <<<<<<<<<<<<<<
 * 
 *         for i in [i for i in j if type(i) != int]:
 */
      __pyx_t_8 = __Pyx_GetItemInt_List(__pyx_v_s, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_8, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __pyx_cur_scope->__pyx_v_i;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_4 = PyList_Insert(__pyx_cur_scope->__pyx_v_t, __pyx_t_9, __pyx_t_6); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":705
 *         s.insert(0, tuple())
 * 
 *         for i in sorted([i for i in j if type(i) == int]):             # <<<<<<<<<<<<<<
 *             s[0] += (sum(1 for q in t if q < i),)
 *             t.insert(s[0][-1], i)
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":709
 *             t.insert(s[0][-1], i)
 * 
 *         for i in [i for i in j if type(i) != int]:             # <<<<<<<<<<<<<<
 *             s[0] += (len(t) + len(c),)
 *             c.append(i)
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_v_j)) || PyTuple_CheckExact(__pyx_v_j)) {
      __pyx_t_6 = __pyx_v_j; __Pyx_INCREF(__pyx_t_6); __pyx_t_3 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_j); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 709, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_6, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_6);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 709, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_cur_scope->__pyx_v_i)), ((PyObject *)(&PyInt_Type)), Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_2) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_cur_scope->__pyx_v_i))) __PYX_ERR(0, 709, __pyx_L1_error)
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_6)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 709, __pyx_L1_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __pyx_t_5 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":710
 * 
 *         for i in [i for i in j if type(i) != int]:
 *             s[0] += (len(t) + len(c),)             # <<<<<<<<<<<<<<
 *             c.append(i)
 * 
 */
      __pyx_t_9 = 0;
      __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_s, __pyx_t_9, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __pyx_cur_scope->__pyx_v_t;
      __Pyx_INCREF(__pyx_t_8);
      if (unlikely(__pyx_t_8 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 710, __pyx_L1_error)
      }
      __pyx_t_10 = PyList_GET_SIZE(__pyx_t_8); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_11 = PyObject_Length(__pyx_v_c); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 710, __pyx_L1_error)
      __pyx_t_8 = PyInt_FromSsize_t((__pyx_t_10 + __pyx_t_11)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(__Pyx_SetItemInt(__pyx_v_s, __pyx_t_9, __pyx_t_8, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1) < 0)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":711
 *         for i in [i for i in j if type(i) != int]:
 *             s[0] += (len(t) + len(c),)
 *             c.append(i)             # <<<<<<<<<<<<<<
 * 
 *     return s
 */
      __pyx_t_4 = __Pyx_PyObject_Append(__pyx_v_c, __pyx_cur_scope->__pyx_v_i); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 711, __pyx_L1_error)

      /* "acqdp/utility/opt_einsum_paths.pyx":709
 *             t.insert(s[0][-1], i)
 * 
 *         for i in [i for i in j if type(i) != int]:             # <<<<<<<<<<<<<<
 *             s[0] += (len(t) + len(c),)
 *             c.append(i)
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":713
 *             c.append(i)
 * 
 *     return s             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_s);
  __pyx_r = __pyx_v_s;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":658
 * 
 * 
 * def _tree_to_sequence(c):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a contraction tree to a contraction path as it has to be
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._tree_to_sequence", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":716
 * 
 * 
 * def _find_disconnected_subgraphs(inputs, output):             # <<<<<<<<<<<<<<
 *     """
 *     Finds disconnected subgraphs in the given list of inputs. Inputs are
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_37_find_disconnected_subgraphs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_36_find_disconnected_subgraphs[] = "\n    Finds disconnected subgraphs in the given list of inputs. Inputs are\n    connected if they share summation indices. Note: Disconnected subgraphs\n    can be contracted independently before forming outer products.\n\n    Parameters\n    ----------\n    inputs : list[set]\n        List of sets that represent the lhs side of the einsum subscript\n    output : set\n        Set that represents the rhs side of the overall einsum subscript\n\n    Returns\n    -------\n    subgraphs : list[set[int]]\n        List containing sets of indices for each subgraph\n\n    Examples\n    --------\n    >>> _find_disconnected_subgraphs([set(\"ab\"), set(\"c\"), set(\"ad\")], set(\"bd\"))\n    [{0, 2}, {1}]\n\n    >>> _find_disconnected_subgraphs([set(\"ab\"), set(\"c\"), set(\"ad\")], set(\"abd\"))\n    [{0}, {1}, {2}]\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_37_find_disconnected_subgraphs = {"_find_disconnected_subgraphs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_37_find_disconnected_subgraphs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_36_find_disconnected_subgraphs};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_37_find_disconnected_subgraphs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_find_disconnected_subgraphs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_find_disconnected_subgraphs", 1, 2, 2, 1); __PYX_ERR(0, 716, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_find_disconnected_subgraphs") < 0)) __PYX_ERR(0, 716, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_find_disconnected_subgraphs", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 716, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._find_disconnected_subgraphs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_36_find_disconnected_subgraphs(__pyx_self, __pyx_v_inputs, __pyx_v_output);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_36_find_disconnected_subgraphs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output) {
  PyObject *__pyx_v_subgraphs = NULL;
  PyObject *__pyx_v_unused_inputs = NULL;
  PyObject *__pyx_v_i_sum = NULL;
  PyObject *__pyx_v_g = NULL;
  PyObject *__pyx_v_q = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_v_i_tmp = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_8genexpr9__pyx_v_k = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_find_disconnected_subgraphs", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":743
 *     """
 * 
 *     subgraphs = []             # <<<<<<<<<<<<<<
 *     unused_inputs = set(range(len(inputs)))
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_subgraphs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":744
 * 
 *     subgraphs = []
 *     unused_inputs = set(range(len(inputs)))             # <<<<<<<<<<<<<<
 * 
 *     i_sum = set.union(*inputs) - output  # all summation indices
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 744, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySet_New(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_unused_inputs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":746
 *     unused_inputs = set(range(len(inputs)))
 * 
 *     i_sum = set.union(*inputs) - output  # all summation indices             # <<<<<<<<<<<<<<
 * 
 *     while len(unused_inputs) > 0:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PySet_Type)), __pyx_n_s_union); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_inputs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_4, __pyx_v_output); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_i_sum = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":748
 *     i_sum = set.union(*inputs) - output  # all summation indices
 * 
 *     while len(unused_inputs) > 0:             # <<<<<<<<<<<<<<
 *         g = set()
 *         q = [unused_inputs.pop()]
 */
  while (1) {
    __pyx_t_2 = PySet_GET_SIZE(__pyx_v_unused_inputs); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 748, __pyx_L1_error)
    __pyx_t_5 = ((__pyx_t_2 > 0) != 0);
    if (!__pyx_t_5) break;

    /* "acqdp/utility/opt_einsum_paths.pyx":749
 * 
 *     while len(unused_inputs) > 0:
 *         g = set()             # <<<<<<<<<<<<<<
 *         q = [unused_inputs.pop()]
 *         while len(q) > 0:
 */
    __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 749, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_g, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":750
 *     while len(unused_inputs) > 0:
 *         g = set()
 *         q = [unused_inputs.pop()]             # <<<<<<<<<<<<<<
 *         while len(q) > 0:
 *             j = q.pop()
 */
    __pyx_t_3 = PySet_Pop(__pyx_v_unused_inputs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_q, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":751
 *         g = set()
 *         q = [unused_inputs.pop()]
 *         while len(q) > 0:             # <<<<<<<<<<<<<<
 *             j = q.pop()
 *             g.add(j)
 */
    while (1) {
      __pyx_t_2 = PyList_GET_SIZE(__pyx_v_q); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 751, __pyx_L1_error)
      __pyx_t_5 = ((__pyx_t_2 > 0) != 0);
      if (!__pyx_t_5) break;

      /* "acqdp/utility/opt_einsum_paths.pyx":752
 *         q = [unused_inputs.pop()]
 *         while len(q) > 0:
 *             j = q.pop()             # <<<<<<<<<<<<<<
 *             g.add(j)
 *             i_tmp = i_sum & inputs[j]
 */
      __pyx_t_4 = __Pyx_PyList_Pop(__pyx_v_q); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 752, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":753
 *         while len(q) > 0:
 *             j = q.pop()
 *             g.add(j)             # <<<<<<<<<<<<<<
 *             i_tmp = i_sum & inputs[j]
 *             n = {k for k in unused_inputs if len(i_tmp & inputs[k]) > 0}
 */
      __pyx_t_6 = PySet_Add(__pyx_v_g, __pyx_v_j); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 753, __pyx_L1_error)

      /* "acqdp/utility/opt_einsum_paths.pyx":754
 *             j = q.pop()
 *             g.add(j)
 *             i_tmp = i_sum & inputs[j]             # <<<<<<<<<<<<<<
 *             n = {k for k in unused_inputs if len(i_tmp & inputs[k]) > 0}
 *             q.extend(n)
 */
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_inputs, __pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyNumber_And(__pyx_v_i_sum, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_i_tmp, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":755
 *             g.add(j)
 *             i_tmp = i_sum & inputs[j]
 *             n = {k for k in unused_inputs if len(i_tmp & inputs[k]) > 0}             # <<<<<<<<<<<<<<
 *             q.extend(n)
 *             unused_inputs.difference_update(n)
 */
      { /* enter inner scope */
        __pyx_t_3 = PySet_New(NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = 0;
        __pyx_t_1 = __Pyx_set_iterator(__pyx_v_unused_inputs, 1, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF(__pyx_t_4);
        __pyx_t_4 = __pyx_t_1;
        __pyx_t_1 = 0;
        while (1) {
          __pyx_t_9 = __Pyx_set_iter_next(__pyx_t_4, __pyx_t_7, &__pyx_t_2, &__pyx_t_1, __pyx_t_8);
          if (unlikely(__pyx_t_9 == 0)) break;
          if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 755, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_k, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_inputs, __pyx_8genexpr9__pyx_v_k); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = PyNumber_And(__pyx_v_i_tmp, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 755, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = PyObject_Length(__pyx_t_10); if (unlikely(__pyx_t_11 == ((Py_ssize_t)-1))) __PYX_ERR(0, 755, __pyx_L9_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_5 = ((__pyx_t_11 > 0) != 0);
          if (__pyx_t_5) {
            if (unlikely(PySet_Add(__pyx_t_3, (PyObject*)__pyx_8genexpr9__pyx_v_k))) __PYX_ERR(0, 755, __pyx_L9_error)
          }
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_k); __pyx_8genexpr9__pyx_v_k = 0;
        goto __pyx_L13_exit_scope;
        __pyx_L9_error:;
        __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_k); __pyx_8genexpr9__pyx_v_k = 0;
        goto __pyx_L1_error;
        __pyx_L13_exit_scope:;
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_n, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":756
 *             i_tmp = i_sum & inputs[j]
 *             n = {k for k in unused_inputs if len(i_tmp & inputs[k]) > 0}
 *             q.extend(n)             # <<<<<<<<<<<<<<
 *             unused_inputs.difference_update(n)
 * 
 */
      __pyx_t_6 = __Pyx_PyList_Extend(__pyx_v_q, __pyx_v_n); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 756, __pyx_L1_error)

      /* "acqdp/utility/opt_einsum_paths.pyx":757
 *             n = {k for k in unused_inputs if len(i_tmp & inputs[k]) > 0}
 *             q.extend(n)
 *             unused_inputs.difference_update(n)             # <<<<<<<<<<<<<<
 * 
 *         subgraphs.append(g)
 */
      __pyx_t_3 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PySet_Type_difference_update, __pyx_v_unused_inputs, __pyx_v_n); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":759
 *             unused_inputs.difference_update(n)
 * 
 *         subgraphs.append(g)             # <<<<<<<<<<<<<<
 * 
 *     return subgraphs
 */
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_subgraphs, __pyx_v_g); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 759, __pyx_L1_error)
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":761
 *         subgraphs.append(g)
 * 
 *     return subgraphs             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_subgraphs);
  __pyx_r = __pyx_v_subgraphs;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":716
 * 
 * 
 * def _find_disconnected_subgraphs(inputs, output):             # <<<<<<<<<<<<<<
 *     """
 *     Finds disconnected subgraphs in the given list of inputs. Inputs are
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._find_disconnected_subgraphs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_subgraphs);
  __Pyx_XDECREF(__pyx_v_unused_inputs);
  __Pyx_XDECREF(__pyx_v_i_sum);
  __Pyx_XDECREF(__pyx_v_g);
  __Pyx_XDECREF(__pyx_v_q);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_i_tmp);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_k);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_40generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":764
 * 
 * 
 * def _bitmapset_indices(s):             # <<<<<<<<<<<<<<
 *     """
 *     Returns a generator object allowing to iterate over the elements contained
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_39_bitmapset_indices(PyObject *__pyx_self, PyObject *__pyx_v_s); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_38_bitmapset_indices[] = "\n    Returns a generator object allowing to iterate over the elements contained\n    in a bitmap set.\n\n    Parameters\n    ----------\n    s : int\n        The bitmap set to iterate over\n\n    Returns\n    -------\n    path : generator\n        Generator object to iterate over the elements in s\n\n    Examples\n    --------\n    >>> type(_bitmapset_indices(0b1001011))\n    generator\n\n    >>> list(_bitmapset_indices(0b1001011))\n    [0, 1, 3, 6]\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_39_bitmapset_indices = {"_bitmapset_indices", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_39_bitmapset_indices, METH_O, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_38_bitmapset_indices};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_39_bitmapset_indices(PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bitmapset_indices (wrapper)", 0);
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_38_bitmapset_indices(__pyx_self, ((PyObject *)__pyx_v_s));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_38_bitmapset_indices(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_bitmapset_indices", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 764, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_s = __pyx_v_s;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_s);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_s);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_40generator, __pyx_codeobj__12, (PyObject *) __pyx_cur_scope, __pyx_n_s_bitmapset_indices, __pyx_n_s_bitmapset_indices, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 764, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths._bitmapset_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_40generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bitmapset_indices", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 764, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":787
 *     [0, 1, 3, 6]
 *     """
 *     j = 0             # <<<<<<<<<<<<<<
 *     while s != 0:
 *         if s & 1 != 0:
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __pyx_cur_scope->__pyx_v_j = __pyx_int_0;

  /* "acqdp/utility/opt_einsum_paths.pyx":788
 *     """
 *     j = 0
 *     while s != 0:             # <<<<<<<<<<<<<<
 *         if s & 1 != 0:
 *             yield j
 */
  while (1) {
    __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_cur_scope->__pyx_v_s, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "acqdp/utility/opt_einsum_paths.pyx":789
 *     j = 0
 *     while s != 0:
 *         if s & 1 != 0:             # <<<<<<<<<<<<<<
 *             yield j
 *         s >>= 1
 */
    __pyx_t_1 = __Pyx_PyInt_AndObjC(__pyx_cur_scope->__pyx_v_s, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 789, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 789, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 789, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "acqdp/utility/opt_einsum_paths.pyx":790
 *     while s != 0:
 *         if s & 1 != 0:
 *             yield j             # <<<<<<<<<<<<<<
 *         s >>= 1
 *         j += 1
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_j);
      __pyx_r = __pyx_cur_scope->__pyx_v_j;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 790, __pyx_L1_error)

      /* "acqdp/utility/opt_einsum_paths.pyx":789
 *     j = 0
 *     while s != 0:
 *         if s & 1 != 0:             # <<<<<<<<<<<<<<
 *             yield j
 *         s >>= 1
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":791
 *         if s & 1 != 0:
 *             yield j
 *         s >>= 1             # <<<<<<<<<<<<<<
 *         j += 1
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_RshiftObjC(__pyx_cur_scope->__pyx_v_s, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":792
 *             yield j
 *         s >>= 1
 *         j += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_j, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_j);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_j, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "acqdp/utility/opt_einsum_paths.pyx":764
 * 
 * 
 * def _bitmapset_indices(s):             # <<<<<<<<<<<<<<
 *     """
 *     Returns a generator object allowing to iterate over the elements contained
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("_bitmapset_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":811
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_1__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer___call__[] = "\n        Parameters\n        ----------\n        inputs : list\n            List of sets that represent the lhs side of the einsum subscript\n        output : set\n            Set that represents the rhs side of the overall einsum subscript\n        size_dict : dictionary\n            Dictionary of index sizes\n        memory_limit : int\n            The maximum number of elements in a temporary array\n\n        Returns\n        -------\n        path : list\n            The contraction order (a list of tuples of ints).\n\n        Examples\n        --------\n        >>> n_in = 3  # exponential scaling\n        >>> n_out = 2 # linear scaling\n        >>> s = dict()\n        >>> i_all = []\n        >>> for _ in range(n_out):\n        >>>     i = [set() for _ in range(n_in)]\n        >>>     for j in range(n_in):\n        >>>         for k in range(j+1, n_in):\n        >>>             c = oe.get_symbol(len(s))\n        >>>             i[j].add(c)\n        >>>             i[k].add(c)\n        >>>             s[c] = 2\n        >>>     i_all.extend(i)\n        >>> o = DynamicProgrammingOptimizer()\n        >>> o(i_all, set(), s)\n        [(1, 2), (0, 4), (1, 2), (0, 2), (0, 1)]\n        ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_1__call__ = {"__call__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_1__call__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer___call__};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_1__call__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  PyObject *__pyx_v_memory_limit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[4] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 1); __PYX_ERR(0, 811, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 2); __PYX_ERR(0, 811, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, 3); __PYX_ERR(0, 811, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__call__") < 0)) __PYX_ERR(0, 811, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_self = values[0];
    __pyx_v_inputs = values[1];
    __pyx_v_output = values[2];
    __pyx_v_size_dict = values[3];
    __pyx_v_memory_limit = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 811, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer___call__(__pyx_self, __pyx_v_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___2generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":851
 *         # convert all indices to integers (makes set operations ~10 % faster)
 *         symbol2int = {c: j for j, c in enumerate(set.union(*inputs) | output)}
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]             # <<<<<<<<<<<<<<
 *         output = set(symbol2int[c] for c in output)
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 851, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___2generator7, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_DynamicProgrammingOptimizer___ca, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___2generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 851, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_i)) { __Pyx_RaiseClosureNameError("i"); __PYX_ERR(0, 851, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_i)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_i)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_i; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 851, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 851, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 851, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 851, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_symbol2int)) { __Pyx_RaiseClosureNameError("symbol2int"); __PYX_ERR(0, 851, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_symbol2int == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 851, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_symbol2int, __pyx_cur_scope->__pyx_v_c); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___5generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":852
 *         symbol2int = {c: j for j, c in enumerate(set.union(*inputs) | output)}
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]
 *         output = set(symbol2int[c] for c in output)             # <<<<<<<<<<<<<<
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}
 *         size_dict = [size_dict[j] for j in range(len(size_dict))]
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 852, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___5generator8, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_DynamicProgrammingOptimizer___ca, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___5generator8(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 852, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output)) { __Pyx_RaiseClosureNameError("output"); __PYX_ERR(0, 852, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 852, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 852, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 852, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 852, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 852, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_symbol2int)) { __Pyx_RaiseClosureNameError("symbol2int"); __PYX_ERR(0, 852, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_symbol2int == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 852, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_symbol2int, __pyx_cur_scope->__pyx_v_c); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 852, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___8generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___7genexpr_2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":859
 *         i_single = set(
 *             c for c in set.union(*inputs) - output
 *             if sum(1 for i in inputs if c in i) == 1             # <<<<<<<<<<<<<<
 *         )
 * 
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___7genexpr_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 859, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___7genexpr_2generator10, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_DynamicProgrammingOptimizer___ca_2, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.genexpr.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___7genexpr_2generator10(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 859, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 859, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_inputs)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_inputs)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_inputs; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_outer_scope->__pyx_v_inputs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 859, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 859, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 859, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_c)) { __Pyx_RaiseClosureNameError("c"); __PYX_ERR(0, 859, __pyx_L1_error) }
    __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_c, __pyx_cur_scope->__pyx_v_i, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 859, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_int_1);
      __pyx_r = __pyx_int_1;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 859, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":858
 *         # all summation indices occurring exactly in one input:
 *         i_single = set(
 *             c for c in set.union(*inputs) - output             # <<<<<<<<<<<<<<
 *             if sum(1 for i in inputs if c in i) == 1
 *         )
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___6genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 858, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___8generator9, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_DynamicProgrammingOptimizer___ca, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___8generator9(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 858, __pyx_L1_error)
  __pyx_r = PySet_New(NULL); if (unlikely(!__pyx_r)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PySet_Type)), __pyx_n_s_union); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 858, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_inputs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_output)) { __Pyx_RaiseClosureNameError("output"); __PYX_ERR(0, 858, __pyx_L1_error) }
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_3, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_output); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 858, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 858, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 858, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 858, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 858, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 858, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_c);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_c, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":859
 *         i_single = set(
 *             c for c in set.union(*inputs) - output
 *             if sum(1 for i in inputs if c in i) == 1             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    __pyx_t_2 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___7genexpr_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_6) {

      /* "acqdp/utility/opt_einsum_paths.pyx":858
 *         # all summation indices occurring exactly in one input:
 *         i_single = set(
 *             c for c in set.union(*inputs) - output             # <<<<<<<<<<<<<<
 *             if sum(1 for i in inputs if c in i) == 1
 *         )
 */
      if (unlikely(PySet_Add(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_c))) __PYX_ERR(0, 858, __pyx_L1_error)

      /* "acqdp/utility/opt_einsum_paths.pyx":859
 *         i_single = set(
 *             c for c in set.union(*inputs) - output
 *             if sum(1 for i in inputs if c in i) == 1             # <<<<<<<<<<<<<<
 *         )
 * 
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":858
 *         # all summation indices occurring exactly in one input:
 *         i_single = set(
 *             c for c in set.union(*inputs) - output             # <<<<<<<<<<<<<<
 *             if sum(1 for i in inputs if c in i) == 1
 *         )
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":894
 * 
 *             # convert set of tensors g to a bitmap set:
 *             g = functools.reduce(lambda x, y: x | y, (1 << j for j in g))             # <<<<<<<<<<<<<<
 * 
 *             # the bitmap set of all tensors is computed as it is needed to
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___9lambda11(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___9lambda11 = {"lambda11", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___9lambda11, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___9lambda11(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_y = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda11 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda11", 1, 2, 2, 1); __PYX_ERR(0, 894, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lambda11") < 0)) __PYX_ERR(0, 894, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = values[0];
    __pyx_v_y = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda11", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 894, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.lambda11", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda11(__pyx_self, __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda11(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda11", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Or(__pyx_v_x, __pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.lambda11", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___12generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___10genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 894, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___12generator11, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_DynamicProgrammingOptimizer___ca, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___12generator11(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 894, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_g)) { __Pyx_RaiseClosureNameError("g"); __PYX_ERR(0, 894, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_g)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_g)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_g; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_g); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 894, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 894, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 894, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 894, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 894, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_j);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_j, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_cur_scope->__pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 894, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___15generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":906
 *             # output index dimensions as initial cost_cap
 *             cost_cap = helpers.compute_size_by_dict(
 *                 set.union(*(inputs[j] for j in _bitmapset_indices(g))) & output,             # <<<<<<<<<<<<<<
 *                 size_dict
 *             )
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___13genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 906, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___15generator12, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_DynamicProgrammingOptimizer___ca, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___15generator12(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bitmapset_indices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_g)) { __Pyx_RaiseClosureNameError("g"); __PYX_ERR(0, 906, __pyx_L1_error) }
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_g) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_g);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 906, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 906, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 906, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 906, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_j);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_j, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 906, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_v_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 906, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___18generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "acqdp/utility/opt_einsum_paths.pyx":939
 * 
 *                                                     # indices of remaining indices:
 *                                                     i_r = (set.union(*(inputs[j] for j in _bitmapset_indices(r)))             # <<<<<<<<<<<<<<
 *                                                            if r != 0 else set())
 * 
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___16genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 939, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___18generator13, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_DynamicProgrammingOptimizer___ca, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!gen)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___18generator13(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *__pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_bitmapset_indices); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_r)) { __Pyx_RaiseClosureNameError("r"); __PYX_ERR(0, 939, __pyx_L1_error) }
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_r) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_r);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 939, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 939, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 939, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 939, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_j);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_j, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_inputs)) { __Pyx_RaiseClosureNameError("inputs"); __PYX_ERR(0, 939, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_v_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 939, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":966
 * 
 *         # build the final contraction tree
 *         tree = functools.reduce(lambda x, y: (x, y), subgraph_contractions)             # <<<<<<<<<<<<<<
 * 
 *         return _tree_to_sequence(tree)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___19lambda15(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___19lambda15 = {"lambda15", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___19lambda15, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___19lambda15(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_y = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda15 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lambda15", 1, 2, 2, 1); __PYX_ERR(0, 966, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lambda15") < 0)) __PYX_ERR(0, 966, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = values[0];
    __pyx_v_y = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda15", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 966, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.lambda15", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda15(__pyx_self, __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda15(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda15", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_x);
  __Pyx_GIVEREF(__pyx_v_x);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_x);
  __Pyx_INCREF(__pyx_v_y);
  __Pyx_GIVEREF(__pyx_v_y);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_y);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__.lambda15", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":811
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer___call__(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_cur_scope;
  PyObject *__pyx_v_i_single = NULL;
  PyObject *__pyx_v_inputs_done = NULL;
  PyObject *__pyx_v_inputs_contractions = NULL;
  PyObject *__pyx_v_subgraph_contractions = NULL;
  PyObject *__pyx_v_subgraph_contractions_size = NULL;
  PyObject *__pyx_v_x = NULL;
  Py_ssize_t __pyx_v_all_tensors;
  PyObject *__pyx_v_cost_cap = NULL;
  PyObject *__pyx_v_n = NULL;
  PyObject *__pyx_v_xn = NULL;
  PyObject *__pyx_v_m = NULL;
  PyObject *__pyx_v_s1 = NULL;
  PyObject *__pyx_v_i1 = NULL;
  PyObject *__pyx_v_cost1 = NULL;
  PyObject *__pyx_v_cntrct1 = NULL;
  PyObject *__pyx_v_s2 = NULL;
  PyObject *__pyx_v_i2 = NULL;
  PyObject *__pyx_v_cost2 = NULL;
  PyObject *__pyx_v_cntrct2 = NULL;
  PyObject *__pyx_v_i1_cut_i2_wo_output = NULL;
  PyObject *__pyx_v_i1_union_i2 = NULL;
  PyObject *__pyx_v_cost = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_i_r = NULL;
  PyObject *__pyx_v_i_cntrct = NULL;
  PyObject *__pyx_v_mem = NULL;
  PyObject *__pyx_v_contraction = NULL;
  PyObject *__pyx_v_tree = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_j = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_c = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_c = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_v = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  Py_ssize_t __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  Py_ssize_t __pyx_t_19;
  PyObject *(*__pyx_t_20)(PyObject *);
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  Py_ssize_t __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  PyObject *__pyx_t_26 = NULL;
  Py_ssize_t __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);
  __pyx_cur_scope = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *)__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__(__pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 811, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_inputs = __pyx_v_inputs;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_inputs);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_inputs);
  __pyx_cur_scope->__pyx_v_output = __pyx_v_output;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_INCREF(__pyx_v_size_dict);

  /* "acqdp/utility/opt_einsum_paths.pyx":850
 * 
 *         # convert all indices to integers (makes set operations ~10 % faster)
 *         symbol2int = {c: j for j, c in enumerate(set.union(*inputs) | output)}             # <<<<<<<<<<<<<<
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]
 *         output = set(symbol2int[c] for c in output)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_2 = __pyx_int_0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PySet_Type)), __pyx_n_s_union); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 850, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_cur_scope->__pyx_v_inputs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 850, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 850, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Or(__pyx_t_5, __pyx_cur_scope->__pyx_v_output); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 850, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 850, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 850, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 850, __pyx_L5_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 850, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 850, __pyx_L5_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 850, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 850, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_c, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_j, __pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 850, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_4;
      __pyx_t_4 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_9genexpr10__pyx_v_c, (PyObject*)__pyx_9genexpr10__pyx_v_j))) __PYX_ERR(0, 850, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_c); __pyx_9genexpr10__pyx_v_c = 0;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_j); __pyx_9genexpr10__pyx_v_j = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_c); __pyx_9genexpr10__pyx_v_c = 0;
    __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_j); __pyx_9genexpr10__pyx_v_j = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_symbol2int = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":851
 *         # convert all indices to integers (makes set operations ~10 % faster)
 *         symbol2int = {c: j for j, c in enumerate(set.union(*inputs) | output)}
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]             # <<<<<<<<<<<<<<
 *         output = set(symbol2int[c] for c in output)
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_inputs)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_inputs)) {
    __pyx_t_2 = __pyx_cur_scope->__pyx_v_inputs; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_inputs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 851, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 851, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 851, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 851, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 851, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_inputs);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_inputs, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":852
 *         symbol2int = {c: j for j, c in enumerate(set.union(*inputs) | output)}
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]
 *         output = set(symbol2int[c] for c in output)             # <<<<<<<<<<<<<<
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}
 *         size_dict = [size_dict[j] for j in range(len(size_dict))]
 */
  __pyx_t_1 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_output);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_output, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":853
 *         inputs = [set(symbol2int[c] for c in i) for i in inputs]
 *         output = set(symbol2int[c] for c in output)
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}             # <<<<<<<<<<<<<<
 *         size_dict = [size_dict[j] for j in range(len(size_dict))]
 * 
 */
  { /* enter inner scope */
    __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 853, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_size_dict, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 853, __pyx_L13_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 853, __pyx_L13_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 853, __pyx_L13_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 853, __pyx_L13_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 853, __pyx_L13_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 853, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 853, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 853, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 853, __pyx_L13_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L17_unpacking_done;
        __pyx_L16_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 853, __pyx_L13_error)
        __pyx_L17_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_c, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_v, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_10 = (__Pyx_PyDict_ContainsTF(__pyx_9genexpr13__pyx_v_c, __pyx_cur_scope->__pyx_v_symbol2int, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 853, __pyx_L13_error)
      __pyx_t_11 = (__pyx_t_10 != 0);
      if (__pyx_t_11) {
        __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_cur_scope->__pyx_v_symbol2int, __pyx_9genexpr13__pyx_v_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(PyDict_SetItem(__pyx_t_2, (PyObject*)__pyx_t_1, (PyObject*)__pyx_9genexpr13__pyx_v_v))) __PYX_ERR(0, 853, __pyx_L13_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_c); __pyx_9genexpr13__pyx_v_c = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_v); __pyx_9genexpr13__pyx_v_v = 0;
    goto __pyx_L19_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_c); __pyx_9genexpr13__pyx_v_c = 0;
    __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_v); __pyx_9genexpr13__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L19_exit_scope:;
  } /* exit inner scope */
  __Pyx_DECREF_SET(__pyx_v_size_dict, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":854
 *         output = set(symbol2int[c] for c in output)
 *         size_dict = {symbol2int[c]: v for c, v in size_dict.items() if c in symbol2int}
 *         size_dict = [size_dict[j] for j in range(len(size_dict))]             # <<<<<<<<<<<<<<
 * 
 *         # all summation indices occurring exactly in one input:
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyObject_Length(__pyx_v_size_dict); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 854, __pyx_L1_error)
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_4 = __pyx_t_1; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 854, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 854, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 854, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 854, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_size_dict, __pyx_v_j); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 854, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_size_dict, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":858
 *         # all summation indices occurring exactly in one input:
 *         i_single = set(
 *             c for c in set.union(*inputs) - output             # <<<<<<<<<<<<<<
 *             if sum(1 for i in inputs if c in i) == 1
 *         )
 */
  __pyx_t_2 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___6genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "acqdp/utility/opt_einsum_paths.pyx":857
 * 
 *         # all summation indices occurring exactly in one input:
 *         i_single = set(             # <<<<<<<<<<<<<<
 *             c for c in set.union(*inputs) - output
 *             if sum(1 for i in inputs if c in i) == 1
 */
  __pyx_t_4 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_i_single = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":864
 *         # contraction expressions for all inputs that have already been
 *         # reduced to scalars:
 *         inputs_done = [             # <<<<<<<<<<<<<<
 *             (j,) for j, i in enumerate(inputs)
 *             if len(i - i_single) == 0
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;

  /* "acqdp/utility/opt_einsum_paths.pyx":865
 *         # reduced to scalars:
 *         inputs_done = [
 *             (j,) for j, i in enumerate(inputs)             # <<<<<<<<<<<<<<
 *             if len(i - i_single) == 0
 *         ]
 */
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_inputs)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_inputs)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_inputs; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_inputs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 865, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 865, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 865, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 865, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":866
 *         inputs_done = [
 *             (j,) for j, i in enumerate(inputs)
 *             if len(i - i_single) == 0             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
    __pyx_t_3 = PyNumber_Subtract(__pyx_cur_scope->__pyx_v_i, __pyx_v_i_single); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_12 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 866, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = ((__pyx_t_12 == 0) != 0);
    if (__pyx_t_11) {

      /* "acqdp/utility/opt_einsum_paths.pyx":865
 *         # reduced to scalars:
 *         inputs_done = [
 *             (j,) for j, i in enumerate(inputs)             # <<<<<<<<<<<<<<
 *             if len(i - i_single) == 0
 *         ]
 */
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_j);
      __Pyx_GIVEREF(__pyx_v_j);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_j);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 864, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":866
 *         inputs_done = [
 *             (j,) for j, i in enumerate(inputs)
 *             if len(i - i_single) == 0             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":865
 *         # reduced to scalars:
 *         inputs_done = [
 *             (j,) for j, i in enumerate(inputs)             # <<<<<<<<<<<<<<
 *             if len(i - i_single) == 0
 *         ]
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_inputs_done = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":872
 *         # indices from i_single are removed and if a single-tensor contraction
 *         # is performed, the contraction expression is (j,) instead of j;
 *         inputs, inputs_contractions = zip(*[             # <<<<<<<<<<<<<<
 *             (i - i_single, j if i.isdisjoint(i_single) else (j,))
 *             for j, i in enumerate(inputs)
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_2 = __pyx_int_0;

  /* "acqdp/utility/opt_einsum_paths.pyx":874
 *         inputs, inputs_contractions = zip(*[
 *             (i - i_single, j if i.isdisjoint(i_single) else (j,))
 *             for j, i in enumerate(inputs)             # <<<<<<<<<<<<<<
 *             if len(i - i_single) > 0
 *         ])
 */
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_inputs)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_inputs)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_inputs; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_inputs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 874, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 874, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 874, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 874, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 874, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 874, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":875
 *             (i - i_single, j if i.isdisjoint(i_single) else (j,))
 *             for j, i in enumerate(inputs)
 *             if len(i - i_single) > 0             # <<<<<<<<<<<<<<
 *         ])
 * 
 */
    __pyx_t_3 = PyNumber_Subtract(__pyx_cur_scope->__pyx_v_i, __pyx_v_i_single); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_12 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 875, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = ((__pyx_t_12 > 0) != 0);
    if (__pyx_t_11) {

      /* "acqdp/utility/opt_einsum_paths.pyx":873
 *         # is performed, the contraction expression is (j,) instead of j;
 *         inputs, inputs_contractions = zip(*[
 *             (i - i_single, j if i.isdisjoint(i_single) else (j,))             # <<<<<<<<<<<<<<
 *             for j, i in enumerate(inputs)
 *             if len(i - i_single) > 0
 */
      __pyx_t_3 = PyNumber_Subtract(__pyx_cur_scope->__pyx_v_i, __pyx_v_i_single); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_i, __pyx_n_s_isdisjoint); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
        }
      }
      __pyx_t_8 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_v_i_single) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_i_single);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_11) {
        __Pyx_INCREF(__pyx_v_j);
        __pyx_t_5 = __pyx_v_j;
      } else {
        __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 873, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_j);
        __Pyx_GIVEREF(__pyx_v_j);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_j);
        __pyx_t_5 = __pyx_t_8;
        __pyx_t_8 = 0;
      }
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 873, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_5);
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 872, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":875
 *             (i - i_single, j if i.isdisjoint(i_single) else (j,))
 *             for j, i in enumerate(inputs)
 *             if len(i - i_single) > 0             # <<<<<<<<<<<<<<
 *         ])
 * 
 */
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":874
 *         inputs, inputs_contractions = zip(*[
 *             (i - i_single, j if i.isdisjoint(i_single) else (j,))
 *             for j, i in enumerate(inputs)             # <<<<<<<<<<<<<<
 *             if len(i - i_single) > 0
 *         ])
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":872
 *         # indices from i_single are removed and if a single-tensor contraction
 *         # is performed, the contraction expression is (j,) instead of j;
 *         inputs, inputs_contractions = zip(*[             # <<<<<<<<<<<<<<
 *             (i - i_single, j if i.isdisjoint(i_single) else (j,))
 *             for j, i in enumerate(inputs)
 */
  __pyx_t_2 = PySequence_Tuple(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
    PyObject* sequence = __pyx_t_4;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 872, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_2)) goto __pyx_L28_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_1)) goto __pyx_L28_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 872, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L29_unpacking_done;
    __pyx_L28_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 872, __pyx_L1_error)
    __pyx_L29_unpacking_done:;
  }
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_inputs);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_inputs, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_v_inputs_contractions = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":880
 *         # a list of all neccessary contraction expressions for each of the
 *         # disconnected subgraphs and their size
 *         subgraph_contractions = inputs_done             # <<<<<<<<<<<<<<
 *         subgraph_contractions_size = [1]*len(inputs_done)
 * 
 */
  __Pyx_INCREF(__pyx_v_inputs_done);
  __pyx_v_subgraph_contractions = __pyx_v_inputs_done;

  /* "acqdp/utility/opt_einsum_paths.pyx":881
 *         # disconnected subgraphs and their size
 *         subgraph_contractions = inputs_done
 *         subgraph_contractions_size = [1]*len(inputs_done)             # <<<<<<<<<<<<<<
 * 
 *         for g in _find_disconnected_subgraphs(inputs, output):
 */
  __pyx_t_6 = PyList_GET_SIZE(__pyx_v_inputs_done); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 881, __pyx_L1_error)
  __pyx_t_4 = PyList_New(1 * ((__pyx_t_6<0) ? 0:__pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_t_6; __pyx_temp++) {
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyList_SET_ITEM(__pyx_t_4, __pyx_temp, __pyx_int_1);
    }
  }
  __pyx_v_subgraph_contractions_size = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":883
 *         subgraph_contractions_size = [1]*len(inputs_done)
 * 
 *         for g in _find_disconnected_subgraphs(inputs, output):             # <<<<<<<<<<<<<<
 * 
 *             # dynamic programming approach to compute x[n] for subgraph g;
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_find_disconnected_subgraphs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_15 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_15 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_cur_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_v_output};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_cur_scope->__pyx_v_inputs, __pyx_cur_scope->__pyx_v_output};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_inputs);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_inputs);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_15, __pyx_cur_scope->__pyx_v_inputs);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_output);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_15, __pyx_cur_scope->__pyx_v_output);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
    __pyx_t_1 = __pyx_t_4; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 883, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 883, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 883, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 883, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 883, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_g);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_g, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":890
 *             # tensor j is in the set, e.g. 0b100101 = {0,2,5}; set unions
 *             # (intersections) can then be computed by bitwise or (and);
 *             x = [None]*2 + [dict() for j in range(len(g)-1)]             # <<<<<<<<<<<<<<
 *             x[1] = {1 << j: (inputs[j], 0, inputs_contractions[j]) for j in g}
 * 
 */
    __pyx_t_4 = PyList_New(1 * 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < 2; __pyx_temp++) {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyList_SET_ITEM(__pyx_t_4, __pyx_temp, Py_None);
      }
    }
    __pyx_t_8 = PyList_New(0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = __pyx_cur_scope->__pyx_v_g;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_12 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_12 - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_2 = __pyx_t_5; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
      __pyx_t_16 = NULL;
    } else {
      __pyx_t_12 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_16 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 890, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_16)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_5); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 890, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 890, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_5); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 890, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 890, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_16(__pyx_t_2);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 890, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_8, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Add(__pyx_t_4, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_x, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":891
 *             # (intersections) can then be computed by bitwise or (and);
 *             x = [None]*2 + [dict() for j in range(len(g)-1)]
 *             x[1] = {1 << j: (inputs[j], 0, inputs_contractions[j]) for j in g}             # <<<<<<<<<<<<<<
 * 
 *             # convert set of tensors g to a bitmap set:
 */
    { /* enter inner scope */
      __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 891, __pyx_L36_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_g)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_g)) {
        __pyx_t_8 = __pyx_cur_scope->__pyx_v_g; __Pyx_INCREF(__pyx_t_8); __pyx_t_12 = 0;
        __pyx_t_16 = NULL;
      } else {
        __pyx_t_12 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_g); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 891, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_16 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 891, __pyx_L36_error)
      }
      for (;;) {
        if (likely(!__pyx_t_16)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 891, __pyx_L36_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L36_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_12); __Pyx_INCREF(__pyx_t_4); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 891, __pyx_L36_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_8, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L36_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_16(__pyx_t_8);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 891, __pyx_L36_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_j, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = PyNumber_Lshift(__pyx_int_1, __pyx_9genexpr16__pyx_v_j); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 891, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_inputs, __pyx_9genexpr16__pyx_v_j); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 891, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_inputs_contractions, __pyx_9genexpr16__pyx_v_j); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 891, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 891, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_5);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_int_0);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_t_3);
        __pyx_t_5 = 0;
        __pyx_t_3 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_2, (PyObject*)__pyx_t_4, (PyObject*)__pyx_t_13))) __PYX_ERR(0, 891, __pyx_L36_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_j); __pyx_9genexpr16__pyx_v_j = 0;
      goto __pyx_L39_exit_scope;
      __pyx_L36_error:;
      __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_j); __pyx_9genexpr16__pyx_v_j = 0;
      goto __pyx_L1_error;
      __pyx_L39_exit_scope:;
    } /* exit inner scope */
    if (unlikely(__Pyx_SetItemInt(__pyx_v_x, 1, __pyx_t_2, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":894
 * 
 *             # convert set of tensors g to a bitmap set:
 *             g = functools.reduce(lambda x, y: x | y, (1 << j for j in g))             # <<<<<<<<<<<<<<
 * 
 *             # the bitmap set of all tensors is computed as it is needed to
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_functools); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_reduce); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___9lambda11, 0, __pyx_n_s_DynamicProgrammingOptimizer___ca_3, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___10genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 894, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_8, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_8, __pyx_t_4};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_15, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_15, __pyx_t_4);
      __pyx_t_8 = 0;
      __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_g);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_g, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":899
 *             # compute set differences: s1 - s2 transforms into
 *             # s1 & (all_tensors ^ s2)
 *             all_tensors = (1 << len(inputs)) - 1             # <<<<<<<<<<<<<<
 * 
 *             # try to find contraction with cost <= cost_cap and increase
 */
    __pyx_t_2 = __pyx_cur_scope->__pyx_v_inputs;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_12 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 899, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_all_tensors = ((1 << __pyx_t_12) - 1);

    /* "acqdp/utility/opt_einsum_paths.pyx":905
 *             # this is a major performance improvement; start with product of
 *             # output index dimensions as initial cost_cap
 *             cost_cap = helpers.compute_size_by_dict(             # <<<<<<<<<<<<<<
 *                 set.union(*(inputs[j] for j in _bitmapset_indices(g))) & output,
 *                 size_dict
 */
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_helpers); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 905, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 905, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":906
 *             # output index dimensions as initial cost_cap
 *             cost_cap = helpers.compute_size_by_dict(
 *                 set.union(*(inputs[j] for j in _bitmapset_indices(g))) & output,             # <<<<<<<<<<<<<<
 *                 size_dict
 *             )
 */
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PySet_Type)), __pyx_n_s_union); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_4 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___13genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PySequence_Tuple(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyNumber_And(__pyx_t_4, __pyx_cur_scope->__pyx_v_output); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 906, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":907
 *             cost_cap = helpers.compute_size_by_dict(
 *                 set.union(*(inputs[j] for j in _bitmapset_indices(g))) & output,
 *                 size_dict             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_4 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_v_size_dict};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_8, __pyx_v_size_dict};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_15, __pyx_t_8);
      __Pyx_INCREF(__pyx_v_size_dict);
      __Pyx_GIVEREF(__pyx_v_size_dict);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_15, __pyx_v_size_dict);
      __pyx_t_8 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_cost_cap, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":910
 *             )
 * 
 *             while len(x[-1]) == 0:             # <<<<<<<<<<<<<<
 *                 for n in range(2, len(x[1]) + 1):
 *                     xn = x[n]
 */
    while (1) {
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_x, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = ((__pyx_t_12 == 0) != 0);
      if (!__pyx_t_11) break;

      /* "acqdp/utility/opt_einsum_paths.pyx":911
 * 
 *             while len(x[-1]) == 0:
 *                 for n in range(2, len(x[1]) + 1):             # <<<<<<<<<<<<<<
 *                     xn = x[n]
 * 
 */
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_x, 1, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_12 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_12 + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_int_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_12 = 0;
        __pyx_t_16 = NULL;
      } else {
        __pyx_t_12 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 911, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_16 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 911, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_16)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_12); __Pyx_INCREF(__pyx_t_2); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 911, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_12); __Pyx_INCREF(__pyx_t_2); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 911, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 911, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_16(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 911, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_n, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":912
 *             while len(x[-1]) == 0:
 *                 for n in range(2, len(x[1]) + 1):
 *                     xn = x[n]             # <<<<<<<<<<<<<<
 * 
 *                     # try to combine solutions from x[m] and x[n-m]
 */
        __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_x, __pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_xn, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":915
 * 
 *                     # try to combine solutions from x[m] and x[n-m]
 *                     for m in range(1, n // 2 + 1):             # <<<<<<<<<<<<<<
 *                         for s1, (i1, cost1, cntrct1) in x[m].items():
 *                             for s2, (i2, cost2, cntrct2) in x[n-m].items():
 */
        __pyx_t_2 = __Pyx_PyInt_FloorDivideObjC(__pyx_v_n, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_13 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_int_1);
        __Pyx_GIVEREF(__pyx_int_1);
        PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_1);
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_13);
        __pyx_t_13 = 0;
        __pyx_t_13 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (likely(PyList_CheckExact(__pyx_t_13)) || PyTuple_CheckExact(__pyx_t_13)) {
          __pyx_t_2 = __pyx_t_13; __Pyx_INCREF(__pyx_t_2); __pyx_t_17 = 0;
          __pyx_t_18 = NULL;
        } else {
          __pyx_t_17 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 915, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_18 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 915, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        for (;;) {
          if (likely(!__pyx_t_18)) {
            if (likely(PyList_CheckExact(__pyx_t_2))) {
              if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_13 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_17); __Pyx_INCREF(__pyx_t_13); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 915, __pyx_L1_error)
              #else
              __pyx_t_13 = PySequence_ITEM(__pyx_t_2, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 915, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
            } else {
              if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_17); __Pyx_INCREF(__pyx_t_13); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 915, __pyx_L1_error)
              #else
              __pyx_t_13 = PySequence_ITEM(__pyx_t_2, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 915, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
            }
          } else {
            __pyx_t_13 = __pyx_t_18(__pyx_t_2);
            if (unlikely(!__pyx_t_13)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 915, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_13);
          }
          __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_13);
          __pyx_t_13 = 0;

          /* "acqdp/utility/opt_einsum_paths.pyx":916
 *                     # try to combine solutions from x[m] and x[n-m]
 *                     for m in range(1, n // 2 + 1):
 *                         for s1, (i1, cost1, cntrct1) in x[m].items():             # <<<<<<<<<<<<<<
 *                             for s2, (i2, cost2, cntrct2) in x[n-m].items():
 * 
 */
          __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_x, __pyx_v_m); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 916, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_items); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 916, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_13 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 916, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (likely(PyList_CheckExact(__pyx_t_13)) || PyTuple_CheckExact(__pyx_t_13)) {
            __pyx_t_4 = __pyx_t_13; __Pyx_INCREF(__pyx_t_4); __pyx_t_19 = 0;
            __pyx_t_20 = NULL;
          } else {
            __pyx_t_19 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 916, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_20 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 916, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          for (;;) {
            if (likely(!__pyx_t_20)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_19 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_13 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_19); __Pyx_INCREF(__pyx_t_13); __pyx_t_19++; if (unlikely(0 < 0)) __PYX_ERR(0, 916, __pyx_L1_error)
                #else
                __pyx_t_13 = PySequence_ITEM(__pyx_t_4, __pyx_t_19); __pyx_t_19++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 916, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_13);
                #endif
              } else {
                if (__pyx_t_19 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_19); __Pyx_INCREF(__pyx_t_13); __pyx_t_19++; if (unlikely(0 < 0)) __PYX_ERR(0, 916, __pyx_L1_error)
                #else
                __pyx_t_13 = PySequence_ITEM(__pyx_t_4, __pyx_t_19); __pyx_t_19++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 916, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_13);
                #endif
              }
            } else {
              __pyx_t_13 = __pyx_t_20(__pyx_t_4);
              if (unlikely(!__pyx_t_13)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 916, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_13);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_13))) || (PyList_CheckExact(__pyx_t_13))) {
              PyObject* sequence = __pyx_t_13;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 916, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_3);
              #else
              __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 916, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 916, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              #endif
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_14 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 916, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_t_9 = Py_TYPE(__pyx_t_14)->tp_iternext;
              index = 0; __pyx_t_8 = __pyx_t_9(__pyx_t_14); if (unlikely(!__pyx_t_8)) goto __pyx_L48_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_8);
              index = 1; __pyx_t_3 = __pyx_t_9(__pyx_t_14); if (unlikely(!__pyx_t_3)) goto __pyx_L48_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_3);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_14), 2) < 0) __PYX_ERR(0, 916, __pyx_L1_error)
              __pyx_t_9 = NULL;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              goto __pyx_L49_unpacking_done;
              __pyx_L48_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_9 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 916, __pyx_L1_error)
              __pyx_L49_unpacking_done:;
            }
            __Pyx_XDECREF_SET(__pyx_v_s1, __pyx_t_8);
            __pyx_t_8 = 0;
            if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
              PyObject* sequence = __pyx_t_3;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 3)) {
                if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 916, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_14 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_21 = PyTuple_GET_ITEM(sequence, 1); 
                __pyx_t_22 = PyTuple_GET_ITEM(sequence, 2); 
              } else {
                __pyx_t_14 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_21 = PyList_GET_ITEM(sequence, 1); 
                __pyx_t_22 = PyList_GET_ITEM(sequence, 2); 
              }
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_21);
              __Pyx_INCREF(__pyx_t_22);
              #else
              __pyx_t_14 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 916, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_21 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 916, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_21);
              __pyx_t_22 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 916, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_22);
              #endif
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_23 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 916, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_23);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_9 = Py_TYPE(__pyx_t_23)->tp_iternext;
              index = 0; __pyx_t_14 = __pyx_t_9(__pyx_t_23); if (unlikely(!__pyx_t_14)) goto __pyx_L50_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_14);
              index = 1; __pyx_t_21 = __pyx_t_9(__pyx_t_23); if (unlikely(!__pyx_t_21)) goto __pyx_L50_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_21);
              index = 2; __pyx_t_22 = __pyx_t_9(__pyx_t_23); if (unlikely(!__pyx_t_22)) goto __pyx_L50_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_22);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_23), 3) < 0) __PYX_ERR(0, 916, __pyx_L1_error)
              __pyx_t_9 = NULL;
              __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
              goto __pyx_L51_unpacking_done;
              __pyx_L50_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
              __pyx_t_9 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 916, __pyx_L1_error)
              __pyx_L51_unpacking_done:;
            }
            __Pyx_XDECREF_SET(__pyx_v_i1, __pyx_t_14);
            __pyx_t_14 = 0;
            __Pyx_XDECREF_SET(__pyx_v_cost1, __pyx_t_21);
            __pyx_t_21 = 0;
            __Pyx_XDECREF_SET(__pyx_v_cntrct1, __pyx_t_22);
            __pyx_t_22 = 0;

            /* "acqdp/utility/opt_einsum_paths.pyx":917
 *                     for m in range(1, n // 2 + 1):
 *                         for s1, (i1, cost1, cntrct1) in x[m].items():
 *                             for s2, (i2, cost2, cntrct2) in x[n-m].items():             # <<<<<<<<<<<<<<
 * 
 *                                 # only if s1 and s2 are disjoint
 */
            __pyx_t_3 = PyNumber_Subtract(__pyx_v_n, __pyx_v_m); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_v_x, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 917, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_8 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_13 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 917, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (likely(PyList_CheckExact(__pyx_t_13)) || PyTuple_CheckExact(__pyx_t_13)) {
              __pyx_t_3 = __pyx_t_13; __Pyx_INCREF(__pyx_t_3); __pyx_t_24 = 0;
              __pyx_t_25 = NULL;
            } else {
              __pyx_t_24 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_25 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 917, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            for (;;) {
              if (likely(!__pyx_t_25)) {
                if (likely(PyList_CheckExact(__pyx_t_3))) {
                  if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_3)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_13 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_24); __Pyx_INCREF(__pyx_t_13); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 917, __pyx_L1_error)
                  #else
                  __pyx_t_13 = PySequence_ITEM(__pyx_t_3, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 917, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  #endif
                } else {
                  if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_24); __Pyx_INCREF(__pyx_t_13); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 917, __pyx_L1_error)
                  #else
                  __pyx_t_13 = PySequence_ITEM(__pyx_t_3, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 917, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  #endif
                }
              } else {
                __pyx_t_13 = __pyx_t_25(__pyx_t_3);
                if (unlikely(!__pyx_t_13)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 917, __pyx_L1_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_13);
              }
              if ((likely(PyTuple_CheckExact(__pyx_t_13))) || (PyList_CheckExact(__pyx_t_13))) {
                PyObject* sequence = __pyx_t_13;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 2)) {
                  if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 917, __pyx_L1_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
                  __pyx_t_22 = PyTuple_GET_ITEM(sequence, 1); 
                } else {
                  __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
                  __pyx_t_22 = PyList_GET_ITEM(sequence, 1); 
                }
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(__pyx_t_22);
                #else
                __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 917, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_8);
                __pyx_t_22 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 917, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_22);
                #endif
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              } else {
                Py_ssize_t index = -1;
                __pyx_t_21 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 917, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_21);
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                __pyx_t_9 = Py_TYPE(__pyx_t_21)->tp_iternext;
                index = 0; __pyx_t_8 = __pyx_t_9(__pyx_t_21); if (unlikely(!__pyx_t_8)) goto __pyx_L54_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_8);
                index = 1; __pyx_t_22 = __pyx_t_9(__pyx_t_21); if (unlikely(!__pyx_t_22)) goto __pyx_L54_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_22);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_21), 2) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
                __pyx_t_9 = NULL;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                goto __pyx_L55_unpacking_done;
                __pyx_L54_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
                __pyx_t_9 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 917, __pyx_L1_error)
                __pyx_L55_unpacking_done:;
              }
              __Pyx_XDECREF_SET(__pyx_v_s2, __pyx_t_8);
              __pyx_t_8 = 0;
              if ((likely(PyTuple_CheckExact(__pyx_t_22))) || (PyList_CheckExact(__pyx_t_22))) {
                PyObject* sequence = __pyx_t_22;
                Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                if (unlikely(size != 3)) {
                  if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                  else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                  __PYX_ERR(0, 917, __pyx_L1_error)
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                if (likely(PyTuple_CheckExact(sequence))) {
                  __pyx_t_21 = PyTuple_GET_ITEM(sequence, 0); 
                  __pyx_t_14 = PyTuple_GET_ITEM(sequence, 1); 
                  __pyx_t_23 = PyTuple_GET_ITEM(sequence, 2); 
                } else {
                  __pyx_t_21 = PyList_GET_ITEM(sequence, 0); 
                  __pyx_t_14 = PyList_GET_ITEM(sequence, 1); 
                  __pyx_t_23 = PyList_GET_ITEM(sequence, 2); 
                }
                __Pyx_INCREF(__pyx_t_21);
                __Pyx_INCREF(__pyx_t_14);
                __Pyx_INCREF(__pyx_t_23);
                #else
                __pyx_t_21 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 917, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_21);
                __pyx_t_14 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 917, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_23 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 917, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_23);
                #endif
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              } else {
                Py_ssize_t index = -1;
                __pyx_t_26 = PyObject_GetIter(__pyx_t_22); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 917, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_26);
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                __pyx_t_9 = Py_TYPE(__pyx_t_26)->tp_iternext;
                index = 0; __pyx_t_21 = __pyx_t_9(__pyx_t_26); if (unlikely(!__pyx_t_21)) goto __pyx_L56_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_21);
                index = 1; __pyx_t_14 = __pyx_t_9(__pyx_t_26); if (unlikely(!__pyx_t_14)) goto __pyx_L56_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_14);
                index = 2; __pyx_t_23 = __pyx_t_9(__pyx_t_26); if (unlikely(!__pyx_t_23)) goto __pyx_L56_unpacking_failed;
                __Pyx_GOTREF(__pyx_t_23);
                if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_26), 3) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
                __pyx_t_9 = NULL;
                __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
                goto __pyx_L57_unpacking_done;
                __pyx_L56_unpacking_failed:;
                __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
                __pyx_t_9 = NULL;
                if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                __PYX_ERR(0, 917, __pyx_L1_error)
                __pyx_L57_unpacking_done:;
              }
              __Pyx_XDECREF_SET(__pyx_v_i2, __pyx_t_21);
              __pyx_t_21 = 0;
              __Pyx_XDECREF_SET(__pyx_v_cost2, __pyx_t_14);
              __pyx_t_14 = 0;
              __Pyx_XDECREF_SET(__pyx_v_cntrct2, __pyx_t_23);
              __pyx_t_23 = 0;

              /* "acqdp/utility/opt_einsum_paths.pyx":920
 * 
 *                                 # only if s1 and s2 are disjoint
 *                                 if s1 & s2 == 0:             # <<<<<<<<<<<<<<
 * 
 *                                     # avoid e.g. s1={0}, s2={1} and s1={1}, s2={0}
 */
              __pyx_t_13 = PyNumber_And(__pyx_v_s1, __pyx_v_s2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 920, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              __pyx_t_22 = __Pyx_PyInt_EqObjC(__pyx_t_13, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 920, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_22);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_22); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 920, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
              if (__pyx_t_11) {

                /* "acqdp/utility/opt_einsum_paths.pyx":923
 * 
 *                                     # avoid e.g. s1={0}, s2={1} and s1={1}, s2={0}
 *                                     if m != n - m or s1 < s2:             # <<<<<<<<<<<<<<
 * 
 *                                         i1_cut_i2_wo_output = (i1 & i2) - output
 */
                __pyx_t_22 = PyNumber_Subtract(__pyx_v_n, __pyx_v_m); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 923, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_22);
                __pyx_t_13 = PyObject_RichCompare(__pyx_v_m, __pyx_t_22, Py_NE); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 923, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 923, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                if (!__pyx_t_10) {
                } else {
                  __pyx_t_11 = __pyx_t_10;
                  goto __pyx_L60_bool_binop_done;
                }
                __pyx_t_13 = PyObject_RichCompare(__pyx_v_s1, __pyx_v_s2, Py_LT); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 923, __pyx_L1_error)
                __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 923, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                __pyx_t_11 = __pyx_t_10;
                __pyx_L60_bool_binop_done:;
                if (__pyx_t_11) {

                  /* "acqdp/utility/opt_einsum_paths.pyx":925
 *                                     if m != n - m or s1 < s2:
 * 
 *                                         i1_cut_i2_wo_output = (i1 & i2) - output             # <<<<<<<<<<<<<<
 * 
 *                                         # ignore outer products:
 */
                  __pyx_t_13 = PyNumber_And(__pyx_v_i1, __pyx_v_i2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 925, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_13);
                  __pyx_t_22 = PyNumber_Subtract(__pyx_t_13, __pyx_cur_scope->__pyx_v_output); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 925, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_22);
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                  __Pyx_XDECREF_SET(__pyx_v_i1_cut_i2_wo_output, __pyx_t_22);
                  __pyx_t_22 = 0;

                  /* "acqdp/utility/opt_einsum_paths.pyx":928
 * 
 *                                         # ignore outer products:
 *                                         if len(i1_cut_i2_wo_output) > 0:             # <<<<<<<<<<<<<<
 * 
 *                                             i1_union_i2 = i1 | i2
 */
                  __pyx_t_27 = PyObject_Length(__pyx_v_i1_cut_i2_wo_output); if (unlikely(__pyx_t_27 == ((Py_ssize_t)-1))) __PYX_ERR(0, 928, __pyx_L1_error)
                  __pyx_t_11 = ((__pyx_t_27 > 0) != 0);
                  if (__pyx_t_11) {

                    /* "acqdp/utility/opt_einsum_paths.pyx":930
 *                                         if len(i1_cut_i2_wo_output) > 0:
 * 
 *                                             i1_union_i2 = i1 | i2             # <<<<<<<<<<<<<<
 *                                             cost = cost1 + cost2 + helpers.compute_size_by_dict(i1_union_i2, size_dict)
 *                                             if cost <= cost_cap:
 */
                    __pyx_t_22 = PyNumber_Or(__pyx_v_i1, __pyx_v_i2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 930, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    __Pyx_XDECREF_SET(__pyx_v_i1_union_i2, __pyx_t_22);
                    __pyx_t_22 = 0;

                    /* "acqdp/utility/opt_einsum_paths.pyx":931
 * 
 *                                             i1_union_i2 = i1 | i2
 *                                             cost = cost1 + cost2 + helpers.compute_size_by_dict(i1_union_i2, size_dict)             # <<<<<<<<<<<<<<
 *                                             if cost <= cost_cap:
 *                                                 s = s1 | s2
 */
                    __pyx_t_22 = PyNumber_Add(__pyx_v_cost1, __pyx_v_cost2); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 931, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_22);
                    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_helpers); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 931, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 931, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_23);
                    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __pyx_t_8 = NULL;
                    __pyx_t_15 = 0;
                    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
                      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_23);
                      if (likely(__pyx_t_8)) {
                        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
                        __Pyx_INCREF(__pyx_t_8);
                        __Pyx_INCREF(function);
                        __Pyx_DECREF_SET(__pyx_t_23, function);
                        __pyx_t_15 = 1;
                      }
                    }
                    #if CYTHON_FAST_PYCALL
                    if (PyFunction_Check(__pyx_t_23)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_i1_union_i2, __pyx_v_size_dict};
                      __pyx_t_13 = __Pyx_PyFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 931, __pyx_L1_error)
                      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                      __Pyx_GOTREF(__pyx_t_13);
                    } else
                    #endif
                    #if CYTHON_FAST_PYCCALL
                    if (__Pyx_PyFastCFunction_Check(__pyx_t_23)) {
                      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_i1_union_i2, __pyx_v_size_dict};
                      __pyx_t_13 = __Pyx_PyCFunction_FastCall(__pyx_t_23, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 931, __pyx_L1_error)
                      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                      __Pyx_GOTREF(__pyx_t_13);
                    } else
                    #endif
                    {
                      __pyx_t_14 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 931, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_14);
                      if (__pyx_t_8) {
                        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_8); __pyx_t_8 = NULL;
                      }
                      __Pyx_INCREF(__pyx_v_i1_union_i2);
                      __Pyx_GIVEREF(__pyx_v_i1_union_i2);
                      PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_15, __pyx_v_i1_union_i2);
                      __Pyx_INCREF(__pyx_v_size_dict);
                      __Pyx_GIVEREF(__pyx_v_size_dict);
                      PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_15, __pyx_v_size_dict);
                      __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_t_14, NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 931, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_13);
                      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                    }
                    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                    __pyx_t_23 = PyNumber_Add(__pyx_t_22, __pyx_t_13); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 931, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_23);
                    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                    __Pyx_XDECREF_SET(__pyx_v_cost, __pyx_t_23);
                    __pyx_t_23 = 0;

                    /* "acqdp/utility/opt_einsum_paths.pyx":932
 *                                             i1_union_i2 = i1 | i2
 *                                             cost = cost1 + cost2 + helpers.compute_size_by_dict(i1_union_i2, size_dict)
 *                                             if cost <= cost_cap:             # <<<<<<<<<<<<<<
 *                                                 s = s1 | s2
 *                                                 if s not in xn or cost < xn[s][1]:
 */
                    __pyx_t_23 = PyObject_RichCompare(__pyx_v_cost, __pyx_v_cost_cap, Py_LE); __Pyx_XGOTREF(__pyx_t_23); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 932, __pyx_L1_error)
                    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_23); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 932, __pyx_L1_error)
                    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                    if (__pyx_t_11) {

                      /* "acqdp/utility/opt_einsum_paths.pyx":933
 *                                             cost = cost1 + cost2 + helpers.compute_size_by_dict(i1_union_i2, size_dict)
 *                                             if cost <= cost_cap:
 *                                                 s = s1 | s2             # <<<<<<<<<<<<<<
 *                                                 if s not in xn or cost < xn[s][1]:
 *                                                     # set of remaining tensors (=g-s)
 */
                      __pyx_t_23 = PyNumber_Or(__pyx_v_s1, __pyx_v_s2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 933, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_23);
                      __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_23);
                      __pyx_t_23 = 0;

                      /* "acqdp/utility/opt_einsum_paths.pyx":934
 *                                             if cost <= cost_cap:
 *                                                 s = s1 | s2
 *                                                 if s not in xn or cost < xn[s][1]:             # <<<<<<<<<<<<<<
 *                                                     # set of remaining tensors (=g-s)
 *                                                     r = g & (all_tensors ^ s)
 */
                      __pyx_t_10 = (__Pyx_PySequence_ContainsTF(__pyx_v_s, __pyx_v_xn, Py_NE)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 934, __pyx_L1_error)
                      __pyx_t_28 = (__pyx_t_10 != 0);
                      if (!__pyx_t_28) {
                      } else {
                        __pyx_t_11 = __pyx_t_28;
                        goto __pyx_L65_bool_binop_done;
                      }
                      __pyx_t_23 = __Pyx_PyObject_GetItem(__pyx_v_xn, __pyx_v_s); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 934, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_23);
                      __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_23, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 934, __pyx_L1_error)
                      __Pyx_GOTREF(__pyx_t_13);
                      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                      __pyx_t_23 = PyObject_RichCompare(__pyx_v_cost, __pyx_t_13, Py_LT); __Pyx_XGOTREF(__pyx_t_23); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 934, __pyx_L1_error)
                      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                      __pyx_t_28 = __Pyx_PyObject_IsTrue(__pyx_t_23); if (unlikely(__pyx_t_28 < 0)) __PYX_ERR(0, 934, __pyx_L1_error)
                      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                      __pyx_t_11 = __pyx_t_28;
                      __pyx_L65_bool_binop_done:;
                      if (__pyx_t_11) {

                        /* "acqdp/utility/opt_einsum_paths.pyx":936
 *                                                 if s not in xn or cost < xn[s][1]:
 *                                                     # set of remaining tensors (=g-s)
 *                                                     r = g & (all_tensors ^ s)             # <<<<<<<<<<<<<<
 * 
 *                                                     # indices of remaining indices:
 */
                        __pyx_t_23 = PyInt_FromSsize_t(__pyx_v_all_tensors); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 936, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_23);
                        __pyx_t_13 = PyNumber_Xor(__pyx_t_23, __pyx_v_s); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 936, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_13);
                        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                        __pyx_t_23 = PyNumber_And(__pyx_cur_scope->__pyx_v_g, __pyx_t_13); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 936, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_23);
                        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
                        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_23);
                        __Pyx_GIVEREF(__pyx_t_23);
                        __pyx_t_23 = 0;

                        /* "acqdp/utility/opt_einsum_paths.pyx":940
 *                                                     # indices of remaining indices:
 *                                                     i_r = (set.union(*(inputs[j] for j in _bitmapset_indices(r)))
 *                                                            if r != 0 else set())             # <<<<<<<<<<<<<<
 * 
 *                                                     # contraction indices:
 */
                        __pyx_t_13 = __Pyx_PyInt_NeObjC(__pyx_cur_scope->__pyx_v_r, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 940, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_13);
                        __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 940, __pyx_L1_error)
                        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                        if (__pyx_t_11) {

                          /* "acqdp/utility/opt_einsum_paths.pyx":939
 * 
 *                                                     # indices of remaining indices:
 *                                                     i_r = (set.union(*(inputs[j] for j in _bitmapset_indices(r)))             # <<<<<<<<<<<<<<
 *                                                            if r != 0 else set())
 * 
 */
                          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PySet_Type)), __pyx_n_s_union); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 939, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_13);
                          __pyx_t_22 = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___16genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 939, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_22);
                          __pyx_t_14 = __Pyx_PySequence_Tuple(__pyx_t_22); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 939, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_14);
                          __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                          __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_14, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 939, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_22);
                          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                          __pyx_t_23 = __pyx_t_22;
                          __pyx_t_22 = 0;
                        } else {
                          __pyx_t_22 = PySet_New(0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 940, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_22);
                          __pyx_t_23 = __pyx_t_22;
                          __pyx_t_22 = 0;
                        }
                        __Pyx_XDECREF_SET(__pyx_v_i_r, __pyx_t_23);
                        __pyx_t_23 = 0;

                        /* "acqdp/utility/opt_einsum_paths.pyx":943
 * 
 *                                                     # contraction indices:
 *                                                     i_cntrct = i1_cut_i2_wo_output - i_r             # <<<<<<<<<<<<<<
 * 
 *                                                     i = i1_union_i2 - i_cntrct
 */
                        __pyx_t_23 = PyNumber_Subtract(__pyx_v_i1_cut_i2_wo_output, __pyx_v_i_r); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 943, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_23);
                        __Pyx_XDECREF_SET(__pyx_v_i_cntrct, __pyx_t_23);
                        __pyx_t_23 = 0;

                        /* "acqdp/utility/opt_einsum_paths.pyx":945
 *                                                     i_cntrct = i1_cut_i2_wo_output - i_r
 * 
 *                                                     i = i1_union_i2 - i_cntrct             # <<<<<<<<<<<<<<
 *                                                     mem = helpers.compute_size_by_dict(i, size_dict)
 *                                                     if memory_limit is None or mem <= memory_limit:
 */
                        __pyx_t_23 = PyNumber_Subtract(__pyx_v_i1_union_i2, __pyx_v_i_cntrct); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 945, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_23);
                        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
                        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_23);
                        __Pyx_GIVEREF(__pyx_t_23);
                        __pyx_t_23 = 0;

                        /* "acqdp/utility/opt_einsum_paths.pyx":946
 * 
 *                                                     i = i1_union_i2 - i_cntrct
 *                                                     mem = helpers.compute_size_by_dict(i, size_dict)             # <<<<<<<<<<<<<<
 *                                                     if memory_limit is None or mem <= memory_limit:
 *                                                         xn[s] = (i, cost, (cntrct1, cntrct2))
 */
                        __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_n_s_helpers); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 946, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_22);
                        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 946, __pyx_L1_error)
                        __Pyx_GOTREF(__pyx_t_14);
                        __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
                        __pyx_t_22 = NULL;
                        __pyx_t_15 = 0;
                        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
                          __pyx_t_22 = PyMethod_GET_SELF(__pyx_t_14);
                          if (likely(__pyx_t_22)) {
                            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                            __Pyx_INCREF(__pyx_t_22);
                            __Pyx_INCREF(function);
                            __Pyx_DECREF_SET(__pyx_t_14, function);
                            __pyx_t_15 = 1;
                          }
                        }
                        #if CYTHON_FAST_PYCALL
                        if (PyFunction_Check(__pyx_t_14)) {
                          PyObject *__pyx_temp[3] = {__pyx_t_22, __pyx_cur_scope->__pyx_v_i, __pyx_v_size_dict};
                          __pyx_t_23 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 946, __pyx_L1_error)
                          __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                          __Pyx_GOTREF(__pyx_t_23);
                        } else
                        #endif
                        #if CYTHON_FAST_PYCCALL
                        if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
                          PyObject *__pyx_temp[3] = {__pyx_t_22, __pyx_cur_scope->__pyx_v_i, __pyx_v_size_dict};
                          __pyx_t_23 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 946, __pyx_L1_error)
                          __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
                          __Pyx_GOTREF(__pyx_t_23);
                        } else
                        #endif
                        {
                          __pyx_t_13 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 946, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_13);
                          if (__pyx_t_22) {
                            __Pyx_GIVEREF(__pyx_t_22); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_22); __pyx_t_22 = NULL;
                          }
                          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_i);
                          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_i);
                          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_15, __pyx_cur_scope->__pyx_v_i);
                          __Pyx_INCREF(__pyx_v_size_dict);
                          __Pyx_GIVEREF(__pyx_v_size_dict);
                          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_15, __pyx_v_size_dict);
                          __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_13, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 946, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_23);
                          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                        }
                        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                        __Pyx_XDECREF_SET(__pyx_v_mem, __pyx_t_23);
                        __pyx_t_23 = 0;

                        /* "acqdp/utility/opt_einsum_paths.pyx":947
 *                                                     i = i1_union_i2 - i_cntrct
 *                                                     mem = helpers.compute_size_by_dict(i, size_dict)
 *                                                     if memory_limit is None or mem <= memory_limit:             # <<<<<<<<<<<<<<
 *                                                         xn[s] = (i, cost, (cntrct1, cntrct2))
 * 
 */
                        __pyx_t_28 = (__pyx_v_memory_limit == Py_None);
                        __pyx_t_10 = (__pyx_t_28 != 0);
                        if (!__pyx_t_10) {
                        } else {
                          __pyx_t_11 = __pyx_t_10;
                          goto __pyx_L68_bool_binop_done;
                        }
                        __pyx_t_23 = PyObject_RichCompare(__pyx_v_mem, __pyx_v_memory_limit, Py_LE); __Pyx_XGOTREF(__pyx_t_23); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 947, __pyx_L1_error)
                        __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_23); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 947, __pyx_L1_error)
                        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
                        __pyx_t_11 = __pyx_t_10;
                        __pyx_L68_bool_binop_done:;
                        if (__pyx_t_11) {

                          /* "acqdp/utility/opt_einsum_paths.pyx":948
 *                                                     mem = helpers.compute_size_by_dict(i, size_dict)
 *                                                     if memory_limit is None or mem <= memory_limit:
 *                                                         xn[s] = (i, cost, (cntrct1, cntrct2))             # <<<<<<<<<<<<<<
 * 
 *                 # increase cost cap for next iteration:
 */
                          __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 948, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_23);
                          __Pyx_INCREF(__pyx_v_cntrct1);
                          __Pyx_GIVEREF(__pyx_v_cntrct1);
                          PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_v_cntrct1);
                          __Pyx_INCREF(__pyx_v_cntrct2);
                          __Pyx_GIVEREF(__pyx_v_cntrct2);
                          PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_v_cntrct2);
                          __pyx_t_14 = PyTuple_New(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 948, __pyx_L1_error)
                          __Pyx_GOTREF(__pyx_t_14);
                          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_i);
                          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_i);
                          PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_cur_scope->__pyx_v_i);
                          __Pyx_INCREF(__pyx_v_cost);
                          __Pyx_GIVEREF(__pyx_v_cost);
                          PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_cost);
                          __Pyx_GIVEREF(__pyx_t_23);
                          PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_t_23);
                          __pyx_t_23 = 0;
                          if (unlikely(PyObject_SetItem(__pyx_v_xn, __pyx_v_s, __pyx_t_14) < 0)) __PYX_ERR(0, 948, __pyx_L1_error)
                          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

                          /* "acqdp/utility/opt_einsum_paths.pyx":947
 *                                                     i = i1_union_i2 - i_cntrct
 *                                                     mem = helpers.compute_size_by_dict(i, size_dict)
 *                                                     if memory_limit is None or mem <= memory_limit:             # <<<<<<<<<<<<<<
 *                                                         xn[s] = (i, cost, (cntrct1, cntrct2))
 * 
 */
                        }

                        /* "acqdp/utility/opt_einsum_paths.pyx":934
 *                                             if cost <= cost_cap:
 *                                                 s = s1 | s2
 *                                                 if s not in xn or cost < xn[s][1]:             # <<<<<<<<<<<<<<
 *                                                     # set of remaining tensors (=g-s)
 *                                                     r = g & (all_tensors ^ s)
 */
                      }

                      /* "acqdp/utility/opt_einsum_paths.pyx":932
 *                                             i1_union_i2 = i1 | i2
 *                                             cost = cost1 + cost2 + helpers.compute_size_by_dict(i1_union_i2, size_dict)
 *                                             if cost <= cost_cap:             # <<<<<<<<<<<<<<
 *                                                 s = s1 | s2
 *                                                 if s not in xn or cost < xn[s][1]:
 */
                    }

                    /* "acqdp/utility/opt_einsum_paths.pyx":928
 * 
 *                                         # ignore outer products:
 *                                         if len(i1_cut_i2_wo_output) > 0:             # <<<<<<<<<<<<<<
 * 
 *                                             i1_union_i2 = i1 | i2
 */
                  }

                  /* "acqdp/utility/opt_einsum_paths.pyx":923
 * 
 *                                     # avoid e.g. s1={0}, s2={1} and s1={1}, s2={0}
 *                                     if m != n - m or s1 < s2:             # <<<<<<<<<<<<<<
 * 
 *                                         i1_cut_i2_wo_output = (i1 & i2) - output
 */
                }

                /* "acqdp/utility/opt_einsum_paths.pyx":920
 * 
 *                                 # only if s1 and s2 are disjoint
 *                                 if s1 & s2 == 0:             # <<<<<<<<<<<<<<
 * 
 *                                     # avoid e.g. s1={0}, s2={1} and s1={1}, s2={0}
 */
              }

              /* "acqdp/utility/opt_einsum_paths.pyx":917
 *                     for m in range(1, n // 2 + 1):
 *                         for s1, (i1, cost1, cntrct1) in x[m].items():
 *                             for s2, (i2, cost2, cntrct2) in x[n-m].items():             # <<<<<<<<<<<<<<
 * 
 *                                 # only if s1 and s2 are disjoint
 */
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "acqdp/utility/opt_einsum_paths.pyx":916
 *                     # try to combine solutions from x[m] and x[n-m]
 *                     for m in range(1, n // 2 + 1):
 *                         for s1, (i1, cost1, cntrct1) in x[m].items():             # <<<<<<<<<<<<<<
 *                             for s2, (i2, cost2, cntrct2) in x[n-m].items():
 * 
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "acqdp/utility/opt_einsum_paths.pyx":915
 * 
 *                     # try to combine solutions from x[m] and x[n-m]
 *                     for m in range(1, n // 2 + 1):             # <<<<<<<<<<<<<<
 *                         for s1, (i1, cost1, cntrct1) in x[m].items():
 *                             for s2, (i2, cost2, cntrct2) in x[n-m].items():
 */
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "acqdp/utility/opt_einsum_paths.pyx":911
 * 
 *             while len(x[-1]) == 0:
 *                 for n in range(2, len(x[1]) + 1):             # <<<<<<<<<<<<<<
 *                     xn = x[n]
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "acqdp/utility/opt_einsum_paths.pyx":951
 * 
 *                 # increase cost cap for next iteration:
 *                 cost_cap = min(size_dict) * cost_cap             # <<<<<<<<<<<<<<
 * 
 *             i, cost, contraction = list(x[-1].values())[0]
 */
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_v_size_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyNumber_Multiply(__pyx_t_5, __pyx_v_cost_cap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_cost_cap, __pyx_t_2);
      __pyx_t_2 = 0;
    }

    /* "acqdp/utility/opt_einsum_paths.pyx":953
 *                 cost_cap = min(size_dict) * cost_cap
 * 
 *             i, cost, contraction = list(x[-1].values())[0]             # <<<<<<<<<<<<<<
 *             subgraph_contractions.append(contraction)
 *             subgraph_contractions_size.append(helpers.compute_size_by_dict(i, size_dict))
 */
    __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_x, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_values); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 953, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_14 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_14)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_9(__pyx_t_14); if (unlikely(!__pyx_t_4)) goto __pyx_L70_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_5 = __pyx_t_9(__pyx_t_14); if (unlikely(!__pyx_t_5)) goto __pyx_L70_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 2; __pyx_t_3 = __pyx_t_9(__pyx_t_14); if (unlikely(!__pyx_t_3)) goto __pyx_L70_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_14), 3) < 0) __PYX_ERR(0, 953, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L71_unpacking_done;
      __pyx_L70_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 953, __pyx_L1_error)
      __pyx_L71_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_cost, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_contraction, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":954
 * 
 *             i, cost, contraction = list(x[-1].values())[0]
 *             subgraph_contractions.append(contraction)             # <<<<<<<<<<<<<<
 *             subgraph_contractions_size.append(helpers.compute_size_by_dict(i, size_dict))
 * 
 */
    __pyx_t_29 = __Pyx_PyList_Append(__pyx_v_subgraph_contractions, __pyx_v_contraction); if (unlikely(__pyx_t_29 == ((int)-1))) __PYX_ERR(0, 954, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":955
 *             i, cost, contraction = list(x[-1].values())[0]
 *             subgraph_contractions.append(contraction)
 *             subgraph_contractions_size.append(helpers.compute_size_by_dict(i, size_dict))             # <<<<<<<<<<<<<<
 * 
 *         # sort the subgraph contractions by the size of the subgraphs in
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_helpers); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_compute_size_by_dict); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_cur_scope->__pyx_v_i, __pyx_v_size_dict};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_cur_scope->__pyx_v_i, __pyx_v_size_dict};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_i);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_15, __pyx_cur_scope->__pyx_v_i);
      __Pyx_INCREF(__pyx_v_size_dict);
      __Pyx_GIVEREF(__pyx_v_size_dict);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_15, __pyx_v_size_dict);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 955, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_29 = __Pyx_PyList_Append(__pyx_v_subgraph_contractions_size, __pyx_t_2); if (unlikely(__pyx_t_29 == ((int)-1))) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":883
 *         subgraph_contractions_size = [1]*len(inputs_done)
 * 
 *         for g in _find_disconnected_subgraphs(inputs, output):             # <<<<<<<<<<<<<<
 * 
 *             # dynamic programming approach to compute x[n] for subgraph g;
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":960
 *         # ascending order (will give the cheapest contractions); note that
 *         # outer products should be performed pairwise (to use BLAS functions)
 *         subgraph_contractions = [             # <<<<<<<<<<<<<<
 *             subgraph_contractions[j]
 *             for j in np.argsort(subgraph_contractions_size)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "acqdp/utility/opt_einsum_paths.pyx":962
 *         subgraph_contractions = [
 *             subgraph_contractions[j]
 *             for j in np.argsort(subgraph_contractions_size)             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_argsort); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_subgraph_contractions_size) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_subgraph_contractions_size);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 962, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 962, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 962, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 962, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 962, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 962, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":961
 *         # outer products should be performed pairwise (to use BLAS functions)
 *         subgraph_contractions = [
 *             subgraph_contractions[j]             # <<<<<<<<<<<<<<
 *             for j in np.argsort(subgraph_contractions_size)
 *         ]
 */
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_subgraph_contractions, __pyx_v_j); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 961, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 960, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":962
 *         subgraph_contractions = [
 *             subgraph_contractions[j]
 *             for j in np.argsort(subgraph_contractions_size)             # <<<<<<<<<<<<<<
 *         ]
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_subgraph_contractions, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":966
 * 
 *         # build the final contraction tree
 *         tree = functools.reduce(lambda x, y: (x, y), subgraph_contractions)             # <<<<<<<<<<<<<<
 * 
 *         return _tree_to_sequence(tree)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_functools); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_reduce); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_8__call___19lambda15, 0, __pyx_n_s_DynamicProgrammingOptimizer___ca_3, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_15 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_15 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_v_subgraph_contractions};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_v_subgraph_contractions};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_15, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_subgraph_contractions);
    __Pyx_GIVEREF(__pyx_v_subgraph_contractions);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_15, __pyx_v_subgraph_contractions);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_tree = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":968
 *         tree = functools.reduce(lambda x, y: (x, y), subgraph_contractions)
 * 
 *         return _tree_to_sequence(tree)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_tree_to_sequence); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_tree) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_tree);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":811
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.DynamicProgrammingOptimizer.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i_single);
  __Pyx_XDECREF(__pyx_v_inputs_done);
  __Pyx_XDECREF(__pyx_v_inputs_contractions);
  __Pyx_XDECREF(__pyx_v_subgraph_contractions);
  __Pyx_XDECREF(__pyx_v_subgraph_contractions_size);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_cost_cap);
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_xn);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF(__pyx_v_s1);
  __Pyx_XDECREF(__pyx_v_i1);
  __Pyx_XDECREF(__pyx_v_cost1);
  __Pyx_XDECREF(__pyx_v_cntrct1);
  __Pyx_XDECREF(__pyx_v_s2);
  __Pyx_XDECREF(__pyx_v_i2);
  __Pyx_XDECREF(__pyx_v_cost2);
  __Pyx_XDECREF(__pyx_v_cntrct2);
  __Pyx_XDECREF(__pyx_v_i1_cut_i2_wo_output);
  __Pyx_XDECREF(__pyx_v_i1_union_i2);
  __Pyx_XDECREF(__pyx_v_cost);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_i_r);
  __Pyx_XDECREF(__pyx_v_i_cntrct);
  __Pyx_XDECREF(__pyx_v_mem);
  __Pyx_XDECREF(__pyx_v_contraction);
  __Pyx_XDECREF(__pyx_v_tree);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_j);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_c);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_c);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_j);
  __Pyx_XDECREF(__pyx_v_size_dict);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":971
 * 
 * 
 * def dynamic_programming(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     optimizer = DynamicProgrammingOptimizer()
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_42dynamic_programming(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_42dynamic_programming = {"dynamic_programming", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_42dynamic_programming, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_42dynamic_programming(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  PyObject *__pyx_v_memory_limit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dynamic_programming (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamic_programming", 0, 3, 4, 1); __PYX_ERR(0, 971, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dynamic_programming", 0, 3, 4, 2); __PYX_ERR(0, 971, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dynamic_programming") < 0)) __PYX_ERR(0, 971, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
    __pyx_v_size_dict = values[2];
    __pyx_v_memory_limit = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dynamic_programming", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 971, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.dynamic_programming", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_41dynamic_programming(__pyx_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_41dynamic_programming(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit) {
  PyObject *__pyx_v_optimizer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dynamic_programming", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":972
 * 
 * def dynamic_programming(inputs, output, size_dict, memory_limit=None):
 *     optimizer = DynamicProgrammingOptimizer()             # <<<<<<<<<<<<<<
 *     return optimizer(inputs, output, size_dict, memory_limit)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_DynamicProgrammingOptimizer); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_optimizer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":973
 * def dynamic_programming(inputs, output, size_dict, memory_limit=None):
 *     optimizer = DynamicProgrammingOptimizer()
 *     return optimizer(inputs, output, size_dict, memory_limit)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_optimizer);
  __pyx_t_2 = __pyx_v_optimizer; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_inputs);
    __Pyx_GIVEREF(__pyx_v_inputs);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_inputs);
    __Pyx_INCREF(__pyx_v_output);
    __Pyx_GIVEREF(__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_output);
    __Pyx_INCREF(__pyx_v_size_dict);
    __Pyx_GIVEREF(__pyx_v_size_dict);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_size_dict);
    __Pyx_INCREF(__pyx_v_memory_limit);
    __Pyx_GIVEREF(__pyx_v_memory_limit);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_memory_limit);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":971
 * 
 * 
 * def dynamic_programming(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     optimizer = DynamicProgrammingOptimizer()
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.dynamic_programming", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_optimizer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":987
 * 
 * 
 * def auto(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """Finds the contraction path by automatically choosing the method based on
 *     how many input arguments there are.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_44auto(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_43auto[] = "Finds the contraction path by automatically choosing the method based on\n    how many input arguments there are.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_44auto = {"auto", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_44auto, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_43auto};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_44auto(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_inputs = 0;
  PyObject *__pyx_v_output = 0;
  PyObject *__pyx_v_size_dict = 0;
  PyObject *__pyx_v_memory_limit = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("auto (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_inputs,&__pyx_n_s_output,&__pyx_n_s_size_dict,&__pyx_n_s_memory_limit,0};
    PyObject* values[4] = {0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_inputs)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_output)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auto", 0, 3, 4, 1); __PYX_ERR(0, 987, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("auto", 0, 3, 4, 2); __PYX_ERR(0, 987, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory_limit);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "auto") < 0)) __PYX_ERR(0, 987, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_inputs = values[0];
    __pyx_v_output = values[1];
    __pyx_v_size_dict = values[2];
    __pyx_v_memory_limit = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("auto", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 987, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.auto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_43auto(__pyx_self, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_43auto(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputs, PyObject *__pyx_v_output, PyObject *__pyx_v_size_dict, PyObject *__pyx_v_memory_limit) {
  Py_ssize_t __pyx_v_N;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("auto", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":991
 *     how many input arguments there are.
 *     """
 *     N = len(inputs)             # <<<<<<<<<<<<<<
 *     return _AUTO_CHOICES.get(N, greedy)(inputs, output, size_dict, memory_limit)
 * 
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_inputs); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 991, __pyx_L1_error)
  __pyx_v_N = __pyx_t_1;

  /* "acqdp/utility/opt_einsum_paths.pyx":992
 *     """
 *     N = len(inputs)
 *     return _AUTO_CHOICES.get(N, greedy)(inputs, output, size_dict, memory_limit)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_AUTO_CHOICES); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_N); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_greedy); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_4, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_4, __pyx_t_6};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_6);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_inputs, __pyx_v_output, __pyx_v_size_dict, __pyx_v_memory_limit};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 4+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(4+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_v_inputs);
    __Pyx_GIVEREF(__pyx_v_inputs);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_inputs);
    __Pyx_INCREF(__pyx_v_output);
    __Pyx_GIVEREF(__pyx_v_output);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_output);
    __Pyx_INCREF(__pyx_v_size_dict);
    __Pyx_GIVEREF(__pyx_v_size_dict);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_size_dict);
    __Pyx_INCREF(__pyx_v_memory_limit);
    __Pyx_GIVEREF(__pyx_v_memory_limit);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_v_memory_limit);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":987
 * 
 * 
 * def auto(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """Finds the contraction path by automatically choosing the method based on
 *     how many input arguments there are.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.auto", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":1009
 * 
 * 
 * def register_path_fn(name, fn):             # <<<<<<<<<<<<<<
 *     """Add path finding function ``fn`` as an option with ``name``.
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_46register_path_fn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_45register_path_fn[] = "Add path finding function ``fn`` as an option with ``name``.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_46register_path_fn = {"register_path_fn", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_46register_path_fn, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_45register_path_fn};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_46register_path_fn(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_fn = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("register_path_fn (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_fn,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("register_path_fn", 1, 2, 2, 1); __PYX_ERR(0, 1009, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "register_path_fn") < 0)) __PYX_ERR(0, 1009, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_name = values[0];
    __pyx_v_fn = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("register_path_fn", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1009, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.register_path_fn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_45register_path_fn(__pyx_self, __pyx_v_name, __pyx_v_fn);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_45register_path_fn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_name, PyObject *__pyx_v_fn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("register_path_fn", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":1012
 *     """Add path finding function ``fn`` as an option with ``name``.
 *     """
 *     if name in _PATH_OPTIONS:             # <<<<<<<<<<<<<<
 *         raise KeyError("Path optimizer '{}' already exists.".format(name))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PATH_OPTIONS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "acqdp/utility/opt_einsum_paths.pyx":1013
 *     """
 *     if name in _PATH_OPTIONS:
 *         raise KeyError("Path optimizer '{}' already exists.".format(name))             # <<<<<<<<<<<<<<
 * 
 *     _PATH_OPTIONS[name.lower()] = fn
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Path_optimizer_already_exists, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_name);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1013, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":1012
 *     """Add path finding function ``fn`` as an option with ``name``.
 *     """
 *     if name in _PATH_OPTIONS:             # <<<<<<<<<<<<<<
 *         raise KeyError("Path optimizer '{}' already exists.".format(name))
 * 
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":1015
 *         raise KeyError("Path optimizer '{}' already exists.".format(name))
 * 
 *     _PATH_OPTIONS[name.lower()] = fn             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PATH_OPTIONS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_lower); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_t_1, __pyx_v_fn) < 0)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1009
 * 
 * 
 * def register_path_fn(name, fn):             # <<<<<<<<<<<<<<
 *     """Add path finding function ``fn`` as an option with ``name``.
 *     """
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.register_path_fn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "acqdp/utility/opt_einsum_paths.pyx":1018
 * 
 * 
 * def get_path_fn(path_type):             # <<<<<<<<<<<<<<
 *     """Get the correct path finding function from str ``path_type``.
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_48get_path_fn(PyObject *__pyx_self, PyObject *__pyx_v_path_type); /*proto*/
static char __pyx_doc_5acqdp_7utility_16opt_einsum_paths_47get_path_fn[] = "Get the correct path finding function from str ``path_type``.\n    ";
static PyMethodDef __pyx_mdef_5acqdp_7utility_16opt_einsum_paths_48get_path_fn = {"get_path_fn", (PyCFunction)__pyx_pw_5acqdp_7utility_16opt_einsum_paths_48get_path_fn, METH_O, __pyx_doc_5acqdp_7utility_16opt_einsum_paths_47get_path_fn};
static PyObject *__pyx_pw_5acqdp_7utility_16opt_einsum_paths_48get_path_fn(PyObject *__pyx_self, PyObject *__pyx_v_path_type) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_path_fn (wrapper)", 0);
  __pyx_r = __pyx_pf_5acqdp_7utility_16opt_einsum_paths_47get_path_fn(__pyx_self, ((PyObject *)__pyx_v_path_type));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5acqdp_7utility_16opt_einsum_paths_47get_path_fn(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_path_fn", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":1021
 *     """Get the correct path finding function from str ``path_type``.
 *     """
 *     if path_type not in _PATH_OPTIONS:             # <<<<<<<<<<<<<<
 *         raise KeyError("Path optimizer '{}' not found, valid options are {}."
 *                        .format(path_type, set(_PATH_OPTIONS.keys())))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PATH_OPTIONS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_path_type, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "acqdp/utility/opt_einsum_paths.pyx":1023
 *     if path_type not in _PATH_OPTIONS:
 *         raise KeyError("Path optimizer '{}' not found, valid options are {}."
 *                        .format(path_type, set(_PATH_OPTIONS.keys())))             # <<<<<<<<<<<<<<
 * 
 *     return _PATH_OPTIONS[path_type]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Path_optimizer_not_found_valid_o, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_PATH_OPTIONS); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_keys); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PySet_New(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_path_type, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_path_type, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_path_type);
      __Pyx_GIVEREF(__pyx_v_path_type);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_8, __pyx_v_path_type);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_8, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":1022
 *     """
 *     if path_type not in _PATH_OPTIONS:
 *         raise KeyError("Path optimizer '{}' not found, valid options are {}."             # <<<<<<<<<<<<<<
 *                        .format(path_type, set(_PATH_OPTIONS.keys())))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1022, __pyx_L1_error)

    /* "acqdp/utility/opt_einsum_paths.pyx":1021
 *     """Get the correct path finding function from str ``path_type``.
 *     """
 *     if path_type not in _PATH_OPTIONS:             # <<<<<<<<<<<<<<
 *         raise KeyError("Path optimizer '{}' not found, valid options are {}."
 *                        .format(path_type, set(_PATH_OPTIONS.keys())))
 */
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":1025
 *                        .format(path_type, set(_PATH_OPTIONS.keys())))
 * 
 *     return _PATH_OPTIONS[path_type]             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PATH_OPTIONS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_v_path_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1018
 * 
 * 
 * def get_path_fn(path_type):             # <<<<<<<<<<<<<<
 *     """Get the correct path finding function from str ``path_type``.
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("acqdp.utility.opt_einsum_paths.get_path_fn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ids);
  Py_CLEAR(p->__pyx_v_ssa_ids);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *)o;
  if (p->__pyx_v_ids) {
    e = (*v)(p->__pyx_v_ids, a); if (e) return e;
  }
  if (p->__pyx_v_ssa_ids) {
    e = (*v)(p->__pyx_v_ssa_ids, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear *)o;
  tmp = ((PyObject*)p->__pyx_v_ids);
  p->__pyx_v_ids = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_ssa_ids);
  p->__pyx_v_ssa_ids = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct__ssa_to_linear", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_ssa_id);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_ssa_id) {
    e = (*v)(p->__pyx_v_ssa_id, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ids);
  Py_CLEAR(p->__pyx_v_linear_to_ssa);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *)o;
  if (p->__pyx_v_ids) {
    e = (*v)(p->__pyx_v_ids, a); if (e) return e;
  }
  if (p->__pyx_v_linear_to_ssa) {
    e = (*v)(p->__pyx_v_linear_to_ssa, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa *)o;
  tmp = ((PyObject*)p->__pyx_v_ids);
  p->__pyx_v_ids = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_linear_to_ssa);
  p->__pyx_v_linear_to_ssa = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_2_linear_to_ssa", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_id_);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_id_) {
    e = (*v)(p->__pyx_v_id_, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v__optimal_iterate);
  Py_CLEAR(p->__pyx_v_best);
  Py_CLEAR(p->__pyx_v_memory_limit);
  Py_CLEAR(p->__pyx_v_output);
  Py_CLEAR(p->__pyx_v_result_cache);
  Py_CLEAR(p->__pyx_v_size_cache);
  Py_CLEAR(p->__pyx_v_size_dict);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *)o;
  if (p->__pyx_v__optimal_iterate) {
    e = (*v)(p->__pyx_v__optimal_iterate, a); if (e) return e;
  }
  if (p->__pyx_v_best) {
    e = (*v)(p->__pyx_v_best, a); if (e) return e;
  }
  if (p->__pyx_v_memory_limit) {
    e = (*v)(p->__pyx_v_memory_limit, a); if (e) return e;
  }
  if (p->__pyx_v_output) {
    e = (*v)(p->__pyx_v_output, a); if (e) return e;
  }
  if (p->__pyx_v_result_cache) {
    e = (*v)(p->__pyx_v_result_cache, a); if (e) return e;
  }
  if (p->__pyx_v_size_cache) {
    e = (*v)(p->__pyx_v_size_cache, a); if (e) return e;
  }
  if (p->__pyx_v_size_dict) {
    e = (*v)(p->__pyx_v_size_dict, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal *)o;
  tmp = ((PyObject*)p->__pyx_v__optimal_iterate);
  p->__pyx_v__optimal_iterate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_best);
  p->__pyx_v_best = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_memory_limit);
  p->__pyx_v_memory_limit = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_output);
  p->__pyx_v_output = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_result_cache);
  p->__pyx_v_result_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_size_cache);
  p->__pyx_v_size_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_size_dict);
  p->__pyx_v_size_dict = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_4_optimal", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v__branch_iterate);
  Py_CLEAR(p->__pyx_v_memory_limit);
  Py_CLEAR(p->__pyx_v_output);
  Py_CLEAR(p->__pyx_v_result_cache);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_size_cache);
  Py_CLEAR(p->__pyx_v_size_dict);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *)o;
  if (p->__pyx_v__branch_iterate) {
    e = (*v)(p->__pyx_v__branch_iterate, a); if (e) return e;
  }
  if (p->__pyx_v_memory_limit) {
    e = (*v)(p->__pyx_v_memory_limit, a); if (e) return e;
  }
  if (p->__pyx_v_output) {
    e = (*v)(p->__pyx_v_output, a); if (e) return e;
  }
  if (p->__pyx_v_result_cache) {
    e = (*v)(p->__pyx_v_result_cache, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(p->__pyx_v_self, a); if (e) return e;
  }
  if (p->__pyx_v_size_cache) {
    e = (*v)(p->__pyx_v_size_cache, a); if (e) return e;
  }
  if (p->__pyx_v_size_dict) {
    e = (*v)(p->__pyx_v_size_dict, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *)o;
  tmp = ((PyObject*)p->__pyx_v__branch_iterate);
  p->__pyx_v__branch_iterate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_memory_limit);
  p->__pyx_v_memory_limit = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_output);
  p->__pyx_v_output = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_result_cache);
  p->__pyx_v_result_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_size_cache);
  p->__pyx_v_size_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_size_dict);
  p->__pyx_v_size_dict = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_5___call__", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_flops);
  Py_CLEAR(p->__pyx_v_inputs);
  Py_CLEAR(p->__pyx_v_path);
  Py_CLEAR(p->__pyx_v_remaining);
  Py_CLEAR(p->__pyx_v_size);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_flops) {
    e = (*v)(p->__pyx_v_flops, a); if (e) return e;
  }
  if (p->__pyx_v_inputs) {
    e = (*v)(p->__pyx_v_inputs, a); if (e) return e;
  }
  if (p->__pyx_v_path) {
    e = (*v)(p->__pyx_v_path, a); if (e) return e;
  }
  if (p->__pyx_v_remaining) {
    e = (*v)(p->__pyx_v_remaining, a); if (e) return e;
  }
  if (p->__pyx_v_size) {
    e = (*v)(p->__pyx_v_size, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_flops);
  p->__pyx_v_flops = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_inputs);
  p->__pyx_v_inputs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_path);
  p->__pyx_v_path = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_remaining);
  p->__pyx_v_remaining = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_size);
  p->__pyx_v_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_6__branch_iterate", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_cost_fn);
  Py_CLEAR(p->__pyx_v_dim_ref_counts);
  Py_CLEAR(p->__pyx_v_footprints);
  Py_CLEAR(p->__pyx_v_k1);
  Py_CLEAR(p->__pyx_v_k2s);
  Py_CLEAR(p->__pyx_v_output);
  Py_CLEAR(p->__pyx_v_remaining);
  Py_CLEAR(p->__pyx_v_sizes);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *)o;
  if (p->__pyx_v_cost_fn) {
    e = (*v)(p->__pyx_v_cost_fn, a); if (e) return e;
  }
  if (p->__pyx_v_dim_ref_counts) {
    e = (*v)(p->__pyx_v_dim_ref_counts, a); if (e) return e;
  }
  if (p->__pyx_v_footprints) {
    e = (*v)(p->__pyx_v_footprints, a); if (e) return e;
  }
  if (p->__pyx_v_k1) {
    e = (*v)(p->__pyx_v_k1, a); if (e) return e;
  }
  if (p->__pyx_v_k2s) {
    e = (*v)(p->__pyx_v_k2s, a); if (e) return e;
  }
  if (p->__pyx_v_output) {
    e = (*v)(p->__pyx_v_output, a); if (e) return e;
  }
  if (p->__pyx_v_remaining) {
    e = (*v)(p->__pyx_v_remaining, a); if (e) return e;
  }
  if (p->__pyx_v_sizes) {
    e = (*v)(p->__pyx_v_sizes, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate *)o;
  tmp = ((PyObject*)p->__pyx_v_cost_fn);
  p->__pyx_v_cost_fn = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_dim_ref_counts);
  p->__pyx_v_dim_ref_counts = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_footprints);
  p->__pyx_v_footprints = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_k1);
  p->__pyx_v_k1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_k2s);
  p->__pyx_v_k2s = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_output);
  p->__pyx_v_output = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_remaining);
  p->__pyx_v_remaining = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_sizes);
  p->__pyx_v_sizes = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_7__push_candidate", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_k2);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_k2) {
    e = (*v)(p->__pyx_v_k2, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_8_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_8genexpr4__pyx_v_count);
  Py_CLEAR(p->__pyx_v_dim_to_keys);
  Py_CLEAR(p->__pyx_v_k1);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *)o;
  if (p->__pyx_8genexpr4__pyx_v_count) {
    e = (*v)(p->__pyx_8genexpr4__pyx_v_count, a); if (e) return e;
  }
  if (p->__pyx_v_dim_to_keys) {
    e = (*v)(p->__pyx_v_dim_to_keys, a); if (e) return e;
  }
  if (p->__pyx_v_k1) {
    e = (*v)(p->__pyx_v_k1, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize *)o;
  tmp = ((PyObject*)p->__pyx_8genexpr4__pyx_v_count);
  p->__pyx_8genexpr4__pyx_v_count = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_dim_to_keys);
  p->__pyx_v_dim_to_keys = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_k1);
  p->__pyx_v_k1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_9_ssa_greedy_optimize", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_dim);
  Py_CLEAR(p->__pyx_v_keys);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_dim) {
    e = (*v)(p->__pyx_v_dim, a); if (e) return e;
  }
  if (p->__pyx_v_keys) {
    e = (*v)(p->__pyx_v_keys, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_10_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_dim);
  Py_CLEAR(p->__pyx_v_k2);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_dim) {
    e = (*v)(p->__pyx_v_dim, a); if (e) return e;
  }
  if (p->__pyx_v_k2) {
    e = (*v)(p->__pyx_v_k2, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_11_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_v_t);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *)o;
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_v_t) {
    e = (*v)(p->__pyx_v_t, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence *)o;
  tmp = ((PyObject*)p->__pyx_v_i);
  p->__pyx_v_i = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_t);
  p->__pyx_v_t = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_12__tree_to_sequence", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_q);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_q) {
    e = (*v)(p->__pyx_v_q, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_13_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_j);
  Py_CLEAR(p->__pyx_v_s);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices *)o;
  if (p->__pyx_v_j) {
    e = (*v)(p->__pyx_v_j, a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_14__bitmapset_indices", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_g);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_v_inputs);
  Py_CLEAR(p->__pyx_v_output);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_v_symbol2int);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *)o;
  if (p->__pyx_v_g) {
    e = (*v)(p->__pyx_v_g, a); if (e) return e;
  }
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_v_inputs) {
    e = (*v)(p->__pyx_v_inputs, a); if (e) return e;
  }
  if (p->__pyx_v_output) {
    e = (*v)(p->__pyx_v_output, a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_v_symbol2int) {
    e = (*v)(p->__pyx_v_symbol2int, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ *)o;
  tmp = ((PyObject*)p->__pyx_v_g);
  p->__pyx_v_g = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_i);
  p->__pyx_v_i = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_inputs);
  p->__pyx_v_inputs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_output);
  p->__pyx_v_output = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_r);
  p->__pyx_v_r = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_symbol2int);
  p->__pyx_v_symbol2int = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_15___call__", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__, /*tp_traverse*/
  __pyx_tp_clear_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_16_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_17_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_c);
  Py_CLEAR(p->__pyx_v_genexpr);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_c) {
    e = (*v)(p->__pyx_v_c, a); if (e) return e;
  }
  if (p->__pyx_v_genexpr) {
    e = (*v)(p->__pyx_v_genexpr, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_18_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_19_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_j);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_j) {
    e = (*v)(p->__pyx_v_j, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_20_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_j);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_j) {
    e = (*v)(p->__pyx_v_j, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_21_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr[8];
static int __pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr = 0;

static PyObject *__pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr)))) {
    o = (PyObject*)__pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr[--__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr(PyObject *o) {
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_j);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr)))) {
    __pyx_freelist_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr[__pyx_freecount_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr++] = ((struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *p = (struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_j) {
    e = (*v)(p->__pyx_v_j, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "acqdp.utility.opt_einsum_paths.__pyx_scope_struct_22_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_opt_einsum_paths(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_opt_einsum_paths},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "opt_einsum_paths",
    __pyx_k_Contains_the_path_technology_be, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AUTO_CHOICES, __pyx_k_AUTO_CHOICES, sizeof(__pyx_k_AUTO_CHOICES), 0, 0, 1, 1},
  {&__pyx_n_s_BETTER_FNS, __pyx_k_BETTER_FNS, sizeof(__pyx_k_BETTER_FNS), 0, 0, 1, 1},
  {&__pyx_kp_s_Base_class_for_different_path_op, __pyx_k_Base_class_for_different_path_op, sizeof(__pyx_k_Base_class_for_different_path_op), 0, 0, 1, 0},
  {&__pyx_n_s_BranchBound, __pyx_k_BranchBound, sizeof(__pyx_k_BranchBound), 0, 0, 1, 1},
  {&__pyx_n_s_BranchBound___call, __pyx_k_BranchBound___call, sizeof(__pyx_k_BranchBound___call), 0, 0, 1, 1},
  {&__pyx_kp_u_BranchBound___call___line_316, __pyx_k_BranchBound___call___line_316, sizeof(__pyx_k_BranchBound___call___line_316), 0, 1, 0, 0},
  {&__pyx_n_s_BranchBound___call___locals__bra, __pyx_k_BranchBound___call___locals__bra, sizeof(__pyx_k_BranchBound___call___locals__bra), 0, 0, 1, 1},
  {&__pyx_n_s_BranchBound___call___locals__bra_2, __pyx_k_BranchBound___call___locals__bra_2, sizeof(__pyx_k_BranchBound___call___locals__bra_2), 0, 0, 1, 1},
  {&__pyx_n_s_BranchBound___init, __pyx_k_BranchBound___init, sizeof(__pyx_k_BranchBound___init), 0, 0, 1, 1},
  {&__pyx_n_s_BranchBound___init___locals_lamb, __pyx_k_BranchBound___init___locals_lamb, sizeof(__pyx_k_BranchBound___init___locals_lamb), 0, 0, 1, 1},
  {&__pyx_n_s_BranchBound_path, __pyx_k_BranchBound_path, sizeof(__pyx_k_BranchBound_path), 0, 0, 1, 1},
  {&__pyx_n_s_COST_FNS, __pyx_k_COST_FNS, sizeof(__pyx_k_COST_FNS), 0, 0, 1, 1},
  {&__pyx_kp_u_Computes_all_possible_pair_cont, __pyx_k_Computes_all_possible_pair_cont, sizeof(__pyx_k_Computes_all_possible_pair_cont), 0, 1, 0, 0},
  {&__pyx_kp_u_Convert_a_path_with_recycled_li, __pyx_k_Convert_a_path_with_recycled_li, sizeof(__pyx_k_Convert_a_path_with_recycled_li), 0, 1, 0, 0},
  {&__pyx_kp_u_Convert_a_path_with_static_sing, __pyx_k_Convert_a_path_with_static_sing, sizeof(__pyx_k_Convert_a_path_with_static_sing), 0, 1, 0, 0},
  {&__pyx_kp_u_Converts_a_contraction_tree_to, __pyx_k_Converts_a_contraction_tree_to, sizeof(__pyx_k_Converts_a_contraction_tree_to), 0, 1, 0, 0},
  {&__pyx_n_s_DynamicProgrammingOptimizer, __pyx_k_DynamicProgrammingOptimizer, sizeof(__pyx_k_DynamicProgrammingOptimizer), 0, 0, 1, 1},
  {&__pyx_n_s_DynamicProgrammingOptimizer___ca, __pyx_k_DynamicProgrammingOptimizer___ca, sizeof(__pyx_k_DynamicProgrammingOptimizer___ca), 0, 0, 1, 1},
  {&__pyx_n_s_DynamicProgrammingOptimizer___ca_2, __pyx_k_DynamicProgrammingOptimizer___ca_2, sizeof(__pyx_k_DynamicProgrammingOptimizer___ca_2), 0, 0, 1, 1},
  {&__pyx_n_s_DynamicProgrammingOptimizer___ca_3, __pyx_k_DynamicProgrammingOptimizer___ca_3, sizeof(__pyx_k_DynamicProgrammingOptimizer___ca_3), 0, 0, 1, 1},
  {&__pyx_n_s_DynamicProgrammingOptimizer___ca_4, __pyx_k_DynamicProgrammingOptimizer___ca_4, sizeof(__pyx_k_DynamicProgrammingOptimizer___ca_4), 0, 0, 1, 1},
  {&__pyx_kp_u_DynamicProgrammingOptimizer___ca_5, __pyx_k_DynamicProgrammingOptimizer___ca_5, sizeof(__pyx_k_DynamicProgrammingOptimizer___ca_5), 0, 1, 0, 0},
  {&__pyx_kp_s_Explores_possible_pair_contract, __pyx_k_Explores_possible_pair_contract, sizeof(__pyx_k_Explores_possible_pair_contract), 0, 0, 1, 0},
  {&__pyx_kp_u_Finds_disconnected_subgraphs_in, __pyx_k_Finds_disconnected_subgraphs_in, sizeof(__pyx_k_Finds_disconnected_subgraphs_in), 0, 1, 0, 0},
  {&__pyx_kp_s_Finds_the_optimal_path_of_pairw, __pyx_k_Finds_the_optimal_path_of_pairw, sizeof(__pyx_k_Finds_the_optimal_path_of_pairw), 0, 0, 1, 0},
  {&__pyx_kp_u_Finds_the_path_by_a_three_stage, __pyx_k_Finds_the_path_by_a_three_stage, sizeof(__pyx_k_Finds_the_path_by_a_three_stage), 0, 1, 0, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_N, __pyx_k_N, sizeof(__pyx_k_N), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_PATH_OPTIONS, __pyx_k_PATH_OPTIONS, sizeof(__pyx_k_PATH_OPTIONS), 0, 0, 1, 1},
  {&__pyx_kp_u_Parameters_input_sets_list_List, __pyx_k_Parameters_input_sets_list_List, sizeof(__pyx_k_Parameters_input_sets_list_List), 0, 1, 0, 0},
  {&__pyx_kp_u_Parameters_inputs_list_List_of, __pyx_k_Parameters_inputs_list_List_of, sizeof(__pyx_k_Parameters_inputs_list_List_of), 0, 1, 0, 0},
  {&__pyx_n_s_PathOptimizer, __pyx_k_PathOptimizer, sizeof(__pyx_k_PathOptimizer), 0, 0, 1, 1},
  {&__pyx_n_s_PathOptimizer___call, __pyx_k_PathOptimizer___call, sizeof(__pyx_k_PathOptimizer___call), 0, 0, 1, 1},
  {&__pyx_kp_s_Path_optimizer_already_exists, __pyx_k_Path_optimizer_already_exists, sizeof(__pyx_k_Path_optimizer_already_exists), 0, 0, 1, 0},
  {&__pyx_kp_s_Path_optimizer_not_found_valid_o, __pyx_k_Path_optimizer_not_found_valid_o, sizeof(__pyx_k_Path_optimizer_not_found_valid_o), 0, 0, 1, 0},
  {&__pyx_kp_u_Returns_a_generator_object_allo, __pyx_k_Returns_a_generator_object_allo, sizeof(__pyx_k_Returns_a_generator_object_allo), 0, 1, 0, 0},
  {&__pyx_n_s_UNLIMITED_MEM, __pyx_k_UNLIMITED_MEM, sizeof(__pyx_k_UNLIMITED_MEM), 0, 0, 1, 1},
  {&__pyx_n_s__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 0, 1, 1},
  {&__pyx_n_s__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 1},
  {&__pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_k_acqdp_utility_opt_einsum_paths, sizeof(__pyx_k_acqdp_utility_opt_einsum_paths), 0, 0, 1, 1},
  {&__pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_k_acqdp_utility_opt_einsum_paths_p, sizeof(__pyx_k_acqdp_utility_opt_einsum_paths_p), 0, 0, 1, 0},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_all_tensors, __pyx_k_all_tensors, sizeof(__pyx_k_all_tensors), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_arange, __pyx_k_arange, sizeof(__pyx_k_arange), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_argsort, __pyx_k_argsort, sizeof(__pyx_k_argsort), 0, 0, 1, 1},
  {&__pyx_n_s_assess_candidate, __pyx_k_assess_candidate, sizeof(__pyx_k_assess_candidate), 0, 0, 1, 1},
  {&__pyx_n_s_auto, __pyx_k_auto, sizeof(__pyx_k_auto), 0, 0, 1, 1},
  {&__pyx_n_s_best, __pyx_k_best, sizeof(__pyx_k_best), 0, 0, 1, 1},
  {&__pyx_n_s_best_flops, __pyx_k_best_flops, sizeof(__pyx_k_best_flops), 0, 0, 1, 1},
  {&__pyx_n_s_best_progress, __pyx_k_best_progress, sizeof(__pyx_k_best_progress), 0, 0, 1, 1},
  {&__pyx_n_s_best_size, __pyx_k_best_size, sizeof(__pyx_k_best_size), 0, 0, 1, 1},
  {&__pyx_n_s_better, __pyx_k_better, sizeof(__pyx_k_better), 0, 0, 1, 1},
  {&__pyx_n_s_better_flops_first, __pyx_k_better_flops_first, sizeof(__pyx_k_better_flops_first), 0, 0, 1, 1},
  {&__pyx_n_s_better_size_first, __pyx_k_better_size_first, sizeof(__pyx_k_better_size_first), 0, 0, 1, 1},
  {&__pyx_n_s_bi, __pyx_k_bi, sizeof(__pyx_k_bi), 0, 0, 1, 1},
  {&__pyx_n_s_bitmapset_indices, __pyx_k_bitmapset_indices, sizeof(__pyx_k_bitmapset_indices), 0, 0, 1, 1},
  {&__pyx_kp_u_bitmapset_indices_line_764, __pyx_k_bitmapset_indices_line_764, sizeof(__pyx_k_bitmapset_indices_line_764), 0, 1, 0, 0},
  {&__pyx_n_s_branch, __pyx_k_branch, sizeof(__pyx_k_branch), 0, 0, 1, 1},
  {&__pyx_n_s_branch_1, __pyx_k_branch_1, sizeof(__pyx_k_branch_1), 0, 0, 1, 1},
  {&__pyx_kp_s_branch_1_2, __pyx_k_branch_1_2, sizeof(__pyx_k_branch_1_2), 0, 0, 1, 0},
  {&__pyx_n_s_branch_2, __pyx_k_branch_2, sizeof(__pyx_k_branch_2), 0, 0, 1, 1},
  {&__pyx_kp_s_branch_2_2, __pyx_k_branch_2_2, sizeof(__pyx_k_branch_2_2), 0, 0, 1, 0},
  {&__pyx_n_s_branch_all, __pyx_k_branch_all, sizeof(__pyx_k_branch_all), 0, 0, 1, 1},
  {&__pyx_kp_s_branch_all_2, __pyx_k_branch_all_2, sizeof(__pyx_k_branch_all_2), 0, 0, 1, 0},
  {&__pyx_n_s_branch_iterate, __pyx_k_branch_iterate, sizeof(__pyx_k_branch_iterate), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_calc_k12_flops, __pyx_k_calc_k12_flops, sizeof(__pyx_k_calc_k12_flops), 0, 0, 1, 1},
  {&__pyx_n_s_call, __pyx_k_call, sizeof(__pyx_k_call), 0, 0, 1, 1},
  {&__pyx_n_s_candidate, __pyx_k_candidate, sizeof(__pyx_k_candidate), 0, 0, 1, 1},
  {&__pyx_n_s_candidates, __pyx_k_candidates, sizeof(__pyx_k_candidates), 0, 0, 1, 1},
  {&__pyx_n_s_choose_fn, __pyx_k_choose_fn, sizeof(__pyx_k_choose_fn), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cntrct1, __pyx_k_cntrct1, sizeof(__pyx_k_cntrct1), 0, 0, 1, 1},
  {&__pyx_n_s_cntrct2, __pyx_k_cntrct2, sizeof(__pyx_k_cntrct2), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_combinations, __pyx_k_combinations, sizeof(__pyx_k_combinations), 0, 0, 1, 1},
  {&__pyx_n_s_compute_oversize_flops, __pyx_k_compute_oversize_flops, sizeof(__pyx_k_compute_oversize_flops), 0, 0, 1, 1},
  {&__pyx_n_s_compute_size_by_dict, __pyx_k_compute_size_by_dict, sizeof(__pyx_k_compute_size_by_dict), 0, 0, 1, 1},
  {&__pyx_n_s_con, __pyx_k_con, sizeof(__pyx_k_con), 0, 0, 1, 1},
  {&__pyx_n_s_contraction, __pyx_k_contraction, sizeof(__pyx_k_contraction), 0, 0, 1, 1},
  {&__pyx_n_s_cost, __pyx_k_cost, sizeof(__pyx_k_cost), 0, 0, 1, 1},
  {&__pyx_n_s_cost1, __pyx_k_cost1, sizeof(__pyx_k_cost1), 0, 0, 1, 1},
  {&__pyx_n_s_cost2, __pyx_k_cost2, sizeof(__pyx_k_cost2), 0, 0, 1, 1},
  {&__pyx_n_s_cost_cap, __pyx_k_cost_cap, sizeof(__pyx_k_cost_cap), 0, 0, 1, 1},
  {&__pyx_n_s_cost_fn, __pyx_k_cost_fn, sizeof(__pyx_k_cost_fn), 0, 0, 1, 1},
  {&__pyx_n_s_cost_memory_removed, __pyx_k_cost_memory_removed, sizeof(__pyx_k_cost_memory_removed), 0, 0, 1, 1},
  {&__pyx_n_s_cost_memory_removed_jitter, __pyx_k_cost_memory_removed_jitter, sizeof(__pyx_k_cost_memory_removed_jitter), 0, 0, 1, 1},
  {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
  {&__pyx_n_s_cutoff_flops_factor, __pyx_k_cutoff_flops_factor, sizeof(__pyx_k_cutoff_flops_factor), 0, 0, 1, 1},
  {&__pyx_n_s_defaultdict, __pyx_k_defaultdict, sizeof(__pyx_k_defaultdict), 0, 0, 1, 1},
  {&__pyx_n_s_difference_update, __pyx_k_difference_update, sizeof(__pyx_k_difference_update), 0, 0, 1, 1},
  {&__pyx_n_s_dim, __pyx_k_dim, sizeof(__pyx_k_dim), 0, 0, 1, 1},
  {&__pyx_n_s_dim_ref_counts, __pyx_k_dim_ref_counts, sizeof(__pyx_k_dim_ref_counts), 0, 0, 1, 1},
  {&__pyx_n_s_dim_to_keys, __pyx_k_dim_to_keys, sizeof(__pyx_k_dim_to_keys), 0, 0, 1, 1},
  {&__pyx_n_s_dims, __pyx_k_dims, sizeof(__pyx_k_dims), 0, 0, 1, 1},
  {&__pyx_n_s_discard, __pyx_k_discard, sizeof(__pyx_k_discard), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dp, __pyx_k_dp, sizeof(__pyx_k_dp), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dynamic_programming, __pyx_k_dynamic_programming, sizeof(__pyx_k_dynamic_programming), 0, 0, 1, 1},
  {&__pyx_kp_s_dynamic_programming_2, __pyx_k_dynamic_programming_2, sizeof(__pyx_k_dynamic_programming_2), 0, 0, 1, 0},
  {&__pyx_n_s_eager, __pyx_k_eager, sizeof(__pyx_k_eager), 0, 0, 1, 1},
  {&__pyx_n_s_either, __pyx_k_either, sizeof(__pyx_k_either), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_find_disconnected_subgraphs, __pyx_k_find_disconnected_subgraphs, sizeof(__pyx_k_find_disconnected_subgraphs), 0, 0, 1, 1},
  {&__pyx_kp_u_find_disconnected_subgraphs_lin, __pyx_k_find_disconnected_subgraphs_lin, sizeof(__pyx_k_find_disconnected_subgraphs_lin), 0, 1, 0, 0},
  {&__pyx_n_s_flop_count, __pyx_k_flop_count, sizeof(__pyx_k_flop_count), 0, 0, 1, 1},
  {&__pyx_n_s_flops, __pyx_k_flops, sizeof(__pyx_k_flops), 0, 0, 1, 1},
  {&__pyx_n_s_flops12, __pyx_k_flops12, sizeof(__pyx_k_flops12), 0, 0, 1, 1},
  {&__pyx_n_s_fn, __pyx_k_fn, sizeof(__pyx_k_fn), 0, 0, 1, 1},
  {&__pyx_n_s_footprints, __pyx_k_footprints, sizeof(__pyx_k_footprints), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_functools, __pyx_k_functools, sizeof(__pyx_k_functools), 0, 0, 1, 1},
  {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
  {&__pyx_n_s_gauss, __pyx_k_gauss, sizeof(__pyx_k_gauss), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_better_fn, __pyx_k_get_better_fn, sizeof(__pyx_k_get_better_fn), 0, 0, 1, 1},
  {&__pyx_n_s_get_candidate, __pyx_k_get_candidate, sizeof(__pyx_k_get_candidate), 0, 0, 1, 1},
  {&__pyx_n_s_get_path_fn, __pyx_k_get_path_fn, sizeof(__pyx_k_get_path_fn), 0, 0, 1, 1},
  {&__pyx_n_s_getitem, __pyx_k_getitem, sizeof(__pyx_k_getitem), 0, 0, 1, 1},
  {&__pyx_n_s_greedy, __pyx_k_greedy, sizeof(__pyx_k_greedy), 0, 0, 1, 1},
  {&__pyx_kp_u_greedy_line_612, __pyx_k_greedy_line_612, sizeof(__pyx_k_greedy_line_612), 0, 1, 0, 0},
  {&__pyx_n_s_heapify, __pyx_k_heapify, sizeof(__pyx_k_heapify), 0, 0, 1, 1},
  {&__pyx_n_s_heappop, __pyx_k_heappop, sizeof(__pyx_k_heappop), 0, 0, 1, 1},
  {&__pyx_n_s_heappush, __pyx_k_heappush, sizeof(__pyx_k_heappush), 0, 0, 1, 1},
  {&__pyx_n_s_heappushpop, __pyx_k_heappushpop, sizeof(__pyx_k_heappushpop), 0, 0, 1, 1},
  {&__pyx_n_s_heapq, __pyx_k_heapq, sizeof(__pyx_k_heapq), 0, 0, 1, 1},
  {&__pyx_n_s_helpers, __pyx_k_helpers, sizeof(__pyx_k_helpers), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_i1, __pyx_k_i1, sizeof(__pyx_k_i1), 0, 0, 1, 1},
  {&__pyx_n_s_i1_cut_i2_wo_output, __pyx_k_i1_cut_i2_wo_output, sizeof(__pyx_k_i1_cut_i2_wo_output), 0, 0, 1, 1},
  {&__pyx_n_s_i1_union_i2, __pyx_k_i1_union_i2, sizeof(__pyx_k_i1_union_i2), 0, 0, 1, 1},
  {&__pyx_n_s_i2, __pyx_k_i2, sizeof(__pyx_k_i2), 0, 0, 1, 1},
  {&__pyx_n_s_i_cntrct, __pyx_k_i_cntrct, sizeof(__pyx_k_i_cntrct), 0, 0, 1, 1},
  {&__pyx_n_s_i_r, __pyx_k_i_r, sizeof(__pyx_k_i_r), 0, 0, 1, 1},
  {&__pyx_n_s_i_single, __pyx_k_i_single, sizeof(__pyx_k_i_single), 0, 0, 1, 1},
  {&__pyx_n_s_i_sum, __pyx_k_i_sum, sizeof(__pyx_k_i_sum), 0, 0, 1, 1},
  {&__pyx_n_s_i_tmp, __pyx_k_i_tmp, sizeof(__pyx_k_i_tmp), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_id1, __pyx_k_id1, sizeof(__pyx_k_id1), 0, 0, 1, 1},
  {&__pyx_n_s_id2, __pyx_k_id2, sizeof(__pyx_k_id2), 0, 0, 1, 1},
  {&__pyx_n_s_ids, __pyx_k_ids, sizeof(__pyx_k_ids), 0, 0, 1, 1},
  {&__pyx_n_s_idx_contraction, __pyx_k_idx_contraction, sizeof(__pyx_k_idx_contraction), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_inner, __pyx_k_inner, sizeof(__pyx_k_inner), 0, 0, 1, 1},
  {&__pyx_n_s_inputs, __pyx_k_inputs, sizeof(__pyx_k_inputs), 0, 0, 1, 1},
  {&__pyx_n_s_inputs_contractions, __pyx_k_inputs_contractions, sizeof(__pyx_k_inputs_contractions), 0, 0, 1, 1},
  {&__pyx_n_s_inputs_done, __pyx_k_inputs_done, sizeof(__pyx_k_inputs_done), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_intersection, __pyx_k_intersection, sizeof(__pyx_k_intersection), 0, 0, 1, 1},
  {&__pyx_n_s_isdisjoint, __pyx_k_isdisjoint, sizeof(__pyx_k_isdisjoint), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_k1, __pyx_k_k1, sizeof(__pyx_k_k1), 0, 0, 1, 1},
  {&__pyx_n_s_k12, __pyx_k_k12, sizeof(__pyx_k_k12), 0, 0, 1, 1},
  {&__pyx_n_s_k2, __pyx_k_k2, sizeof(__pyx_k_k2), 0, 0, 1, 1},
  {&__pyx_n_s_k2s, __pyx_k_k2s, sizeof(__pyx_k_k2s), 0, 0, 1, 1},
  {&__pyx_n_s_keep, __pyx_k_keep, sizeof(__pyx_k_keep), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_linear_to_ssa, __pyx_k_linear_to_ssa, sizeof(__pyx_k_linear_to_ssa), 0, 0, 1, 1},
  {&__pyx_kp_u_linear_to_ssa_line_69, __pyx_k_linear_to_ssa_line_69, sizeof(__pyx_k_linear_to_ssa_line_69), 0, 1, 0, 0},
  {&__pyx_n_s_linear_to_ssa_locals_genexpr, __pyx_k_linear_to_ssa_locals_genexpr, sizeof(__pyx_k_linear_to_ssa_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_lower, __pyx_k_lower, sizeof(__pyx_k_lower), 0, 0, 1, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_mem, __pyx_k_mem, sizeof(__pyx_k_mem), 0, 0, 1, 1},
  {&__pyx_n_s_memory_limit, __pyx_k_memory_limit, sizeof(__pyx_k_memory_limit), 0, 0, 1, 1},
  {&__pyx_kp_s_memory_removed, __pyx_k_memory_removed, sizeof(__pyx_k_memory_removed), 0, 0, 1, 0},
  {&__pyx_kp_s_memory_removed_jitter, __pyx_k_memory_removed_jitter, sizeof(__pyx_k_memory_removed_jitter), 0, 0, 1, 0},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_minimize, __pyx_k_minimize, sizeof(__pyx_k_minimize), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_nbranch, __pyx_k_nbranch, sizeof(__pyx_k_nbranch), 0, 0, 1, 1},
  {&__pyx_n_s_new_flops, __pyx_k_new_flops, sizeof(__pyx_k_new_flops), 0, 0, 1, 1},
  {&__pyx_n_s_new_ids, __pyx_k_new_ids, sizeof(__pyx_k_new_ids), 0, 0, 1, 1},
  {&__pyx_n_s_new_size, __pyx_k_new_size, sizeof(__pyx_k_new_size), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_num_inputs, __pyx_k_num_inputs, sizeof(__pyx_k_num_inputs), 0, 0, 1, 1},
  {&__pyx_n_s_num_terms, __pyx_k_num_terms, sizeof(__pyx_k_num_terms), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_one, __pyx_k_one, sizeof(__pyx_k_one), 0, 0, 1, 1},
  {&__pyx_n_s_opportunistic, __pyx_k_opportunistic, sizeof(__pyx_k_opportunistic), 0, 0, 1, 1},
  {&__pyx_n_s_opt_einsum_helpers, __pyx_k_opt_einsum_helpers, sizeof(__pyx_k_opt_einsum_helpers), 0, 0, 1, 1},
  {&__pyx_n_s_optimal, __pyx_k_optimal, sizeof(__pyx_k_optimal), 0, 0, 1, 1},
  {&__pyx_n_s_optimal_iterate, __pyx_k_optimal_iterate, sizeof(__pyx_k_optimal_iterate), 0, 0, 1, 1},
  {&__pyx_kp_u_optimal_line_141, __pyx_k_optimal_line_141, sizeof(__pyx_k_optimal_line_141), 0, 1, 0, 0},
  {&__pyx_n_s_optimal_locals__optimal_iterate, __pyx_k_optimal_locals__optimal_iterate, sizeof(__pyx_k_optimal_locals__optimal_iterate), 0, 0, 1, 1},
  {&__pyx_n_s_optimizer, __pyx_k_optimizer, sizeof(__pyx_k_optimizer), 0, 0, 1, 1},
  {&__pyx_n_s_optimizer_kwargs, __pyx_k_optimizer_kwargs, sizeof(__pyx_k_optimizer_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_output, __pyx_k_output, sizeof(__pyx_k_output), 0, 0, 1, 1},
  {&__pyx_n_s_partial, __pyx_k_partial, sizeof(__pyx_k_partial), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_path_type, __pyx_k_path_type, sizeof(__pyx_k_path_type), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_property, __pyx_k_property, sizeof(__pyx_k_property), 0, 0, 1, 1},
  {&__pyx_n_s_push_all, __pyx_k_push_all, sizeof(__pyx_k_push_all), 0, 0, 1, 1},
  {&__pyx_n_s_push_candidate, __pyx_k_push_candidate, sizeof(__pyx_k_push_candidate), 0, 0, 1, 1},
  {&__pyx_n_s_push_candidate_locals_genexpr, __pyx_k_push_candidate_locals_genexpr, sizeof(__pyx_k_push_candidate_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_queue, __pyx_k_queue, sizeof(__pyx_k_queue), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_register_path_fn, __pyx_k_register_path_fn, sizeof(__pyx_k_register_path_fn), 0, 0, 1, 1},
  {&__pyx_n_s_remaining, __pyx_k_remaining, sizeof(__pyx_k_remaining), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_result_cache, __pyx_k_result_cache, sizeof(__pyx_k_result_cache), 0, 0, 1, 1},
  {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_s1, __pyx_k_s1, sizeof(__pyx_k_s1), 0, 0, 1, 1},
  {&__pyx_n_s_s2, __pyx_k_s2, sizeof(__pyx_k_s2), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_shared, __pyx_k_shared, sizeof(__pyx_k_shared), 0, 0, 1, 1},
  {&__pyx_n_s_simple_chooser, __pyx_k_simple_chooser, sizeof(__pyx_k_simple_chooser), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_size1, __pyx_k_size1, sizeof(__pyx_k_size1), 0, 0, 1, 1},
  {&__pyx_n_s_size12, __pyx_k_size12, sizeof(__pyx_k_size12), 0, 0, 1, 1},
  {&__pyx_n_s_size2, __pyx_k_size2, sizeof(__pyx_k_size2), 0, 0, 1, 1},
  {&__pyx_n_s_size_cache, __pyx_k_size_cache, sizeof(__pyx_k_size_cache), 0, 0, 1, 1},
  {&__pyx_n_s_size_dict, __pyx_k_size_dict, sizeof(__pyx_k_size_dict), 0, 0, 1, 1},
  {&__pyx_n_s_sizes, __pyx_k_sizes, sizeof(__pyx_k_sizes), 0, 0, 1, 1},
  {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_greedy_optimize, __pyx_k_ssa_greedy_optimize, sizeof(__pyx_k_ssa_greedy_optimize), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_greedy_optimize_locals_genex, __pyx_k_ssa_greedy_optimize_locals_genex, sizeof(__pyx_k_ssa_greedy_optimize_locals_genex), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_id, __pyx_k_ssa_id, sizeof(__pyx_k_ssa_id), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_id1, __pyx_k_ssa_id1, sizeof(__pyx_k_ssa_id1), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_id12, __pyx_k_ssa_id12, sizeof(__pyx_k_ssa_id12), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_id2, __pyx_k_ssa_id2, sizeof(__pyx_k_ssa_id2), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_ids, __pyx_k_ssa_ids, sizeof(__pyx_k_ssa_ids), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_path, __pyx_k_ssa_path, sizeof(__pyx_k_ssa_path), 0, 0, 1, 1},
  {&__pyx_n_s_ssa_to_linear, __pyx_k_ssa_to_linear, sizeof(__pyx_k_ssa_to_linear), 0, 0, 1, 1},
  {&__pyx_kp_u_ssa_to_linear_line_52, __pyx_k_ssa_to_linear_line_52, sizeof(__pyx_k_ssa_to_linear_line_52), 0, 1, 0, 0},
  {&__pyx_n_s_ssa_to_linear_locals_genexpr, __pyx_k_ssa_to_linear_locals_genexpr, sizeof(__pyx_k_ssa_to_linear_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_subgraph_contractions, __pyx_k_subgraph_contractions, sizeof(__pyx_k_subgraph_contractions), 0, 0, 1, 1},
  {&__pyx_n_s_subgraph_contractions_size, __pyx_k_subgraph_contractions_size, sizeof(__pyx_k_subgraph_contractions_size), 0, 0, 1, 1},
  {&__pyx_n_s_subgraphs, __pyx_k_subgraphs, sizeof(__pyx_k_subgraphs), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_symbol2int, __pyx_k_symbol2int, sizeof(__pyx_k_symbol2int), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tree, __pyx_k_tree, sizeof(__pyx_k_tree), 0, 0, 1, 1},
  {&__pyx_n_s_tree_to_sequence, __pyx_k_tree_to_sequence, sizeof(__pyx_k_tree_to_sequence), 0, 0, 1, 1},
  {&__pyx_kp_u_tree_to_sequence_line_658, __pyx_k_tree_to_sequence_line_658, sizeof(__pyx_k_tree_to_sequence_line_658), 0, 1, 0, 0},
  {&__pyx_n_s_tree_to_sequence_locals_genexpr, __pyx_k_tree_to_sequence_locals_genexpr, sizeof(__pyx_k_tree_to_sequence_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_two, __pyx_k_two, sizeof(__pyx_k_two), 0, 0, 1, 1},
  {&__pyx_n_s_union, __pyx_k_union, sizeof(__pyx_k_union), 0, 0, 1, 1},
  {&__pyx_n_s_unused_inputs, __pyx_k_unused_inputs, sizeof(__pyx_k_unused_inputs), 0, 0, 1, 1},
  {&__pyx_n_s_update_ref_counts, __pyx_k_update_ref_counts, sizeof(__pyx_k_update_ref_counts), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_xn, __pyx_k_xn, sizeof(__pyx_k_xn), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_builtin_property = __Pyx_GetBuiltinName(__pyx_n_s_property); if (!__pyx_builtin_property) __PYX_ERR(0, 312, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 977, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 49, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 77, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 83, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 476, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 534, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 872, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "acqdp/utility/opt_einsum_paths.pyx":179
 *     result_cache = {}
 * 
 *     def _optimal_iterate(path, remaining, inputs, flops):             # <<<<<<<<<<<<<<
 * 
 *         # reached end of path (only ever get here if flops is best found so far)
 */
  __pyx_tuple_ = PyTuple_Pack(11, __pyx_n_s_path, __pyx_n_s_remaining, __pyx_n_s_inputs, __pyx_n_s_flops, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_key, __pyx_n_s_k12, __pyx_n_s_flops12, __pyx_n_s_new_flops, __pyx_n_s_size12); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_codeobj__2 = (PyObject*)__Pyx_PyCode_New(4, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple_, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_optimal_iterate, 179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__2)) __PYX_ERR(0, 179, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":264
 *     and thus jumbles the contractions a bit.
 *     """
 *     return random.gauss(1.0, 0.01) * (size12 - size1 - size2)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(2, __pyx_float_1_0, __pyx_float_0_01); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "acqdp/utility/opt_einsum_paths.pyx":359
 *                 return
 * 
 *             def _assess_candidate(k1, k2, i, j):             # <<<<<<<<<<<<<<
 *                 # find resulting indices and flops
 *                 try:
 */
  __pyx_tuple__4 = PyTuple_Pack(12, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_k12, __pyx_n_s_flops12, __pyx_n_s_size12, __pyx_n_s_new_flops, __pyx_n_s_new_size, __pyx_n_s_size1, __pyx_n_s_size2, __pyx_n_s_cost); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(4, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_assess_candidate, 359, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 359, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":350
 *         result_cache = {}
 * 
 *         def _branch_iterate(path, inputs, remaining, flops, size):             # <<<<<<<<<<<<<<
 * 
 *             # reached end of path (only ever get here if flops is best found so far)
 */
  __pyx_tuple__7 = PyTuple_Pack(18, __pyx_n_s_path, __pyx_n_s_inputs, __pyx_n_s_remaining, __pyx_n_s_flops, __pyx_n_s_size, __pyx_n_s_assess_candidate, __pyx_n_s_assess_candidate, __pyx_n_s_candidates, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_candidate, __pyx_n_s_bi, __pyx_n_s__6, __pyx_n_s_new_flops, __pyx_n_s_new_size, __pyx_n_s_k12); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_codeobj__8 = (PyObject*)__Pyx_PyCode_New(5, 0, 18, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__7, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_branch_iterate, 350, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__8)) __PYX_ERR(0, 350, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":510
 *     if len(inputs) == 1:
 *         # Perform a single contraction to match output shape.
 *         return [(0,)]             # <<<<<<<<<<<<<<
 * 
 *     # set the function that assigns a heuristic cost to a possible contraction
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "acqdp/utility/opt_einsum_paths.pyx":552
 *     dim_ref_counts = {
 *         count: set(dim for dim, keys in dim_to_keys.items() if len(keys) >= count) - output
 *         for count in [2, 3]}             # <<<<<<<<<<<<<<
 * 
 *     # Compute separable part of the objective function for contractions.
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "acqdp/utility/opt_einsum_paths.pyx":561
 *     for dim, keys in dim_to_keys.items():
 *         keys = sorted(keys, key=remaining.__getitem__)
 *         for i, k1 in enumerate(keys[:-1]):             # <<<<<<<<<<<<<<
 *             k2s = keys[1 + i:]
 *             _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn)
 */
  __pyx_slice__11 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);

  /* "acqdp/utility/opt_einsum_paths.pyx":24
 * 
 * 
 * class PathOptimizer(object):             # <<<<<<<<<<<<<<
 *     """Base class for different path optimizers to inherit from.
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "acqdp/utility/opt_einsum_paths.pyx":48
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_tuple__15 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_call, 48, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 48, __pyx_L1_error)
  __pyx_tuple__17 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "acqdp/utility/opt_einsum_paths.pyx":52
 * 
 * 
 * def ssa_to_linear(ssa_path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with static single assignment ids to a path with recycled
 */
  __pyx_tuple__18 = PyTuple_Pack(7, __pyx_n_s_ssa_path, __pyx_n_s_ids, __pyx_n_s_path, __pyx_n_s_ssa_ids, __pyx_n_s_ssa_id, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_ssa_to_linear, 52, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 52, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":69
 * 
 * 
 * def linear_to_ssa(path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with recycled linear ids to a path with static single
 */
  __pyx_tuple__20 = PyTuple_Pack(9, __pyx_n_s_path, __pyx_n_s_num_inputs, __pyx_n_s_linear_to_ssa, __pyx_n_s_new_ids, __pyx_n_s_ssa_path, __pyx_n_s_ids, __pyx_n_s_id, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_linear_to_ssa, 69, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 69, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":89
 * 
 * 
 * def calc_k12_flops(inputs, output, remaining, int i, int j, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Calculate the resulting indices and flops for a potential pairwise
 */
  __pyx_tuple__22 = PyTuple_Pack(13, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_remaining, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_size_dict, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_either, __pyx_n_s_shared, __pyx_n_s_keep, __pyx_n_s_k12, __pyx_n_s_cost); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(6, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_calc_k12_flops, 89, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 89, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":130
 * 
 * 
 * def _compute_oversize_flops(inputs, remaining, output, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Compute the flop count for a contraction of all remaining arguments. This
 */
  __pyx_tuple__24 = PyTuple_Pack(7, __pyx_n_s_inputs, __pyx_n_s_remaining, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_idx_contraction, __pyx_n_s_inner, __pyx_n_s_num_terms); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(4, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_compute_oversize_flops, 130, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 130, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":141
 * 
 * 
 * def optimal(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """
 *     Computes all possible pair contractions in a depth-first recursive manner,
 */
  __pyx_tuple__26 = PyTuple_Pack(9, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit, __pyx_n_s_best, __pyx_n_s_size_cache, __pyx_n_s_result_cache, __pyx_n_s_optimal_iterate, __pyx_n_s_optimal_iterate); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(4, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_optimal, 141, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 141, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":233
 * # functions for comparing which of two paths is 'better'
 * 
 * def better_flops_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (flops, size) < (best_flops, best_size)
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(4, __pyx_n_s_flops, __pyx_n_s_size, __pyx_n_s_best_flops, __pyx_n_s_best_size); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_better_flops_first, 233, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 233, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":237
 * 
 * 
 * def better_size_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (size, flops) < (best_size, best_flops)
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(4, __pyx_n_s_flops, __pyx_n_s_size, __pyx_n_s_best_flops, __pyx_n_s_best_size); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_better_size_first, 237, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 237, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":247
 * 
 * 
 * def get_better_fn(key):             # <<<<<<<<<<<<<<
 *     return _BETTER_FNS[key]
 * 
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_n_s_key); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_get_better_fn, 247, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 247, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":253
 * # functions for assigning a heuristic 'cost' to a potential contraction
 * 
 * def cost_memory_removed(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """The default heuristic cost, corresponding to the total reduction in
 *     memory of performing a contraction.
 */
  __pyx_tuple__34 = PyTuple_Pack(6, __pyx_n_s_size12, __pyx_n_s_size1, __pyx_n_s_size2, __pyx_n_s_k12, __pyx_n_s_k1, __pyx_n_s_k2); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_cost_memory_removed, 253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 253, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":260
 * 
 * 
 * def cost_memory_removed_jitter(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """Like memory-removed, but with a slight amount of noise that breaks ties
 *     and thus jumbles the contractions a bit.
 */
  __pyx_tuple__36 = PyTuple_Pack(6, __pyx_n_s_size12, __pyx_n_s_size1, __pyx_n_s_size2, __pyx_n_s_k12, __pyx_n_s_k1, __pyx_n_s_k2); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(6, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_cost_memory_removed_jitter, 260, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 260, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":302
 *     """
 * 
 *     def __init__(self, nbranch=None, cutoff_flops_factor=4, minimize='flops', cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *         self.nbranch = nbranch
 *         self.cutoff_flops_factor = cutoff_flops_factor
 */
  __pyx_tuple__38 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_nbranch, __pyx_n_s_cutoff_flops_factor, __pyx_n_s_minimize, __pyx_n_s_cost_fn); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(5, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_init, 302, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 302, __pyx_L1_error)
  __pyx_tuple__40 = PyTuple_Pack(4, ((PyObject *)Py_None), ((PyObject *)__pyx_int_4), ((PyObject*)__pyx_n_s_flops), ((PyObject*)__pyx_kp_s_memory_removed)); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "acqdp/utility/opt_einsum_paths.pyx":313
 * 
 *     @property
 *     def path(self):             # <<<<<<<<<<<<<<
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_path, 313, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 313, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":316
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 * 
 */
  __pyx_tuple__43 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit, __pyx_n_s_size_cache, __pyx_n_s_result_cache, __pyx_n_s_branch_iterate, __pyx_n_s_branch_iterate, __pyx_n_s_k); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(5, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_call, 316, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 316, __pyx_L1_error)
  __pyx_tuple__45 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "acqdp/utility/opt_einsum_paths.pyx":445
 * 
 * 
 * def branch(inputs, output, size_dict, memory_limit=None, **optimizer_kwargs):             # <<<<<<<<<<<<<<
 *     optimizer = BranchBound(**optimizer_kwargs)
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */
  __pyx_tuple__46 = PyTuple_Pack(6, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit, __pyx_n_s_optimizer_kwargs, __pyx_n_s_optimizer); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(4, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_branch, 445, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 445, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":455
 * 
 * 
 * def _get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn):             # <<<<<<<<<<<<<<
 *     cdef int id1, id2
 *     either = k1 | k2
 */
  __pyx_tuple__48 = PyTuple_Pack(15, __pyx_n_s_output, __pyx_n_s_sizes, __pyx_n_s_remaining, __pyx_n_s_footprints, __pyx_n_s_dim_ref_counts, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_cost_fn, __pyx_n_s_id1, __pyx_n_s_id2, __pyx_n_s_either, __pyx_n_s_two, __pyx_n_s_one, __pyx_n_s_k12, __pyx_n_s_cost); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(8, 0, 15, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_get_candidate, 455, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 455, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":469
 * 
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):             # <<<<<<<<<<<<<<
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:
 */
  __pyx_tuple__50 = PyTuple_Pack(14, __pyx_n_s_output, __pyx_n_s_sizes, __pyx_n_s_remaining, __pyx_n_s_footprints, __pyx_n_s_dim_ref_counts, __pyx_n_s_k1, __pyx_n_s_k2s, __pyx_n_s_queue, __pyx_n_s_push_all, __pyx_n_s_cost_fn, __pyx_n_s_candidates, __pyx_n_s_candidate, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(10, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_push_candidate, 469, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 469, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":479
 * 
 * 
 * def _update_ref_counts(dim_to_keys, dim_ref_counts, dims):             # <<<<<<<<<<<<<<
 *     for dim in dims:
 *         count = len(dim_to_keys[dim])
 */
  __pyx_tuple__52 = PyTuple_Pack(5, __pyx_n_s_dim_to_keys, __pyx_n_s_dim_ref_counts, __pyx_n_s_dims, __pyx_n_s_dim, __pyx_n_s_count); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_update_ref_counts, 479, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 479, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":493
 * 
 * 
 * def _simple_chooser(queue, remaining):             # <<<<<<<<<<<<<<
 *     """Default contraction chooser that simply takes the minimum cost option.
 *     """
 */
  __pyx_tuple__54 = PyTuple_Pack(6, __pyx_n_s_queue, __pyx_n_s_remaining, __pyx_n_s_cost, __pyx_n_s_k1, __pyx_n_s_k2, __pyx_n_s_k12); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_simple_chooser, 493, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 493, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":502
 * 
 * 
 * def ssa_greedy_optimize(inputs, output, sizes, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     This is the core function for :func:`greedy` but produces a path with
 */
  __pyx_tuple__56 = PyTuple_Pack(33, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_sizes, __pyx_n_s_choose_fn, __pyx_n_s_cost_fn, __pyx_n_s_push_all, __pyx_n_s_remaining, __pyx_n_s_ssa_ids, __pyx_n_s_ssa_path, __pyx_n_s_ssa_id, __pyx_n_s_key, __pyx_n_s_dim_to_keys, __pyx_n_s_dim, __pyx_n_s_dim_ref_counts, __pyx_n_s_footprints, __pyx_n_s_queue, __pyx_n_s_keys, __pyx_n_s_i, __pyx_n_s_k1, __pyx_n_s_k2s, __pyx_n_s_con, __pyx_n_s_cost, __pyx_n_s_k2, __pyx_n_s_k12, __pyx_n_s_ssa_id1, __pyx_n_s_ssa_id2, __pyx_n_s__6, __pyx_n_s_ssa_id12, __pyx_n_s_count, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_key, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(5, 0, 33, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_ssa_greedy_optimize, 502, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 502, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":612
 * 
 * 
 * def greedy(inputs, output, size_dict, memory_limit=None, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     Finds the path by a three stage algorithm:
 */
  __pyx_tuple__58 = PyTuple_Pack(7, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit, __pyx_n_s_choose_fn, __pyx_n_s_cost_fn, __pyx_n_s_ssa_path); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(6, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_greedy, 612, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 612, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":658
 * 
 * 
 * def _tree_to_sequence(c):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a contraction tree to a contraction path as it has to be
 */
  __pyx_tuple__60 = PyTuple_Pack(7, __pyx_n_s_c, __pyx_n_s_t, __pyx_n_s_s, __pyx_n_s_j, __pyx_n_s_i, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_tree_to_sequence, 658, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 658, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":716
 * 
 * 
 * def _find_disconnected_subgraphs(inputs, output):             # <<<<<<<<<<<<<<
 *     """
 *     Finds disconnected subgraphs in the given list of inputs. Inputs are
 */
  __pyx_tuple__62 = PyTuple_Pack(11, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_subgraphs, __pyx_n_s_unused_inputs, __pyx_n_s_i_sum, __pyx_n_s_g, __pyx_n_s_q, __pyx_n_s_j, __pyx_n_s_i_tmp, __pyx_n_s_n, __pyx_n_s_k); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(2, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__62, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_find_disconnected_subgraphs, 716, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 716, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":764
 * 
 * 
 * def _bitmapset_indices(s):             # <<<<<<<<<<<<<<
 *     """
 *     Returns a generator object allowing to iterate over the elements contained
 */
  __pyx_tuple__64 = PyTuple_Pack(2, __pyx_n_s_s, __pyx_n_s_j); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_bitmapset_indices, 764, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 764, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":811
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */
  __pyx_tuple__65 = PyTuple_Pack(50, __pyx_n_s_self, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit, __pyx_n_s_symbol2int, __pyx_n_s_i_single, __pyx_n_s_inputs_done, __pyx_n_s_inputs_contractions, __pyx_n_s_subgraph_contractions, __pyx_n_s_subgraph_contractions_size, __pyx_n_s_g, __pyx_n_s_x, __pyx_n_s_all_tensors, __pyx_n_s_cost_cap, __pyx_n_s_n, __pyx_n_s_xn, __pyx_n_s_m, __pyx_n_s_s1, __pyx_n_s_i1, __pyx_n_s_cost1, __pyx_n_s_cntrct1, __pyx_n_s_s2, __pyx_n_s_i2, __pyx_n_s_cost2, __pyx_n_s_cntrct2, __pyx_n_s_i1_cut_i2_wo_output, __pyx_n_s_i1_union_i2, __pyx_n_s_cost, __pyx_n_s_s, __pyx_n_s_r, __pyx_n_s_i_r, __pyx_n_s_i_cntrct, __pyx_n_s_i, __pyx_n_s_mem, __pyx_n_s_contraction, __pyx_n_s_tree, __pyx_n_s_j, __pyx_n_s_c, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_c, __pyx_n_s_v, __pyx_n_s_j, __pyx_n_s_genexpr, __pyx_n_s_j, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(5, 0, 50, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_call, 811, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 811, __pyx_L1_error)
  __pyx_tuple__67 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "acqdp/utility/opt_einsum_paths.pyx":971
 * 
 * 
 * def dynamic_programming(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     optimizer = DynamicProgrammingOptimizer()
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */
  __pyx_tuple__68 = PyTuple_Pack(5, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit, __pyx_n_s_optimizer); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_dynamic_programming, 971, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 971, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":987
 * 
 * 
 * def auto(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """Finds the contraction path by automatically choosing the method based on
 *     how many input arguments there are.
 */
  __pyx_tuple__70 = PyTuple_Pack(5, __pyx_n_s_inputs, __pyx_n_s_output, __pyx_n_s_size_dict, __pyx_n_s_memory_limit, __pyx_n_s_N); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(4, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_auto, 987, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 987, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":1009
 * 
 * 
 * def register_path_fn(name, fn):             # <<<<<<<<<<<<<<
 *     """Add path finding function ``fn`` as an option with ``name``.
 *     """
 */
  __pyx_tuple__72 = PyTuple_Pack(2, __pyx_n_s_name, __pyx_n_s_fn); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__72, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_register_path_fn, 1009, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(0, 1009, __pyx_L1_error)

  /* "acqdp/utility/opt_einsum_paths.pyx":1018
 * 
 * 
 * def get_path_fn(path_type):             # <<<<<<<<<<<<<<
 *     """Get the correct path finding function from str ``path_type``.
 *     """
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_n_s_path_type); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__74, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_acqdp_utility_opt_einsum_paths_p, __pyx_n_s_get_path_fn, 1018, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  __pyx_umethod_PySet_Type_difference_update.type = (PyObject*)&PySet_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_1_0 = PyFloat_FromDouble(1.0); if (unlikely(!__pyx_float_1_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_01 = PyFloat_FromDouble(0.01); if (unlikely(!__pyx_float_0_01)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct__ssa_to_linear;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa) < 0) __PYX_ERR(0, 69, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_2_linear_to_ssa;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_3_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_4_optimal;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__) < 0) __PYX_ERR(0, 316, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__ = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_5___call__;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate) < 0) __PYX_ERR(0, 350, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_6__branch_iterate;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate) < 0) __PYX_ERR(0, 469, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_7__push_candidate;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr) < 0) __PYX_ERR(0, 470, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_8_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize) < 0) __PYX_ERR(0, 502, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_9_ssa_greedy_optimize;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr) < 0) __PYX_ERR(0, 551, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_10_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr) < 0) __PYX_ERR(0, 591, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_11_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence) < 0) __PYX_ERR(0, 658, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_12__tree_to_sequence;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr) < 0) __PYX_ERR(0, 706, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_13_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices) < 0) __PYX_ERR(0, 764, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_14__bitmapset_indices;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__) < 0) __PYX_ERR(0, 811, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__ = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_15___call__;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr) < 0) __PYX_ERR(0, 851, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_16_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr) < 0) __PYX_ERR(0, 852, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_17_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr) < 0) __PYX_ERR(0, 858, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_18_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr) < 0) __PYX_ERR(0, 859, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_19_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr) < 0) __PYX_ERR(0, 894, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_20_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr) < 0) __PYX_ERR(0, 906, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_21_genexpr;
  if (PyType_Ready(&__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr.tp_dictoffset && __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr = &__pyx_type_5acqdp_7utility_16opt_einsum_paths___pyx_scope_struct_22_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initopt_einsum_paths(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initopt_einsum_paths(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_opt_einsum_paths(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_opt_einsum_paths(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_opt_einsum_paths(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  long __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'opt_einsum_paths' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_opt_einsum_paths(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("opt_einsum_paths", __pyx_methods, __pyx_k_Contains_the_path_technology_be, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_acqdp__utility__opt_einsum_paths) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "acqdp.utility.opt_einsum_paths")) {
      if (unlikely(PyDict_SetItemString(modules, "acqdp.utility.opt_einsum_paths", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "acqdp/utility/opt_einsum_paths.pyx":5
 * """
 * 
 * import functools             # <<<<<<<<<<<<<<
 * import heapq
 * import itertools
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_functools, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_functools, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":6
 * 
 * import functools
 * import heapq             # <<<<<<<<<<<<<<
 * import itertools
 * import random
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_heapq, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_heapq, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":7
 * import functools
 * import heapq
 * import itertools             # <<<<<<<<<<<<<<
 * import random
 * from collections import defaultdict
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_itertools, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itertools, __pyx_t_1) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":8
 * import heapq
 * import itertools
 * import random             # <<<<<<<<<<<<<<
 * from collections import defaultdict
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_random, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":9
 * import itertools
 * import random
 * from collections import defaultdict             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_defaultdict);
  __Pyx_GIVEREF(__pyx_n_s_defaultdict);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_defaultdict);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultdict, __pyx_t_1) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":11
 * from collections import defaultdict
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * from . import opt_einsum_helpers as helpers
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":13
 * import numpy as np
 * 
 * from . import opt_einsum_helpers as helpers             # <<<<<<<<<<<<<<
 * 
 * __all__ = [
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_opt_einsum_helpers);
  __Pyx_GIVEREF(__pyx_n_s_opt_einsum_helpers);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_opt_einsum_helpers);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s__13, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_opt_einsum_helpers); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_helpers, __pyx_t_2) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":15
 * from . import opt_einsum_helpers as helpers
 * 
 * __all__ = [             # <<<<<<<<<<<<<<
 *     "optimal", "BranchBound", "branch", "greedy", "auto", "get_path_fn",
 *     "DynamicProgrammingOptimizer", "dynamic_programming"
 */
  __pyx_t_1 = PyList_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_optimal);
  __Pyx_GIVEREF(__pyx_n_s_optimal);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_optimal);
  __Pyx_INCREF(__pyx_n_s_BranchBound);
  __Pyx_GIVEREF(__pyx_n_s_BranchBound);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_BranchBound);
  __Pyx_INCREF(__pyx_n_s_branch);
  __Pyx_GIVEREF(__pyx_n_s_branch);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_branch);
  __Pyx_INCREF(__pyx_n_s_greedy);
  __Pyx_GIVEREF(__pyx_n_s_greedy);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_greedy);
  __Pyx_INCREF(__pyx_n_s_auto);
  __Pyx_GIVEREF(__pyx_n_s_auto);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_auto);
  __Pyx_INCREF(__pyx_n_s_get_path_fn);
  __Pyx_GIVEREF(__pyx_n_s_get_path_fn);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_s_get_path_fn);
  __Pyx_INCREF(__pyx_n_s_DynamicProgrammingOptimizer);
  __Pyx_GIVEREF(__pyx_n_s_DynamicProgrammingOptimizer);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_s_DynamicProgrammingOptimizer);
  __Pyx_INCREF(__pyx_n_s_dynamic_programming);
  __Pyx_GIVEREF(__pyx_n_s_dynamic_programming);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_s_dynamic_programming);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_1) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":21
 * 
 * 
 * _UNLIMITED_MEM = {-1, None, float('inf')}             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_n_s_inf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_int_neg_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, Py_None) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_t_1) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UNLIMITED_MEM, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":24
 * 
 * 
 * class PathOptimizer(object):             # <<<<<<<<<<<<<<
 *     """Base class for different path optimizers to inherit from.
 * 
 */
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple__14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_tuple__14, __pyx_n_s_PathOptimizer, __pyx_n_s_PathOptimizer, (PyObject *) NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_kp_s_Base_class_for_different_path_op); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "acqdp/utility/opt_einsum_paths.pyx":48
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_13PathOptimizer_1__call__, 0, __pyx_n_s_PathOptimizer___call, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__16)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__17);
  if (__Pyx_SetNameInClass(__pyx_t_1, __pyx_n_s_call, __pyx_t_3) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":24
 * 
 * 
 * class PathOptimizer(object):             # <<<<<<<<<<<<<<
 *     """Base class for different path optimizers to inherit from.
 * 
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_PathOptimizer, __pyx_tuple__14, __pyx_t_1, NULL, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PathOptimizer, __pyx_t_3) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":52
 * 
 * 
 * def ssa_to_linear(ssa_path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with static single assignment ids to a path with recycled
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_1ssa_to_linear, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ssa_to_linear, __pyx_t_2) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":69
 * 
 * 
 * def linear_to_ssa(path):             # <<<<<<<<<<<<<<
 *     """
 *     Convert a path with recycled linear ids to a path with static single
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_3linear_to_ssa, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_linear_to_ssa, __pyx_t_2) < 0) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":89
 * 
 * 
 * def calc_k12_flops(inputs, output, remaining, int i, int j, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Calculate the resulting indices and flops for a potential pairwise
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_5calc_k12_flops, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_calc_k12_flops, __pyx_t_2) < 0) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":130
 * 
 * 
 * def _compute_oversize_flops(inputs, remaining, output, size_dict):             # <<<<<<<<<<<<<<
 *     """
 *     Compute the flop count for a contraction of all remaining arguments. This
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_7_compute_oversize_flops, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compute_oversize_flops, __pyx_t_2) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":141
 * 
 * 
 * def optimal(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """
 *     Computes all possible pair contractions in a depth-first recursive manner,
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_9optimal, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_optimal, __pyx_t_2) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":233
 * # functions for comparing which of two paths is 'better'
 * 
 * def better_flops_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (flops, size) < (best_flops, best_size)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11better_flops_first, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_better_flops_first, __pyx_t_2) < 0) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":237
 * 
 * 
 * def better_size_first(flops, size, best_flops, best_size):             # <<<<<<<<<<<<<<
 *     return (size, flops) < (best_size, best_flops)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_13better_size_first, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_better_size_first, __pyx_t_2) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":242
 * 
 * _BETTER_FNS = {
 *     'flops': better_flops_first,             # <<<<<<<<<<<<<<
 *     'size': better_size_first,
 * }
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_better_flops_first); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_flops, __pyx_t_1) < 0) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":243
 * _BETTER_FNS = {
 *     'flops': better_flops_first,
 *     'size': better_size_first,             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_better_size_first); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_size, __pyx_t_1) < 0) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BETTER_FNS, __pyx_t_2) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":247
 * 
 * 
 * def get_better_fn(key):             # <<<<<<<<<<<<<<
 *     return _BETTER_FNS[key]
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_15get_better_fn, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_better_fn, __pyx_t_2) < 0) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":253
 * # functions for assigning a heuristic 'cost' to a potential contraction
 * 
 * def cost_memory_removed(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """The default heuristic cost, corresponding to the total reduction in
 *     memory of performing a contraction.
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_17cost_memory_removed, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cost_memory_removed, __pyx_t_2) < 0) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":260
 * 
 * 
 * def cost_memory_removed_jitter(size12, size1, size2, k12, k1, k2):             # <<<<<<<<<<<<<<
 *     """Like memory-removed, but with a slight amount of noise that breaks ties
 *     and thus jumbles the contractions a bit.
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_19cost_memory_removed_jitter, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cost_memory_removed_jitter, __pyx_t_2) < 0) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":268
 * 
 * _COST_FNS = {
 *     'memory-removed': cost_memory_removed,             # <<<<<<<<<<<<<<
 *     'memory-removed-jitter': cost_memory_removed_jitter,
 * }
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_cost_memory_removed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_memory_removed, __pyx_t_1) < 0) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":269
 * _COST_FNS = {
 *     'memory-removed': cost_memory_removed,
 *     'memory-removed-jitter': cost_memory_removed_jitter,             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_cost_memory_removed_jitter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_memory_removed_jitter, __pyx_t_1) < 0) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_COST_FNS, __pyx_t_2) < 0) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":273
 * 
 * 
 * class BranchBound(PathOptimizer):             # <<<<<<<<<<<<<<
 *     """
 *     Explores possible pair contractions in a depth-first recursive manner like
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PathOptimizer); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_BranchBound, __pyx_n_s_BranchBound, (PyObject *) NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_kp_s_Explores_possible_pair_contract); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "acqdp/utility/opt_einsum_paths.pyx":302
 *     """
 * 
 *     def __init__(self, nbranch=None, cutoff_flops_factor=4, minimize='flops', cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *         self.nbranch = nbranch
 *         self.cutoff_flops_factor = cutoff_flops_factor
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_1__init__, 0, __pyx_n_s_BranchBound___init, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__40);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_init, __pyx_t_4) < 0) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":313
 * 
 *     @property
 *     def path(self):             # <<<<<<<<<<<<<<
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_3path, 0, __pyx_n_s_BranchBound_path, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "acqdp/utility/opt_einsum_paths.pyx":312
 *         self.best_progress = defaultdict(lambda: float('inf'))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def path(self):
 *         return ssa_to_linear(self.best['ssa_path'])
 */
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_property, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_path, __pyx_t_5) < 0) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":316
 *         return ssa_to_linear(self.best['ssa_path'])
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_11BranchBound_5__call__, 0, __pyx_n_s_BranchBound___call, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__45);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_call, __pyx_t_5) < 0) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":273
 * 
 * 
 * class BranchBound(PathOptimizer):             # <<<<<<<<<<<<<<
 *     """
 *     Explores possible pair contractions in a depth-first recursive manner like
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_BranchBound, __pyx_t_1, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BranchBound, __pyx_t_5) < 0) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":445
 * 
 * 
 * def branch(inputs, output, size_dict, memory_limit=None, **optimizer_kwargs):             # <<<<<<<<<<<<<<
 *     optimizer = BranchBound(**optimizer_kwargs)
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_21branch, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_branch, __pyx_t_1) < 0) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":450
 * 
 * 
 * branch_all = functools.partial(branch, nbranch=None)             # <<<<<<<<<<<<<<
 * branch_2 = functools.partial(branch, nbranch=2)
 * branch_1 = functools.partial(branch, nbranch=1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_functools); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_partial); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_branch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_nbranch, Py_None) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_branch_all, __pyx_t_5) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":451
 * 
 * branch_all = functools.partial(branch, nbranch=None)
 * branch_2 = functools.partial(branch, nbranch=2)             # <<<<<<<<<<<<<<
 * branch_1 = functools.partial(branch, nbranch=1)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_functools); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_partial); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_branch); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_nbranch, __pyx_int_2) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_branch_2, __pyx_t_2) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":452
 * branch_all = functools.partial(branch, nbranch=None)
 * branch_2 = functools.partial(branch, nbranch=2)
 * branch_1 = functools.partial(branch, nbranch=1)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_functools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_partial); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_branch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_nbranch, __pyx_int_1) < 0) __PYX_ERR(0, 452, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_branch_1, __pyx_t_1) < 0) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":455
 * 
 * 
 * def _get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn):             # <<<<<<<<<<<<<<
 *     cdef int id1, id2
 *     either = k1 | k2
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_23_get_candidate, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_candidate, __pyx_t_1) < 0) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":469
 * 
 * 
 * def _push_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2s, queue, push_all, cost_fn):             # <<<<<<<<<<<<<<
 *     candidates = (_get_candidate(output, sizes, remaining, footprints, dim_ref_counts, k1, k2, cost_fn) for k2 in k2s)
 *     if push_all:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_25_push_candidate, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_push_candidate, __pyx_t_1) < 0) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":479
 * 
 * 
 * def _update_ref_counts(dim_to_keys, dim_ref_counts, dims):             # <<<<<<<<<<<<<<
 *     for dim in dims:
 *         count = len(dim_to_keys[dim])
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27_update_ref_counts, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_update_ref_counts, __pyx_t_1) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":493
 * 
 * 
 * def _simple_chooser(queue, remaining):             # <<<<<<<<<<<<<<
 *     """Default contraction chooser that simply takes the minimum cost option.
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_29_simple_chooser, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_simple_chooser, __pyx_t_1) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":502
 * 
 * 
 * def ssa_greedy_optimize(inputs, output, sizes, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     This is the core function for :func:`greedy` but produces a path with
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_31ssa_greedy_optimize, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ssa_greedy_optimize, __pyx_t_1) < 0) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":612
 * 
 * 
 * def greedy(inputs, output, size_dict, memory_limit=None, choose_fn=None, cost_fn='memory-removed'):             # <<<<<<<<<<<<<<
 *     """
 *     Finds the path by a three stage algorithm:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_33greedy, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_greedy, __pyx_t_1) < 0) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":658
 * 
 * 
 * def _tree_to_sequence(c):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a contraction tree to a contraction path as it has to be
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_35_tree_to_sequence, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_tree_to_sequence, __pyx_t_1) < 0) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":716
 * 
 * 
 * def _find_disconnected_subgraphs(inputs, output):             # <<<<<<<<<<<<<<
 *     """
 *     Finds disconnected subgraphs in the given list of inputs. Inputs are
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_37_find_disconnected_subgraphs, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_find_disconnected_subgraphs, __pyx_t_1) < 0) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":764
 * 
 * 
 * def _bitmapset_indices(s):             # <<<<<<<<<<<<<<
 *     """
 *     Returns a generator object allowing to iterate over the elements contained
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_39_bitmapset_indices, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bitmapset_indices, __pyx_t_1) < 0) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":795
 * 
 * 
 * class DynamicProgrammingOptimizer(PathOptimizer):             # <<<<<<<<<<<<<<
 *     """
 *     Finds the optimal path of pairwise contractions without intermediate outer
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PathOptimizer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_DynamicProgrammingOptimizer, __pyx_n_s_DynamicProgrammingOptimizer, (PyObject *) NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_kp_s_Finds_the_optimal_path_of_pairw); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "acqdp/utility/opt_einsum_paths.pyx":811
 *     """
 * 
 *     def __call__(self, inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *         """
 *         Parameters
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_27DynamicProgrammingOptimizer_1__call__, 0, __pyx_n_s_DynamicProgrammingOptimizer___ca_4, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__67);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_call, __pyx_t_5) < 0) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":795
 * 
 * 
 * class DynamicProgrammingOptimizer(PathOptimizer):             # <<<<<<<<<<<<<<
 *     """
 *     Finds the optimal path of pairwise contractions without intermediate outer
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_DynamicProgrammingOptimizer, __pyx_t_2, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DynamicProgrammingOptimizer, __pyx_t_5) < 0) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":971
 * 
 * 
 * def dynamic_programming(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     optimizer = DynamicProgrammingOptimizer()
 *     return optimizer(inputs, output, size_dict, memory_limit)
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_42dynamic_programming, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dynamic_programming, __pyx_t_2) < 0) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":976
 * 
 * 
 * _AUTO_CHOICES = {}             # <<<<<<<<<<<<<<
 * for i in range(1, 5):
 *     _AUTO_CHOICES[i] = optimal
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AUTO_CHOICES, __pyx_t_2) < 0) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":977
 * 
 * _AUTO_CHOICES = {}
 * for i in range(1, 5):             # <<<<<<<<<<<<<<
 *     _AUTO_CHOICES[i] = optimal
 * for i in range(5, 7):
 */
  for (__pyx_t_6 = 1; __pyx_t_6 < 5; __pyx_t_6+=1) {
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_i, __pyx_t_2) < 0) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":978
 * _AUTO_CHOICES = {}
 * for i in range(1, 5):
 *     _AUTO_CHOICES[i] = optimal             # <<<<<<<<<<<<<<
 * for i in range(5, 7):
 *     _AUTO_CHOICES[i] = branch_all
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_optimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AUTO_CHOICES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < 0)) __PYX_ERR(0, 978, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":979
 * for i in range(1, 5):
 *     _AUTO_CHOICES[i] = optimal
 * for i in range(5, 7):             # <<<<<<<<<<<<<<
 *     _AUTO_CHOICES[i] = branch_all
 * for i in range(7, 9):
 */
  for (__pyx_t_6 = 5; __pyx_t_6 < 7; __pyx_t_6+=1) {
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 979, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_i, __pyx_t_2) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":980
 *     _AUTO_CHOICES[i] = optimal
 * for i in range(5, 7):
 *     _AUTO_CHOICES[i] = branch_all             # <<<<<<<<<<<<<<
 * for i in range(7, 9):
 *     _AUTO_CHOICES[i] = branch_2
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_branch_all); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_AUTO_CHOICES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_3, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 980, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":981
 * for i in range(5, 7):
 *     _AUTO_CHOICES[i] = branch_all
 * for i in range(7, 9):             # <<<<<<<<<<<<<<
 *     _AUTO_CHOICES[i] = branch_2
 * for i in range(9, 15):
 */
  for (__pyx_t_6 = 7; __pyx_t_6 < 9; __pyx_t_6+=1) {
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_i, __pyx_t_2) < 0) __PYX_ERR(0, 981, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":982
 *     _AUTO_CHOICES[i] = branch_all
 * for i in range(7, 9):
 *     _AUTO_CHOICES[i] = branch_2             # <<<<<<<<<<<<<<
 * for i in range(9, 15):
 *     _AUTO_CHOICES[i] = branch_1
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_branch_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AUTO_CHOICES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < 0)) __PYX_ERR(0, 982, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":983
 * for i in range(7, 9):
 *     _AUTO_CHOICES[i] = branch_2
 * for i in range(9, 15):             # <<<<<<<<<<<<<<
 *     _AUTO_CHOICES[i] = branch_1
 * 
 */
  for (__pyx_t_6 = 9; __pyx_t_6 < 15; __pyx_t_6+=1) {
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_i, __pyx_t_2) < 0) __PYX_ERR(0, 983, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "acqdp/utility/opt_einsum_paths.pyx":984
 *     _AUTO_CHOICES[i] = branch_2
 * for i in range(9, 15):
 *     _AUTO_CHOICES[i] = branch_1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_branch_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_AUTO_CHOICES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(PyObject_SetItem(__pyx_t_3, __pyx_t_1, __pyx_t_2) < 0)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "acqdp/utility/opt_einsum_paths.pyx":987
 * 
 * 
 * def auto(inputs, output, size_dict, memory_limit=None):             # <<<<<<<<<<<<<<
 *     """Finds the contraction path by automatically choosing the method based on
 *     how many input arguments there are.
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_44auto, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_auto, __pyx_t_2) < 0) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":996
 * 
 * _PATH_OPTIONS = {
 *     'auto': auto,             # <<<<<<<<<<<<<<
 *     'optimal': optimal,
 *     'branch-all': branch_all,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_auto); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_auto, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":997
 * _PATH_OPTIONS = {
 *     'auto': auto,
 *     'optimal': optimal,             # <<<<<<<<<<<<<<
 *     'branch-all': branch_all,
 *     'branch-2': branch_2,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_optimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_optimal, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":998
 *     'auto': auto,
 *     'optimal': optimal,
 *     'branch-all': branch_all,             # <<<<<<<<<<<<<<
 *     'branch-2': branch_2,
 *     'branch-1': branch_1,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_branch_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_branch_all_2, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":999
 *     'optimal': optimal,
 *     'branch-all': branch_all,
 *     'branch-2': branch_2,             # <<<<<<<<<<<<<<
 *     'branch-1': branch_1,
 *     'greedy': greedy,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_branch_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_branch_2_2, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1000
 *     'branch-all': branch_all,
 *     'branch-2': branch_2,
 *     'branch-1': branch_1,             # <<<<<<<<<<<<<<
 *     'greedy': greedy,
 *     'eager': greedy,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_branch_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_branch_1_2, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1001
 *     'branch-2': branch_2,
 *     'branch-1': branch_1,
 *     'greedy': greedy,             # <<<<<<<<<<<<<<
 *     'eager': greedy,
 *     'opportunistic': greedy,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_greedy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_greedy, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1002
 *     'branch-1': branch_1,
 *     'greedy': greedy,
 *     'eager': greedy,             # <<<<<<<<<<<<<<
 *     'opportunistic': greedy,
 *     'dp': dynamic_programming,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_greedy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_eager, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1003
 *     'greedy': greedy,
 *     'eager': greedy,
 *     'opportunistic': greedy,             # <<<<<<<<<<<<<<
 *     'dp': dynamic_programming,
 *     'dynamic-programming': dynamic_programming
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_greedy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_opportunistic, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1004
 *     'eager': greedy,
 *     'opportunistic': greedy,
 *     'dp': dynamic_programming,             # <<<<<<<<<<<<<<
 *     'dynamic-programming': dynamic_programming
 * }
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_dynamic_programming); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dp, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1005
 *     'opportunistic': greedy,
 *     'dp': dynamic_programming,
 *     'dynamic-programming': dynamic_programming             # <<<<<<<<<<<<<<
 * }
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_dynamic_programming); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_s_dynamic_programming_2, __pyx_t_1) < 0) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PATH_OPTIONS, __pyx_t_2) < 0) __PYX_ERR(0, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1009
 * 
 * 
 * def register_path_fn(name, fn):             # <<<<<<<<<<<<<<
 *     """Add path finding function ``fn`` as an option with ``name``.
 *     """
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_46register_path_fn, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_register_path_fn, __pyx_t_2) < 0) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1018
 * 
 * 
 * def get_path_fn(path_type):             # <<<<<<<<<<<<<<
 *     """Get the correct path finding function from str ``path_type``.
 *     """
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_5acqdp_7utility_16opt_einsum_paths_48get_path_fn, NULL, __pyx_n_s_acqdp_utility_opt_einsum_paths); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_path_fn, __pyx_t_2) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "acqdp/utility/opt_einsum_paths.pyx":1
 * """             # <<<<<<<<<<<<<<
 * Contains the path technology behind opt_einsum in addition to several path helpers
 * """
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_ssa_to_linear_line_52, __pyx_kp_u_Convert_a_path_with_static_sing) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_linear_to_ssa_line_69, __pyx_kp_u_Convert_a_path_with_recycled_li) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_optimal_line_141, __pyx_kp_u_Computes_all_possible_pair_cont) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_BranchBound___call___line_316, __pyx_kp_u_Parameters_input_sets_list_List) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_greedy_line_612, __pyx_kp_u_Finds_the_path_by_a_three_stage) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_tree_to_sequence_line_658, __pyx_kp_u_Converts_a_contraction_tree_to) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_find_disconnected_subgraphs_lin, __pyx_kp_u_Finds_disconnected_subgraphs_in) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_bitmapset_indices_line_764, __pyx_kp_u_Returns_a_generator_object_allo) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_kp_u_DynamicProgrammingOptimizer___ca_5, __pyx_kp_u_Parameters_inputs_list_List_of) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init acqdp.utility.opt_einsum_paths", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init acqdp.utility.opt_einsum_paths");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddCObj(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        const long a = intval;
        long x;
        long b = PyInt_AS_LONG(op2);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        const long a = intval;
        long b, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG lla = intval;
        PY_LONG_LONG llb, llx;
#endif
        const digit* digits = ((PyLongObject*)op2)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op2);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            b = likely(size) ? digits[0] : 0;
            if (size == -1) b = -b;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        const long a = intval;
        double b = PyFloat_AS_DOUBLE(op2);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_Occurred();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    PyErr_Format(PyExc_TypeError,
        "%.200s object is not an iterator", Py_TYPE(iterator)->tp_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS
        next = iternext(iterator);
        if (likely(next))
            return next;
        #if PY_VERSION_HEX >= 0x02070000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
        #endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_items */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* pop_index */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix) {
    PyObject *r;
    if (unlikely(!py_ix)) return NULL;
    r = __Pyx__PyObject_PopIndex(L, py_ix);
    Py_DECREF(py_ix);
    return r;
}
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix) {
    return __Pyx_PyObject_CallMethod1(L, __pyx_n_s_pop, py_ix);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix) {
    Py_ssize_t size = PyList_GET_SIZE(L);
    if (likely(size > (((PyListObject*)L)->allocated >> 1))) {
        Py_ssize_t cix = ix;
        if (cix < 0) {
            cix += size;
        }
        if (likely(__Pyx_is_valid_index(cix, size))) {
            PyObject* v = PyList_GET_ITEM(L, cix);
            __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
            size -= 1;
            memmove(&PyList_GET_ITEM(L, cix), &PyList_GET_ITEM(L, cix+1), (size_t)(size-cix)*sizeof(PyObject*));
            return v;
        }
    }
    if (py_ix == Py_None) {
        return __Pyx__PyObject_PopNewIndex(L, PyInt_FromSsize_t(ix));
    } else {
        return __Pyx__PyObject_PopIndex(L, py_ix);
    }
}
#endif

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    (void)is_set;
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        (void)orig_length;
        (void)ppos;
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a & b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_and(op1, op2);
            }
        }
                x = a & b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla & llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceAnd : PyNumber_And)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RshiftObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a >> b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_rshift(op1, op2);
            }
        }
                x = a >> b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla >> llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceRshift : PyNumber_Rshift)(op1, op2);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
            if (unlikely(b == -1 && ((unsigned long)a) == 0-(unsigned long)a))
                return PyInt_Type.tp_as_number->nb_floor_divide(op1, op2);
            else {
                long q, r;
                q = a / b;
                r = a - q*b;
                q -= ((r != 0) & ((r ^ b) < 0));
                x = q;
            }
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_floor_divide(op1, op2);
            }
        }
                __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
                {
                    long q, r;
                    q = a / b;
                    r = a - q*b;
                    q -= ((r != 0) & ((r ^ b) < 0));
                    x = q;
                }
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                {
                    PY_LONG_LONG q, r;
                    q = lla / llb;
                    r = lla - q*llb;
                    q -= ((r != 0) & ((r ^ llb) < 0));
                    llx = q;
                }
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceFloorDivide : PyNumber_FloorDivide)(op1, op2);
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
